#!/bin/bash
# copyleft (0) xiaohu. 2010
#
# wrapper script for convert Ttyuitls DocBook document to
# other format.
#
echo "Ttyutils Document generator V 0.0.0"
echo "Copyleft (0) 2010 xiaohu417@gmail.com."
echo ""

if [ "$#" -lt 2 ] ; then
    echo "Usage:   $0 source-path format"
    echo ""
    echo "Options:"
    echo "  source-path : path to .docbook file"
    echo "  format      : html, pdf, man"
    echo ""
    exit 1
fi

# xsl
#
man_xsl=/usr/share/sgml/docbook/xsl-stylesheets-1.73.2/manpages/docbook.xsl
htm_xsl=/usr/share/sgml/docbook/xsl-stylesheets-1.73.2/html/docbook.xsl
fo_xsl=/usr/share/sgml/docbook/xsl-stylesheets-1.73.2/fo/docbook.xsl

# arguments
#
srcdir=$1
format=$2

# files to process
#
mans="ttyrcd.docbook ttylook.docbook ttyplay.docbook ttyagent.docbook ttyutils.docbook"
htms="$mans regex.docbook"
pdfs="$mans"

case $format in
    htm) files=$htms ;;
    pdf) files=$pdfs ;;
    man) files=$mans ;;
    *) "Bad args" ;;
esac


# . start
#
for f in $files ; do
    # prepend the source path, so we can run the command outside
    # current directory.
    # the output file place in build directory, so them can
    # got install.
    #
    file=$srcdir/$f
    base=`basename $f ".docbook"`

    # verify input document with xmllint.
    # xmllint is part of libxml2.
    #
    #xmllint --nonet --noout --xinclude "$file"

    # generate document
    #
    case $format in
	man)
	    if [ ${base}.1 -ot $file ] ; then
		xsltproc --novalid --nonet --xinclude $man_xsl $file
	    fi
	    ;;
	htm)
	    if [ ${base}.html -ot $file ] ; then
		xsltproc --novalid --nonet --xinclude --output "${base}.html" $htm_xsl $file
	    fi
	    ;;
	pdf)
	    if [ ${base}.pdf -ot $file ] ; then
		xsltproc --novalid --nonet --xinclude --output "${base}.fo" $fo_xsl $file
		fop -fo "${base}.fo" -pdf "${base}.pdf"
	    fi
	    ;;
	*)
	    echo "format '$format' unrecogizable."
	    ;;
    esac
done
# . end
