/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include <gdk/gdkkeysyms.h>

#include "glook-accel-map.h"

enum
{
  PROP_0,
  PROP_SETTINGS,
  PROP_NEW_TAB,
  PROP_NEW_WINDOW,
  PROP_PLAY_FILE,
  PROP_SHOW_CONTROL,
  PROP_CLOSE_WINDOW,
  PROP_QUIT,
  PROP_COPY,
  PROP_PASTE,
  PROP_SELECT_ALL,
  PROP_RELOAD,
  PROP_ZOOM_IN,
  PROP_ZOOM_OUT,
  PROP_ZOOM_100,
  PROP_WRITABLE,
  PROP_SAVE_SCREEN,
  PROP_SUSPENSION,
  PROP_DISCONNECT,
  PROP_CLOSE_TAB,
  PROP_TAB_1,
  PROP_TAB_2,
  PROP_TAB_3,
  PROP_TAB_4,
  PROP_TAB_5,
  PROP_TAB_6,
  PROP_TAB_7,
  PROP_TAB_8,
  PROP_TAB_9,
  PROP_LEFT_TAB,
  PROP_RIGHT_TAB,
  PROP_GET_START,
  PROP_RESOURCES,
  PROP_ADDRESS_BAR,
  PROP_PLAYBACK,
  PROP_FORWARD,
  PROP_BACKWARD,
};

typedef struct _GLookAccelMapEntry	GLookAccelMapEntry;

struct _GLookAccelMapPriv
{
  GSettings *settings;
};
#define _ACCEL_MAP_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_ACCEL_MAP, GLookAccelMapPriv))

struct _GLookAccelMapEntry
{
  const gchar        *accel_path;
  guint               accel_key;
  GdkModifierType     accel_mods;
  const gchar        *settings_key;
  const gchar        *prop_name;
  GSettingsBindFlags  flags;
};

G_DEFINE_TYPE (GLookAccelMap, glook_accel_map, G_TYPE_OBJECT);

static GLookAccelMapEntry	_accel_entries[] =
{
    { _GLOOK_ACCEL_PATH_NEW_TAB, 't', GDK_CONTROL_MASK,
      "new-tab-accel", "new-tab", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_NEW_WINDOW, 'n', GDK_CONTROL_MASK,
      "new-window-accel", "new-window", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_PLAY_FILE, 'o', GDK_CONTROL_MASK,
      "play-file-accel", "play-file", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_SHOW_CONTROL, GDK_F12, 0,
      "show-control-accel", "show-control", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_CLOSE_WINDOW, 'w', GDK_CONTROL_MASK,
      "close-window-accel", "close-window", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_QUIT, 'q', GDK_CONTROL_MASK,
      "quit-accel", "quit", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_COPY, 'c', GDK_CONTROL_MASK,
      "copy-accel", "copy", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_PASTE, 'v', GDK_CONTROL_MASK,
      "paste-accel", "paste", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_SELECT_ALL, 'a', GDK_CONTROL_MASK,
      "select-all-accel", "select-all", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_RELOAD, GDK_F5, 0,
      "reload-accel", "reload", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_ZOOM_IN, GDK_plus, GDK_CONTROL_MASK,
      "zoom-in-accel", "zoom-in", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_ZOOM_OUT, GDK_minus, GDK_CONTROL_MASK,
      "zoom-out-accel", "zoom-out", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_ZOOM_100, '0', GDK_CONTROL_MASK,
      "zoom-100-accel", "zoom-100", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_WRITABLE, GDK_Insert, 0,
      "writable-accel", "writable", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_SAVE_SCREEN, 's', GDK_CONTROL_MASK,
      "save-screen-accel", "save-screen", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_SUSPENSION, GDK_Pause, 0,
      "suspension-accel", "suspension", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_DISCONNECT, '\\', GDK_CONTROL_MASK,
      "disconnect-accel", "disconnect", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_CLOSE_TAB, GDK_F4, GDK_CONTROL_MASK,
      "close-tab-accel", "close-tab", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_TAB_1, '1', GDK_META_MASK,
      "tab-1-accel", "tab-1", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_TAB_2, '2', GDK_META_MASK,
      "tab-2-accel", "tab-2", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_TAB_3, '3', GDK_META_MASK,
      "tab-3-accel", "tab-3", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_TAB_4, '4', GDK_META_MASK,
      "tab-4-accel", "tab-4", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_TAB_5, '5', GDK_META_MASK,
      "tab-5-accel", "tab-5", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_TAB_6, '6', GDK_META_MASK,
      "tab-6-accel", "tab-6", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_TAB_7, '7', GDK_META_MASK,
      "tab-7-accel", "tab-7", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_TAB_8, '8', GDK_META_MASK,
      "tab-8-accel", "tab-8", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_TAB_9, '9', GDK_META_MASK,
      "tab-9-accel", "tab-9", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_LEFT_TAB, GDK_Left, GDK_META_MASK,
      "left-tab-accel", "left-tab", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_RIGHT_TAB, GDK_Right, GDK_META_MASK,
      "right-tab-accel", "right-tab", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_GET_START, GDK_F1, 0,
      "get-start-accel", "get-start", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_RESOURCES, 'h', GDK_CONTROL_MASK,
      "resources-accel", "resources", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_ADDRESS_BAR, 'l', GDK_CONTROL_MASK,
      "address-bar-accel", "address-bar", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_PLAYBACK, GDK_space, 0,
      "playback-accel", "playback", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_FORWARD, GDK_Right, GDK_CONTROL_MASK,
      "forward-accel", "forward", G_SETTINGS_BIND_DEFAULT, },
    { _GLOOK_ACCEL_PATH_BACKWARD, GDK_Left, GDK_CONTROL_MASK,
      "backward-accel", "backward", G_SETTINGS_BIND_DEFAULT, },
};

/**
 * glook_accel_map_reset:
 */
void
glook_accel_map_reset (void)
{
  gint i;

  for (i = 0; i < G_N_ELEMENTS (_accel_entries); i++)
    {
      gtk_accel_map_change_entry (_accel_entries[i].accel_path,
				  _accel_entries[i].accel_key,
				  _accel_entries[i].accel_mods,
				  TRUE);
    }
}


static void	glook_accel_map_set_settings	(GLookAccelMap   *map,
						 GSettings       *settings);

/**
 * glook_accel_map_init:
 */
static void
glook_accel_map_init (GLookAccelMap *map)
{
  map->priv = _ACCEL_MAP_GET_PRIVATE (map);

  map->priv->settings = NULL;
}

/**
 * glook_accel_map_finalize:
 */
static void
glook_accel_map_finalize (GObject *object)
{
  GLookAccelMap *map = GLOOK_ACCEL_MAP (object);

  if (map->priv->settings)
    {
      g_object_unref (map->priv->settings);
      map->priv->settings = NULL;
    }
  G_OBJECT_CLASS (glook_accel_map_parent_class)->finalize (object);
}

/**
 * glook_accel_map_get_property:
 */
static void
glook_accel_map_get_property (GObject    *object,
			      guint       property_id,
			      GValue     *value,
			      GParamSpec *pspec)
{
  GLookAccelMap *map = GLOOK_ACCEL_MAP (object);

  switch (property_id)
    {
    case PROP_SETTINGS:
      g_value_set_object (value, glook_accel_map_get_settings (map));
      break;
    case PROP_NEW_TAB:
      g_value_set_string (value, glook_accel_map_get_new_tab (map));
      break;
    case PROP_NEW_WINDOW:
      g_value_set_string (value, glook_accel_map_get_new_window (map));
      break;
    case PROP_PLAY_FILE:
      g_value_set_string (value, glook_accel_map_get_play_file (map));
      break;
    case PROP_SHOW_CONTROL:
      g_value_set_string (value, glook_accel_map_get_show_control (map));
      break;
    case PROP_CLOSE_WINDOW:
      g_value_set_string (value, glook_accel_map_get_close_window (map));
      break;
    case PROP_QUIT:
      g_value_set_string (value, glook_accel_map_get_quit (map));
      break;
    case PROP_COPY:
      g_value_set_string (value, glook_accel_map_get_copy (map));
      break;
    case PROP_PASTE:
      g_value_set_string (value, glook_accel_map_get_paste (map));
      break;
    case PROP_SELECT_ALL:
      g_value_set_string (value, glook_accel_map_get_select_all (map));
      break;
    case PROP_RELOAD:
      g_value_set_string (value, glook_accel_map_get_reload (map));
      break;
    case PROP_ZOOM_IN:
      g_value_set_string (value, glook_accel_map_get_zoom_in (map));
      break;
    case PROP_ZOOM_OUT:
      g_value_set_string (value, glook_accel_map_get_zoom_out (map));
      break;
    case PROP_ZOOM_100:
      g_value_set_string (value, glook_accel_map_get_zoom_100 (map));
      break;
    case PROP_WRITABLE:
      g_value_set_string (value, glook_accel_map_get_writable (map));
      break;
    case PROP_SAVE_SCREEN:
      g_value_set_string (value, glook_accel_map_get_save_screen (map));
      break;
    case PROP_SUSPENSION:
      g_value_set_string (value, glook_accel_map_get_suspension (map));
      break;
    case PROP_DISCONNECT:
      g_value_set_string (value, glook_accel_map_get_disconnect (map));
      break;
    case PROP_CLOSE_TAB:
      g_value_set_string (value, glook_accel_map_get_close_tab (map));
      break;
    case PROP_TAB_1:
      g_value_set_string (value, glook_accel_map_get_tab_1 (map));
      break;
    case PROP_TAB_2:
      g_value_set_string (value, glook_accel_map_get_tab_2 (map));
      break;
    case PROP_TAB_3:
      g_value_set_string (value, glook_accel_map_get_tab_3 (map));
      break;
    case PROP_TAB_4:
      g_value_set_string (value, glook_accel_map_get_tab_4 (map));
      break;
    case PROP_TAB_5:
      g_value_set_string (value, glook_accel_map_get_tab_5 (map));
      break;
    case PROP_TAB_6:
      g_value_set_string (value, glook_accel_map_get_tab_6 (map));
      break;
    case PROP_TAB_7:
      g_value_set_string (value, glook_accel_map_get_tab_7 (map));
      break;
    case PROP_TAB_8:
      g_value_set_string (value, glook_accel_map_get_tab_8 (map));
      break;
    case PROP_TAB_9:
      g_value_set_string (value, glook_accel_map_get_tab_9 (map));
      break;
    case PROP_LEFT_TAB:
      g_value_set_string (value, glook_accel_map_get_left_tab (map));
      break;
    case PROP_RIGHT_TAB:
      g_value_set_string (value, glook_accel_map_get_right_tab (map));
      break;
    case PROP_GET_START:
      g_value_set_string (value, glook_accel_map_get_get_start (map));
      break;
    case PROP_RESOURCES:
      g_value_set_string (value, glook_accel_map_get_resources (map));
      break;
    case PROP_ADDRESS_BAR:
      g_value_set_string (value, glook_accel_map_get_address_bar (map));
      break;
    case PROP_PLAYBACK:
      g_value_set_string (value, glook_accel_map_get_playback (map));
      break;
    case PROP_FORWARD:
      g_value_set_string (value, glook_accel_map_get_forward (map));
      break;
    case PROP_BACKWARD:
      g_value_set_string (value, glook_accel_map_get_backward (map));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_accel_map_set_property:
 */
static void
glook_accel_map_set_property (GObject      *object,
			      guint         property_id,
			      const GValue *value,
			      GParamSpec   *pspec)
{
  GLookAccelMap *map = GLOOK_ACCEL_MAP (object);

  switch (property_id)
    {
    case PROP_SETTINGS:
      glook_accel_map_set_settings (map, g_value_get_object (value));
      break;
    case PROP_NEW_TAB:
      glook_accel_map_set_new_tab (map, g_value_get_string (value));
      break;
    case PROP_NEW_WINDOW:
      glook_accel_map_set_new_window (map, g_value_get_string (value));
      break;
    case PROP_PLAY_FILE:
      glook_accel_map_set_play_file (map, g_value_get_string (value));
      break;
    case PROP_SHOW_CONTROL:
      glook_accel_map_set_show_control (map, g_value_get_string (value));
      break;
    case PROP_CLOSE_WINDOW:
      glook_accel_map_set_close_window (map, g_value_get_string (value));
      break;
    case PROP_QUIT:
      glook_accel_map_set_quit (map, g_value_get_string (value));
      break;
    case PROP_COPY:
      glook_accel_map_set_copy (map, g_value_get_string (value));
      break;
    case PROP_PASTE:
      glook_accel_map_set_paste (map, g_value_get_string (value));
      break;
    case PROP_SELECT_ALL:
      glook_accel_map_set_select_all (map, g_value_get_string (value));
      break;
    case PROP_RELOAD:
      glook_accel_map_set_reload (map, g_value_get_string (value));
      break;
    case PROP_ZOOM_IN:
      glook_accel_map_set_zoom_in (map, g_value_get_string (value));
      break;
    case PROP_ZOOM_OUT:
      glook_accel_map_set_zoom_out (map, g_value_get_string (value));
      break;
    case PROP_ZOOM_100:
      glook_accel_map_set_zoom_100 (map, g_value_get_string (value));
      break;
    case PROP_WRITABLE:
      glook_accel_map_set_writable (map, g_value_get_string (value));
      break;
    case PROP_SAVE_SCREEN:
      glook_accel_map_set_save_screen (map, g_value_get_string (value));
      break;
    case PROP_SUSPENSION:
      glook_accel_map_set_suspension (map, g_value_get_string (value));
      break;
    case PROP_DISCONNECT:
      glook_accel_map_set_disconnect (map, g_value_get_string (value));
      break;
    case PROP_CLOSE_TAB:
      glook_accel_map_set_close_tab (map, g_value_get_string (value));
      break;
    case PROP_TAB_1:
      glook_accel_map_set_tab_1 (map, g_value_get_string (value));
      break;
    case PROP_TAB_2:
      glook_accel_map_set_tab_2 (map, g_value_get_string (value));
      break;
    case PROP_TAB_3:
      glook_accel_map_set_tab_3 (map, g_value_get_string (value));
      break;
    case PROP_TAB_4:
      glook_accel_map_set_tab_4 (map, g_value_get_string (value));
      break;
    case PROP_TAB_5:
      glook_accel_map_set_tab_5 (map, g_value_get_string (value));
      break;
    case PROP_TAB_6:
      glook_accel_map_set_tab_6 (map, g_value_get_string (value));
      break;
    case PROP_TAB_7:
      glook_accel_map_set_tab_7 (map, g_value_get_string (value));
      break;
    case PROP_TAB_8:
      glook_accel_map_set_tab_8 (map, g_value_get_string (value));
      break;
    case PROP_TAB_9:
      glook_accel_map_set_tab_9 (map, g_value_get_string (value));
      break;
    case PROP_LEFT_TAB:
      glook_accel_map_set_left_tab (map, g_value_get_string (value));
      break;
    case PROP_RIGHT_TAB:
      glook_accel_map_set_right_tab (map, g_value_get_string (value));
      break;
    case PROP_GET_START:
      glook_accel_map_set_get_start (map, g_value_get_string (value));
      break;
    case PROP_RESOURCES:
      glook_accel_map_set_resources (map, g_value_get_string (value));
      break;
    case PROP_ADDRESS_BAR:
      glook_accel_map_set_address_bar (map, g_value_get_string (value));
      break;
    case PROP_PLAYBACK:
      glook_accel_map_set_playback (map, g_value_get_string (value));
      break;
    case PROP_FORWARD:
      glook_accel_map_set_forward (map, g_value_get_string (value));
      break;
    case PROP_BACKWARD:
      glook_accel_map_set_backward (map, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_accel_map_class_init:
 */
static void
glook_accel_map_class_init (GLookAccelMapClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookAccelMapPriv));

  gobject_class->get_property = glook_accel_map_get_property;
  gobject_class->set_property = glook_accel_map_set_property;
  gobject_class->finalize = glook_accel_map_finalize;

  g_object_class_install_property (gobject_class, PROP_SETTINGS,
				   g_param_spec_object ("settings", "settings", "gsettings object", G_TYPE_SETTINGS,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
  g_object_class_install_property (gobject_class, PROP_NEW_TAB,
				   g_param_spec_string ("new-tab", "new-tab", "accelerator for new tab", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_NEW_WINDOW,
				   g_param_spec_string ("new-window", "new-window", "accelerator for new window", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_PLAY_FILE,
				   g_param_spec_string ("play-file", "play-file", "accelerator for play file", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_SHOW_CONTROL,
				   g_param_spec_string ("show-control", "show-control", "accelerator for show control", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_CLOSE_WINDOW,
				   g_param_spec_string ("close-window", "close-window", "accelerator for close window", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_QUIT,
				   g_param_spec_string ("quit", "quit", "accelerator for quit", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_COPY,
				   g_param_spec_string ("copy", "copy", "accelerator for copy", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_PASTE,
				   g_param_spec_string ("paste", "paste", "accelerator for paste", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_SELECT_ALL,
				   g_param_spec_string ("select-all", "select-all", "accelerator for select all", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_RELOAD,
				   g_param_spec_string ("reload", "reload", "accelerator for reload", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_ZOOM_IN,
				   g_param_spec_string ("zoom-in", "zoom-in", "accelerator for zoom in", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_ZOOM_OUT,
				   g_param_spec_string ("zoom-out", "zoom-out", "accelerator for zoom out", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_ZOOM_100,
				   g_param_spec_string ("zoom-100", "zoom-100", "accelerator for zoom 100", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_WRITABLE,
				   g_param_spec_string ("writable", "writable", "accelerator for writable", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_SAVE_SCREEN,
				   g_param_spec_string ("save-screen", "save-screen", "accelerator for save terminal screen", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_SUSPENSION,
				   g_param_spec_string ("suspension", "suspension", "accelerator for suspension", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_DISCONNECT,
				   g_param_spec_string ("disconnect", "disconnect", "accelerator for disconnect", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_TAB_1,
				   g_param_spec_string ("tab-1", "tab-1", "accelerator for tab 1", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_TAB_2,
				   g_param_spec_string ("tab-2", "tab-2", "accelerator for tab 2", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_TAB_3,
				   g_param_spec_string ("tab-3", "tab-3", "accelerator for tab 3", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_TAB_4,
				   g_param_spec_string ("tab-4", "tab-4", "accelerator for tab 4", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_TAB_5,
				   g_param_spec_string ("tab-5", "tab-5", "accelerator for tab 5", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_TAB_6,
				   g_param_spec_string ("tab-6", "tab-6", "accelerator for tab 6", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_TAB_7,
				   g_param_spec_string ("tab-7", "tab-7", "accelerator for tab 7", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_TAB_8,
				   g_param_spec_string ("tab-8", "tab-8", "accelerator for tab 8", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_TAB_9,
				   g_param_spec_string ("tab-9", "tab-9", "accelerator for tab 9", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_LEFT_TAB,
				   g_param_spec_string ("left-tab", "left-tab", "accelerator for left tab", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_RIGHT_TAB,
				   g_param_spec_string ("right-tab", "right-tab", "accelerator for right tab", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_CLOSE_TAB,
				   g_param_spec_string ("close-tab", "close-tab", "accelerator for close tab", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_GET_START,
				   g_param_spec_string ("get-start", "get-start", "accelerator for get start", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_RESOURCES,
				   g_param_spec_string ("resources", "resources", "accelerator for online resources", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_ADDRESS_BAR,
				   g_param_spec_string ("address-bar", "address-bar", "accelerator for address bar", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_PLAYBACK,
				   g_param_spec_string ("playback", "playback", "accelerator for playback", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_FORWARD,
				   g_param_spec_string ("forward", "forward", "accelerator for forward", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_BACKWARD,
				   g_param_spec_string ("backward", "backward", "accelerator for backward", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
}

/**
 * glook_accel_map_new:
 */
GLookAccelMap*
glook_accel_map_new (GSettings *accel_settings)
{
  return g_object_new (GLOOK_TYPE_ACCEL_MAP,
		       "settings", accel_settings, NULL);
}

/**
 * on_accel_map_changed:
 */
static void
on_accel_map_changed (GtkAccelMap     *accel_map,
		      const gchar     *accel_path,
		      guint            accel_key,
		      GdkModifierType  accel_mods,
		      GLookAccelMap   *map)
{
  const gchar *prop_name;
  GLookAccelMapClass *klass;

  g_return_if_fail (GTK_IS_ACCEL_MAP (accel_map));
  g_return_if_fail (accel_path);
  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));

  /* skip unconcerned accelerator path. */
  if (!g_str_has_prefix (accel_path, GLOOK_ACCEL_PATH) ||
      strlen (accel_path) <= strlen (GLOOK_ACCEL_PATH))
    return;

  /* get property name from accel path. */
  prop_name = accel_path + strlen (GLOOK_ACCEL_PATH);

  klass = GLOOK_ACCEL_MAP_GET_CLASS (map);

  /* change map object property(also settings since we have bind
   * map property with settings). */
  if (g_object_class_find_property (G_OBJECT_CLASS (klass), prop_name))
    {
      gchar *accelerator = gtk_accelerator_name (accel_key, accel_mods);
      g_object_set (map, prop_name, accelerator, NULL);
      g_free (accelerator);
    }
}

/**
 * glook_accel_map_set_settings:
 */
static void
glook_accel_map_set_settings (GLookAccelMap *map,
			      GSettings     *settings)
{
  gint i;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (G_IS_SETTINGS (settings));

  map->priv->settings = g_object_ref (settings);

  for (i = 0; i < G_N_ELEMENTS (_accel_entries); i++)
    {
      g_settings_bind (settings, _accel_entries[i].settings_key,
		       map, _accel_entries[i].prop_name,
		       _accel_entries[i].flags);
    }
  /* when accel map changed, we need update settings. */
  g_signal_connect (gtk_accel_map_get (), "changed",
		    G_CALLBACK (on_accel_map_changed), map);
}

/**
 * glook_accel_map_get_settings:
 */
GSettings*
glook_accel_map_get_settings (GLookAccelMap *map)
{
  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);
  return map->priv->settings;
}

/**
 * glook_accel_map_set_new_tab:
 */
void
glook_accel_map_set_new_tab (GLookAccelMap *map,
			     const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_NEW_TAB, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "new-tab");
}

/**
 * glook_accel_map_get_new_tab:
 */
gchar*
glook_accel_map_get_new_tab (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_NEW_TAB, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_new_window:
 */
void
glook_accel_map_set_new_window (GLookAccelMap *map,
				const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_NEW_WINDOW, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "new-window");
}

/**
 * glook_accel_map_get_new_window:
 */
gchar*
glook_accel_map_get_new_window (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_NEW_WINDOW, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_play_file:
 */
void
glook_accel_map_set_play_file (GLookAccelMap *map,
			       const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_PLAY_FILE, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "play-file");
}

/**
 * glook_accel_map_get_play_file:
 */
gchar*
glook_accel_map_get_play_file (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_PLAY_FILE, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_show_control:
 */
void
glook_accel_map_set_show_control (GLookAccelMap *map,
				  const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_SHOW_CONTROL, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "show-control");
}

/**
 * glook_accel_map_get_show_control:
 */
gchar*
glook_accel_map_get_show_control (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_SHOW_CONTROL, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_close_window:
 */
void
glook_accel_map_set_close_window (GLookAccelMap *map,
				  const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_CLOSE_WINDOW, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "close-window");
}

/**
 * glook_accel_map_get_close_window:
 */
gchar*
glook_accel_map_get_close_window (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_CLOSE_WINDOW, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_quit:
 */
void
glook_accel_map_set_quit (GLookAccelMap *map,
			  const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_QUIT, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "quit");
}

/**
 * glook_accel_map_get_quit:
 */
gchar*
glook_accel_map_get_quit (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_QUIT, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_copy:
 */
void
glook_accel_map_set_copy (GLookAccelMap *map,
			  const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_COPY, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "copy");
}

/**
 * glook_accel_map_get_copy:
 */
gchar*
glook_accel_map_get_copy (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_COPY, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_paste:
 */
void
glook_accel_map_set_paste (GLookAccelMap *map,
			   const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_PASTE, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "paste");
}

/**
 * glook_accel_map_get_paste:
 */
gchar*
glook_accel_map_get_paste (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_PASTE, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_select_all:
 */
void
glook_accel_map_set_select_all (GLookAccelMap *map,
				const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_SELECT_ALL, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "select-all");
}

/**
 * glook_accel_map_get_select_all:
 */
gchar*
glook_accel_map_get_select_all (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_SELECT_ALL, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_reload:
 */
void
glook_accel_map_set_reload (GLookAccelMap *map,
			    const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_RELOAD, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "reload");
}

/**
 * glook_accel_map_get_reload:
 */
gchar*
glook_accel_map_get_reload (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_RELOAD, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_zoom_in:
 */
void
glook_accel_map_set_zoom_in (GLookAccelMap *map,
			     const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_ZOOM_IN, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "zoom-in");
}

/**
 * glook_accel_map_get_zoom_in:
 */
gchar*
glook_accel_map_get_zoom_in (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_ZOOM_IN, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_zoom_out:
 */
void
glook_accel_map_set_zoom_out (GLookAccelMap *map,
			      const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_ZOOM_OUT, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "zoom-out");
}

/**
 * glook_accel_map_get_zoom_out:
 */
gchar*
glook_accel_map_get_zoom_out (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_ZOOM_OUT, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_zoom_100:
 */
void
glook_accel_map_set_zoom_100 (GLookAccelMap *map,
			      const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_ZOOM_100, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "zoom-100");
}

/**
 * glook_accel_map_get_zoom_100:
 */
gchar*
glook_accel_map_get_zoom_100 (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_ZOOM_100, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_writable:
 */
void
glook_accel_map_set_writable (GLookAccelMap *map,
			      const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_WRITABLE, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "writable");
}

/**
 * glook_accel_map_get_writable:
 */
gchar*
glook_accel_map_get_writable (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_WRITABLE, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_save_screen:
 */
void
glook_accel_map_set_save_screen (GLookAccelMap *map,
				 const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_SAVE_SCREEN, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "save-screen");
}

/**
 * glook_accel_map_get_save_screen:
 */
gchar*
glook_accel_map_get_save_screen (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_SAVE_SCREEN, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_suspension:
 */
void
glook_accel_map_set_suspension (GLookAccelMap *map,
				const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_SUSPENSION, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "suspension");
}

/**
 * glook_accel_map_get_suspension:
 */
gchar*
glook_accel_map_get_suspension (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_SUSPENSION, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_disconnect:
 */
void
glook_accel_map_set_disconnect (GLookAccelMap *map,
				const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_DISCONNECT, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "disconnect");
}

/**
 * glook_accel_map_get_disconnect:
 */
gchar*
glook_accel_map_get_disconnect (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_DISCONNECT, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_tab_1:
 */
void
glook_accel_map_set_tab_1 (GLookAccelMap *map,
			   const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_TAB_1, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "tab-1");
}

/**
 * glook_accel_map_get_tab_1:
 */
gchar*
glook_accel_map_get_tab_1 (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_TAB_1, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_tab_2:
 */
void
glook_accel_map_set_tab_2 (GLookAccelMap *map,
			   const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_TAB_2, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "tab-2");
}

/**
 * glook_accel_map_get_tab_2:
 */
gchar*
glook_accel_map_get_tab_2 (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_TAB_2, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_tab_3:
 */
void
glook_accel_map_set_tab_3 (GLookAccelMap *map,
			   const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_TAB_3, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "tab-3");
}

/**
 * glook_accel_map_get_tab_3:
 */
gchar*
glook_accel_map_get_tab_3 (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_TAB_3, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_tab_4:
 */
void
glook_accel_map_set_tab_4 (GLookAccelMap *map,
			   const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_TAB_4, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "tab-4");
}

/**
 * glook_accel_map_get_tab_4:
 */
gchar*
glook_accel_map_get_tab_4 (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_TAB_4, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_tab_5:
 */
void
glook_accel_map_set_tab_5 (GLookAccelMap *map,
			   const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_TAB_5, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "tab-5");
}

/**
 * glook_accel_map_get_tab_5:
 */
gchar*
glook_accel_map_get_tab_5 (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_TAB_5, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_tab_6:
 */
void
glook_accel_map_set_tab_6 (GLookAccelMap *map,
			   const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_TAB_6, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "tab-6");
}

/**
 * glook_accel_map_get_tab_6:
 */
gchar*
glook_accel_map_get_tab_6 (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_TAB_6, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_tab_7:
 */
void
glook_accel_map_set_tab_7 (GLookAccelMap *map,
			   const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_TAB_7, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "tab-7");
}

/**
 * glook_accel_map_get_tab_7:
 */
gchar*
glook_accel_map_get_tab_7 (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_TAB_7, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_tab_8:
 */
void
glook_accel_map_set_tab_8 (GLookAccelMap *map,
			   const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_TAB_8, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "tab-8");
}

/**
 * glook_accel_map_get_tab_8:
 */
gchar*
glook_accel_map_get_tab_8 (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_TAB_8, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_tab_9:
 */
void
glook_accel_map_set_tab_9 (GLookAccelMap *map,
			   const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_TAB_9, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "tab-9");
}

/**
 * glook_accel_map_get_tab_9:
 */
gchar*
glook_accel_map_get_tab_9 (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_TAB_9, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_left_tab:
 */
void
glook_accel_map_set_left_tab (GLookAccelMap *map,
			      const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_LEFT_TAB, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "left-tab");
}

/**
 * glook_accel_map_get_left_tab:
 */
gchar*
glook_accel_map_get_left_tab (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_LEFT_TAB, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_right_tab:
 */
void
glook_accel_map_set_right_tab (GLookAccelMap *map,
			       const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_RIGHT_TAB, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "right-tab");
}

/**
 * glook_accel_map_get_right_tab:
 */
gchar*
glook_accel_map_get_right_tab (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_RIGHT_TAB, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_close_tab:
 */
void
glook_accel_map_set_close_tab (GLookAccelMap *map,
			       const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_CLOSE_TAB, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "close-tab");
}

/**
 * glook_accel_map_get_close_tab:
 */
gchar*
glook_accel_map_get_close_tab (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_CLOSE_TAB, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_get_start:
 */
void
glook_accel_map_set_get_start (GLookAccelMap *map,
			       const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_GET_START, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "get-start");
}

/**
 * glook_accel_map_get_get_start:
 */
gchar*
glook_accel_map_get_get_start (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_GET_START, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_resources:
 */
void
glook_accel_map_set_resources (GLookAccelMap *map,
			       const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_RESOURCES, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "resources");
}

/**
 * glook_accel_map_get_resources:
 */
gchar*
glook_accel_map_get_resources (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_RESOURCES, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_address_bar:
 */
void
glook_accel_map_set_address_bar (GLookAccelMap *map,
				 const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_ADDRESS_BAR, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "address-bar");
}

/**
 * glook_accel_map_get_address_bar:
 */
gchar*
glook_accel_map_get_address_bar (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_ADDRESS_BAR, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_playback:
 */
void
glook_accel_map_set_playback (GLookAccelMap *map,
			      const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_PLAYBACK, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "playback");
}

/**
 * glook_accel_map_get_playback:
 */
gchar*
glook_accel_map_get_playback (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_PLAYBACK, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_forward:
 */
void
glook_accel_map_set_forward (GLookAccelMap *map,
			     const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_FORWARD, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "forward");
}

/**
 * glook_accel_map_get_forward:
 */
gchar*
glook_accel_map_get_forward (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_FORWARD, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

/**
 * glook_accel_map_set_backward:
 */
void
glook_accel_map_set_backward (GLookAccelMap *map,
			      const gchar   *accelerator)
{
  guint key;
  GdkModifierType mods;

  g_return_if_fail (GLOOK_IS_ACCEL_MAP (map));
  g_return_if_fail (accelerator);

  gtk_accelerator_parse (accelerator, &key, &mods);
  gtk_accel_map_change_entry (GLOOK_ACCEL_PATH_BACKWARD, key, mods, TRUE);
  g_object_notify (G_OBJECT (map), "backward");
}

/**
 * glook_accel_map_get_backward:
 */
gchar*
glook_accel_map_get_backward (GLookAccelMap *map)
{
  GtkAccelKey accel_key;

  g_return_val_if_fail (GLOOK_IS_ACCEL_MAP (map), NULL);

  if (!gtk_accel_map_lookup_entry (GLOOK_ACCEL_PATH_BACKWARD, &accel_key))
    return NULL;

  return gtk_accelerator_name (accel_key.accel_key, accel_key.accel_mods);
}

