/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-app.h"
#include "glook-toolbar.h"
#include "glook-addrbar.h"
#include "glook-notebook.h"
#include "glook-terminal.h"
#include "glook-tab.h"
#include "glook-agent-tab.h"
#include "glook-look-tab.h"
#include "glook-play-tab.h"
#include "glook-blank-tab.h"
#include "glook-terminal-tab.h"
#include "glook-local-preference.h"
#include "glook-uri.h"
#include "glook-misc.h"
#include "glook-ui.h"
#include "glook-window.h"
#include "glook-actions.h"

/* add bookmark dialog ui file. */
#ifdef DEBUG
# define ADD_BOOKMARK_UI_FILE	SRCDIR G_DIR_SEPARATOR_S "glade" G_DIR_SEPARATOR_S "add-bookmark-dialog.xml"
#else
# define ADD_BOOKMARK_UI_FILE	GLADEDIR G_DIR_SEPARATOR_S "add-bookmark-dialog.xml"
#endif

/**
 * on_app_agent_server_activated:
 *
 * connect agent server to list all look servers.
 */
void
on_app_open_agent_tab_activated (GtkAction   *action,
				 GLookWindow *window)
{
  GSettings *settings;
  gchar *agent_address, *string;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_GENERAL);

  /* read agent server address from settings. */
  agent_address = g_settings_get_string (settings, "agent-address");

  /* fallback to default address. */
  if (!agent_address || strlen (agent_address) == 0)
    {
      g_free (agent_address);
      agent_address = g_strdup ("proto=tcp,host=localhost,port=5433");
    }
  /* construct a uri to open agent tab. */
  string = g_strdup_printf ("agent:%s", agent_address);
  g_free (agent_address);

  glook_window_open_uri_string (window, string);
  g_free (string);
}

/**
 * on_app_address_bar_activated:
 *
 * grab address bar focus.
 */
void
on_app_address_bar_activated (GtkAction   *action,
			      GLookWindow *window)
{
  GtkWidget *addrbar;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  addrbar = glook_window_get_addrbar (window);
  gtk_widget_grab_focus (addrbar);
}

/**
 * on_app_new_tab_activated:
 *
 * called when NewTab action is activated.
 *
 * this function will create a new _blank_ tab widget
 * and insert it to current window's notebook.
 */
void
on_app_new_tab_activated (GtkAction   *action,
			  GLookWindow *window)
{
  GtkWidget *notebook;
  GtkWidget *tab;
  gint n;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  tab = glook_blank_tab_new ();
  glook_blank_tab_set_uri (GLOOK_BLANK_TAB (tab));
  gtk_widget_show (tab);

  notebook = glook_window_get_notebook (window);
  n = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));

  glook_notebook_insert (GLOOK_NOTEBOOK (notebook), tab, n + 1);
  glook_notebook_set_current_n (GLOOK_NOTEBOOK (notebook), n + 1);
}

/**
 * on_app_new_window_activated:
 *
 * create a new main window.
 */
void
on_app_new_window_activated (GtkAction   *action,
			     GLookWindow *window)
{
  GtkWidget *newwin;
  GLookApp *app = glook_app_get ();

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  newwin = glook_app_create_window (app, TRUE);
  gtk_window_present (GTK_WINDOW (newwin));
}

/**
 * glook_open_record_file:
 */
void
glook_open_record_file (GLookWindow *window)
{
  static gchar *lastdir = NULL;
  GtkWidget *dialog;
  GtkFileChooser *chooser;
  GtkFileFilter *filter;

  g_return_if_fail (GTK_IS_WINDOW (window));

  dialog = gtk_file_chooser_dialog_new (_("Open record file"),
					GTK_WINDOW (window),
					GTK_FILE_CHOOSER_ACTION_OPEN,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
					NULL);
  chooser = GTK_FILE_CHOOSER (dialog);

  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("Terminal session record file"));
  gtk_file_filter_add_mime_type (filter, "application/x-ttyrcd");
  gtk_file_chooser_add_filter (chooser, filter);
  gtk_file_chooser_set_filter (chooser, filter);

  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("All files"));
  gtk_file_filter_add_pattern (filter, "*");
  gtk_file_chooser_add_filter (chooser, filter);

  if (lastdir)
    gtk_file_chooser_set_current_folder (chooser, lastdir);

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      gchar *string, *path;

      path = gtk_file_chooser_get_filename (chooser);
      lastdir = g_path_get_dirname (path);

      string = g_strdup_printf ("play:path=%s", path);
      g_free (path);

      glook_window_open_uri_string (GLOOK_WINDOW (window), string);
      g_free (string);
    }
  gtk_widget_destroy (dialog);
}

/**
 * on_app_play_file_activated:
 */
void
on_app_play_file_activated (GtkAction   *action,
			    GLookWindow *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  glook_open_record_file (window);
}

/**
 * on_app_play_recent_activated:
 */
void
on_app_play_recent_activated (GtkRecentChooser *chooser,
			      GLookWindow      *window)
{
  gchar *uri, *string;

  g_return_if_fail (GTK_IS_RECENT_CHOOSER (chooser));
  g_return_if_fail (GTK_IS_WINDOW (window));

  uri = gtk_recent_chooser_get_current_uri (chooser);
  string = g_strdup_printf ("play:uri=%s", uri);
  g_free (uri);

  glook_window_open_uri_string (GLOOK_WINDOW (window), string);
  g_free (string);
}

/**
 * on_app_open_terminal_activated:
 */
#ifdef G_OS_UNIX
void
on_app_open_terminal_activated (GtkAction   *action,
				GLookWindow *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  glook_window_open_uri_string (GLOOK_WINDOW (window), "terminal:");
}
#endif

/**
 * resize_window_idle:
 */
static gboolean
resize_window_idle (gpointer data)
{
  GtkWidget *window = data;
  GdkWindow *gdkwindow;
  GdkWindowState state;

  g_return_val_if_fail (GTK_IS_WINDOW (window), FALSE);

  gdkwindow = gtk_widget_get_window (window);
  if (gdkwindow && GDK_IS_WINDOW (gdkwindow))
    {
      state = gdk_window_get_state (gdkwindow);

      if ((state & (GDK_WINDOW_STATE_MAXIMIZED |
		    GDK_WINDOW_STATE_FULLSCREEN)) == 0)
	{
	  gtk_window_resize (GTK_WINDOW (window), 1, 1);
	  gtk_widget_queue_resize (GTK_WIDGET (window));
	}
    }
  return FALSE;
}

/**
 * on_app_show_control_toggled:
 *
 * show or hide menubar, toolbar, statusbar and so on.
 */
void
on_app_show_control_toggled (GtkToggleAction *action,
			     GLookWindow     *window)
{
  GtkWidget *menubar;
  GtkWidget *toolbar;
  GtkWidget *notebook;
  GtkWidget *statusbar;
  gboolean visible;
  gint n, i;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  menubar = glook_window_get_menubar (window);
  toolbar = glook_window_get_toolbar (window);
  notebook = glook_window_get_notebook (window);
  statusbar = glook_window_get_statusbar (window);

  visible = gtk_toggle_action_get_active (action);

  gtk_widget_set_visible (menubar, visible);
  gtk_widget_set_visible (toolbar, visible);
  gtk_widget_set_visible (statusbar, visible);

  n = gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook));
  gtk_notebook_set_show_tabs (GTK_NOTEBOOK (notebook), visible && n > 1);

  /* resize window. */
  for (i = 0; i < n; i++)
    {
      GtkWidget *tab;

      tab = gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), i);
      if (GLOOK_IS_LOOK_TAB (tab) || GLOOK_IS_PLAY_TAB (tab))
	{
	  g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
			   resize_window_idle,
			   g_object_ref (window),
			   (GDestroyNotify) g_object_unref);
	  break;
	}
    }
}

/**
 * on_app_bookmarks_activated:
 */
void
on_app_bookmarks_activated (GtkAction   *action,
			    GLookWindow *window)
{
  const gchar *uri = "local:path=bookmark.xml";

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  glook_window_open_uri_string (window, uri);
}

/**
 * on_app_preferene_activated:
 *
 * open preference tab.
 */
void
on_app_preference_activated (GtkAction    *action,
			     GLookWindow  *window)
{
  const gchar *uri = "local:path=preference.xml";

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  glook_window_open_uri_string (window, uri);
}

/**
 * on_app_log_activated:
 */
void
on_app_log_activated (GtkAction   *action,
		      GLookWindow *window)
{
  const gchar *uri = "local:path=log.xml";

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  glook_window_open_uri_string (window, uri);
}

/**
 * on_app_close_window_activated:
 *
 * close current window. all tabs of current window
 * will be destroy.
 */
void
on_app_close_window_activated (GtkAction   *action,
			       GLookWindow *window)
{
  GSettings *settings;
  GtkWidget *notebook;
  gint n;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  /* popup warning dialog when close multiple open tabs. */
  settings = glook_app_get_settings (glook_app_get (), SETTINGS_GENERAL);
  if (g_settings_get_boolean (settings, "close-window-warning"))
    {
      notebook = glook_window_get_notebook (window);
      n = gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook));
      if (n > 1)
	{
	  GtkWidget *dialog;

	  dialog = gtk_message_dialog_new (GTK_WINDOW (window),
					   GTK_DIALOG_MODAL |
					   GTK_DIALOG_DESTROY_WITH_PARENT,
					   GTK_MESSAGE_QUESTION,
					   GTK_BUTTONS_YES_NO,
					   _("Close all tabs?"));
	  gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
						    _("Close the window will close "
						      "all open %d tabs."), n);
	  if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_YES)
	    {
	      gtk_widget_destroy (dialog);
	      return;
	    }
	  gtk_widget_destroy (dialog);
	}
    }
  gtk_widget_set_sensitive (GTK_WIDGET (window), FALSE);

  while (gtk_events_pending ())
    gtk_main_iteration ();

  glook_window_delete (window);
  gtk_widget_destroy (GTK_WIDGET (window));
}

/**
 * on_app_quit_activated:
 *
 * quit the application.
 */
void
on_app_quit_activated (GtkAction   *action,
		       GLookWindow *window)
{
  glook_app_quit (glook_app_get ());
}

/**
 * on_tab_copy_activated:
 */
void
on_tab_copy_activated (GtkAction    *action,
		       GLookWindow  *window)
{
  GtkWidget *widget;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  widget = gtk_window_get_focus (GTK_WINDOW (window));
  if (!widget)
    return;

  if (GTK_IS_ENTRY (widget))
    {
      gtk_editable_copy_clipboard (GTK_EDITABLE (widget));
      return;
    }
  if (GTK_IS_TEXT_VIEW (widget))
    {
      GtkTextBuffer *buffer;
      GtkClipboard *clipboard;

      buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (widget));
      if (buffer)
	{
	  clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
	  gtk_text_buffer_copy_clipboard (buffer, clipboard);
	}
      return;
    }
  if (VTE_IS_TERMINAL (widget))
    {
      vte_terminal_copy_clipboard (VTE_TERMINAL (widget));
      return;
    }
}

/**
 * on_tab_paste_activated:
 */
void
on_tab_paste_activated (GtkAction    *action,
			GLookWindow  *window)
{
  GtkWidget *widget;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  widget = gtk_window_get_focus (GTK_WINDOW (window));
  if (!widget)
    return;

  if (GTK_IS_ENTRY (widget))
    {
      /* we can paste to entry only if it is editable. */
      if (gtk_editable_get_editable (GTK_EDITABLE (widget)))
	gtk_editable_paste_clipboard (GTK_EDITABLE (widget));

      return;
    }
  if (GTK_IS_TEXT_VIEW (widget))
    {
      /* we can paste to textview only if it is editable. */
      if (gtk_text_view_get_editable (GTK_TEXT_VIEW (widget)))
	{
	  GtkTextBuffer *buffer;
	  GtkClipboard *clipboard;

	  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (widget));
	  if (buffer)
	    {
	      clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
	      gtk_text_buffer_paste_clipboard (buffer, clipboard, NULL, TRUE);
	    }
	}
      return;
    }
  /* paste data from clipboard to terminal means send data
   * to look server. */
  if (GLOOK_IS_TERMINAL (widget))
    {
      GtkWidget *tab;

      tab = gtk_widget_get_ancestor (widget, GLOOK_TYPE_LOOK_TAB);
      if (tab)
	{
	  glook_look_tab_paste_clipboard (GLOOK_LOOK_TAB (tab));
	  return;
	}
#ifdef G_OS_UNIX
      tab = gtk_widget_get_ancestor (widget, GLOOK_TYPE_TERMINAL_TAB);
      if (tab)
	{
	  vte_terminal_paste_clipboard (VTE_TERMINAL (widget));
	  return;
	}
#endif
      return;
    }
}

/**
 * on_tab_select_all_activated:
 *
 * not all widget support select all.
 */
void
on_tab_select_all_activated (GtkAction    *action,
			     GLookWindow  *window)
{
  GtkWidget *widget;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  widget = gtk_window_get_focus (GTK_WINDOW (window));
  if (!widget)
    return;

  /* Entry support select all operation. */
  if (GTK_IS_ENTRY (widget))
    {
      gtk_editable_select_region (GTK_EDITABLE (widget), 0, -1);
      return;
    }
  /* Textview support select all operation. */
  if (GTK_IS_TEXT_VIEW (widget))
    {
      GtkTextBuffer *buffer;
      GtkTextIter start_iter, end_iter;

      buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (widget));
      if (buffer)
	{
	  gtk_text_buffer_get_bounds (buffer, &start_iter, &end_iter);
	  gtk_text_buffer_select_range (buffer, &start_iter, &end_iter);
	}
      return;
    }
  /* Treeview maybe support select all operation. */
  if (GTK_IS_TREE_VIEW (widget))
    {
      GtkTreeSelection *selection;

      selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (widget));
      if (gtk_tree_selection_get_mode (selection) == GTK_SELECTION_MULTIPLE)
	gtk_tree_selection_select_all (selection);
      return;
    }
  /* Terminal support select all operation. */
  if (VTE_IS_TERMINAL (widget))
    {
      vte_terminal_select_all (VTE_TERMINAL (widget));
      return;
    }
}

/**
 * on_tab_close_tab_activated:
 */
void
on_tab_close_tab_activated (GtkAction    *action,
			    GLookWindow  *window)
{
  GtkWidget *notebook;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  notebook = glook_window_get_notebook (window);
  glook_notebook_remove_current (GLOOK_NOTEBOOK (notebook));
}

/**
 * on_tab_zoom_in_activated:
 */
void
on_tab_zoom_in_activated (GtkAction    *action,
			  GLookWindow  *window)
{
  GtkWidget *toolbar;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  toolbar = glook_window_get_toolbar (window);
  glook_toolbar_zoom_in (GLOOK_TOOLBAR (toolbar));
}

/**
 * on_tab_zoom_out_activated:
 */
void
on_tab_zoom_out_activated (GtkAction    *action,
			   GLookWindow  *window)
{
  GtkWidget *toolbar;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  toolbar = glook_window_get_toolbar (window);
  glook_toolbar_zoom_out (GLOOK_TOOLBAR (toolbar));
}

/**
 * on_tab_zoom_100_activated:
 */
void
on_tab_zoom_100_activated (GtkAction   *action,
			   GLookWindow *window)
{
  GtkWidget *toolbar;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  toolbar = glook_window_get_toolbar (window);
  glook_toolbar_zoom_100 (GLOOK_TOOLBAR (toolbar));
}

/**
 * on_tab_reload_activated:
 *
 * reload current tab. not all tab support reload.
 */
void
on_tab_reload_activated (GtkAction   *action,
			 GLookWindow *window)
{
  GtkWidget *notebook;
  GtkWidget *tab;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  notebook = glook_window_get_notebook (window);
  tab = glook_notebook_get_current (GLOOK_NOTEBOOK (notebook));
  glook_tab_reload (GLOOK_TAB (tab));
}

/**
 * on_tab_switch_activated:
 *
 * switch to specified tab(from 0 to 8, 9 means last tab).
 */
void
on_tab_switch_activated (GtkAction   *action,
			 GLookWindow *window)
{
  const gchar *p;
  GtkWidget *notebook;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  /* get the tab number from action name. */
  p = gtk_action_get_name (action);
  while (*p != '\0')
    {
      if (g_ascii_isdigit (*p))
	{
	  gint n = *p - '0' - 1;

	  notebook = glook_window_get_notebook (window);

	  if (n < 8)
	    gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), n);
	  else
	    gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), -1);
	}
       p++;
    }
}

/**
 * on_tab_switch_left_activated:
 *
 * switch to previous tab.
 */
void
on_tab_switch_left_activated (GtkAction   *action,
			      GLookWindow *window)
{
  GtkWidget *notebook;
  gint n;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  notebook = glook_window_get_notebook (window);

  n = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));
  if (n > 0)
    gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), n - 1);
}

/**
 * on_tab_switch_right_activated:
 *
 * switch to next tab.
 */
void
on_tab_switch_right_activated (GtkAction   *action,
			       GLookWindow *window)
{
  GtkWidget *notebook;
  gint n;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  notebook = glook_window_get_notebook (window);

  n = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));
  gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), n + 1);
}

/**
 * glook_open_manpage_with_term:
 */
static void
glook_open_manpage (GLookWindow *window,
		    const gchar *manual,
		    const gchar *title)
{
  gchar buff[128];
  gchar *file;

  g_return_if_fail (GLOOK_IS_WINDOW (window));
  g_return_if_fail (manual);

  /* first try open manpage with terminal. */
#ifdef G_OS_UNIX
  g_snprintf (buff, sizeof (buff), "man %s", manual);
  if (glook_launch_terminal (GTK_WIDGET (window), buff, title))
    return;
#endif

  /* open html manual with web browser. */
  g_snprintf (buff, sizeof (buff), "%s.html", manual);
#ifdef DEBUG
  file = g_build_filename ("..", "..", "docs", buff, NULL);
#else
  file = g_build_filename (DOCDIR, buff, NULL);
#endif
  glook_open_url (GTK_WIDGET (window), file);
  g_free (file);
}

/**
 * on_help_getstart_activated:
 */
void
on_help_getstart_activated (GtkAction   *action,
			    GLookWindow *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  glook_open_manpage (window, "ttyutils", "ttyutils");
}

/**
 * on_help_ttylook_activated:
 */
void
on_help_ttylook_activated (GtkAction   *action,
			   GLookWindow *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  glook_open_manpage (window, "ttylook", "ttylook");
}

/**
 * on_help_ttyplay_activated:
 */
void
on_help_ttyplay_activated (GtkAction   *action,
			   GLookWindow *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  glook_open_manpage (window, "ttyplay", "ttyplay");
}

/**
 * on_help_ttyrcd_activated:
 */
void
on_help_ttyrcd_activated (GtkAction   *action,
			  GLookWindow *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  glook_open_manpage (window, "ttyrcd", "ttyrcd");
}

/**
 * on_help_ttyagent_activated:
 */
void
on_help_ttyagent_activated (GtkAction   *action,
			    GLookWindow *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  glook_open_manpage (window, "ttyagent", "ttyagent");
}

/**
 * on_help_content_activated:
 *
 * show user tutorial contents.
 */
void
on_help_online_activated (GtkAction   *action,
			  GLookWindow *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  glook_open_url (GTK_WIDGET (window), WEBSITE);
}

/**
 * on_help_about_activated:
 *
 * open a dialog to display application information.
 */
void
on_help_about_activated (GtkAction   *action,
			 GLookWindow *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  /* show about dialog. */
  glook_about (GTK_WIDGET (window));
}

/**
 * on_window_toolbar_zoom:
 *
 * zoom current tab. not all tab support zoom.
 */
void
on_window_toolbar_zoom (GtkWidget    *toolbar,
			gint          factor,
			GLookWindow  *window)
{
  GtkWidget *notebook;
  GtkWidget *tab;

  g_return_if_fail (GLOOK_IS_TOOLBAR (toolbar));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  notebook = glook_window_get_notebook (window);
  tab = glook_notebook_get_current (GLOOK_NOTEBOOK (notebook));

  if (GLOOK_TAB_ZOOM_SUPPORT (tab))
    {
      GtkWidget *combo;

      /* if zoom fail, we need scroll back zoom combo. */
      if (!glook_tab_zoom (GLOOK_TAB (tab), factor))
	{
	  /* block the handler to prevent loop. */
	  g_signal_handlers_block_by_func (toolbar, on_window_toolbar_zoom, window);

	  /* scroll back zoom combo. */
	  combo = glook_toolbar_get_zoom_combo (GLOOK_TOOLBAR (toolbar));
	  gtk_combo_box_set_active (GTK_COMBO_BOX (combo),
				    glook_tab_get_zoom_index (GLOOK_TAB (tab)));
	  /* unblock the handler. */
	  g_signal_handlers_unblock_by_func (toolbar, on_window_toolbar_zoom, window);
	}
    }
}

/**
 * on_window_addrbar_erase:
 *
 * this function will be called when user press the ESC
 * key in the address bar.
 */
void
on_window_addrbar_erase (GtkEntry    *addrbar,
			 GLookWindow *window)
{
  GtkWidget *notebook;
  GtkWidget *tab;
  GLookUri  *uri;

  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  notebook = glook_window_get_notebook (window);
  tab = glook_notebook_get_current (GLOOK_NOTEBOOK (notebook));

  /* current tab is not active. clear the address bar */
  if (!tab || !glook_tab_is_active (GLOOK_TAB (tab)))
    {
      gtk_entry_set_text (addrbar, "");
      return;
    }
  /* switch between tab uri and empty. */
  if (gtk_entry_get_text_length (addrbar) > 0)
    {
      gtk_entry_set_text (addrbar, "");
    }
  else
    {
      /* set address bar to tab uri and select it. */
      uri = glook_tab_get_uri (GLOOK_TAB (tab));
      if (uri)
	{
	  gchar *uristr = glook_uri_to_string (uri);
	  gtk_entry_set_text (addrbar, uristr);
	  g_free (uristr);
	}
      gtk_editable_select_region (GTK_EDITABLE (addrbar), 0, -1);
    }
}

/**
 * on_window_addrbar_go:
 *
 * this function will be called when address bar get
 * activated(press the ENTER key or click the right go icon).
 */
void
on_window_addrbar_go (GtkEntry     *addrbar,
		      const gchar  *address,
		      GLookWindow  *window)
{
  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  /* get the address and open it */

  glook_window_open_uri_string (window, address);
}

/**
 * on_window_addrbar_stop:
 */
void
on_window_addrbar_stop (GtkEntry    *addrbar,
			GLookWindow *window)
{
  GtkWidget *notebook;
  GtkWidget *tab;

  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  /* cancel current tab operation. */
  notebook = glook_window_get_notebook (GLOOK_WINDOW (window));
  tab = glook_notebook_get_current (GLOOK_NOTEBOOK (notebook));
  if (tab && GLOOK_IS_TAB (tab))
    glook_tab_stop (GLOOK_TAB (tab));
}

/**
 * on_add_bookmark_dialog_responsed:
 *
 * signal handler from add bookmark ui file.
 */
void
on_add_bookmark_dialog_responsed (GtkWidget   *dialog,
				  gint         respid,
				  GLookWindow *window)
{
  GtkBuilder *builder;
  GObject *object;
  const gchar *address, *title, *desc;

  g_return_if_fail (GTK_IS_DIALOG (dialog));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (dialog), "builder");

  /* get bookmark address. */
  object = gtk_builder_get_object (builder, "BookmarkAddressLabel");
  address = gtk_label_get_text (GTK_LABEL (object));

  /* get bookmark title. */
  object = gtk_builder_get_object (builder, "BookmarkTitleEntry");
  title = gtk_entry_get_text (GTK_ENTRY (object));

  /* get bookmark description. */
  object = gtk_builder_get_object (builder, "BookmarkDescEntry");
  desc = gtk_entry_get_text (GTK_ENTRY (object));

  glook_bookmark_add (glook_app_get_bookmark (glook_app_get ()),
		      address, title, desc);

  /* destroy the dialog. */
  g_object_set_data (G_OBJECT (window), "add-bookmark-dialog", NULL);
}

/**
 * on_manage_bookmark_label_activate_link:
 *
 * signal handler from add bookmark ui file.
 * open bookmark manager.
 */
gboolean
on_manage_bookmark_label_activate_link (GtkWidget   *label,
					const gchar *uri,
					GLookWindow *window)
{
  const gchar *address = "local:path=bookmark.xml";
  GtkWidget *dialog;

  g_return_val_if_fail (GLOOK_IS_WINDOW (window), FALSE);

  /* open bookmark manager. */
  glook_window_open_uri_string (window, address);

  /* close the dialog. */
  dialog = gtk_widget_get_toplevel (label);
  gtk_dialog_response (GTK_DIALOG (dialog), GTK_RESPONSE_CLOSE);

  return TRUE;
}

/**
 * do_add_bookmark_dialog:
 */
static void
do_add_bookmark_dialog (GLookWindow  *window,
			GLookAddrbar *addrbar,
			const gchar  *title)
{
  GtkBuilder *builder;
  GObject *dialog;
  GObject *object;
  const gchar *text;
  GError *error = NULL;

  g_return_if_fail (GLOOK_IS_WINDOW (window));
  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  /* we just want has a add bookmark dialog for a window. */
  dialog = g_object_get_data (G_OBJECT (window), "add-bookmark-dialog");
  if (dialog)
    {
      gtk_window_present (GTK_WINDOW (dialog));
      return;
    }
  /* load ui from glade file. */
  builder = gtk_builder_new ();
  gtk_builder_add_from_file (builder, ADD_BOOKMARK_UI_FILE, &error);
  if (error)
    {
      g_critical (_("Failed to load ui file `%s': %s"),
		  ADD_BOOKMARK_UI_FILE, error->message);
      g_error_free (error);
      return;
    }
  /* connect object signals. */
  gtk_builder_connect_signals (builder, window);

  /* set address label from address bar. */
  object = gtk_builder_get_object (builder, "BookmarkAddressLabel");
  text = gtk_entry_get_text (GTK_ENTRY (addrbar));
  gtk_label_set_text (GTK_LABEL (object), text);

  /* set title from tab label. */
  object = gtk_builder_get_object (builder, "BookmarkTitleEntry");
  gtk_entry_set_text (GTK_ENTRY (object), title);

  /* when dialog close, the builder will be free. */
  dialog = gtk_builder_get_object (builder, "AddBookmarkDialog");
  g_object_set_data_full (dialog, "builder", builder, g_object_unref);

  /* to prevent multiple add bookmark dialog. */
  g_object_set_data_full (G_OBJECT (window), "add-bookmark-dialog",
			  dialog, (GDestroyNotify) gtk_widget_destroy);
  /* show the dialog. */
  gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
  gtk_widget_show (GTK_WIDGET (dialog));
}

/**
 * on_window_addrbar_bookmark:
 *
 * @button: mouse button, 1 for left button, 3 for right button.
 *
 * when button == 3, we popup a dialog.
 */
void
on_window_addrbar_bookmark (GtkEntry    *addrbar,
			    gint         button,
			    GLookWindow *window)
{
  GtkWidget *notebook;
  GtkWidget *tab;
  const gchar *text, *title;

  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  notebook = glook_window_get_notebook (window);
  tab = glook_notebook_get_current (GLOOK_NOTEBOOK (notebook));
  title = glook_tab_get_label_text (GLOOK_TAB (tab));

  switch (button)
    {
    case 1:	/* left button */
      /* add bookmark without user interaction. */
      text = gtk_entry_get_text (GTK_ENTRY (addrbar));
      glook_bookmark_add (glook_app_get_bookmark (glook_app_get ()),
			  text, title, NULL);
      break;
    case 3:	/* right button */
      do_add_bookmark_dialog (window, GLOOK_ADDRBAR (addrbar), title);
      break;
    }
}

/**
 * glook_window_close_idle:
 */
static gboolean
glook_window_close_idle (gpointer data)
{
  if (GLOOK_IS_WINDOW (data))
    {
      glook_window_delete (GLOOK_WINDOW (data));
      gtk_widget_destroy (GTK_WIDGET (data));
    }
  return FALSE;
}

/**
 * on_notebook_tab_progress:
 */
static void
on_notebook_tab_progress (GLookTab    *tab,
			  GLookWindow *window)
{
  GtkWidget *notebook;

  g_return_if_fail (GLOOK_IS_TAB (tab));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  notebook = glook_window_get_notebook (window);

  /* only update the address bar if tab is current actived */
  if (glook_notebook_is_current (GLOOK_NOTEBOOK (notebook),
				 GTK_WIDGET (tab)))
    {
      GtkWidget *addrbar;

      addrbar = glook_window_get_addrbar (GLOOK_WINDOW (window));

      /* update address bar progress status */
      glook_addrbar_set_progress (GLOOK_ADDRBAR (addrbar),
				  glook_tab_get_progress (tab));
    }
}

/**
 * on_notebook_tab_finished:
 */
static void
on_notebook_tab_finished (GLookTab    *tab,
			  GLookWindow *window)
{
  GtkWidget *notebook;

  g_return_if_fail (GLOOK_IS_TAB (tab));

  if (!GLOOK_IS_WINDOW (window))
    return;

  notebook = glook_window_get_notebook (window);

  /* only update the address bar if tab is current actived */
  if (glook_notebook_is_current (GLOOK_NOTEBOOK (notebook),
				 GTK_WIDGET (tab)))
    {
      GtkWidget *addrbar;
      gchar title[512];

      addrbar = glook_window_get_addrbar (GLOOK_WINDOW (window));

      /* update address bar progress status */
      glook_addrbar_set_progress (GLOOK_ADDRBAR (addrbar),
				  glook_tab_get_progress (tab));

      /* update window title */
      g_snprintf (title, sizeof (title), "%s - Ttyutils GLook %s",
		  glook_tab_get_label_text (tab), VERSION);
      gtk_window_set_title (GTK_WINDOW (window), title);
    }
}

/**
 * on_window_notebook_page_added:
 */
void
on_window_notebook_page_added (GtkNotebook *notebook,
			       GtkWidget   *child,
			       guint        page_num,
			       GLookWindow *window)
{
  GtkWidget *tab;

  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  tab = gtk_notebook_get_nth_page (notebook, page_num);

  /* connect to these signals to update address progress state */
  g_signal_connect (tab, "progress",
		    G_CALLBACK (on_notebook_tab_progress), window);
  g_signal_connect (tab, "finished",
		    G_CALLBACK (on_notebook_tab_finished), window);
}

/**
 * on_window_notebook_page_removed:
 */
void
on_window_notebook_page_removed (GtkNotebook *notebook,
				 GtkWidget   *child,
				 guint        page_num,
				 GLookWindow *window)
{
  static guint resize_id = 0;
  gint n, i;
  GtkWidget *tab;

  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  /* remove previous resize event first. */
  if (resize_id != 0)
    {
      g_source_remove (resize_id);
      resize_id = 0;
    }
  /* if there no more tabs, destroy window */
  n = gtk_notebook_get_n_pages (notebook);
  if (n == 0)
    {
      g_idle_add_full (G_PRIORITY_HIGH_IDLE,
		       glook_window_close_idle, window, NULL);
      return;
    }
  /* resize window. */
  for (i = 0; i < n; i++)
    {
      tab = gtk_notebook_get_nth_page (notebook, i);
      if (GLOOK_IS_LOOK_TAB (tab) || GLOOK_IS_PLAY_TAB (tab))
	{
	  resize_id = g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
				       resize_window_idle,
				       g_object_ref (window),
				       (GDestroyNotify) g_object_unref);
	  break;
	}
    }
}

/**
 * update_action_sensitive_by_tab:
 *
 * not all tab support all actions appear in main menu or
 * toolbar. when tab been switched, we need update the action
 * sensitive in order to give user hint.
 */
static void
update_action_sensitive_by_tab (GLookWindow *window,
				GLookTab    *tab)
{
  GLookUi *uiman;

  g_return_if_fail (GLOOK_IS_WINDOW (window));
  g_return_if_fail (GLOOK_IS_TAB (tab));

  uiman = GLOOK_UI (glook_window_get_uiman (window));

  if (GLOOK_IS_UI (uiman))
    {
      GtkAction *action;
      GtkWidget *combo;
      gboolean sensitive;

      /* reload action */
      sensitive = GLOOK_TAB_REFRESH_SUPPORT (tab);
      action = glook_ui_get_window_action (uiman, "Reload");
      gtk_action_set_sensitive (action, sensitive);

      /* zoom action and toolbar zoom combo */
      sensitive = GLOOK_TAB_ZOOM_SUPPORT (tab);
      action = glook_ui_get_window_action (uiman, "ZoomMenu");
      gtk_action_set_sensitive (action, sensitive);
      action = glook_ui_get_window_action (uiman, "ZoomIn");
      gtk_action_set_sensitive (action, sensitive);
      action = glook_ui_get_window_action (uiman, "ZoomOut");
      gtk_action_set_sensitive (action, sensitive);
      combo = glook_window_get_zoom_combo (window);
      gtk_widget_set_sensitive (combo, sensitive);
    }
}

/**
 * on_window_notebook_page_switched:
 *
 * will be called when notebook start switch page. the @page_num
 * reference the new current page, and the @page still is current
 * page(will become non-current after this function return).
 */
void
on_window_notebook_page_switched (GtkNotebook     *notebook,
				  GtkNotebookPage *page,
				  guint            page_num,
				  GLookWindow     *window)
{
  GtkWidget *tab, *oldtab;

  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  /* the current notebook page stay not changed. but the
   * @page_num ref the new page no */
  oldtab = glook_notebook_get_current (GLOOK_NOTEBOOK (notebook));
  tab = gtk_notebook_get_nth_page (notebook, page_num);
  if (tab)
    {
      gchar title[256];
      GLookUri *uri;
      gchar *uristr;
      GtkWidget *toolbar;
      GtkWidget *addrbar;
      GtkWidget *combo;

      /* change window title */
      g_snprintf (title, sizeof (title), "%s - Ttyutils GLook %s",
		  glook_tab_get_label_text (GLOOK_TAB (tab)), VERSION);
      gtk_window_set_title (GTK_WINDOW (window), title);

      /* update address bar progress status. */
      addrbar = glook_window_get_addrbar (window);
      glook_addrbar_set_progress (GLOOK_ADDRBAR (addrbar),
				  glook_tab_get_progress (GLOOK_TAB (tab)));
      /* update address bar uri. */
      if (!GLOOK_IS_BLANK_TAB (tab))
	{
	  uri = glook_tab_get_uri (GLOOK_TAB (tab));
	  uristr = glook_uri_to_string (uri);
	  gtk_entry_set_text (GTK_ENTRY (addrbar), uristr);
	  g_free (uristr);
	}
      else
	gtk_entry_set_text (GTK_ENTRY (addrbar), "");

      /* update window geometry hint. */
      glook_tab_geometry_hint (GLOOK_TAB (tab));

      toolbar = glook_window_get_toolbar (window);

      /* update zoom combo index.
       * zoom combo change will cause toolbar's zoom signal
       * been emit, here we has not actual finish the page switch.
       * so we block the handler. */
      g_signal_handlers_block_by_func (toolbar,
				       on_window_toolbar_zoom, window);

      combo = glook_window_get_zoom_combo (window);
      gtk_combo_box_set_active (GTK_COMBO_BOX (combo),
				glook_tab_get_zoom_index (GLOOK_TAB (tab)));

      g_signal_handlers_unblock_by_func (toolbar,
					 on_window_toolbar_zoom, window);
      /* update actions sensitive */
      update_action_sensitive_by_tab (window, GLOOK_TAB (tab));

      /* unmerge old tab's ui, and merge new tab's ui. */
      glook_tab_unmerge_ui (GLOOK_TAB (oldtab), GTK_WIDGET (window));
      glook_tab_merge_ui (GLOOK_TAB (tab), GTK_WIDGET (window));
    }
}
