/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-uri.h"
#include "glook-bookmark.h"

enum
{
  PROP_0,
  PROP_PATH,
};

struct _GLookBookmarkPriv
{
  gchar         *path;
  GBookmarkFile *file;
};
#define _BOOKMARK_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_BOOKMARK, GLookBookmarkPriv))

G_DEFINE_TYPE (GLookBookmark, glook_bookmark, G_TYPE_OBJECT);

/**
 * glook_bookmark_init:
 */
static void
glook_bookmark_init (GLookBookmark *bookmark)
{
  bookmark->priv = _BOOKMARK_GET_PRIVATE (bookmark);

  bookmark->priv->path = NULL;
  bookmark->priv->file = NULL;
}

/**
 * glook_bookmark_dispose:
 */
static void
glook_bookmark_dispose (GObject *object)
{
#if 0
  GLookBookmark *bookmark = GLOOK_BOOKMARK (object);

#endif
  G_OBJECT_CLASS (glook_bookmark_parent_class)->dispose (object);
}

/**
 * glook_bookmark_finalize:
 */
static void
glook_bookmark_finalize (GObject *object)
{
  GLookBookmark *bookmark = GLOOK_BOOKMARK (object);

  if (bookmark->priv->file)
    {
      g_bookmark_file_free (bookmark->priv->file);
      bookmark->priv->file = NULL;
    }
  if (bookmark->priv->path)
    {
      g_free (bookmark->priv->path);
      bookmark->priv->path = NULL;
    }
  G_OBJECT_CLASS (glook_bookmark_parent_class)->finalize (object);
}

/**
 * glook_bookmark_get_property:
 */
static void
glook_bookmark_get_property (GObject    *object,
			     guint       property_id,
			     GValue     *value,
			     GParamSpec *pspec)
{
  GLookBookmark *bookmark = GLOOK_BOOKMARK (object);

  switch (property_id)
    {
    case PROP_PATH:
      g_value_set_string (value, glook_bookmark_get_path (bookmark));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_bookmark_set_property:
 */
static void
glook_bookmark_set_property (GObject      *object,
			     guint         property_id,
			     const GValue *value,
			     GParamSpec   *pspec)
{
  GLookBookmark *bookmark = GLOOK_BOOKMARK (object);

  switch (property_id)
    {
    case PROP_PATH:
      glook_bookmark_set_path (bookmark, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_bookmark_class_init:
 */
static void
glook_bookmark_class_init (GLookBookmarkClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookBookmarkPriv));

  gobject_class->get_property = glook_bookmark_get_property;
  gobject_class->set_property = glook_bookmark_set_property;
  gobject_class->dispose  = glook_bookmark_dispose;
  gobject_class->finalize = glook_bookmark_finalize;

  g_object_class_install_property (gobject_class, PROP_PATH,
				   g_param_spec_string ("path", "path", "Bookmark file path", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_signal_new ("added", GLOOK_TYPE_BOOKMARK, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__STRING,
		G_TYPE_NONE, 1, G_TYPE_STRING);
  g_signal_new ("removed", GLOOK_TYPE_BOOKMARK, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__STRING,
		G_TYPE_NONE, 1, G_TYPE_STRING);
  g_signal_new ("changed", GLOOK_TYPE_BOOKMARK, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__STRING,
		G_TYPE_NONE, 1, G_TYPE_STRING);
}

/**
 * glook_bookmark_new:
 */
GLookBookmark*
glook_bookmark_new (void)
{
  return g_object_new (GLOOK_TYPE_BOOKMARK, NULL);
}

/**
 * glook_bookmark_get_file:
 *
 * we expose the bookmark file object. application should only
 * use this object to access and modify uri related info such as 
 * title. use GLookBbookmark object add and remove method to add
 * and remove uri to emit signal.
 */
GBookmarkFile*
glook_bookmark_get_file (GLookBookmark *bookmark)
{
  g_return_val_if_fail (GLOOK_IS_BOOKMARK (bookmark), NULL);
  return bookmark->priv->file;
}

/**
 * glook_bookmark_get_path:
 */
const gchar*
glook_bookmark_get_path (GLookBookmark *bookmark)
{
  g_return_val_if_fail (GLOOK_IS_BOOKMARK (bookmark), NULL);
  return bookmark->priv->path;
}

/**
 * glook_bookmark_set_path:
 */
void
glook_bookmark_set_path (GLookBookmark *bookmark,
			 const gchar   *path)
{
  GBookmarkFile *file;
  GError *error = NULL;

  g_return_if_fail (GLOOK_IS_BOOKMARK (bookmark));
  g_return_if_fail (path);

  /* create a new bookmark file object. */
  file = g_bookmark_file_new ();
  g_bookmark_file_load_from_file (file, path, &error);
  if (error)
    {
      if (error->code != G_FILE_ERROR_NOENT)
	{
	  g_warning (_("Failed to load bookmark file `%s': %s"),
		     path, error->message);
	}
      g_error_free (error);
    }
  /* set bookmark file title. */
  g_bookmark_file_set_title (file, NULL, _("Ttyutils bookmark file"));

  /* save bookmark file object and path. */
  if (bookmark->priv->file)
    g_bookmark_file_free (bookmark->priv->file);

  bookmark->priv->file = file;

  g_free (bookmark->priv->path);
  bookmark->priv->path = g_strdup (path);

  g_object_notify (G_OBJECT (bookmark), "path");
}

/**
 * glook_bookmark_lookup_icon:
 */
static const gchar*
glook_bookmark_lookup_icon (GtkIconTheme *theme,
			    const gchar  *icon_name)
{
  GList *windows;
  GtkTextDirection direction = GTK_TEXT_DIR_LTR;
  static gchar name[64];

  windows = gtk_window_list_toplevels ();
  if (windows && GTK_IS_WIDGET (windows->data))
    {
      direction = gtk_widget_get_direction (GTK_WIDGET (windows->data));
      g_list_free (windows);
    }
  switch (direction)
    {
    case GTK_TEXT_DIR_LTR:
      g_snprintf (name, sizeof (name), "%s-ltr", icon_name);
      break;
    case GTK_TEXT_DIR_RTL:
      g_snprintf (name, sizeof (name), "%s-rtl", icon_name);
      break;
    default:
      break;
    }
  if (gtk_icon_theme_has_icon (theme, name))
    return name;

  g_message (_("Icon `%s' not found"), icon_name);
  return GTK_STOCK_MISSING_IMAGE;
}

/**
 * glook_bookmark_add:
 *
 * add new bookmark.
 */
gboolean
glook_bookmark_add (GLookBookmark *bookmark,
		    const gchar   *address,
		    const gchar   *title,
		    const gchar   *desc)
{
  GLookUri *uri;
  const gchar *icon_name;
  GtkIconTheme *theme;
  const gchar *group;

  g_return_val_if_fail (GLOOK_IS_BOOKMARK (bookmark), FALSE);
  g_return_val_if_fail (address, FALSE);
  g_return_val_if_fail (bookmark->priv->file, FALSE);

  uri = glook_uri_new_from_string (address);
  if (!uri)
    return FALSE;

  g_bookmark_file_set_title (bookmark->priv->file, address,
			     title ? title : _("Untitle"));
  g_bookmark_file_set_description (bookmark->priv->file, address,
				   desc ? desc : "");
  /* mime type and icon. */
  switch (glook_uri_get_scheme_type (uri))
    {
    case GLOOK_URI_SCHEME_LOOK:
      icon_name = "application-x-ttylook";
      group = "Look";
      break;
    case GLOOK_URI_SCHEME_LOCAL:
      icon_name = "ttyutils-local";
      group = _("Local");
      break;
    case GLOOK_URI_SCHEME_AGENT:
      icon_name = "application-x-ttylook";
      group = _("Sessions");
      break;
    case GLOOK_URI_SCHEME_PLAY:
      icon_name = GTK_STOCK_MEDIA_PLAY;
      group = _("Playback");
      break;
    default:
      icon_name = GTK_STOCK_FILE;
      group = _("Other");
      break;
    }
  glook_uri_free (uri);

  /* if icon name not found, we try to use -ltr/-rtl instead. */
  theme = gtk_icon_theme_get_default ();
  if (!gtk_icon_theme_has_icon (theme, icon_name))
    {
      icon_name = glook_bookmark_lookup_icon (theme, icon_name);
    }
  g_bookmark_file_set_icon (bookmark->priv->file, address,
			    icon_name, "application/x-ttylook");
  g_bookmark_file_set_mime_type (bookmark->priv->file, address,
				 "application/x-ttylook");
  /* timestamps */
  g_bookmark_file_set_added (bookmark->priv->file, address, time (NULL));
  g_bookmark_file_set_modified (bookmark->priv->file, address, time (NULL));
  g_bookmark_file_set_visited (bookmark->priv->file, address, time (NULL));

  /* pass NULL here to get the application name and exec from
   * g_get_appliation_name() and g_get_prgname() */
  g_bookmark_file_add_application (bookmark->priv->file, address, NULL, NULL);

  g_bookmark_file_add_group (bookmark->priv->file, address, group);
  g_bookmark_file_set_is_private (bookmark->priv->file, address, TRUE);

  g_signal_emit_by_name (bookmark, "added", address);
  return TRUE;
}

/**
 * glook_bookmark_remove:
 */
gboolean
glook_bookmark_remove (GLookBookmark *bookmark,
		       const gchar   *address)
{
  GError *error = NULL;

  g_return_val_if_fail (GLOOK_IS_BOOKMARK (bookmark), FALSE);
  g_return_val_if_fail (address, FALSE);
  g_return_val_if_fail (bookmark->priv->file, FALSE);

  g_bookmark_file_remove_item (bookmark->priv->file, address, &error);
  if (error)
    {
      g_message (_("Failed to remove bookmark `%s': %s"),
		 address, error->message);
      g_error_free (error);
      return FALSE;
    }
  g_signal_emit_by_name (bookmark, "removed", address);
  return TRUE;
}

/**
 * glook_bookmark_has:
 */
gboolean
glook_bookmark_has (GLookBookmark *bookmark,
		    const gchar   *address)
{
  g_return_val_if_fail (GLOOK_IS_BOOKMARK (bookmark), FALSE);
  g_return_val_if_fail (address, FALSE);
  g_return_val_if_fail (bookmark->priv->file, FALSE);

  return g_bookmark_file_has_item (bookmark->priv->file, address);
}

/**
 * glook_bookmark_update_visited:
 *
 * update bookmark visited time.
 */
void
glook_bookmark_update_visited (GLookBookmark *bookmark,
			       const gchar   *address)
{
  g_return_if_fail (GLOOK_IS_BOOKMARK (bookmark));
  g_return_if_fail (address);
  g_return_if_fail (bookmark->priv->file);

  if (g_bookmark_file_has_item (bookmark->priv->file, address))
    {
      g_bookmark_file_set_visited (bookmark->priv->file, address, time (NULL));
      g_signal_emit_by_name (bookmark, "changed", address);
    }
}

/**
 * glook_bookmark_save_to_file:
 *
 * backup bookmark file.
 */
gboolean
glook_bookmark_save_to_file (GLookBookmark *bookmark,
			     const gchar   *pathname)
{
  GError *error = NULL;

  g_return_val_if_fail (GLOOK_IS_BOOKMARK (bookmark), FALSE);
  g_return_val_if_fail (bookmark->priv->file, FALSE);

  if (!pathname)
    pathname = bookmark->priv->path;

  g_bookmark_file_to_file (bookmark->priv->file, pathname, &error);
  if (error)
    {
      g_warning (_("Unable to save bookmark to `%s': %s"),
		 pathname, error->message);
      g_error_free (error);
      return FALSE;
    }
  return TRUE;
}

/**
 * glook_bookmark_add_address:
 */
static void
glook_bookmark_add_address (GLookBookmark  *bookmark,
			    GBookmarkFile  *file,
			    const gchar    *address,
			    GLookUri       *uri)
{
  gchar *title, *desc, *icon_name;

  title = g_bookmark_file_get_title (file, address, NULL);
  if (title)
    {
      g_bookmark_file_set_title (bookmark->priv->file, address, title);
      g_free (title);
    }
  desc = g_bookmark_file_get_description (file, address, NULL);
  if (desc)
    {
      g_bookmark_file_set_description (bookmark->priv->file, address, desc);
      g_free (desc);
    }
  g_bookmark_file_set_added (bookmark->priv->file, address,
			     g_bookmark_file_get_added (file, address, NULL));
  g_bookmark_file_set_modified (bookmark->priv->file, address,
				g_bookmark_file_get_modified (file, address, NULL));
  g_bookmark_file_set_visited (bookmark->priv->file, address,
			       g_bookmark_file_get_visited (file, address, NULL));
  /* mime type and icon. */
  switch (glook_uri_get_scheme_type (uri))
    {
    case GLOOK_URI_SCHEME_LOOK:
      icon_name = "application-x-ttylook";
      break;
    case GLOOK_URI_SCHEME_LOCAL:
      icon_name = "ttyutils-local";
      break;
    case GLOOK_URI_SCHEME_AGENT:
      icon_name = "ttyutils-ttyagent";
      break;
    default:
      icon_name = GTK_STOCK_FILE;
      break;
    }
  g_bookmark_file_set_icon (bookmark->priv->file, address,
			    icon_name, "application/x-ttylook");
  g_bookmark_file_set_mime_type (bookmark->priv->file, address,
				 "application/x-ttylook");
  /* applicatoin and group info. */
  g_bookmark_file_add_application (bookmark->priv->file, address, NULL, NULL);
  g_bookmark_file_add_group (bookmark->priv->file, address, "Ttyutils");
  g_bookmark_file_set_is_private (bookmark->priv->file, address, TRUE);

  g_signal_emit_by_name (bookmark, "added", address);
}

/**
 * glook_bookmark_add_from_file:
 */
gboolean
glook_bookmark_add_from_file (GLookBookmark *bookmark,
			      const gchar   *pathname)
{
  GBookmarkFile *file;
  gchar **uris;
  gsize length, i;
  GError *error = NULL;

  g_return_val_if_fail (GLOOK_IS_BOOKMARK (bookmark), FALSE);
  g_return_val_if_fail (pathname, FALSE);
  g_return_val_if_fail (bookmark->priv->file, FALSE);

  file = g_bookmark_file_new ();
  g_bookmark_file_load_from_file (file, pathname, &error);
  if (error)
    {
      g_warning ("%s", error->message);
      g_error_free (error);
      return FALSE;
    }
  uris = g_bookmark_file_get_uris (file, &length); 
  for (i = 0; i < length; i++)
    {
      GLookUri *uri;

      /* skip exists address. */
      if (g_bookmark_file_has_item (bookmark->priv->file, uris[i]))
	{
	  g_message (_("Bookmark `%s' exists, skip."), uris[i]);
	  continue;
	}
      uri = glook_uri_new_from_string (uris[i]);
      if (uri)
	{
	  glook_bookmark_add_address (bookmark, file, uris[i], uri);
	  glook_uri_free (uri);
	}
    }
  g_strfreev (uris);

  return TRUE;
}
