/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-local-tab.h"
#include "glook-local-log.h"

/* toplevel widget name. */
#define TOPLEVEL_WIDGET_NAME    "LocalToplevelWidget"

/* we will try to find the glade file in this directory. */
#ifdef DEBUG
# define LOCAL_GLADE_PATH	SRCDIR G_DIR_SEPARATOR_S "glade" G_DIR_SEPARATOR_S "local"
#else
# define LOCAL_GLADE_PATH	GLADEDIR G_DIR_SEPARATOR_S "local"
#endif

typedef enum
{
  LOCAL_BUILTIN_INVALID,
  LOCAL_BUILTIN_PREFERENCE,
  LOCAL_BUILTIN_LOG,
  LOCAL_BUILTIN_BOOKMARK,
} GLookLocalBuiltin;

typedef struct _GLookLocalInternal	GLookLocalInternal;
struct _GLookLocalInternal
{
  GLookLocalBuiltin  builtin;
  const gchar       *pathname;
};

struct _GLookLocalTabPriv
{
  GtkWidget         *info_bar;
  GtkWidget         *info_image;
  GtkWidget         *info_close_button;
  GtkWidget         *info_yes_button;
  GtkWidget         *info_label;
  gchar             *pathname;
  GtkBuilder        *builder;
  GtkWidget         *toplevel;
  GLookLocalBuiltin  builtin;
};
#define _LOCAL_TAB_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_LOCAL_TAB, GLookLocalTabPriv))

G_DEFINE_TYPE (GLookLocalTab, glook_local_tab, GLOOK_TYPE_TAB);

/* all builtin files. */
static GLookLocalInternal	_internal_files[] =
{
    { LOCAL_BUILTIN_PREFERENCE,	LOCAL_GLADE_PATH G_DIR_SEPARATOR_S "preference.xml", },
    { LOCAL_BUILTIN_LOG,	LOCAL_GLADE_PATH G_DIR_SEPARATOR_S "log.xml", },
    { LOCAL_BUILTIN_BOOKMARK,	LOCAL_GLADE_PATH G_DIR_SEPARATOR_S "bookmark.xml", },
};

static void	glook_local_tab_open		(GLookTab      *tab);
static void	glook_local_tab_stop		(GLookTab      *tab);
static void	glook_local_tab_reload		(GLookTab      *tab);
static gboolean	glook_local_tab_zoom		(GLookTab      *tab,
						 gint           factor);
static gboolean	glook_local_tab_is_active	(GLookTab      *tab);

static void	on_info_bar_responsed		(GtkInfoBar    *info_bar,
						 gint           resp_id,
						 GLookLocalTab *tab);

/**
 * glook_local_tab_init:
 */
static void
glook_local_tab_init (GLookLocalTab *tab)
{
  GtkWidget *content_area;

  tab->priv = _LOCAL_TAB_GET_PRIVATE (tab);

  g_object_set (tab, "image", "ttyutils-local", "label", "Untitle", NULL);

  tab->priv->pathname = NULL;
  tab->priv->builder  = NULL;
  tab->priv->toplevel = NULL;
  tab->priv->builtin  = LOCAL_BUILTIN_INVALID;

  /* create the info bar */
  tab->priv->info_bar = gtk_info_bar_new ();
  gtk_widget_hide (tab->priv->info_bar);

  /* info bar action buttons */
  tab->priv->info_close_button = gtk_button_new_with_mnemonic (_("_Close tab"));
  gtk_widget_show (tab->priv->info_close_button);
  gtk_info_bar_add_action_widget (GTK_INFO_BAR (tab->priv->info_bar),
				  tab->priv->info_close_button,
				  GTK_RESPONSE_CLOSE);
  tab->priv->info_yes_button = gtk_button_new_from_stock (GTK_STOCK_YES);
  gtk_info_bar_add_action_widget (GTK_INFO_BAR (tab->priv->info_bar),
				  tab->priv->info_yes_button,
				  GTK_RESPONSE_YES);
  /* info bar label. */
  tab->priv->info_label = gtk_label_new (NULL);
  gtk_misc_set_alignment (GTK_MISC (tab->priv->info_label), 0.0, 0.5);
  g_object_set (tab->priv->info_label, "wrap", TRUE,
		"wrap-mode", PANGO_WRAP_WORD_CHAR, NULL);
  gtk_widget_show (tab->priv->info_label);

  tab->priv->info_image = gtk_image_new_from_stock (GTK_STOCK_DIALOG_QUESTION,
						    GTK_ICON_SIZE_DIALOG);

  content_area = gtk_info_bar_get_content_area (GTK_INFO_BAR (tab->priv->info_bar));
  gtk_box_pack_start (GTK_BOX (content_area), tab->priv->info_image, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (content_area), tab->priv->info_label, TRUE, TRUE, 0);

  gtk_box_pack_start (GTK_BOX (tab), GTK_WIDGET (tab->priv->info_bar), FALSE, TRUE, 0);

  g_signal_connect (tab->priv->info_bar, "close",
		    G_CALLBACK (gtk_widget_hide), tab);
  g_signal_connect (tab->priv->info_bar, "response",
		    G_CALLBACK (on_info_bar_responsed), tab);
}

/**
 * glook_local_tab_dispose:
 */
static void
glook_local_tab_dispose (GObject *object)
{
  GLookLocalTab *tab = GLOOK_LOCAL_TAB (object);

  if (tab->priv->builder)
    {
      g_object_unref (tab->priv->builder);
      tab->priv->builder = NULL;
    }
  G_OBJECT_CLASS (glook_local_tab_parent_class)->dispose (object);
}

/**
 * glook_local_tab_finalize:
 */
static void
glook_local_tab_finalize (GObject *object)
{
  GLookLocalTab *tab = GLOOK_LOCAL_TAB (object);

  if (tab->priv->pathname)
    {
      g_free (tab->priv->pathname);
      tab->priv->pathname = NULL;
    }
  G_OBJECT_CLASS (glook_local_tab_parent_class)->finalize (object);
}

/**
 * glook_local_tab_class_init:
 */
static void
glook_local_tab_class_init (GLookLocalTabClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GLookTabClass *parent_class = GLOOK_TAB_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookLocalTabPriv));

  gobject_class->dispose  = glook_local_tab_dispose;
  gobject_class->finalize = glook_local_tab_finalize;

  parent_class->open      = glook_local_tab_open;
  parent_class->stop      = glook_local_tab_stop;
  parent_class->reload    = glook_local_tab_reload;
  parent_class->zoom      = glook_local_tab_zoom;
  parent_class->is_active = glook_local_tab_is_active;
}

/**
 * on_info_bar_responsed:
 */
static void
on_info_bar_responsed (GtkInfoBar    *info_bar,
		       gint           resp_id,
		       GLookLocalTab *tab)
{
  g_return_if_fail (GTK_IS_INFO_BAR (info_bar));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  if (resp_id == GTK_RESPONSE_CLOSE)
    glook_tab_close (GLOOK_TAB (tab));
}

/**
 * glook_local_tab_show_info:
 *
 * show message in info bar. if message type is question,
 * we will show a yes button and a close button. otherwise
 * only show the close button.
 */
static void
glook_local_tab_show_info (GLookLocalTab  *tab,
			   GtkMessageType  type,
			   const gchar    *format,
			   ...)
{
  va_list args;
  gchar buff[1024];

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  va_start (args, format);
  g_vsnprintf (buff, sizeof (buff), format, args);
  va_end (args);

  gtk_label_set_label (GTK_LABEL (tab->priv->info_label), buff);
  gtk_info_bar_set_message_type (GTK_INFO_BAR (tab->priv->info_bar), type);

  if (type == GTK_MESSAGE_QUESTION)
    {
      gtk_widget_show (tab->priv->info_image);
      gtk_widget_show (tab->priv->info_yes_button);
    }
  else
    {
      gtk_widget_hide (tab->priv->info_image);
      gtk_widget_hide (tab->priv->info_yes_button);
    }
  gtk_widget_show (tab->priv->info_bar);
  gtk_widget_grab_focus (tab->priv->info_close_button);
}

/**
 * glook_local_tab_open_file:
 *
 * open glade file.
 */
static void
glook_local_tab_open_file (GLookLocalTab *tab)
{
  const gchar *title;
  GError *error = NULL;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* remove child widget from box first. */
  if (tab->priv->toplevel)
    {
      gtk_container_remove (GTK_CONTAINER (tab), tab->priv->toplevel);
      tab->priv->toplevel = NULL;
    }
  if (tab->priv->builder)
    {
      g_object_unref (tab->priv->builder);
      tab->priv->builder = NULL;
    }
  tab->priv->builder = gtk_builder_new ();

  /* load ui from glade file. */
  gtk_builder_add_from_file (tab->priv->builder, tab->priv->pathname, &error);
  if (error)
    {
      glook_local_tab_show_info (tab, GTK_MESSAGE_ERROR,
				 _("File `%s' format error."), tab->priv->pathname);
      g_warning (_("Unable to load glade file `%s': %s"),
		 tab->priv->pathname, error->message);
      g_error_free (error);

      g_object_unref (tab->priv->builder);
      tab->priv->builder = NULL;
      return;
    }
  gtk_builder_connect_signals (tab->priv->builder, tab);

  /* ui must has a toplevel widget named TOPLEVE_WIDGET_NAME.
   * here get toplevel widget and packing it into tab. */
  tab->priv->toplevel = GTK_WIDGET (gtk_builder_get_object (tab->priv->builder,
							    TOPLEVEL_WIDGET_NAME));
  if (!GTK_IS_WIDGET (tab->priv->toplevel) ||
      gtk_widget_get_parent (tab->priv->toplevel))
    {
      glook_local_tab_show_info (tab, GTK_MESSAGE_ERROR,
				 _("File `%s' format error."),
				 tab->priv->pathname);
      g_warning (_("File `%s' missing toplevel widget `%s' or toplevel "
		   "widget already inside other container."),
		 tab->priv->pathname, TOPLEVEL_WIDGET_NAME);

      g_object_unref (tab->priv->builder);
      tab->priv->builder = NULL;
      tab->priv->toplevel = NULL;
      return;
    }
  /* use toplevel widget `name' property as tab label. */
  title = gtk_widget_get_name (tab->priv->toplevel);
  glook_tab_set_label (GLOOK_TAB (tab), title ? title : tab->priv->pathname);

  /* add toplevel widget to tab. */
  gtk_box_pack_end (GTK_BOX (tab), tab->priv->toplevel, TRUE, TRUE, 0);

  /* notify external we are ready(change tab and window title, and so on.) */
  g_signal_emit_by_name (tab, "finished");
}

/**
 * on_open_external_file_responsed:
 */
static void
on_open_external_file_responsed (GtkInfoBar    *info_bar,
				 gint           resp_id,
				 GLookLocalTab *tab)
{
  g_return_if_fail (GTK_IS_INFO_BAR (info_bar));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* disconnect the signal handlers. */
  g_signal_handlers_disconnect_by_func (info_bar, on_open_external_file_responsed, tab);
  g_object_set_data (G_OBJECT (info_bar), 
		     "open-external-file-signal", GUINT_TO_POINTER (0));

  /* turn on auto hide(ESC key can hide the info bar). */
  g_signal_handlers_unblock_by_func (tab->priv->info_bar, gtk_widget_hide, tab);

  /* if user select yes, we open the external file.
   * if user select close, the tab will be closed in above response handler. */
  if (resp_id == GTK_RESPONSE_YES)
    {
      gtk_widget_hide (GTK_WIDGET (info_bar));
      glook_local_tab_open_file (tab);
    }
}

/**
 * glook_local_tab_open:
 *
 * try to load ui from glade file.
 */
static void
glook_local_tab_open (GLookTab *tab)
{
  GLookLocalTab *local_tab = GLOOK_LOCAL_TAB (tab);
  GLookUri *uri;
  const gchar *path;
  guint i, sigid;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  uri = glook_tab_get_uri (tab);
  g_assert (uri);

  /* if info bar is visible, we need hide it first. */
  if (gtk_widget_get_visible (local_tab->priv->info_bar))
    {
      /* disconnect open external file signal handler.
       * this is possible beacuse when user try to open other file without
       * response the info bar. */
      sigid = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (local_tab->priv->info_bar), 
						   "open-external-file-signal"));
      if (sigid > 0)
	{
	  /* response info bar with NO. */
	  gtk_info_bar_response (GTK_INFO_BAR (local_tab->priv->info_bar),
				 GTK_RESPONSE_NO);
	}
      gtk_widget_hide (local_tab->priv->info_bar);
    }
  /* get path attribute from uri. */
  path = glook_uri_lookup_attr (uri, "path");
  if (!path || strlen (path) == 0)
    {
      glook_local_tab_show_info (local_tab, GTK_MESSAGE_WARNING,
				 _("Uri missing attribute `path'."));
      return;
    }
  /* append default prefix if path is not absoluted. */
  g_free (local_tab->priv->pathname);
  if (!g_path_is_absolute (path))
    local_tab->priv->pathname = g_build_filename (LOCAL_GLADE_PATH, path, NULL);
  else
    local_tab->priv->pathname = g_strdup (path);

  /* check whether file exists. */
  if (!g_file_test (local_tab->priv->pathname, G_FILE_TEST_IS_REGULAR))
    {
      glook_local_tab_show_info (local_tab, GTK_MESSAGE_ERROR,
				 _("File `%s' not found."),
				 local_tab->priv->pathname);
      return;
    }
  /* check whether or not open internal file. */
  for (i = 0; i < G_N_ELEMENTS (_internal_files); i++)
    {
      if (g_strcmp0 (local_tab->priv->pathname,
		     _internal_files[i].pathname) == 0)
	{
	  local_tab->priv->builtin = _internal_files[i].builtin;
	  glook_local_tab_open_file (local_tab);
	  return;
	}
    }
  /* we are try to open external file, this need user premit. */
  glook_local_tab_show_info (local_tab, GTK_MESSAGE_QUESTION,
			     _("You are try to open file `%s' that is not "
			       "Ttyutils GLook provides, are you really want "
			       "to open it?"), local_tab->priv->pathname);
  /* turn off auto hide(ESC key cannot hide the info bar). */
  g_signal_handlers_block_by_func (local_tab->priv->info_bar,
				   gtk_widget_hide, local_tab);
  /* connect response signal to wait for user response.
   * and save the signal connect id. */
  sigid = g_signal_connect (local_tab->priv->info_bar, "response",
			    G_CALLBACK (on_open_external_file_responsed), tab);
  g_object_set_data (G_OBJECT (local_tab->priv->info_bar),
		     "open-external-file-signal", GUINT_TO_POINTER (sigid));
}

/**
 * glook_local_tab_stop:
 */
static void
glook_local_tab_stop (GLookTab *tab)
{
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));
}

/**
 * glook_local_tab_reload:
 *
 * some tab support reload, and some not.
 */
static void
glook_local_tab_reload (GLookTab *tab)
{
  GLookLocalTab *local_tab = GLOOK_LOCAL_TAB (tab);

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  if (!glook_local_tab_is_active (tab))
    {
      glook_local_tab_open (tab);
      return;
    }
  switch (local_tab->priv->builtin)
    {
    case LOCAL_BUILTIN_LOG:
      glook_local_log_refresh (local_tab);
      break;
    case LOCAL_BUILTIN_PREFERENCE:
    default:
      break;
    }
}

/**
 * glook_local_tab_zoom:
 *
 * some tab support zoom, and some not.
 */
static gboolean
glook_local_tab_zoom (GLookTab *tab,
		      gint      factor)
{
  GLookLocalTab *local_tab = GLOOK_LOCAL_TAB (tab);

  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (tab), FALSE);

  if (!glook_local_tab_is_active (tab))
    return FALSE;

  switch (local_tab->priv->builtin)
    {
    case LOCAL_BUILTIN_LOG:
      return glook_local_log_zoom (local_tab, factor);
    case LOCAL_BUILTIN_PREFERENCE:
    default:
      break;
    }
  return FALSE;
}

/**
 * glook_local_tab_is_active:
 */
static gboolean
glook_local_tab_is_active (GLookTab *tab)
{
  GLookLocalTab *local_tab = GLOOK_LOCAL_TAB (tab);

  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (tab), FALSE);

  return (local_tab->priv->builder && local_tab->priv->toplevel);
}

/**
 * glook_local_tab_new:
 */
GtkWidget*
glook_local_tab_new (void)
{
  return g_object_new (GLOOK_TYPE_LOCAL_TAB, NULL);
}

/**
 * glook_local_tab_get_builder:
 */
GtkBuilder*
glook_local_tab_get_builder (GLookLocalTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (tab), NULL);
  return tab->priv->builder;
}

/**
 * glook_local_tab_get_toplevel:
 */
GtkWidget*
glook_local_tab_get_toplevel (GLookLocalTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (tab), NULL);
  return tab->priv->toplevel;
}

/**
 * glook_local_tab_get_widget:
 */
GtkWidget*
glook_local_tab_get_widget (GLookLocalTab *tab,
			    const gchar   *name)
{
  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (tab), NULL);
  g_return_val_if_fail (name, NULL);

  if (!tab->priv->builder)
    return NULL;

  return GTK_WIDGET (gtk_builder_get_object (tab->priv->builder, name));
}

/**
 * glook_local_tab_get_object:
 */
GObject*
glook_local_tab_get_object (GLookLocalTab *tab,
			    const gchar   *name)
{
  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (tab), NULL);
  g_return_val_if_fail (name, NULL);

  if (!tab->priv->builder)
    return NULL;

  return gtk_builder_get_object (tab->priv->builder, name);
}
