/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include <ttyutils/ttyutils.h>

#include "glook-app.h"
#include "glook-window.h"
#include "glook-notebook.h"
#include "glook-terminal.h"
#include "glook-terminal-actions.h"
#include "glook-uri.h"
#include "glook-misc.h"
#include "glook-accel-map.h"
#include "glook-tab.h"
#include "glook-local-preference.h"
#include "glook-profile-dialog.h"
#include "glook-look-tab.h"

/* paste confirm dialog ui file name. */
#ifdef DEBUG
# define PASTE_CONFIRM_GLADE_FILE	SRCDIR G_DIR_SEPARATOR_S "glade" G_DIR_SEPARATOR_S "look-paste-confirm.xml"
#else
# define PASTE_CONFIRM_GLADE_FILE	GLADEDIR G_DIR_SEPARATOR_S "look-paste-confirm.xml"
#endif

enum
{
  PROP_0,
  PROP_ENCODING_UI,
};

enum
{
  LOOK_PASTE_NONE  = 0,
  LOOK_PASTE_LINE  = 1,
  LOOK_PASTE_ALL   = 2,
  LOOK_PASTE_UNSET = 3,
};

struct _GLookLookTabPriv
{
  TLook          *look;
  guint           look_cancel_id;
  GtkWidget      *box_scrolled;
  GtkWidget      *alignment;
  GtkWidget      *term_scrolled;
  GtkWidget      *terminal;
  guint           resize_idle_id;
  GtkActionGroup *ui_actions;
  guint           ui_merge_id;
  GtkAccelGroup  *ui_accel_group;
  GtkUIManager   *popup_menu_ui;
  gboolean        writable;
  gint            paste_remember;
  GtkActionGroup *encoding_actions;
  gchar          *encoding_ui;
  guint           encoding_merge_id;
  GtkWidget      *status_box;
  GtkWidget      *status_connected_image;
  GtkWidget      *status_writable_viewport;
  GtkWidget      *status_writable_image;
  GtkWidget      *status_suspension_viewport;
  GtkWidget      *status_suspension_label;
};
#define _LOOK_TAB_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_LOOK_TAB, GLookLookTabPriv))

G_DEFINE_TYPE (GLookLookTab, glook_look_tab, GLOOK_TYPE_TAB);

static void	set_terminal_size_request	(GLookLookTab  *tab,
						 guint16        rows,
						 guint16        columns);
static void	set_connect_action_sensitive	(GLookLookTab  *tab,
						 gboolean       sensitive);
static void	glook_look_tab_bind_settings	(GLookLookTab  *tab);
static gboolean	glook_look_tab_build_encoding_ui(GValue        *value,
						 GVariant      *variant,
						 gpointer       user_data);
static void	glook_look_tab_set_encoding_ui	(GLookLookTab  *tab,
						 const gchar   *encoding_ui);
const gchar*	glook_look_tab_get_encoding_ui	(GLookLookTab *tab);

static void	glook_look_tab_paste_with_check	(GLookLookTab  *tab,
						 const gchar   *text);

/* override parent methods */
static void	glook_look_tab_progress		(GLookTab      *tab);
static void	glook_look_tab_open		(GLookTab      *tab);
static void	glook_look_tab_stop		(GLookTab      *tab);
static void	glook_look_tab_reload		(GLookTab      *tab);
static gboolean	glook_look_tab_is_active	(GLookTab      *tab);
static gboolean	glook_look_tab_zoom		(GLookTab      *tab,
						 gint           factor);
static void	glook_look_tab_merge_ui		(GLookTab      *tab,
						 GtkWidget     *window);
static void	glook_look_tab_unmerge_ui	(GLookTab      *tab,
						 GtkWidget     *window);
static void	glook_look_tab_geometry_hint	(GLookTab      *tab);

/* look object's signal handler */
static void	on_look_connected		(TLook         *look,
						 GLookLookTab  *tab);
static void	on_look_received		(TLook         *look,
						 const gchar   *data,
						 guint          size,
						 GLookLookTab  *tab);
static void	on_look_hangup			(TLook         *look,
						 GLookLookTab  *tab);

static void	on_term_scrolled_window_policy_changed	(GtkScrolledWindow *scrolled,
							 GParamSpec        *spec,
							 GLookLookTab      *tab);

/* terminal's signal handler */
static void	on_terminal_commit		(GLookTerminal *terminal,
						 const gchar   *data,
						 guint          size,
						 GLookLookTab  *tab);
static void	on_terminal_size_changed	(GLookTerminal *terminal,
						 gint           rows,
						 gint           cols,
						 GLookLookTab  *tab);

static void	on_terminal_show_popup_menu	(GLookTerminal          *terminal,
						 GLookTerminalPopupInfo *info,
						 GLookLookTab           *tab);
static void	on_terminal_drag_text_received	(GLookTerminal          *terminal,
						 const gchar            *text,
						 GLookLookTab           *tab);
static void	on_terminal_font_desc_notify	(GLookTerminal          *terminal,
						 GParamSpec             *spec,
						 GLookLookTab           *tab);

/* menuitem signal handler */
static void	on_terminal_paste_activated		(GtkAction       *action,
							 GLookLookTab    *tab);
static void	on_terminal_disconnect_activated	(GtkAction       *action,
							 GLookLookTab    *tab);
static void	on_suspension_receive_activated		(GtkAction       *action,
							 GLookLookTab    *tab);
static void	on_writable_check_toggled		(GtkToggleAction *action,
							 GLookLookTab    *tab);

static void	on_look_tab_mapped			(GLookLookTab    *tab,
							 gpointer         user_data);

static gboolean	on_suspension_viewport_button_pressed	(GtkWidget       *viewport,
							 GdkEventButton  *event,
							 GLookLookTab    *tab);
static gboolean	on_writable_viewport_button_pressed	(GtkWidget       *viewport,
							 GdkEventButton  *event,
							 GLookLookTab    *tab);

/* actions will be merge to tab menu. */
static GtkActionEntry _tabmenu_actions[] =
{
    { "TerminalPaste", GTK_STOCK_PASTE, NULL, NULL,
      N_("Paste text from clipboard to look server"),
      G_CALLBACK (on_terminal_paste_activated),
    },
    { "TerminalDisconnect", GTK_STOCK_DISCONNECT, NULL, NULL,
      N_("Disconnect from look server"),
      G_CALLBACK (on_terminal_disconnect_activated),
    },
    { "TerminalSuspension", GTK_STOCK_MEDIA_PAUSE,
      N_("S_uspension of receiving data"), NULL,
      N_("Temporary stop receiving data from server"),
      G_CALLBACK (on_suspension_receive_activated),
    },
};
static GtkToggleActionEntry _tabmenu_toggle_actions[] =
{
    { "TerminalWritable", NULL, N_("_Writable"), NULL,
      N_("Enable/Disable write data to server"),
      G_CALLBACK (on_writable_check_toggled),
    },
    { "TerminalDisableAccelerator", NULL, N_("Disable acc_elerators"), NULL,
      N_("Disable accelerator keys"), NULL,
    },
    { "TerminalDisableMnemonic", NULL, N_("Disable _mnemonics"), NULL,
      N_("Disable weiget mnemonics"), NULL,
    },
};

static const gchar *_tabmenu_ui_string =
"<ui>"
"  <menubar name='Menubar'>"
"    <menu action='TabMenu'>"
"      <placeholder name='TabMenuPlaceholder'>"
"        <menuitem action='TerminalDisconnect'/>"
"        <menuitem action='TerminalSuspension'/>"
"        <menuitem action='TerminalWritable'/>"
"        <separator />"
"        <menu action='TerminalEncodingMenu'>"
"          <menuitem action='TerminalEncodingCurrent'/>"
"          <placeholder name='EncodingMenuPlaceholder'/>"
"          <separator />"
"          <menuitem action='TerminalEncodingSetting'/>"
"        </menu>"
"        <separator />"
"        <menuitem action='TerminalEditProfile'/>"
"        <menu action='TerminalProfileMenu'>"
"          <menuitem action='TerminalProfile1'/>"
"          <menuitem action='TerminalProfile2'/>"
"          <menuitem action='TerminalProfile3'/>"
"          <menuitem action='TerminalProfile4'/>"
"          <menuitem action='TerminalProfile5'/>"
"          <separator />"
"          <menuitem action='TerminalManageProfile'/>"
"        </menu>"
"        <separator />"
"        <menuitem action='TerminalSaveScreen'/>"
"        <menuitem action='TerminalReset'/>"
"        <menuitem action='TerminalResetClear'/>"
"      </placeholder>"
"    </menu>"
"  </menubar>"
"</ui>";

static const gchar *_popup_menu_ui_string =
"<ui>"
"  <popup name='TerminalPopupMenu'>"
"    <menuitem action='TerminalOpenLink'/>"
"    <menuitem action='TerminalMailto'/>"
"    <menuitem action='TerminalCallto'/>"
"    <menuitem action='TerminalCopyLink'/>"
"    <separator />"
"    <menuitem action='TerminalCopy'/>"
"    <menuitem action='TerminalPaste'/>"
"    <menuitem action='TerminalSelectAll'/>"
"    <separator />"
"    <menuitem action='TerminalDisconnect'/>"
"    <menuitem action='TerminalSuspension'/>"
"    <menuitem action='TerminalWritable'/>"
"    <separator />"
"    <menuitem action='TerminalDisableAccelerator'/>"
"    <menuitem action='TerminalDisableMnemonic'/>"
"    <separator />"
"    <menuitem action='TerminalEditProfile'/>"
"    <menu action='TerminalProfileMenu'>"
"      <menuitem action='TerminalProfile1'/>"
"      <menuitem action='TerminalProfile2'/>"
"      <menuitem action='TerminalProfile3'/>"
"      <menuitem action='TerminalProfile4'/>"
"      <menuitem action='TerminalProfile5'/>"
"      <separator />"
"      <menuitem action='TerminalManageProfile'/>"
"    </menu>"
"    <separator />"
"    <menuitem action='TerminalSaveScreen'/>"
"  </popup>"
"</ui>";

/**
 * glook_look_tab_bind_settings:
 */
static void
glook_look_tab_bind_settings (GLookLookTab *tab)
{
  GSettings *settings;
  GtkAction *action;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_ACCELERATOR);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalDisableAccelerator");
  g_settings_bind (settings, "disable-accelerators", action, "active",
		   G_SETTINGS_BIND_DEFAULT);
  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalDisableMnemonic");
  g_settings_bind (settings, "disable-mnemonics", action, "active",
		   G_SETTINGS_BIND_DEFAULT);
}

/**
 * action_connect_accel_group:
 */
static void
connect_action_accel_group (GLookLookTab *tab)
{
  GtkAction *action;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalDisconnect");
  gtk_action_set_accel_path (action, GLOOK_ACCEL_PATH_DISCONNECT);
  gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
  gtk_action_connect_accelerator (action);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalWritable");
  gtk_action_set_accel_path (action, GLOOK_ACCEL_PATH_WRITABLE);
  gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
  gtk_action_connect_accelerator (action);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalSaveScreen");
  gtk_action_set_accel_path (action, GLOOK_ACCEL_PATH_SAVE_SCREEN);
  gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
  gtk_action_connect_accelerator (action);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalSuspension");
  gtk_action_set_accel_path (action, GLOOK_ACCEL_PATH_SUSPENSION);
  gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
  gtk_action_connect_accelerator (action);
}

/**
 * action_disconnect_accel_group:
 */
static void
disconnect_action_accel_group (GLookLookTab *tab)
{
  GtkAction *action;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  if (GTK_IS_ACTION_GROUP (tab->priv->ui_actions))
    {
      action = gtk_action_group_get_action (tab->priv->ui_actions,
					    "TerminalDisconnect");
      gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
      gtk_action_disconnect_accelerator (action);

      action = gtk_action_group_get_action (tab->priv->ui_actions,
					    "TerminalWritable");
      gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
      gtk_action_disconnect_accelerator (action);

      action = gtk_action_group_get_action (tab->priv->ui_actions,
					    "TerminalSaveScreen");
      gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
      gtk_action_disconnect_accelerator (action);

      action = gtk_action_group_get_action (tab->priv->ui_actions,
					    "TerminalSuspension");
      gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
      gtk_action_disconnect_accelerator (action);
    }
}

/**
 * create_status_box:
 */
static void
create_status_box (GLookLookTab *tab)
{
  GtkWidget *viewport;
  GtkWidget *alignment;

  tab->priv->status_box = gtk_hbox_new (FALSE, 4);

  tab->priv->status_suspension_viewport = gtk_viewport_new (NULL, NULL);
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (tab->priv->status_suspension_viewport),
				GTK_SHADOW_NONE);
  gtk_widget_add_events (tab->priv->status_suspension_viewport, GDK_BUTTON_PRESS_MASK);
  g_signal_connect (tab->priv->status_suspension_viewport, "button-press-event",
		    G_CALLBACK (on_suspension_viewport_button_pressed), tab);
  gtk_box_pack_start (GTK_BOX (tab->priv->status_box),
		      tab->priv->status_suspension_viewport, TRUE, TRUE, 0);
  alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 4, 4, 4, 4);
  gtk_container_add (GTK_CONTAINER (tab->priv->status_suspension_viewport), alignment);
  tab->priv->status_suspension_label = gtk_label_new (_("Receiving"));
  g_object_set (tab->priv->status_suspension_label, "xalign", 0.0, NULL);
  gtk_widget_set_tooltip_text (tab->priv->status_suspension_label,
			       _("Receiving data from server now, double "
				 "click here to suspension"));
  gtk_container_add (GTK_CONTAINER (alignment), tab->priv->status_suspension_label);

  viewport = gtk_viewport_new (NULL, NULL);
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport), GTK_SHADOW_NONE);
  gtk_box_pack_start (GTK_BOX (tab->priv->status_box), viewport, FALSE, FALSE, 0);
  alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 4, 4, 4, 4);
  gtk_container_add (GTK_CONTAINER (viewport), alignment);
  tab->priv->status_connected_image =
   gtk_image_new_from_stock (GTK_STOCK_DISCONNECT, GTK_ICON_SIZE_MENU);
  gtk_widget_set_tooltip_text (tab->priv->status_connected_image,
			       _("Unconnected"));
  gtk_container_add (GTK_CONTAINER (alignment), tab->priv->status_connected_image);

  tab->priv->status_writable_viewport = gtk_viewport_new (NULL, NULL);
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (tab->priv->status_writable_viewport),
				GTK_SHADOW_NONE);
  gtk_widget_add_events (tab->priv->status_writable_viewport, GDK_BUTTON_PRESS_MASK);
  g_signal_connect (tab->priv->status_writable_viewport, "button-press-event",
		    G_CALLBACK (on_writable_viewport_button_pressed), tab);
  gtk_box_pack_start (GTK_BOX (tab->priv->status_box),
		      tab->priv->status_writable_viewport, FALSE, FALSE, 0);
  alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 4, 4, 4, 4);
  gtk_container_add (GTK_CONTAINER (tab->priv->status_writable_viewport), alignment);
  tab->priv->status_writable_image = gtk_image_new_from_stock (GTK_STOCK_EDIT,
							       GTK_ICON_SIZE_MENU);
  gtk_widget_set_sensitive (tab->priv->status_writable_image, FALSE);
  gtk_widget_set_tooltip_text (tab->priv->status_writable_image,
			       _("Writable mode is off, double "
				 "click here to turn on"));
  gtk_container_add (GTK_CONTAINER (alignment), tab->priv->status_writable_image);

  gtk_widget_show_all (tab->priv->status_box);
  g_object_ref_sink (tab->priv->status_box);
}

/**
 * glook_look_tab_init:
 */
static void
glook_look_tab_init (GLookLookTab *tab)
{
  GtkWidget *viewport;
  GSettings *settings;
  gchar *profile;
  GError *error = NULL;

  tab->priv = _LOOK_TAB_GET_PRIVATE (tab);

  g_object_set (tab, "label", _("Unconnected"), "image", "application-x-ttylook", NULL);

  tab->priv->ui_merge_id = 0;
  tab->priv->ui_accel_group = gtk_accel_group_new ();
  tab->priv->writable = FALSE;
  tab->priv->paste_remember = LOOK_PASTE_UNSET;
  tab->priv->encoding_actions = gtk_action_group_new ("LookTabEncodingActionGroup");;
  tab->priv->encoding_ui = NULL;
  tab->priv->encoding_merge_id = 0;
  tab->priv->look = t_look_new ();
  tab->priv->look_cancel_id = 0;
  tab->priv->resize_idle_id = 0;

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_TERMINAL);
  profile = g_settings_get_string (settings, "default-profile");

  /* create terminal. */
  tab->priv->terminal = glook_terminal_new ();
  glook_terminal_set_profile (GLOOK_TERMINAL (tab->priv->terminal), profile);
  gtk_widget_show (tab->priv->terminal);

  /* create a scroll window to packing terminal. */
  tab->priv->term_scrolled = gtk_scrolled_window_new (NULL, NULL);
  g_object_set (tab->priv->term_scrolled,
		"border-width", 0,
		"hscrollbar-policy", GTK_POLICY_NEVER,
		"vscrollbar-policy", GTK_POLICY_NEVER,
		"shadow-type", GTK_SHADOW_IN, NULL);
  gtk_widget_show (tab->priv->term_scrolled);

  /* when scrolled window policy changed, we have to update window size. */
  g_signal_connect (tab->priv->term_scrolled, "notify::vscrollbar-policy",
		    G_CALLBACK (on_term_scrolled_window_policy_changed), tab);

  /* create a alignment to packing scrolled window. */
  tab->priv->alignment = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_container_set_border_width (GTK_CONTAINER (tab->priv->alignment), 0);
  gtk_widget_show (tab->priv->alignment);

  /* create a scroll window to packing aligment. */
  tab->priv->box_scrolled = gtk_scrolled_window_new (NULL, NULL);
  g_object_set (tab->priv->box_scrolled,
		"border-width", 0,
		"hscrollbar-policy", GTK_POLICY_NEVER,
		"vscrollbar-policy", GTK_POLICY_NEVER,
		"shadow-type", GTK_SHADOW_NONE, NULL);
  gtk_widget_show (tab->priv->box_scrolled);

  gtk_container_add (GTK_CONTAINER (tab->priv->term_scrolled), tab->priv->terminal);
  gtk_container_add (GTK_CONTAINER (tab->priv->alignment), tab->priv->term_scrolled);
  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (tab->priv->box_scrolled),
					 tab->priv->alignment);
  gtk_box_pack_start (GTK_BOX (tab), tab->priv->box_scrolled, TRUE, TRUE, 0);

  viewport = gtk_bin_get_child (GTK_BIN (tab->priv->box_scrolled));
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport), GTK_SHADOW_NONE);

  /* create ui merge and popup menu action group. */
  tab->priv->ui_actions = gtk_action_group_new ("LookTabActionGroup");
  gtk_action_group_set_translation_domain (tab->priv->ui_actions,
					   GETTEXT_PACKAGE);
  /* insert terminal actions. */
  glook_terminal_add_actions (tab->priv->ui_actions, tab->priv->terminal,
			      tab->priv->term_scrolled, profile);
  g_free (profile);

  gtk_action_group_add_actions (tab->priv->ui_actions, _tabmenu_actions,
				G_N_ELEMENTS (_tabmenu_actions), tab);
  gtk_action_group_add_toggle_actions (tab->priv->ui_actions, _tabmenu_toggle_actions,
				       G_N_ELEMENTS (_tabmenu_toggle_actions), tab);
  glook_look_tab_bind_settings (tab);

  /* change action accel path conform to predefined accel map.
   * see glook-accel-map.c for detail.
   * and set action's accel group for ui merge. */
  connect_action_accel_group (tab);

  /* create statusbar widgets. */
  create_status_box (tab);

  /* create context menu ui. */
  tab->priv->popup_menu_ui = gtk_ui_manager_new ();
  gtk_ui_manager_insert_action_group (tab->priv->popup_menu_ui,
				      tab->priv->ui_actions, 0);
  gtk_ui_manager_add_ui_from_string (tab->priv->popup_menu_ui,
				     _popup_menu_ui_string, -1, &error);
  if (error)
    {
      g_error ("terminal context menu ui error: %s", error->message);
      return;
    }
  g_signal_connect (tab, "map", G_CALLBACK (on_look_tab_mapped), NULL);

  g_signal_connect (tab->priv->terminal, "commit",
		    G_CALLBACK (on_terminal_commit), tab);
  g_signal_connect (tab->priv->terminal, "show-popup-menu",
		    G_CALLBACK (on_terminal_show_popup_menu), tab);
  g_signal_connect (tab->priv->terminal, "drag-text-received",
		    G_CALLBACK (on_terminal_drag_text_received), tab);
  g_signal_connect (tab->priv->terminal, "size-changed",
		    G_CALLBACK (on_terminal_size_changed), tab);
  g_signal_connect (tab->priv->terminal, "notify::font-desc",
		    G_CALLBACK (on_terminal_font_desc_notify), tab);

  gtk_widget_grab_focus (tab->priv->terminal);
}

/**
 * glook_look_tab_dispose:
 */
static void
glook_look_tab_dispose (GObject *object)
{
  GLookLookTab *tab = GLOOK_LOOK_TAB (object);

  if (tab->priv->look)
    {
      g_signal_handlers_disconnect_by_func (tab->priv->look,
					    on_look_received, tab);
      g_signal_handlers_disconnect_by_func (tab->priv->look,
					    on_look_hangup, tab);
      g_signal_handlers_disconnect_by_func (tab->priv->look,
					    on_look_connected, tab);
      if (t_look_is_connected (tab->priv->look))
	t_look_disconnect (tab->priv->look);

      g_object_unref (tab->priv->look);
      tab->priv->look = NULL;
    }
  disconnect_action_accel_group (tab);

  if (tab->priv->resize_idle_id > 0)
    {
      g_source_remove (tab->priv->resize_idle_id);
      tab->priv->resize_idle_id = 0;
    }
  if (tab->priv->ui_accel_group)
    {
      g_object_unref (tab->priv->ui_accel_group);
      tab->priv->ui_accel_group = NULL;
    }
  if (tab->priv->ui_actions)
    {
      g_object_unref (tab->priv->ui_actions);
      tab->priv->ui_actions = NULL;
    }
  if (tab->priv->encoding_actions)
    {
      g_object_unref (tab->priv->encoding_actions);
      tab->priv->encoding_actions = NULL;
    }
  G_OBJECT_CLASS (glook_look_tab_parent_class)->dispose (object);
}

/**
 * glook_look_tab_finalize:
 */
static void
glook_look_tab_finalize (GObject *object)
{
  GLookLookTab *tab = GLOOK_LOOK_TAB (object);

  if (tab->priv->popup_menu_ui)
    {
      g_object_unref (tab->priv->popup_menu_ui);
      tab->priv->popup_menu_ui = NULL;
    }
  if (tab->priv->encoding_ui)
    {
      g_free (tab->priv->encoding_ui);
      tab->priv->encoding_ui = NULL;
    }
  if (tab->priv->status_box)
    {
      gtk_widget_destroy (tab->priv->status_box);
      tab->priv->status_box = NULL;
    }
  G_OBJECT_CLASS (glook_look_tab_parent_class)->finalize (object);
}

/**
 * glook_look_tab_get_property:
 */
static void
glook_look_tab_get_property (GObject    *object,
			     guint       property_id,
			     GValue     *value,
			     GParamSpec *pspec)
{
  GLookLookTab *tab = GLOOK_LOOK_TAB (object);

  switch (property_id)
    {
    case PROP_ENCODING_UI:
      g_value_set_string (value, glook_look_tab_get_encoding_ui (tab));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_look_tab_set_property:
 */
static void
glook_look_tab_set_property (GObject      *object,
			     guint         property_id,
			     const GValue *value,
			     GParamSpec   *pspec)
{
  GLookLookTab *tab = GLOOK_LOOK_TAB (object);

  switch (property_id)
    {
    case PROP_ENCODING_UI:
      glook_look_tab_set_encoding_ui (tab, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_look_tab_class_init:
 */
static void
glook_look_tab_class_init (GLookLookTabClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GLookTabClass *parent_class = GLOOK_TAB_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookLookTabPriv));

  gobject_class->get_property = glook_look_tab_get_property;
  gobject_class->set_property = glook_look_tab_set_property;
  gobject_class->dispose  = glook_look_tab_dispose;
  gobject_class->finalize = glook_look_tab_finalize;

  parent_class->merge_ui   = glook_look_tab_merge_ui;
  parent_class->unmerge_ui = glook_look_tab_unmerge_ui;
  parent_class->progress   = glook_look_tab_progress;
  parent_class->open       = glook_look_tab_open;
  parent_class->stop       = glook_look_tab_stop;
  parent_class->is_active  = glook_look_tab_is_active;
  parent_class->reload     = glook_look_tab_reload;
  parent_class->zoom       = glook_look_tab_zoom;
  parent_class->geometry_hint = glook_look_tab_geometry_hint;

  g_object_class_install_property (gobject_class, PROP_ENCODING_UI,
				   g_param_spec_string ("encoding-ui", "encoding-ui", "Encoding UI", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
}

/**
 * update_terminal_size_idle:
 */
static gboolean
update_terminal_size_idle (gpointer data)
{
  GLookLookTab *tab = data;
  glong rows, cols;

  g_return_val_if_fail (GLOOK_IS_LOOK_TAB (tab), FALSE);

  glook_terminal_get_size (GLOOK_TERMINAL (tab->priv->terminal),
			   &cols, &rows);
  set_terminal_size_request (tab, rows, cols);

  return FALSE;
}

/**
 * on_term_scrolled_window_policy_changed:
 */
static void
on_term_scrolled_window_policy_changed (GtkScrolledWindow *scrolled,
					GParamSpec        *spec,
					GLookLookTab      *tab)
{
  g_return_if_fail (GTK_IS_SCROLLED_WINDOW (scrolled));
  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  g_idle_add (update_terminal_size_idle, tab);
}

/**
 * on_suspension_viewport_button_pressed:
 */
static gboolean
on_suspension_viewport_button_pressed (GtkWidget      *viewport,
				       GdkEventButton *event,
				       GLookLookTab   *tab)
{
  g_return_val_if_fail (GLOOK_IS_LOOK_TAB (tab), FALSE);

  /* when double click the suspension viewport will
   * start/stop the suspension. */
  if (t_look_is_connected (tab->priv->look) &&
      event->button == 1 && event->type == GDK_2BUTTON_PRESS)
    {
      GtkAction *action;

      action = gtk_action_group_get_action (tab->priv->ui_actions,
					    "TerminalSuspension");
      gtk_action_activate (action);
      return TRUE;
    }
  return FALSE;
}

/**
 * on_writable_viewport_button_pressed:
 */
static gboolean
on_writable_viewport_button_pressed (GtkWidget      *viewport,
				     GdkEventButton *event,
				     GLookLookTab   *tab)
{
  g_return_val_if_fail (GLOOK_IS_LOOK_TAB (tab), FALSE);

  /* when double click the writable viewport will
   * turn on/off writable mode. */
  if (t_look_is_connected (tab->priv->look) &&
      event->button == 1 && event->type == GDK_2BUTTON_PRESS)
    {
      GtkAction *action;

      action = gtk_action_group_get_action (tab->priv->ui_actions,
					    "TerminalWritable");
      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
				    !tab->priv->writable);
      return TRUE;
    }
  return FALSE;
}

/**
 * remove_encoding_action:
 */
static void
remove_encoding_action (GtkAction      *action,
			GtkActionGroup *group)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GTK_IS_ACTION_GROUP (group));

  gtk_action_group_remove_action (group, action);
}

/**
 * on_encoding_action_activated:
 */
static void
on_encoding_action_activated (GtkAction    *action,
			      GLookLookTab *tab)
{
  const gchar *encoding;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  encoding = g_object_get_data (G_OBJECT (action), "encoding");
  vte_terminal_set_encoding (VTE_TERMINAL (tab->priv->terminal), encoding);
}

/**
 * glook_look_tab_build_encoding_ui:
 *
 * convert selected-encoding settings to a mergable ui string.
 */
gboolean
glook_look_tab_build_encoding_ui (GValue   *value,
				  GVariant *variant,
				  gpointer  user_data)
{
  GLookLookTab *tab = user_data;
  GList *actions;
  GString *ui_string;
  GVariantIter iter;
  gchar *language, *encoding;

  g_return_val_if_fail (GLOOK_IS_LOOK_TAB (tab), FALSE);

  /* clear encoding action group first. */
  actions = gtk_action_group_list_actions (tab->priv->encoding_actions);
  if (actions)
    {
      g_list_foreach (actions, (GFunc) remove_encoding_action,
		      tab->priv->encoding_actions);
      g_list_free (actions);
    }
  /* build encoding ui string. */
  ui_string = g_string_new ("<menubar name='Menubar'>"
			    "<menu action='TabMenu'>"
			    "<placeholder name='TabMenuPlaceholder'>"
			    "<menu action='TerminalEncodingMenu'>"
			    "<placeholder name='EncodingMenuPlaceholder'>");

  /* create encoding action and insert into encoding action group. */
  g_variant_iter_init (&iter, variant);
  while (g_variant_iter_next (&iter, "{ss}", &language, &encoding))
    {
      GtkAction *action;
      gchar name[256], label[512];

      g_snprintf (name, sizeof (name), "%sAction", encoding);
      g_snprintf (label, sizeof (label), "%s (_%s)", language, encoding);

      action = gtk_action_new (name, label, NULL, NULL);

      /* add ui string. */
      g_string_append_printf (ui_string, "<menuitem action='%s'/>", name);

      /* associated encoding with action. */
      g_object_set_data_full (G_OBJECT (action), "encoding", encoding, g_free);
      g_free (language);

      /* when action activated, changed terminal encoding. */
      g_signal_connect (action, "activate",
			G_CALLBACK (on_encoding_action_activated), tab);

      gtk_action_group_add_action (tab->priv->encoding_actions, action);
      g_object_unref (action);
    }
  g_string_append (ui_string, "</placeholder></menu></placeholder></menu></menubar>");

  g_value_set_string (value, ui_string->str);
  g_string_free (ui_string, TRUE);

  return TRUE;
}

/**
 * glook_look_tab_set_encoding_ui:
 */
static void
glook_look_tab_set_encoding_ui (GLookLookTab *tab,
				const gchar  *encoding_ui)
{
  gchar *new_ui;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  new_ui = g_strdup (encoding_ui);
  g_free (tab->priv->encoding_ui);
  tab->priv->encoding_ui = new_ui;

  /* remerge encoding ui. */
  if (tab->priv->encoding_merge_id > 0)
    {
      GtkWidget *window = gtk_widget_get_toplevel (GTK_WIDGET (tab));

      glook_window_unmerge_ui (GLOOK_WINDOW (window), NULL,
			       tab->priv->encoding_merge_id);
      if (tab->priv->encoding_ui)
	{
	  tab->priv->encoding_merge_id =
	   glook_window_merge_ui (GLOOK_WINDOW (window),
				  tab->priv->encoding_actions, NULL,
				  tab->priv->encoding_ui, -1);
	  if (tab->priv->encoding_merge_id < 0)
	    g_critical ("Failed to merge look tab encoding ui.");
	}
    }
}

/**
 * glook_look_tab_get_encoding_ui:
 */
const gchar*
glook_look_tab_get_encoding_ui (GLookLookTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_LOOK_TAB (tab), NULL);
  return tab->priv->encoding_ui;
}

/**
 * on_look_tab_mapped:
 */
static void
on_look_tab_mapped (GLookLookTab *tab,
		    gpointer      user_data)
{
  GSettings *settings;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  /* disconnect to prevent bind more than once. */
  g_signal_handlers_disconnect_by_func (tab, on_look_tab_mapped, NULL);

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_GENERAL);

  /* when selected-encoding settings changed, we need rebuild
   * the encoding ui. */
  g_settings_bind_with_mapping (settings, "selected-encodings",
				tab, "encoding-ui", G_SETTINGS_BIND_GET,
				glook_look_tab_build_encoding_ui,
				NULL, tab, NULL);
  set_terminal_size_request (tab, 24, 80);
}

/**
 * set_connect_action_sensitive:
 */
static void
set_connect_action_sensitive (GLookLookTab *tab,
			      gboolean      sensitive)
{
  GtkAction *action;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalDisconnect");
  gtk_action_set_sensitive (action, sensitive);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalWritable");
  gtk_action_set_sensitive (action, sensitive);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalSuspension");
  gtk_action_set_sensitive (action, sensitive);
}

/**
 * get_term_scroll_get_border_size:
 *
 * get terminal scrolled window border total size.
 *
 * border size includes scrolled window's border, scrollbar,
 * and so on.
 */
static void
terminal_scrolled_get_border_size (GLookLookTab *tab,
				   gint         *width,
				   gint         *height)
{
  GtkAllocation alloc1, alloc2;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  /* get terminal scrolled window and terminal allocated size. */
  gtk_widget_get_allocation (tab->priv->term_scrolled, &alloc1);
  gtk_widget_get_allocation (tab->priv->terminal, &alloc2);

  /* terminal scrolled's border =
   * terminal scrolled allocated size - terminal allocaed size. */
  if (width)  *width  = alloc1.width  - alloc2.width;
  if (height) *height = alloc1.height - alloc2.height;
}

/**
 * get_box_scrolled_get_border_size:
 */
#if 0
static void
box_scrolled_get_border_size (GLookLookTab *tab,
			      gint         *width,
			      gint         *height)
{
  GtkWidget *viewport;
  GtkAllocation alloc1, alloc2;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  viewport = gtk_bin_get_child (GTK_BIN (tab->priv->box_scrolled));

  /* get box scrolled window and viewport allocated size. */
  gtk_widget_get_allocation (tab->priv->box_scrolled, &alloc1);
  gtk_widget_get_allocation (viewport, &alloc2);

  /* box scrolled's border =
   * box scrolled allocated size - viewport allocaed size. */
  if (width)  *width  = alloc1.width  - alloc2.width;
  if (height) *height = alloc1.height - alloc2.height;
}
#endif

/**
 * get_terminal_maximum_size:
 */
static void
get_terminal_maximum_size (GLookLookTab *tab,
			   gint         *width,
			   gint         *height)
{
  GtkWidget *window;
  GtkAllocation tab_alloc;
  GdkRectangle window_frame, *workarea;
  gint scroll_width, scroll_height;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  workarea = glook_app_get_workarea (glook_app_get ());

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  if (!GLOOK_IS_WINDOW (window))
    {
      *width  = workarea->width;
      *height = workarea->height;
      return;
    }
  /* take window frame into account. */
  glook_window_get_frame_extents (GLOOK_WINDOW (window), &window_frame);
  gtk_widget_get_allocation (GTK_WIDGET (tab), &tab_alloc);

  if (tab_alloc.width > 1 || tab_alloc.height > 1)
    {
      *width  = workarea->width - (window_frame.width - tab_alloc.width);
      *height = workarea->height - (window_frame.height - tab_alloc.height);
    }
  else
    {
      *width  = workarea->width;
      *height = workarea->height;
      return;
    }
  terminal_scrolled_get_border_size (tab, &scroll_width, &scroll_height);
  *width  -= scroll_width;
  *height -= scroll_height;
}

/**
 * resize_window_idle:
 *
 * resize toplevel window when terminal size changed.
 */
static gboolean
resize_window_idle (gpointer data)
{
  GLookLookTab *tab = data;
  GtkWidget *window;
  GdkWindow *gdkwindow;

  g_return_val_if_fail (GLOOK_IS_TAB (tab), FALSE);

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  if (!GTK_IS_WINDOW (window))
    {
      tab->priv->resize_idle_id = 0;
      return FALSE;
    }
  gdkwindow = gtk_widget_get_window (window);
  if (gdkwindow && GDK_IS_WINDOW (gdkwindow))
    {
      GdkWindowState state = gdk_window_get_state (gdkwindow);

      if ((state & (GDK_WINDOW_STATE_MAXIMIZED |
		    GDK_WINDOW_STATE_FULLSCREEN)) == 0)
	{
	  gtk_window_resize (GTK_WINDOW (window), 1, 1);
	  gtk_widget_queue_resize (GTK_WIDGET (window));
	}
    }
  gtk_widget_grab_focus (tab->priv->terminal);

  tab->priv->resize_idle_id = 0;
  return FALSE;
}

/**
 * set_terminal_size_request:
 * 
 * set terminal request. the terminal will has a fixed
 * minimum size(that is not GTK+ suggest).
 */
static void
set_terminal_size_request (GLookLookTab *tab,
			   guint16       rows,
			   guint16       cols)
{
  VteTerminal *terminal;
  glong char_width, char_height;
  GtkBorder *inner_border;
  gint term_width, term_height, max_width = 0, max_height = 0;
  gint border_width, border_height, scroll_width, scroll_height;

  g_return_if_fail (GLOOK_IS_TAB (tab));

  terminal = VTE_TERMINAL (tab->priv->terminal);

  /* get terminal request width and height. */
  glook_terminal_get_char_size (GLOOK_TERMINAL (terminal),
				&char_width, &char_height);
  term_width  = char_width  * cols;
  term_height = char_height * rows;

  gtk_widget_style_get (GTK_WIDGET (terminal), "inner-border",
			&inner_border, NULL);
  if (inner_border != NULL)
    {
      term_width  += inner_border->left + inner_border->right;
      term_height += inner_border->top  + inner_border->bottom;
      gtk_border_free (inner_border);
    }
  /* terminal size request. */
  gtk_widget_set_size_request (GTK_WIDGET (terminal), term_width, term_height);
  vte_terminal_set_size (terminal, cols, rows); 

  /* get maximum usable size of terminal widget. */
  get_terminal_maximum_size (tab, &max_width, &max_height);

  /* if terminal request size large than usable size,
   * turn box scrolled window into work. */
  if (term_width > max_width || term_height > max_height)
    {
      GtkPolicyType hpolicy = GTK_POLICY_NEVER;
      GtkPolicyType vpolicy = GTK_POLICY_NEVER;

      if (term_width > max_width)
	hpolicy = GTK_POLICY_ALWAYS;

      if (term_height > max_height)
	vpolicy = GTK_POLICY_ALWAYS;

      g_object_set (tab->priv->box_scrolled, "hscrollbar-policy", hpolicy,
		    "vscrollbar-policy", vpolicy, NULL);

      scroll_width  = MIN (max_width, term_width);
      scroll_height = MIN (max_height, term_height);

      terminal_scrolled_get_border_size (tab, &border_width, &border_height);
      scroll_width  += border_width;
      scroll_height += border_height;
    }
  else
    {
      gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (tab->priv->box_scrolled),
				      GTK_POLICY_NEVER, GTK_POLICY_NEVER);
      scroll_width  = term_width;
      scroll_height = term_height;

      terminal_scrolled_get_border_size (tab, &border_width, &border_height);
      scroll_width  += border_width;
      scroll_height += border_height;
    }
  gtk_widget_set_size_request (tab->priv->box_scrolled, scroll_width, scroll_height);

  glook_look_tab_geometry_hint (GLOOK_TAB (tab));

  /* update window size to fit new terminal size. */
  if (tab->priv->resize_idle_id > 0)
    g_source_remove (tab->priv->resize_idle_id);

  tab->priv->resize_idle_id =
   g_idle_add_full (G_PRIORITY_HIGH_IDLE, resize_window_idle,
		    g_object_ref (tab), g_object_unref);
}

/**
 * glook_look_tab_set_label:
 *
 * change tab's title label or color.
 */
static void
glook_look_tab_set_label (GLookLookTab *tab,
			  gboolean      emphasize)
{
  GString *string;
  GLookUri *uri;
  const gchar *host;
  gchar *escape;

  g_return_if_fail (GLOOK_IS_TAB (tab));

  string = g_string_new ("<span");

  if (emphasize)
    g_string_append (string, " foreground='blue2'");

  g_string_append_c (string, '>');

  uri = glook_tab_get_uri (GLOOK_TAB (tab));

  host = glook_uri_lookup_attr (uri, "host");
  escape = g_markup_escape_text (host, -1);
  g_string_append_printf (string, "%s:%d", escape,
			  t_look_get_port (tab->priv->look));
  g_free (escape);

  g_string_append (string, "</span>");

  glook_tab_set_label (GLOOK_TAB (tab), string->str);
  g_string_free (string, TRUE);
}

/**
 * glook_look_tab_finish:
 */
static void
glook_look_tab_finish (GLookLookTab *tab)
{
  GdkWindow *window;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  window = gtk_widget_get_window (tab->priv->terminal);
  if (GDK_IS_WINDOW (window))
    gdk_window_set_cursor (window, NULL);

  glook_look_tab_set_label (tab, FALSE);
  g_signal_emit_by_name (tab, "finished");
}

/**
 * on_look_hangup:
 *
 * called when connection been closed by server or
 * network fault.
 */
static void
on_look_hangup (TLook        *look,
		GLookLookTab *tab)
{
  GtkAction *action;
  const gchar *label;

  g_return_if_fail (T_IS_LOOK (look));

  if (!GLOOK_IS_LOOK_TAB (tab))
    return;

  /* remove the handlers. */
  g_signal_handlers_disconnect_by_func (look, on_look_received, tab);
  g_signal_handlers_disconnect_by_func (look, on_look_hangup, tab);

  /* when connection been close. reset the terminal,
   * clear the screen, and print error information. */
  glook_terminal_show_hangup (GLOOK_TERMINAL (tab->priv->terminal));

  glook_look_tab_set_label (tab, FALSE);

  /* update statusbar widgets. */
  gtk_image_set_from_stock (GTK_IMAGE (tab->priv->status_connected_image),
			    GTK_STOCK_DISCONNECT, GTK_ICON_SIZE_MENU);
  gtk_widget_set_tooltip_text (tab->priv->status_connected_image,
			       _("Disconnected"));

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalWritable");
  gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);

  label = gtk_label_get_label (GTK_LABEL (tab->priv->status_suspension_label));
  if (g_strcmp0 (label, _("Suspension")) == 0)
    {
      action = gtk_action_group_get_action (tab->priv->ui_actions,
					    "TerminalSuspension");
      gtk_action_activate (action);
    }
  gtk_widget_set_sensitive (tab->priv->status_suspension_label, FALSE);

  /* update actions state. */
  set_connect_action_sensitive (tab, FALSE);
}

/**
 * on_page_switch_update_title:
 */
static void
on_page_switch_update_title (GtkNotebook     *notebook,
			     GtkNotebookPage *page,
			     guint            page_num,
			     GLookLookTab    *tab)
{
  GtkWidget *curr_tab;

  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));

  curr_tab = gtk_notebook_get_nth_page (notebook, page_num);

  /* I will become current tab. */
  if ((gpointer) curr_tab == (gpointer) tab)
    {
      glook_look_tab_set_label (tab, FALSE);

      /* disconnect the signal handler. */
      g_signal_handlers_disconnect_by_func (notebook,
					    on_page_switch_update_title, tab);
    }
}

/**
 * window_urgency_clear_timeout:
 */
static gboolean
window_urgency_clear_timeout (gpointer data)
{
  GtkWidget *window = GTK_WIDGET (data);

  g_return_val_if_fail (GTK_IS_WINDOW (window), FALSE);

  if (gtk_window_is_active (GTK_WINDOW (window)))
    {
      gtk_window_set_urgency_hint (GTK_WINDOW (window), FALSE);
      return FALSE;
    }
  return TRUE;
}

/**
 * window_urgency_clear_timeout_destroy:
 */
static void
window_urgency_clear_timeout_destroy (gpointer data)
{
  GtkWidget *window = GTK_WIDGET (data);

  g_return_if_fail (GTK_IS_WINDOW (window));

  g_object_set_data (G_OBJECT (window), "unset-ungency-timeout-id",
		     GUINT_TO_POINTER (0));
  g_object_unref (window);
}

/**
 * remove_window_urgency_clear_timeout:
 */
static void
remove_window_urgency_clear_timeout (GtkWidget *window,
				     gpointer   data)
{
  guint id;

  g_return_if_fail (GTK_IS_WINDOW (window));

  id = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (window),
					    "unset-ungency-timeout-id"));
  if (id > 0)
    g_source_remove (id);
}

/**
 * on_look_received:
 *
 * this function will be called when data recevied from
 * look server.
 */
static void
on_look_received (TLook         *look,
		  const gchar   *data,
		  guint          size,
		  GLookLookTab  *tab)
{
  GtkWidget *window;
  GtkWidget *notebook;
  GSettings *settings;

  g_return_if_fail (T_IS_LOOK (look));
  g_return_if_fail (data);

  if (!GLOOK_IS_LOOK_TAB (tab))
    return;

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  /* feed received data to terminal.
   * TODO: verify character encoding. */
  vte_terminal_feed (VTE_TERMINAL (tab->priv->terminal), data, size);

  /* if current tab is not actived, we change the title vision
   * effect to hint user has data arrived. */
  notebook = glook_window_get_notebook (GLOOK_WINDOW (window));
  if (GLOOK_IS_NOTEBOOK (notebook))
    {
      if (!glook_notebook_is_current (GLOOK_NOTEBOOK (notebook),
				      GTK_WIDGET (tab)))
	{
	  glook_look_tab_set_label (tab, TRUE);

	  /* when notebook page switched, we need update title
	   * vision effect. */
	  g_signal_connect (notebook, "switch-page",
			    G_CALLBACK (on_page_switch_update_title), tab);
	}
    }
  /* when data received, we should asking the desktop environment
   * to draw the users attention to the window. */
  settings = glook_app_get_settings (glook_app_get (), SETTINGS_GENERAL);
  if (g_settings_get_boolean (settings, "window-urgency-hint"))
    {
      if (!gtk_window_is_active (GTK_WINDOW (window)))
	{
	  if (!g_object_get_data (G_OBJECT (window), "unset-ungency-timeout-id"))
	    {
	      guint id;

	      gtk_window_set_urgency_hint (GTK_WINDOW (window), TRUE);

	      /* when window become activated we need clear urgency hint. */
	      id = g_timeout_add_seconds_full (G_PRIORITY_DEFAULT, 1,
					       window_urgency_clear_timeout,
					       g_object_ref (window),
					       window_urgency_clear_timeout_destroy);
	      g_object_set_data (G_OBJECT (window), "unset-urgency-timeout-id",
				 GUINT_TO_POINTER (id));

	      /* when window destroy, remove above timeout event. */
	      g_signal_connect (window, "destroy",
				G_CALLBACK (remove_window_urgency_clear_timeout), NULL);
	    }
	}
    }
}

/**
 * apply_terminal_setting_idle:
 */
static gboolean
apply_terminal_setting_idle (gpointer data)
{
  GLookLookTab *tab = GLOOK_LOOK_TAB (data);
  TLookdInfo *info;

  g_return_val_if_fail (GLOOK_IS_LOOK_TAB (tab), FALSE);

  /* when connected, server will give us some information
   * about terminal settings such as charset and so on.
   * we apply these settings here. */
  info = t_look_get_info (tab->priv->look);

#if 0
  /* we cannot modify emulation unless install other
   * termcaps file in vte search path or /etc/termcap
   * file is present(this is obsolescence by terminfo).
   */
  vte_terminal_set_emulation (terminal), info->term);
#endif
  vte_terminal_set_encoding (VTE_TERMINAL (tab->priv->terminal),
			     info->charset);
  set_terminal_size_request (tab, info->termrows, info->termcols);

  /* clear the screen. */
  vte_terminal_reset (VTE_TERMINAL (tab->priv->terminal), TRUE, TRUE);

  return FALSE;
}

/**
 * connect_failure_idle:
 */
static gboolean
connect_failure_idle (gpointer data)
{
  GLookLookTab *tab = GLOOK_LOOK_TAB (data);

  g_return_val_if_fail (GLOOK_IS_LOOK_TAB (tab), FALSE);

  /* update statusbar widgets. */
  gtk_image_set_from_stock (GTK_IMAGE (tab->priv->status_connected_image),
			    GTK_STOCK_DISCONNECT, GTK_ICON_SIZE_MENU);
  gtk_widget_set_tooltip_text (tab->priv->status_connected_image,
			       _("Disconnected"));
  gtk_widget_set_sensitive (tab->priv->status_suspension_label, FALSE);

  /* print not found information to terminal. */
  glook_terminal_show_not_found (GLOOK_TERMINAL (tab->priv->terminal),
				 t_look_get_host (tab->priv->look));
  return FALSE;
}

/**
 * on_look_connected:
 *
 * this function will be called when connecting
 * look server succeed or failure.
 */
static void
on_look_connected (TLook        *look,
		   GLookLookTab *tab)
{
  VteTerminal *terminal;

  g_return_if_fail (T_IS_LOOK (look));

  if (!GLOOK_IS_LOOK_TAB (tab))
    return;

  terminal = VTE_TERMINAL (tab->priv->terminal);

  /* must check whether the connection is actual created. */
  if (!t_look_is_connected (look))
    {
      glook_look_tab_finish (tab);

      gtk_widget_set_sensitive (GTK_WIDGET (tab), TRUE);
      gtk_widget_grab_focus (GTK_WIDGET (terminal));

      g_idle_add_full (G_PRIORITY_DEFAULT_IDLE, connect_failure_idle,
		       g_object_ref (tab), g_object_unref);
      return;
    }
  /* apply server terminal settings.
   * we use a idle event beacuse until here the tab widget
   * maybe still does not packing into notebook yet. */
  g_idle_add_full (G_PRIORITY_DEFAULT_IDLE, apply_terminal_setting_idle,
		   g_object_ref (tab), g_object_unref);
  glook_look_tab_finish (tab);

  /* update statusbar widgets. */
  gtk_image_set_from_stock (GTK_IMAGE (tab->priv->status_connected_image),
			    GTK_STOCK_CONNECT, GTK_ICON_SIZE_MENU);
  gtk_widget_set_tooltip_text (tab->priv->status_connected_image, _("Connected"));

  gtk_widget_set_sensitive (tab->priv->status_suspension_label, TRUE);

  /* update actions and widget state. */
  set_connect_action_sensitive (tab, TRUE);
  gtk_widget_set_sensitive (GTK_WIDGET (tab), TRUE);
  gtk_widget_grab_focus (GTK_WIDGET (terminal));

  /* start read data from server */
  g_signal_connect (look, "received",
		    G_CALLBACK (on_look_received), tab);
  g_signal_connect (tab->priv->look, "hangup",
		    G_CALLBACK (on_look_hangup), tab);
  t_look_read (look);
}

/**
 * glook_look_connect_idle:
 */
static gboolean
glook_look_connect_idle (gpointer data)
{
  GLookLookTab *tab = data;
  GCancellable *cancellable;

  g_return_val_if_fail (GLOOK_IS_LOOK_TAB (data), FALSE);

  /* remove the event handler */
  if (tab->priv->look_cancel_id > 0)
    {
      cancellable = t_look_get_cancellable (tab->priv->look);
      g_cancellable_disconnect (cancellable, tab->priv->look_cancel_id);
      tab->priv->look_cancel_id = 0;
    }
  gtk_widget_set_tooltip_text (tab->priv->status_connected_image,
			       _("Connecting ..."));

  /* connect look object signals and start the connect. */
  g_signal_connect (tab->priv->look, "connected",
		    G_CALLBACK (on_look_connected), tab);
  t_look_connect (tab->priv->look);

  /* change tab's state. and notify external the operation is
   * cancellable. */
  g_signal_emit_by_name (tab, "progress");

  return FALSE;
}

/**
 * on_look_connection_cancelled:
 */
static void
on_look_connection_cancelled (GCancellable *cancellable,
			      GLookLookTab *tab)
{
  /* disconnect all handlers first. */
  g_signal_handlers_disconnect_by_func (tab->priv->look,
					on_look_connected, tab);
  g_signal_handlers_disconnect_by_func (tab->priv->look,
					on_look_received, tab);
  g_signal_handlers_disconnect_by_func (tab->priv->look,
					on_look_hangup, tab);
  /* update actions state. */
  set_connect_action_sensitive (tab, FALSE);
  gtk_widget_set_sensitive (GTK_WIDGET (tab), FALSE);

  /* change tab title */
  glook_look_tab_set_label (tab, FALSE);

  /* clear terminal screen */
  vte_terminal_reset (VTE_TERMINAL (tab->priv->terminal), TRUE, TRUE);

  /* here we have to use idle event to prevent lock of cancellable */
  g_idle_add_full (G_PRIORITY_HIGH_IDLE, glook_look_connect_idle,
		   g_object_ref (tab), g_object_unref);
}

/**
 * glook_look_tab_progress:
 */
static void
glook_look_tab_progress (GLookTab *tab)
{
  GLookLookTab *look_tab = GLOOK_LOOK_TAB (tab);
  GdkWindow *window;
  GdkCursor *cursor;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  cursor = gdk_cursor_new (GDK_WATCH);

  window = gtk_widget_get_window (look_tab->priv->terminal);
  if (GDK_IS_WINDOW (window))
    gdk_window_set_cursor (window, cursor);
  else
    gdk_cursor_unref (cursor);

  /* call parent class method. */
  GLOOK_TAB_CLASS (glook_look_tab_parent_class)->progress (tab);
}

/**
 * glook_look_tab_open:
 */
static void
glook_look_tab_open (GLookTab *tab)
{
  GLookLookTab *look_tab = GLOOK_LOOK_TAB (tab);
  GLookUri *uri;
  const gchar *host, *port;
  GCancellable *cancellable;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  /* prevent start another connect before previous
   * connecting has not started yet. */
  if (look_tab->priv->look_cancel_id > 0)
    return;

  /* get host and port info from uri. */
  uri = glook_tab_get_uri (tab);
  g_assert (uri);

  /* if host is not specified, try `localhost'. */
  host = glook_uri_lookup_attr (uri, "host");
  if (!host || strlen (host) == 0)
    {
      glook_uri_insert_attr (uri, "host", "localhost");
      host = glook_uri_lookup_attr (uri, "host");
    }
  t_look_set_host (look_tab->priv->look, host);

  /* look service has not a default port. */
  port = glook_uri_lookup_attr (uri, "port");
  if (!port || strlen (port) == 0)
    {
      g_warning (_("Server port unknown."));
      return;
    }
  else
    {
      guint16 nport = g_strtod (port, NULL);
      if (nport == 0)
	{
	  g_warning (_("Server port unknown."));
	  return;
	}
      t_look_set_port (look_tab->priv->look, nport);
    }
  /* before open connection, we have to disconnect exists
   * connection first. and new connect action will be start
   * just after disconnect. */
  cancellable = t_look_get_cancellable (look_tab->priv->look);
  look_tab->priv->look_cancel_id =
   g_cancellable_connect (cancellable,
			  G_CALLBACK (on_look_connection_cancelled),
			  g_object_ref (look_tab), g_object_unref);
  t_look_disconnect (look_tab->priv->look);
}

/**
 * glook_look_tab_stop:
 */
static void
glook_look_tab_stop (GLookTab *tab)
{
  GLookLookTab *look_tab = GLOOK_LOOK_TAB (tab);

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  /* cancel current operation. */
  t_look_disconnect (look_tab->priv->look);

  g_signal_emit_by_name (tab, "finished");
}

/**
 * glook_look_tab_is_active:
 */
static gboolean
glook_look_tab_is_active (GLookTab *tab)
{
  GLookLookTab *look_tab = GLOOK_LOOK_TAB (tab);

  g_return_val_if_fail (GLOOK_IS_LOOK_TAB (tab), FALSE);

  return t_look_is_connected (look_tab->priv->look);
}

/**
 * glook_look_tab_reload:
 *
 * override parent method.
 */
static void
glook_look_tab_reload (GLookTab *tab)
{
  GLookLookTab *look_tab = GLOOK_LOOK_TAB (tab);

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  /* reconnecting only if has not connection yet. */
  if (!t_look_is_connected (look_tab->priv->look))
    glook_look_tab_open (tab);
}

/**
 * glook_look_tab_zoom:
 *
 * override parent method.
 */
static gboolean
glook_look_tab_zoom (GLookTab *tab,
		     gint      factor)
{
  GLookLookTab *look_tab = GLOOK_LOOK_TAB (tab);
  GLookTerminal *terminal;
  glong rows, cols;

  g_return_val_if_fail (GLOOK_IS_LOOK_TAB (tab), FALSE);
  g_return_val_if_fail (factor > 0, FALSE);

  terminal = GLOOK_TERMINAL (look_tab->priv->terminal);

  /* change terminal font size. */
  glook_terminal_zoom (terminal, factor);

  /* changed terminal size request. */
  glook_terminal_get_size (terminal, &cols, &rows);
  set_terminal_size_request (look_tab, rows, cols);

  return TRUE;
}

/**
 * glook_look_tab_merge_ui:
 *
 * merge menu item and statusbar widgets into main window.
 */
static void
glook_look_tab_merge_ui (GLookTab  *tab,
			 GtkWidget *window)
{
  GLookLookTab *look_tab = GLOOK_LOOK_TAB (tab);
  GtkWidget *statusbar;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));
  g_return_if_fail (GLOOK_IS_WINDOW (window));
  g_return_if_fail (look_tab->priv->ui_merge_id == 0);

  /* merge tab menu. */
  look_tab->priv->ui_merge_id =
   glook_window_merge_ui (GLOOK_WINDOW (window),
			  look_tab->priv->ui_actions,
			  look_tab->priv->ui_accel_group,
			  _tabmenu_ui_string, -1);
  if (look_tab->priv->ui_merge_id < 0)
    g_critical ("Failed to merge look tab ui.");

  /* merge encoding menu. */
  if (look_tab->priv->encoding_ui)
    {
      look_tab->priv->encoding_merge_id =
       glook_window_merge_ui (GLOOK_WINDOW (window),
			      look_tab->priv->encoding_actions, NULL,
			      look_tab->priv->encoding_ui, -1);
      if (look_tab->priv->encoding_merge_id < 0)
	g_critical ("Failed to merge look tab encoding ui.");
    }
  /* merge statusbar widgets. */
  if (!gtk_widget_get_parent (look_tab->priv->status_box))
    {
      statusbar = glook_window_get_statusbar (GLOOK_WINDOW (window));
      gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR (statusbar), FALSE);
      gtk_box_pack_end (GTK_BOX (statusbar),
			look_tab->priv->status_box, TRUE, TRUE, 0);
    }
}

/**
 * glook_look_tab_unmerge_ui:
 */
static void
glook_look_tab_unmerge_ui (GLookTab  *tab,
			   GtkWidget *window)
{
  GLookLookTab *look_tab = GLOOK_LOOK_TAB (tab);
  GtkWidget *statusbar;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  if (look_tab->priv->encoding_merge_id > 0)
    {
      glook_window_unmerge_ui (GLOOK_WINDOW (window), NULL,
			       look_tab->priv->encoding_merge_id);
      look_tab->priv->encoding_merge_id = 0;
    }
  if (look_tab->priv->ui_merge_id > 0)
    {
      glook_window_unmerge_ui (GLOOK_WINDOW (window),
			       look_tab->priv->ui_accel_group,
			       look_tab->priv->ui_merge_id);
      look_tab->priv->ui_merge_id = 0;
    }
  /* unmerge statusbar widgets. */
  if (gtk_widget_get_parent (look_tab->priv->status_box))
    {
      statusbar = glook_window_get_statusbar (GLOOK_WINDOW (window));
      gtk_container_remove (GTK_CONTAINER (statusbar),
			    look_tab->priv->status_box);
      gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR (statusbar), TRUE);
    }
}

/**
 * glook_look_tab_geometry_hint:
 */
static void
glook_look_tab_geometry_hint (GLookTab *tab)
{
  GLookLookTab *look_tab = GLOOK_LOOK_TAB (tab);
  VteTerminal *terminal;
  GtkWidget *window;
  GdkGeometry geometry;
  guint width, height;
  glong cols, rows;
  GtkBorder *inner_border;
  gint xpad = 0, ypad = 0;

  g_return_if_fail (GLOOK_IS_TAB (tab));

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  if (!gtk_widget_get_realized (window))
    return;

  terminal = VTE_TERMINAL (look_tab->priv->terminal);

  gtk_widget_style_get (GTK_WIDGET (terminal),
			"inner-border", &inner_border, NULL);
  if (inner_border)
    {
      xpad = inner_border->left + inner_border->right;
      ypad = inner_border->top + inner_border->bottom;
      gtk_border_free (inner_border);
    }
  width  = vte_terminal_get_char_width (terminal);
  height = vte_terminal_get_char_height (terminal);
  glook_terminal_get_size (GLOOK_TERMINAL (terminal), &cols, &rows);

#if 0
  geometry.width_inc   = -1;
  geometry.height_inc  = -1;
  geometry.base_width  = xpad;
  geometry.base_height = ypad;
#endif
  geometry.min_width   = xpad + width * cols;
  geometry.min_height  = ypad + height * rows;

  gtk_window_set_geometry_hints (GTK_WINDOW (window), GTK_WIDGET (terminal),
				 &geometry,
				 //GDK_HINT_RESIZE_INC | GDK_HINT_BASE_SIZE |
				 GDK_HINT_MIN_SIZE);
}

/**
 * on_terminal_commit:
 *
 * process user keyboard input.
 */
static void
on_terminal_commit (GLookTerminal *terminal,
		    const gchar   *data,
		    guint          size,
		    GLookLookTab  *tab)
{
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (data);
  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  /* we can send data to server only if writable
   * mode is turn on. */
  if (tab->priv->writable)
    {
      if (t_look_is_connected (tab->priv->look))
	t_look_write (tab->priv->look, data, size);
    }
}

/**
 * on_terminal_size_changed:
 */
static void
on_terminal_size_changed (GLookTerminal *terminal,
			  gint           rows,
			  gint           cols,
			  GLookLookTab  *tab)
{
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  set_terminal_size_request (tab, rows, cols);
}

/**
 * on_terminal_paste_activated:
 */
static void
on_terminal_paste_activated (GtkAction    *action,
			     GLookLookTab *tab)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  glook_look_tab_paste_clipboard (tab);
}

/**
 * on_terminal_disconnect_activated:
 */
static void
on_terminal_disconnect_activated (GtkAction    *action,
				  GLookLookTab *tab)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  if (t_look_is_connected (tab->priv->look))
    t_look_disconnect (tab->priv->look);
}

/**
 * on_suspension_receive_activated:
 *
 * pause/continue receive data from server.
 *
 * when pause, we still receive data from server,
 * and place into cache rather than discard.
 * when later continue, the data will be output
 * to terminal immediately.
 */
static void
on_suspension_receive_activated (GtkAction    *action,
				 GLookLookTab *tab)
{
  GtkAction *writable;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  writable = gtk_action_group_get_action (tab->priv->ui_actions,
					  "TerminalWritable");
  if (!t_look_is_connected (tab->priv->look))
    {
      gtk_action_set_stock_id (action, GTK_STOCK_MEDIA_PAUSE);
      gtk_action_set_label (action, _("S_uspension of receiving data"));
      gtk_action_set_tooltip (action, _("Temporary stop receiving data"));

      gtk_widget_set_tooltip_text (tab->priv->status_suspension_label,
				   _("Receiving data from server now, double "
				     "click here to suspension"));
      gtk_label_set_label (GTK_LABEL (tab->priv->status_suspension_label),
			   _("Receiving"));
      return;
    }
  if (t_look_is_suspension (tab->priv->look))
    {
      gtk_action_set_sensitive (writable, TRUE);

      t_look_set_suspension (tab->priv->look, FALSE);
      gtk_action_set_stock_id (action, GTK_STOCK_MEDIA_PAUSE);
      gtk_action_set_label (action, _("S_uspension of receiving data"));
      gtk_action_set_tooltip (action, _("Temporary stop receiving data"));

      gtk_widget_set_tooltip_text (tab->priv->status_suspension_label,
				   _("Receiving data from server now, double "
				     "click here to suspension"));
      gtk_label_set_label (GTK_LABEL (tab->priv->status_suspension_label),
			   _("Receiving"));
    }
  else
    {
      /* trun off writable mode also. */
      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (writable), FALSE);
      gtk_action_set_sensitive (writable, FALSE);

      t_look_set_suspension (tab->priv->look, TRUE);
      gtk_action_set_stock_id (action, GTK_STOCK_MEDIA_PLAY);
      gtk_action_set_label (action, _("Contin_ue to receive data"));
      gtk_action_set_tooltip (action, _("Continue to receive data"));

      gtk_widget_set_tooltip_text (tab->priv->status_suspension_label,
				   _("Suspension of receiving data now, double "
				     "click here to continue"));
      gtk_label_set_label (GTK_LABEL (tab->priv->status_suspension_label),
			   _("Suspension"));
    }
}

/**
 * on_writable_check_toggled:
 */
static void
on_writable_check_toggled (GtkToggleAction *action,
			   GLookLookTab    *tab)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  /* toggle writable only if connection is not suspension. */
  if (t_look_is_suspension (tab->priv->look))
    return;

  tab->priv->writable = gtk_toggle_action_get_active (action);

  gtk_widget_set_sensitive (tab->priv->status_writable_image,
			    tab->priv->writable);
  if (tab->priv->writable)
    {
      gtk_widget_set_tooltip_text (tab->priv->status_writable_image,
				   _("Writable mode is on, double "
				     "click here to turn off"));
    }
  else
    {
      gtk_widget_set_tooltip_text (tab->priv->status_writable_image,
				   _("Writable mode is off, double "
				     "click here to turn on"));
    }
}

/**
 * popup_menu_clipboard_targets_received:
 */
static void
popup_menu_clipboard_targets_received (GtkClipboard *clipboard,
				       GdkAtom      *targets,
				       gint          ntargets,
				       gpointer      data)
{
  GLookLookTab *tab = GLOOK_LOOK_TAB (data);
  GLookTerminalPopupInfo *info;
  GtkAction *open_link_action;
  GtkAction *copy_link_action;
  GtkAction *mailto_action;
  GtkAction *callto_action;
  GtkAction *action;
  GtkWidget *menu;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  info = glook_terminal_get_popup_info (GLOOK_TERMINAL (tab->priv->terminal));

  /* change link actions visible. */
  open_link_action = gtk_action_group_get_action (tab->priv->ui_actions,
						  "TerminalOpenLink");
  copy_link_action = gtk_action_group_get_action (tab->priv->ui_actions,
						  "TerminalCopyLink");
  mailto_action = gtk_action_group_get_action (tab->priv->ui_actions,
					       "TerminalMailto");
  callto_action = gtk_action_group_get_action (tab->priv->ui_actions,
					       "TerminalCallto");
  gtk_action_set_visible (open_link_action, FALSE);
  gtk_action_set_visible (copy_link_action, FALSE);
  gtk_action_set_visible (mailto_action, FALSE);
  gtk_action_set_visible (callto_action, FALSE);

  if (info->address)
    {
      switch (info->uri_type)
	{
	case TERMINAL_URI_TYPE_URI:
	case TERMINAL_URI_TYPE_WEB:
	  gtk_action_set_visible (open_link_action, TRUE);
	  gtk_action_set_visible (copy_link_action, TRUE);
	  break;
	case TERMINAL_URI_TYPE_MAIL:
	  gtk_action_set_visible (mailto_action, TRUE);
	  gtk_action_set_visible (copy_link_action, TRUE);
	  break;
	case TERMINAL_URI_TYPE_CALL:
	  gtk_action_set_visible (callto_action, TRUE);
	  gtk_action_set_visible (copy_link_action, TRUE);
	  break;
	default:
	  break;
	}
    }
  /* change copy/paste action sensitive. */
  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalCopy");
  if (vte_terminal_get_has_selection (VTE_TERMINAL (tab->priv->terminal)))
    gtk_action_set_sensitive (action, TRUE);
  else
    gtk_action_set_sensitive (action, FALSE);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalPaste");
  if (tab->priv->writable &&
      targets && gtk_targets_include_text (targets, ntargets))
    gtk_action_set_sensitive (action, TRUE);
  else
    gtk_action_set_sensitive (action, FALSE);

  menu = gtk_ui_manager_get_widget (tab->priv->popup_menu_ui,
				    "/TerminalPopupMenu");
  /* select the first item when popup activate by accelerator. */
  if (info->button == 0)
    gtk_menu_shell_select_first (GTK_MENU_SHELL (menu), FALSE);

  gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 
		  info->button, info->timestamp);
}

/**
 * on_terminal_show_popup_menu:
 */
static void
on_terminal_show_popup_menu (GLookTerminal          *terminal,
			     GLookTerminalPopupInfo *info,
			     GLookLookTab           *tab)
{
  GtkClipboard *clipboard;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (info);
  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  clipboard = gtk_widget_get_clipboard (GTK_WIDGET (terminal),
					GDK_SELECTION_CLIPBOARD);
  gtk_clipboard_request_targets (clipboard,
				 popup_menu_clipboard_targets_received, tab);
}

/**
 * on_terminal_drag_text_received:
 */
static void
on_terminal_drag_text_received (GLookTerminal    *terminal,
				const gchar      *text,
				GLookLookTab     *tab)
{
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));
  g_return_if_fail (text);

  glook_look_tab_paste_with_check (tab, text);
}

/**
 * on_terminal_font_desc_notify:
 */
static void
on_terminal_font_desc_notify (GLookTerminal *terminal,
			      GParamSpec    *spec,
			      GLookLookTab  *tab)
{
  glong rows, cols;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  glook_terminal_get_size (terminal, &cols, &rows);
  set_terminal_size_request (tab, rows, cols);
}

/**
 * glook_look_tab_new:
 */
GtkWidget*
glook_look_tab_new (void)
{
  return g_object_new (GLOOK_TYPE_LOOK_TAB, NULL);
}

/**
 * glook_look_tab_get_terminal:
 */
GtkWidget*
glook_look_tab_get_terminal (GLookLookTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_LOOK_TAB (tab), NULL);
  return tab->priv->terminal;
}

/**
 * glook_look_tab_get_writable:
 */
gboolean
glook_look_tab_get_writable (GLookLookTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_LOOK_TAB (tab), FALSE);
  return tab->priv->writable;
}

/**
 * on_paste_remember_choice_toggled:
 *
 * a callback for glade file signal.
 */
void
on_paste_remember_choice_toggled (GtkToggleButton *button,
				  GLookLookTab    *tab)
{
  GtkBuilder *builder;
  GObject *radio;

  g_return_if_fail (GTK_IS_TOGGLE_BUTTON (button));
  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  if (!gtk_toggle_button_get_active (button))
    {
      tab->priv->paste_remember = LOOK_PASTE_UNSET;
      return;
    }
  builder = g_object_get_data (G_OBJECT (tab), "paste-confirm-dialog-builder");
  g_assert (GTK_IS_BUILDER (builder));

  radio = gtk_builder_get_object (builder, "FirstLineRadio");
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radio)))
    {
      tab->priv->paste_remember = LOOK_PASTE_LINE;
      return;
    }
  radio = gtk_builder_get_object (builder, "AllLinesRadio");
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radio)))
    {
      tab->priv->paste_remember = LOOK_PASTE_ALL;
      return;
    }
  tab->priv->paste_remember = LOOK_PASTE_NONE;
}

/**
 * run_paste_confirm_dialog:
 *
 * when clipboard has multiple line text, we popup a dialog
 * let user select how to paste these text.
 */
static gint
run_paste_confirm_dialog (GLookLookTab *tab)
{
  GtkBuilder *builder;
  GtkWidget *window;
  GtkWidget *dialog;
  gint ret = 0;
  GError *error = NULL;

  if (!g_file_test (PASTE_CONFIRM_GLADE_FILE, G_FILE_TEST_EXISTS))
    {
      g_critical (_("File `%s' not found. please check your installation "
		    "and try again. Quit now."), PASTE_CONFIRM_GLADE_FILE);
      return LOOK_PASTE_NONE;
    }
  builder = gtk_builder_new ();
  gtk_builder_add_from_file (builder, PASTE_CONFIRM_GLADE_FILE, &error);
  if (error)
    {
      g_critical (_("Failed to parse file '%s': %s"),
		  PASTE_CONFIRM_GLADE_FILE, error->message);
      g_critical (_("Please check your installation and try again."));
      g_error_free (error);
      g_object_unref (builder);
      return LOOK_PASTE_NONE;
    }
  /* keep trace to builder object. beacuse other callback may need access
   * the builder object. */
  g_object_set_data_full (G_OBJECT (tab), "paste-confirm-dialog-builder",
			  g_object_ref (builder), g_object_unref);
  /* connect all signals */
  gtk_builder_connect_signals (builder, tab);

  dialog = GTK_WIDGET (gtk_builder_get_object (builder, "ConfirmDialog"));

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
  gtk_window_set_destroy_with_parent (GTK_WINDOW (dialog), TRUE);

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      GtkWidget *button;

      button = GTK_WIDGET (gtk_builder_get_object (builder, "FirstLineRadio"));
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button)))
	ret = LOOK_PASTE_LINE;

      button = GTK_WIDGET (gtk_builder_get_object (builder, "AllLinesRadio"));
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button)))
	ret = LOOK_PASTE_ALL;
    }
  g_object_unref (builder);
  g_object_set_data (G_OBJECT (tab), "paste-confirm-dialog-builder", NULL);

  gtk_widget_destroy (dialog);
  gtk_window_present (GTK_WINDOW (window));

  return ret;
}

/**
 * glook_look_tab_paste_with_check:
 */
static void
glook_look_tab_paste_with_check (GLookLookTab *tab,
				 const gchar  *text)
{
  const gchar *p;
  gint length, paste;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  if (!text || !t_look_is_connected (tab->priv->look) ||
      !tab->priv->writable)
    return;

  length = strlen (text);

  /* we have to check the text to prevent paste multi-line
   * without user permit. */
  p = g_strstr_len (text, length, "\r");
  if (!p)
    p = g_strstr_len (text, length, "\n");

  if (p) /* has line terminate character */
    {
      /* the remember me option is only for current tab. */
      paste = tab->priv->paste_remember;
      if (paste == LOOK_PASTE_UNSET)
	paste = run_paste_confirm_dialog (tab);

      switch (paste)
	{
	case LOOK_PASTE_LINE:	/* paste first line */
	  length = p - text;
	  break;
	case LOOK_PASTE_ALL:	/* paste all lines */
	  break;
	default:		/* no data to paste */
	  return;
	}
    }
  /* send data to server. */
  t_look_write (tab->priv->look, text, length);
}

/**
 * on_request_clipboard_ready:
 *
 * callback for clipboard data ready.
 */
static void
on_request_clipboard_ready (GtkClipboard *clipboard,
			    const gchar  *text,
			    gpointer      data)
{
  GLookLookTab *tab = data;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  glook_look_tab_paste_with_check (tab, text);
}

/**
 * glook_look_tab_paste_clipboard:
 *
 * paste clipboard to look server.
 */
void
glook_look_tab_paste_clipboard (GLookLookTab *tab)
{
  GtkClipboard *clipboard;

  g_return_if_fail (GLOOK_IS_LOOK_TAB (tab));

  /* we cannot paste data if writable is disabled. */
  if (!t_look_is_connected (tab->priv->look) || !tab->priv->writable)
    return;

  /* request text from clipboard and paste to look server. */
  clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
  if (clipboard)
    gtk_clipboard_request_text (clipboard, on_request_clipboard_ready, tab);
}
