/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-app.h"
#include "glook-window.h"
#include "glook-tab.h"
#include "glook-notebook.h"

struct _GLookNotebookPriv
{
  guint     label_notify_id;
  gboolean  is_highlight;
};
#define _NOTEBOOK_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_NOTEBOOK, GLookNotebookPriv))

G_DEFINE_TYPE (GLookNotebook, glook_notebook, GTK_TYPE_NOTEBOOK);

static gboolean	glook_notebook_drag_motion	(GtkWidget      *widget,
						 GdkDragContext *drag_context,
						 gint            x,
						 gint            y,
						 guint           time);
static void	glook_notebook_drag_leave	(GtkWidget      *widget,
						 GdkDragContext *drag_context,
						 guint           time);
static void	on_notebook_page_added		(GtkNotebook *notebook,
						 GtkWidget   *child,
						 guint        page_num,
						 gpointer     data);
static void	on_notebook_page_removed	(GtkNotebook *notebook,
						 GtkWidget   *child,
						 guint        page_num,
						 gpointer     data);

/**
 * glook_notebook_init:
 */
static void
glook_notebook_init (GLookNotebook *notebook)
{
  notebook->priv = _NOTEBOOK_GET_PRIVATE (notebook);

  g_object_set (notebook,
		"group", GUINT_TO_POINTER (2),
		"enable-popup", FALSE,
		"homogeneous", TRUE,
		"scrollable", TRUE,
		"show-tabs", FALSE,
		"show-border", FALSE, NULL);

  notebook->priv->label_notify_id = 0;
  notebook->priv->is_highlight = FALSE;

  /* connect to below 2 signals to auto switch "show-tabs" property */
  g_signal_connect (notebook, "page-added",
		    G_CALLBACK (on_notebook_page_added), NULL);
  g_signal_connect (notebook, "page-removed",
		    G_CALLBACK (on_notebook_page_removed), NULL);
}

/**
 * glook_notebook_dispose:
 */
static void
glook_notebook_dispose (GObject *object)
{
  GLookNotebook *notebook = GLOOK_NOTEBOOK (object);

  if (notebook->priv->label_notify_id > 0)
    {
      g_source_remove (notebook->priv->label_notify_id);
      notebook->priv->label_notify_id = 0;
    }
  G_OBJECT_CLASS (glook_notebook_parent_class)->dispose (object);
}

/**
 * glook_notebook_finalize:
 */
static void
glook_notebook_finalize (GObject *object)
{
#if 0
  GLookNotebook *notebook = GLOOK_NOTEBOOK (object);

#endif
  G_OBJECT_CLASS (glook_notebook_parent_class)->finalize (object);
}

/**
 * on_notebook_create_window:
 */
static GtkNotebook*
on_notebook_create_window (GtkNotebook *source_notebook,
			   GtkWidget   *tab,
			   gint         x,
			   gint         y,
			   gpointer     data)
{
  GtkWidget *window;

  /* create a new window and add to toplevel window group. */
  window = glook_app_create_window (glook_app_get (), FALSE);

  /* move window to mouse pointer position. */
  gtk_window_move (GTK_WINDOW (window), x, y);
  gtk_widget_show (window);

  return GTK_NOTEBOOK (glook_window_get_notebook (GLOOK_WINDOW (window)));
}

/**
 * glook_notebook_class_init:
 */
static void
glook_notebook_class_init (GLookNotebookClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookNotebookPriv));

  gobject_class->dispose  = glook_notebook_dispose;
  gobject_class->finalize = glook_notebook_finalize;

  widget_class->drag_motion = glook_notebook_drag_motion;
  widget_class->drag_leave = glook_notebook_drag_leave;

  /* installs a global function used to create a window when
   * a detached tab is dropped in an empty area. */
  gtk_notebook_set_window_creation_hook (on_notebook_create_window, NULL, NULL);
}

/**
 * glook_notebook_drag_motion:
 */
static gboolean
glook_notebook_drag_motion (GtkWidget      *widget,
			    GdkDragContext *drag_context,
			    gint            x,
			    gint            y,
			    guint           time)
{
  GLookNotebook *notebook = GLOOK_NOTEBOOK (widget);
  GtkWidgetClass *widget_class;

  widget_class = GTK_WIDGET_CLASS (glook_notebook_parent_class);
  if (widget_class->drag_motion)
    if (!widget_class->drag_motion (widget, drag_context, x, y, time))
      return FALSE;

  if (!notebook->priv->is_highlight)
    {
      gtk_drag_highlight (widget);
      notebook->priv->is_highlight = TRUE;
    }
  return TRUE;
}

/**
 * glook_notebook_drag_leave:
 */
static void
glook_notebook_drag_leave (GtkWidget      *widget,
			   GdkDragContext *drag_context,
			   guint           time)
{
  GLookNotebook *notebook = GLOOK_NOTEBOOK (widget);
  GtkWidgetClass *widget_class;

  if (notebook->priv->is_highlight)
    {
      gtk_drag_unhighlight (widget);
      notebook->priv->is_highlight = FALSE;
    }
  widget_class = GTK_WIDGET_CLASS (glook_notebook_parent_class);
  if (widget_class->drag_leave)
    widget_class->drag_leave (widget, drag_context, time);
}

/**
 * on_notebook_page_added:
 */
static void
on_notebook_page_added (GtkNotebook *notebook,
			GtkWidget   *child,
			guint        page_num,
			gpointer     data)
{
  if (gtk_notebook_get_n_pages (notebook) > 1)
    gtk_notebook_set_show_tabs (notebook, TRUE);
}

/**
 * on_notebook_page_removed:
 */
static void
on_notebook_page_removed (GtkNotebook *notebook,
			  GtkWidget   *child,
			  guint        page_num,
			  gpointer     data)
{
  GtkWidget *window;

  g_return_if_fail (GTK_IS_NOTEBOOK (notebook));

  if (gtk_notebook_get_n_pages (notebook) > 1)
    gtk_notebook_set_show_tabs (notebook, TRUE);
  else
    gtk_notebook_set_show_tabs (notebook, FALSE);

  /* unmerge tab's ui. */
  window = gtk_widget_get_toplevel (GTK_WIDGET (notebook));
  if (GTK_IS_WINDOW (window))
    glook_tab_unmerge_ui (GLOOK_TAB (child), window);
}

/**
 * glook_notebook_new:
 */
GtkWidget*
glook_notebook_new (void)
{
  return g_object_new (GLOOK_TYPE_NOTEBOOK, NULL);
}

/**
 * glook_notebook_insert:
 *
 * insert a tab into notebook, differ from gtk_notebook_insert(),
 * this function also set need child properties.
 */
void
glook_notebook_insert (GLookNotebook *notebook,
		       GtkWidget     *tab,
		       gint           n)
{
  GtkWidget *head;

  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));
  g_return_if_fail (GLOOK_IS_TAB (tab));

  head = glook_tab_get_tab_head (GLOOK_TAB (tab));

  gtk_notebook_insert_page (GTK_NOTEBOOK (notebook), tab,
			    g_object_ref (head), n);

  gtk_container_child_set (GTK_CONTAINER (notebook), tab,
			   "reorderable", TRUE,
			   "detachable", TRUE,
			   "tab-expand", TRUE,
			   "tab-fill", TRUE, NULL);
}

/**
 * glook_notebook_append:
 *
 * @tab must be a GLookTab subclass.
 */
void
glook_notebook_append (GLookNotebook *notebook,
		       GtkWidget     *tab)
{
  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));
  g_return_if_fail (GLOOK_IS_TAB (tab));

  glook_notebook_insert (notebook, tab, -1);
}

/**
 * glook_notebook_remove_n:
 */
void
glook_notebook_remove_n (GLookNotebook *notebook,
			 gint           n)
{
  GtkWidget *tab;

  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));

  tab = gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), n);

  if (GLOOK_IS_TAB (tab))
    glook_tab_stop (GLOOK_TAB (tab));

  gtk_notebook_remove_page (GTK_NOTEBOOK (notebook), n);
}

/**
 * glook_notebook_remove:
 */
void
glook_notebook_remove (GLookNotebook *notebook,
		       GtkWidget     *tab)
{
  gint n;

  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));
  g_return_if_fail (GLOOK_IS_TAB (tab));

  n = gtk_notebook_page_num (GTK_NOTEBOOK (notebook), tab);
  glook_notebook_remove_n (notebook, n);
}

/**
 * glook_notebook_remove_others:
 */
void
glook_notebook_remove_others (GLookNotebook *notebook,
			      GtkWidget     *tab)
{
  gint n, i, j;

  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));
  g_return_if_fail (GLOOK_IS_TAB (tab));

  n = gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook));
  for (i = 0, j = 0; i < n; i++)
    {
      GtkWidget *tab2;

      tab2 = gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), j);
      if (tab != tab2)
	{
	  glook_notebook_remove_n (notebook, j);
	  continue;
	}
      j++;
    }
}

/**
 * glook_notebook_remove_current:
 *
 * remove current actived tab.
 */
void
glook_notebook_remove_current (GLookNotebook *notebook)
{
  gint n;

  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));

  n = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));
  glook_notebook_remove_n (notebook, n);
}

/**
 * glook_notebook_get_current:
 *
 * get current actived tab.
 */
GtkWidget*
glook_notebook_get_current (GLookNotebook *notebook)
{
  gint n;

  g_return_val_if_fail (GLOOK_IS_NOTEBOOK (notebook), NULL);

  if (gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook)) == 0)
    return NULL;

  n = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));
  return gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), n);
}

/**
 * switch_page_idle:
 */
static gboolean
switch_page_idle (gpointer data)
{
  GtkNotebook *notebook = data;
  gint n;

  g_return_val_if_fail (GTK_IS_NOTEBOOK (notebook), FALSE);

  n = GPOINTER_TO_UINT (g_object_get_data (data, "switch-page-n"));
  gtk_notebook_set_current_page (notebook, n);

  return FALSE;
}

/**
 * glook_notebook_set_current:
 *
 * set tab as current active tab.
 */
void
glook_notebook_set_current (GLookNotebook *notebook,
			    GtkWidget     *tab)
{
  gint n, i;

  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));
  g_return_if_fail (GLOOK_IS_TAB (tab));

  n = gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook));
  for (i = 0; i < n; i++)
    {
      GtkWidget *tab2;

      tab2 = gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), i);
      if (tab == tab2)
	{
	  g_object_set_data (G_OBJECT (notebook), "switch-page-n",
			     GUINT_TO_POINTER (i));
	  g_idle_add_full (G_PRIORITY_DEFAULT_IDLE, switch_page_idle,
			   g_object_ref (notebook), g_object_unref);
	  break;
	}
    }
}

/**
 * glook_notebook_set_current_n:
 */
void
glook_notebook_set_current_n (GLookNotebook *notebook,
			      gint           n)
{
  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));

  g_object_set_data (G_OBJECT (notebook), "switch-page-n",
		     GUINT_TO_POINTER (n));
  g_idle_add_full (G_PRIORITY_DEFAULT_IDLE, switch_page_idle,
		   g_object_ref (notebook), g_object_unref);
}

/**
 * glook_notebook_is_current:
 */
gboolean
glook_notebook_is_current (GLookNotebook *notebook,
			   GtkWidget     *tab)
{
  GtkWidget *curr_tab;

  g_return_val_if_fail (GLOOK_IS_NOTEBOOK (notebook), FALSE);
  g_return_val_if_fail (GLOOK_IS_TAB (tab), FALSE);

  curr_tab = glook_notebook_get_current (notebook);
  return curr_tab == tab;
}

/**
 * glook_notebook_get_nth:
 */
gint
glook_notebook_get_nth (GLookNotebook *notebook,
			GtkWidget     *tab)
{
  gint n, i;

  g_return_val_if_fail (GLOOK_IS_NOTEBOOK (notebook), 0);
  g_return_val_if_fail (GLOOK_IS_TAB (tab), 0);

  n = gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook));
  for (i = 0; i < n; i++)
    {
      GtkWidget *tab2;

      tab2 = gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), i);
      if (tab == tab2)
	return i;
    }
  return 0;
}

/**
 * glook_notebook_replace:
 *
 * replace current tab with new tab.
 */
void
glook_notebook_replace (GLookNotebook *notebook,
			GtkWidget     *tab)
{
  gint n;

  g_return_if_fail (GLOOK_IS_NOTEBOOK (notebook));
  g_return_if_fail (GLOOK_IS_TAB (tab));

  n = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));

  if (n >= 0)
    {
      /* we have to do insert action before remove. beacuse
       * if notebook has just 1 tab, remove it will cause
       * window been closed. */
      glook_notebook_insert (notebook, tab, n + 1);
      glook_notebook_remove_n (notebook, n);

      gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), n);
    }
}
