/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <time.h>

#include <glib/gi18n.h>

#include "glook-app.h"
#include "glook-misc.h"
#include "glook-accel-map.h"
#include "glook-window.h"
#include "glook-terminal.h"
#include "glook-terminal-actions.h"
#include "glook-profile-dialog.h"
#include "glook-local-preference.h"
#include "glook-tab.h"
#include "glook-play-tab.h"

#include <ttyutils/ttyutils.h>

/* paste confirm dialog ui file name. */
#ifdef DEBUG
# define FILE_INFO_GLADE_FILE	SRCDIR G_DIR_SEPARATOR_S "glade" G_DIR_SEPARATOR_S "play-file-info.xml"
#else
# define FILE_INFO_GLADE_FILE	GLADEDIR G_DIR_SEPARATOR_S "play-file-info.xml"
#endif

enum
{
  PROP_0,
  PROP_ENCODING_UI,
};

struct _GLookPlayTabPriv
{
  TDecoder        *decoder;
  GtkWidget       *box_scrolled;
  GtkWidget       *alignment;
  GtkWidget       *term_scrolled;
  GtkWidget       *terminal;
  guint            resize_idle_id;
  GtkWidget       *control_box;
  GtkWidget       *interval_spin;
  GtkWidget       *info_button;
  GtkWidget       *play_button;
  GtkAdjustment   *elapsed_adjust;
  gboolean         paused;
  guint32          icurr;
  guint            play_timeout;
  guint            play_idle;
  GTimer          *timer;
  GtkWidget       *status_box;
  GtkWidget       *status_state;
  GtkWidget       *status_elapsed;
  GtkWidget       *status_timer;
  guint            status_timer_timeout;
  GtkActionGroup  *ui_actions;
  GtkAccelGroup   *ui_accel_group;
  gint             ui_merge_id;
  GtkUIManager    *popup_menu_ui;
  GtkActionGroup  *encoding_actions;
  gchar           *encoding_ui;
  guint            encoding_merge_id;
};
#define _PLAY_TAB_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_PLAY_TAB, GLookPlayTabPriv))

G_DEFINE_TYPE (GLookPlayTab, glook_play_tab, GLOOK_TYPE_TAB);

static void	glook_play_tab_open		(GLookTab      *tab);
static void	glook_play_tab_stop		(GLookTab      *tab);
static gboolean	glook_play_tab_is_active	(GLookTab      *tab);
static void	glook_play_tab_reload		(GLookTab      *tab);
static gboolean	glook_play_tab_zoom		(GLookTab      *tab,
						 gint           factor);
static void	glook_play_tab_merge_ui		(GLookTab      *tab,
						 GtkWidget     *window);
static void	glook_play_tab_unmerge_ui	(GLookTab      *tab,
						 GtkWidget     *window);
static void	glook_play_tab_geometry_hint	(GLookTab      *tab);

static void	glook_play_tab_set_encoding_ui	(GLookPlayTab  *tab,
						 const gchar   *encoding_ui);
const gchar*	glook_play_tab_get_encoding_ui	(GLookPlayTab *tab);

static void	on_play_tab_mapped		(GLookPlayTab    *tab,
						 gpointer         user_data);

static gboolean	glook_play_tab_play_idle		(gpointer  user_data);
static void	glook_play_tab_play_idle_destroy	(gpointer  user_data);
static gboolean	glook_play_tab_play_timeout		(gpointer  user_data);
static void	glook_play_tab_play_timeout_destroy	(gpointer  user_data);

static void	on_term_scrolled_window_policy_changed	(GtkScrolledWindow *scrolled,
							 GParamSpec        *spec,
							 GLookPlayTab      *tab);

static void	on_elapsed_adjust_value_changed		(GtkAdjustment *adjust,
							 GLookPlayTab  *tab);

static void	on_terminal_show_popup_menu	(GLookTerminal          *terminal,
						 GLookTerminalPopupInfo *info,
						 GLookPlayTab           *tab);
static void	on_terminal_size_changed	(GLookTerminal          *terminal,
						 gint                    rows,
						 gint                    cols,
						 GLookPlayTab           *tab);
static void	on_terminal_font_desc_notify	(GLookTerminal          *terminal,
						 GParamSpec             *spec,
						 GLookPlayTab           *tab);

static void	on_terminal_fileinfo_activated		(GtkAction       *action,
							 GLookPlayTab    *tab);
static void	on_terminal_playback_activated		(GtkAction       *action,
							 GLookPlayTab    *tab);
static void	on_terminal_forward_activated		(GtkAction       *action,
							 GLookPlayTab    *tab);
static void	on_terminal_backward_activated		(GtkAction       *action,
							 GLookPlayTab    *tab);
static void	on_terminal_stop_activated		(GtkAction       *action,
							 GLookPlayTab    *tab);

static void	set_terminal_size_request		(GLookPlayTab    *tab,
							 guint16          rows,
							 guint16          cols);

static GtkActionEntry _tabmenu_actions[] =
{
    { "TerminalFileInfo", GTK_STOCK_INFO, N_("File _info"), NULL,
      N_("View file information"),
      G_CALLBACK (on_terminal_fileinfo_activated),
    },
    { "TerminalPlayback", GTK_STOCK_MEDIA_PLAY, N_("_Playback/Pause"), NULL,
      N_("Switch between playback and pause"),
      G_CALLBACK (on_terminal_playback_activated),
    },
    { "TerminalForward", GTK_STOCK_MEDIA_FORWARD, NULL, NULL,
      N_("Forward"),
      G_CALLBACK (on_terminal_forward_activated),
    },
    { "TerminalBackward", GTK_STOCK_MEDIA_REWIND, NULL, NULL,
      N_("Backward"),
      G_CALLBACK (on_terminal_backward_activated),
    },
    { "TerminalStop", GTK_STOCK_MEDIA_STOP, NULL, NULL,
      N_("Stop playback"),
      G_CALLBACK (on_terminal_stop_activated),
    },
};

static const gchar *_tabmenu_ui_string =
"<ui>"
"  <menubar name='Menubar'>"
"    <menu action='TabMenu'>"
"      <placeholder name='TabMenuPlaceholder'>"
"        <menuitem action='TerminalPlayback'/>"
"        <menuitem action='TerminalForward'/>"
"        <menuitem action='TerminalBackward'/>"
"        <menuitem action='TerminalStop'/>"
"        <separator />"
"        <menu action='TerminalEncodingMenu'>"
"          <menuitem action='TerminalEncodingCurrent'/>"
"          <placeholder name='EncodingMenuPlaceholder'/>"
"          <separator />"
"          <menuitem action='TerminalEncodingSetting'/>"
"        </menu>"
"        <separator />"
"        <menuitem action='TerminalEditProfile'/>"
"        <menu action='TerminalProfileMenu'>"
"          <menuitem action='TerminalProfile1'/>"
"          <menuitem action='TerminalProfile2'/>"
"          <menuitem action='TerminalProfile3'/>"
"          <menuitem action='TerminalProfile4'/>"
"          <menuitem action='TerminalProfile5'/>"
"          <separator />"
"          <menuitem action='TerminalManageProfile'/>"
"        </menu>"
"        <separator />"
"        <menuitem action='TerminalSaveScreen'/>"
"        <menuitem action='TerminalReset'/>"
"        <menuitem action='TerminalResetClear'/>"
"      </placeholder>"
"    </menu>"
"  </menubar>"
"</ui>";

static const gchar *_popup_menu_ui_string =
"<ui>"
"  <popup name='TerminalPopupMenu'>"
"    <menuitem action='TerminalOpenLink'/>"
"    <menuitem action='TerminalMailto'/>"
"    <menuitem action='TerminalCallto'/>"
"    <menuitem action='TerminalCopyLink'/>"
"    <separator />"
"    <menuitem action='TerminalCopy'/>"
"    <menuitem action='TerminalSelectAll'/>"
"    <separator />"
"    <menuitem action='TerminalPlayback'/>"
"    <menuitem action='TerminalForward'/>"
"    <menuitem action='TerminalBackward'/>"
"    <menuitem action='TerminalStop'/>"
"    <separator />"
"    <menuitem action='TerminalFileInfo'/>"
"    <separator />"
"    <menuitem action='TerminalEditProfile'/>"
"    <menu action='TerminalProfileMenu'>"
"      <menuitem action='TerminalProfile1'/>"
"      <menuitem action='TerminalProfile2'/>"
"      <menuitem action='TerminalProfile3'/>"
"      <menuitem action='TerminalProfile4'/>"
"      <menuitem action='TerminalProfile5'/>"
"      <separator />"
"      <menuitem action='TerminalManageProfile'/>"
"    </menu>"
"    <separator />"
"    <menuitem action='TerminalSaveScreen'/>"
"  </popup>"
"</ui>";

/**
 * create_control_bar:
 */
static GtkWidget*
create_control_bar (GLookPlayTab *tab)
{
  GtkWidget *scale;
  GtkAction *action;

  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), NULL);

  tab->priv->control_box = gtk_hbox_new (FALSE, 0);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalFileInfo");
  tab->priv->info_button = gtk_action_create_tool_item (action);
  gtk_box_pack_start (GTK_BOX (tab->priv->control_box),
		      tab->priv->info_button, FALSE, TRUE, 0);

  gtk_box_pack_start (GTK_BOX (tab->priv->control_box),
		      gtk_vseparator_new (), FALSE, TRUE, 0);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalPlayback");
  tab->priv->play_button = gtk_action_create_tool_item (action);
  gtk_box_pack_start (GTK_BOX (tab->priv->control_box),
		      tab->priv->play_button, FALSE, TRUE, 0);

  tab->priv->elapsed_adjust = g_object_new (GTK_TYPE_ADJUSTMENT, NULL);
  g_signal_connect (tab->priv->elapsed_adjust, "value-changed",
		    G_CALLBACK (on_elapsed_adjust_value_changed), tab);

  scale = gtk_hscale_new (tab->priv->elapsed_adjust);
  gtk_scale_set_draw_value (GTK_SCALE (scale), FALSE);
  gtk_box_pack_start (GTK_BOX (tab->priv->control_box), scale, TRUE, TRUE, 0);

  tab->priv->interval_spin = gtk_spin_button_new_with_range (-9000, 9000, 100);
  gtk_spin_button_set_digits (GTK_SPIN_BUTTON (tab->priv->interval_spin), 0);
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (tab->priv->interval_spin), 0);
  gtk_widget_set_tooltip_text (tab->priv->interval_spin,
			       _("Playback speed, in milliseconds"));
  gtk_box_pack_end (GTK_BOX (tab->priv->control_box),
		    tab->priv->interval_spin, FALSE, TRUE, 0);

  gtk_widget_set_sensitive (tab->priv->control_box, FALSE);
  gtk_widget_show_all (tab->priv->control_box);

  return tab->priv->control_box;
}

/**
 * create_status_box:
 */
static void
create_status_box (GLookPlayTab *tab)
{
  GtkWidget *separator;

  tab->priv->status_box = gtk_hbox_new (FALSE, 4);
  g_object_set (tab->priv->status_box, "border-width", 4, NULL);

  tab->priv->status_state = gtk_label_new (_("Stop"));
  g_object_set (tab->priv->status_state, "xalign", 0.0, NULL);
  gtk_box_pack_start (GTK_BOX (tab->priv->status_box),
		      tab->priv->status_state, TRUE, TRUE, 0);

  tab->priv->status_elapsed = gtk_label_new ("-");
  gtk_box_pack_start (GTK_BOX (tab->priv->status_box),
		      tab->priv->status_elapsed, FALSE, FALSE, 0);

  separator = gtk_vseparator_new ();
  gtk_box_pack_start (GTK_BOX (tab->priv->status_box),
		      separator, FALSE, FALSE, 0);

  tab->priv->status_timer = gtk_label_new ("-");
  gtk_box_pack_start (GTK_BOX (tab->priv->status_box),
		      tab->priv->status_timer, FALSE, FALSE, 0);

  gtk_widget_show_all (tab->priv->status_box);
  g_object_ref_sink (tab->priv->status_box);
}

/**
 * action_connect_accel_group:
 */
static void
connect_action_accel_group (GLookPlayTab *tab)
{
  GtkAction *action;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalPlayback");
  gtk_action_set_accel_path (action, GLOOK_ACCEL_PATH_PLAYBACK);
  gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
  gtk_action_connect_accelerator (action);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalForward");
  gtk_action_set_accel_path (action, GLOOK_ACCEL_PATH_FORWARD);
  gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
  gtk_action_connect_accelerator (action);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalBackward");
  gtk_action_set_accel_path (action, GLOOK_ACCEL_PATH_BACKWARD);
  gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
  gtk_action_connect_accelerator (action);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalSaveScreen");
  gtk_action_set_accel_path (action, GLOOK_ACCEL_PATH_SAVE_SCREEN);
  gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
  gtk_action_connect_accelerator (action);
}

/**
 * action_disconnect_accel_group:
 */
static void
disconnect_action_accel_group (GLookPlayTab *tab)
{
  GtkAction *action;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  if (GTK_IS_ACTION_GROUP (tab->priv->ui_actions))
    {
      action = gtk_action_group_get_action (tab->priv->ui_actions,
					    "TerminalPlayback");
      gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
      gtk_action_disconnect_accelerator (action);

      action = gtk_action_group_get_action (tab->priv->ui_actions,
					    "TerminalForward");
      gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
      gtk_action_disconnect_accelerator (action);

      action = gtk_action_group_get_action (tab->priv->ui_actions,
					    "TerminalBackward");
      gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
      gtk_action_disconnect_accelerator (action);

      action = gtk_action_group_get_action (tab->priv->ui_actions,
					    "TerminalSaveScreen");
      gtk_action_set_accel_group (action, tab->priv->ui_accel_group);
      gtk_action_disconnect_accelerator (action);
    }
}

/**
 * glook_play_tab_init:
 */
static void
glook_play_tab_init (GLookPlayTab *tab)
{
  GSettings *settings;
  gchar *profile;
  GtkWidget *viewport;
  GError *error = NULL;

  g_object_set (tab, "stock", GTK_STOCK_MEDIA_PLAY, "label", _("Play"), NULL);

  tab->priv = _PLAY_TAB_GET_PRIVATE (tab);

  tab->priv->decoder = t_decoder_new ();
  tab->priv->icurr = 0;
  tab->priv->play_timeout = 0;
  tab->priv->play_idle = 0;
  tab->priv->timer = g_timer_new ();
  tab->priv->status_timer_timeout = 0;
  tab->priv->ui_merge_id = 0;
  tab->priv->ui_accel_group = gtk_accel_group_new ();
  tab->priv->encoding_actions = gtk_action_group_new ("PlayTabEncodingActionGroup");;
  tab->priv->encoding_ui = NULL;
  tab->priv->encoding_merge_id = 0;
  tab->priv->resize_idle_id = 0;

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_TERMINAL);
  profile = g_settings_get_string (settings, "default-profile");

  /* create terminal widget and set profile. */
  tab->priv->terminal = glook_terminal_new ();
  glook_terminal_set_profile (GLOOK_TERMINAL (tab->priv->terminal), profile);
  gtk_widget_show (tab->priv->terminal);

  tab->priv->term_scrolled = gtk_scrolled_window_new (NULL, NULL);
  g_object_set (tab->priv->term_scrolled,
		"hscrollbar-policy", GTK_POLICY_NEVER,
		"vscrollbar-policy", GTK_POLICY_NEVER,
		"shadow-type", GTK_SHADOW_IN,
		"visible", TRUE, NULL);
  gtk_widget_show (tab->priv->term_scrolled);

  /* when scrolled window policy changed, we have to update window size. */
  g_signal_connect (tab->priv->term_scrolled, "notify::vscrollbar-policy",
		    G_CALLBACK (on_term_scrolled_window_policy_changed), tab);

  /* create a alignment to packing scrolled window. */
  tab->priv->alignment = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_container_set_border_width (GTK_CONTAINER (tab->priv->alignment), 0);
  gtk_widget_show (tab->priv->alignment);

  /* create a scroll window to packing aligment. */
  tab->priv->box_scrolled = gtk_scrolled_window_new (NULL, NULL);
  g_object_set (tab->priv->box_scrolled,
		"border-width", 0,
		"hscrollbar-policy", GTK_POLICY_NEVER,
		"vscrollbar-policy", GTK_POLICY_NEVER,
		"shadow-type", GTK_SHADOW_NONE, NULL);
  gtk_widget_show (tab->priv->box_scrolled);

  gtk_container_add (GTK_CONTAINER (tab->priv->term_scrolled), tab->priv->terminal);
  gtk_container_add (GTK_CONTAINER (tab->priv->alignment), tab->priv->term_scrolled);
  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (tab->priv->box_scrolled),
					 tab->priv->alignment);
  gtk_box_pack_start (GTK_BOX (tab), tab->priv->box_scrolled, TRUE, TRUE, 0);

  viewport = gtk_bin_get_child (GTK_BIN (tab->priv->box_scrolled));
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport), GTK_SHADOW_NONE);

  /* action group for merge ui and popup menu. */
  tab->priv->ui_actions = gtk_action_group_new ("PlayTabActionGroup");
  gtk_action_group_set_translation_domain (tab->priv->ui_actions,
					   GETTEXT_PACKAGE);

  glook_terminal_add_actions (tab->priv->ui_actions, tab->priv->terminal,
			      tab->priv->term_scrolled, profile);
  g_free (profile);

  gtk_action_group_add_actions (tab->priv->ui_actions, _tabmenu_actions,
				G_N_ELEMENTS (_tabmenu_actions), tab);
  /* connect accelerator */
  connect_action_accel_group (tab);

  /* playback control box. */
  gtk_box_pack_end (GTK_BOX (tab), create_control_bar (tab), FALSE, TRUE, 0);

  /* create context menu ui. */
  tab->priv->popup_menu_ui = gtk_ui_manager_new ();
  gtk_ui_manager_insert_action_group (tab->priv->popup_menu_ui,
				      tab->priv->ui_actions, 0);
  gtk_ui_manager_add_ui_from_string (tab->priv->popup_menu_ui,
				     _popup_menu_ui_string, -1, &error);
  if (error)
    {
      g_error ("terminal context menu ui error: %s", error->message);
      return;
    }
  /* statusbar widgets. */
  create_status_box (tab);

  g_signal_connect (tab, "map", G_CALLBACK (on_play_tab_mapped), NULL);

  /* popup context menu or open uri link. */
  g_signal_connect (tab->priv->terminal, "show-popup-menu",
		    G_CALLBACK (on_terminal_show_popup_menu), tab);
  g_signal_connect (tab->priv->terminal, "size-changed",
		    G_CALLBACK (on_terminal_size_changed), tab);
  g_signal_connect (tab->priv->terminal, "notify::font-desc",
		    G_CALLBACK (on_terminal_font_desc_notify), tab);

  gtk_widget_grab_focus (tab->priv->terminal);
}

/**
 * glook_play_tab_get_property:
 */
static void
glook_play_tab_get_property (GObject    *object,
			     guint       property_id,
			     GValue     *value,
			     GParamSpec *pspec)
{
  GLookPlayTab *tab = GLOOK_PLAY_TAB (object);

  switch (property_id)
    {
    case PROP_ENCODING_UI:
      g_value_set_string (value, glook_play_tab_get_encoding_ui (tab));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_play_tab_set_property:
 */
static void
glook_play_tab_set_property (GObject      *object,
			     guint         property_id,
			     const GValue *value,
			     GParamSpec   *pspec)
{
  GLookPlayTab *tab = GLOOK_PLAY_TAB (object);

  switch (property_id)
    {
    case PROP_ENCODING_UI:
      glook_play_tab_set_encoding_ui (tab, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_play_tab_dispose:
 */
static void
glook_play_tab_dispose (GObject *object)
{
  GLookPlayTab *tab = GLOOK_PLAY_TAB (object);

  disconnect_action_accel_group (tab);

  if (tab->priv->play_idle > 0)
    {
      g_source_remove (tab->priv->play_idle);
      tab->priv->play_idle = 0;
    }
  if (tab->priv->play_timeout > 0)
    {
      g_source_remove (tab->priv->play_timeout);
      tab->priv->play_timeout = 0;
    }
  if (tab->priv->status_timer_timeout > 0)
    {
      g_source_remove (tab->priv->status_timer_timeout);
      tab->priv->status_timer_timeout = 0;
    }
  if (tab->priv->resize_idle_id > 0)
    {
      g_source_remove (tab->priv->resize_idle_id);
      tab->priv->resize_idle_id = 0;
    }
  if (tab->priv->status_box)
    {
      gtk_widget_destroy (tab->priv->status_box);
      tab->priv->status_box = NULL;
    }
  if (tab->priv->ui_accel_group)
    {
      g_object_unref (tab->priv->ui_accel_group);
      tab->priv->ui_accel_group = NULL;
    }
  if (tab->priv->ui_actions)
    {
      g_object_unref (tab->priv->ui_actions);
      tab->priv->ui_actions = NULL;
    }
  if (tab->priv->encoding_actions)
    {
      g_object_unref (tab->priv->encoding_actions);
      tab->priv->encoding_actions = NULL;
    }
  G_OBJECT_CLASS (glook_play_tab_parent_class)->dispose (object);
}

/**
 * glook_play_tab_finalize:
 */
static void
glook_play_tab_finalize (GObject *object)
{
  GLookPlayTab *tab = GLOOK_PLAY_TAB (object);

  if (tab->priv->popup_menu_ui)
    {
      g_object_unref (tab->priv->popup_menu_ui);
      tab->priv->popup_menu_ui = NULL;
    }
  if (tab->priv->encoding_ui)
    {
      g_free (tab->priv->encoding_ui);
      tab->priv->encoding_ui = NULL;
    }
  if (tab->priv->timer > 0)
    {
      g_timer_destroy (tab->priv->timer);
      tab->priv->timer = NULL;
    }
  if (tab->priv->decoder)
    {
      g_object_unref (tab->priv->decoder);
      tab->priv->decoder = NULL;
    }
  G_OBJECT_CLASS (glook_play_tab_parent_class)->finalize (object);
}

/**
 * glook_play_tab_class_init:
 */
static void
glook_play_tab_class_init (GLookPlayTabClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GLookTabClass *parent_class = GLOOK_TAB_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookPlayTabPriv));

  gobject_class->get_property = glook_play_tab_get_property;
  gobject_class->set_property = glook_play_tab_set_property;
  gobject_class->dispose  = glook_play_tab_dispose;
  gobject_class->finalize = glook_play_tab_finalize;

  parent_class->open       = glook_play_tab_open;
  parent_class->stop       = glook_play_tab_stop;
  parent_class->is_active  = glook_play_tab_is_active;
  parent_class->reload     = glook_play_tab_reload;
  parent_class->zoom       = glook_play_tab_zoom;
  parent_class->merge_ui   = glook_play_tab_merge_ui;
  parent_class->unmerge_ui = glook_play_tab_unmerge_ui;
  parent_class->geometry_hint = glook_play_tab_geometry_hint;

  g_object_class_install_property (gobject_class, PROP_ENCODING_UI,
				   g_param_spec_string ("encoding-ui", "encoding-ui", "Encoding UI", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
}

/**
 * update_terminal_size_idle:
 */
static gboolean
update_terminal_size_idle (gpointer data)
{
  GLookPlayTab *tab = data;
  glong rows, cols;

  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), FALSE);

  glook_terminal_get_size (GLOOK_TERMINAL (tab->priv->terminal),
			   &cols, &rows);
  set_terminal_size_request (tab, rows, cols);

  return FALSE;
}

/**
 * on_term_scrolled_window_policy_changed:
 */
static void
on_term_scrolled_window_policy_changed (GtkScrolledWindow *scrolled,
					GParamSpec        *spec,
					GLookPlayTab      *tab)
{
  g_return_if_fail (GTK_IS_SCROLLED_WINDOW (scrolled));
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  g_idle_add (update_terminal_size_idle, tab);
}

/**
 * get_term_scroll_get_border_size:
 *
 * get terminal scrolled window border total size.
 *
 * border size includes scrolled window's border, scrollbar,
 * and so on.
 */
static void
terminal_scrolled_get_border_size (GLookPlayTab *tab,
				   gint         *width,
				   gint         *height)
{
  GtkAllocation alloc1, alloc2;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  /* get terminal scrolled window and terminal allocated size. */
  gtk_widget_get_allocation (tab->priv->term_scrolled, &alloc1);
  gtk_widget_get_allocation (tab->priv->terminal, &alloc2);

  /* terminal scrolled's border =
   * terminal scrolled allocated size - terminal allocaed size. */
  if (width)  *width  = alloc1.width  - alloc2.width;
  if (height) *height = alloc1.height - alloc2.height;
}

/**
 * get_box_scrolled_get_border_size:
 */
#if 0
static void
box_scrolled_get_border_size (GLookPlayTab *tab,
			      gint         *width,
			      gint         *height)
{
  GtkWidget *viewport;
  GtkAllocation alloc1, alloc2;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  viewport = gtk_bin_get_child (GTK_BIN (tab->priv->box_scrolled));

  /* get box scrolled window and viewport allocated size. */
  gtk_widget_get_allocation (tab->priv->box_scrolled, &alloc1);
  gtk_widget_get_allocation (viewport, &alloc2);

  /* box scrolled's border =
   * box scrolled allocated size - viewport allocaed size. */
  if (width)  *width  = alloc1.width  - alloc2.width;
  if (height) *height = alloc1.height - alloc2.height;
}
#endif

/**
 * get_terminal_maximum_size:
 */
static void
get_terminal_maximum_size (GLookPlayTab *tab,
			   gint         *width,
			   gint         *height)
{
  GtkWidget *window;
  GtkAllocation alloc;
  GdkRectangle window_frame, *workarea;
  gint scroll_width, scroll_height;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  workarea = glook_app_get_workarea (glook_app_get ());

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  if (!GLOOK_IS_WINDOW (window))
    {
      *width  = workarea->width;
      *height = workarea->height;
      return;
    }
  /* take window frame into account. */
  glook_window_get_frame_extents (GLOOK_WINDOW (window), &window_frame);
  gtk_widget_get_allocation (GTK_WIDGET (tab), &alloc);

  if (alloc.width > 1 || alloc.height > 1)
    {
      *width  = workarea->width - (window_frame.width - alloc.width);
      *height = workarea->height - (window_frame.height - alloc.height);
    }
  else
    {
      *width  = workarea->width;
      *height = workarea->height;
      return;
    }
  gtk_widget_get_allocation (tab->priv->control_box, &alloc);
  *height -= alloc.height;

  terminal_scrolled_get_border_size (tab, &scroll_width, &scroll_height);
  *width  -= scroll_width;
  *height -= scroll_height;
}

/**
 * resize_window_idle:
 *
 * resize toplevel window when terminal size changed.
 */
static gboolean
resize_window_idle (gpointer data)
{
  GLookPlayTab *tab = data;
  GtkWidget *window;
  GdkWindow *gdkwindow;

  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), FALSE);

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  if (!GTK_IS_WINDOW (window))
    {
      tab->priv->resize_idle_id = 0;
      return FALSE;
    }
  gdkwindow = gtk_widget_get_window (window);
  if (gdkwindow && GDK_IS_WINDOW (gdkwindow))
    {
      GdkWindowState state = gdk_window_get_state (gdkwindow);

      if ((state & (GDK_WINDOW_STATE_MAXIMIZED |
		    GDK_WINDOW_STATE_FULLSCREEN)) == 0)
	{
	  gtk_window_resize (GTK_WINDOW (window), 1, 1);
	  gtk_widget_queue_resize (GTK_WIDGET (window));
	}
    }
  gtk_widget_grab_focus (tab->priv->terminal);

  tab->priv->resize_idle_id = 0;
  return FALSE;
}

/**
 * set_terminal_size_request:
 * 
 * set terminal request. the terminal will has a fixed
 * minimum size(that is not GTK+ suggest).
 */
static void
set_terminal_size_request (GLookPlayTab *tab,
			   guint16       rows,
			   guint16       cols)
{
  VteTerminal *terminal;
  glong char_width, char_height;
  GtkBorder *inner_border;
  gint term_width, term_height, max_width = 0, max_height = 0;
  gint border_width, border_height, scroll_width, scroll_height;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  terminal = VTE_TERMINAL (tab->priv->terminal);

  glook_terminal_get_char_size (GLOOK_TERMINAL (terminal),
				&char_width, &char_height);
  gtk_widget_style_get (GTK_WIDGET (terminal), "inner-border",
			&inner_border, NULL);

  /* get terminal request width and height. */
  term_width  = char_width  * cols;
  term_height = char_height * rows;
  if (inner_border != NULL)
    {
      term_width  += inner_border->left + inner_border->right;
      term_height += inner_border->top  + inner_border->bottom;
      gtk_border_free (inner_border);
    }
  /* minimum size request. */
  gtk_widget_set_size_request (GTK_WIDGET (terminal), term_width, term_height);
  vte_terminal_set_size (terminal, cols, rows); 

  /* if terminal request size large than desktop workarea size.
   * let box scroll window work. */
  get_terminal_maximum_size (tab, &max_width, &max_height);
  if (term_width > max_width || term_height > max_height)
    {
      GtkPolicyType hpolicy = GTK_POLICY_NEVER;
      GtkPolicyType vpolicy = GTK_POLICY_NEVER;

      if (term_width > max_width)
	hpolicy = GTK_POLICY_ALWAYS;

      if (term_height > max_height)
	vpolicy = GTK_POLICY_ALWAYS;

      g_object_set (tab->priv->box_scrolled, "hscrollbar-policy", hpolicy,
		    "vscrollbar-policy", vpolicy, NULL);

      scroll_width  = MIN (max_width, term_width);
      scroll_height = MIN (max_height, term_height);

      terminal_scrolled_get_border_size (tab, &border_width, &border_height);
      scroll_width  += border_width;
      scroll_height += border_height;
    }
  else
    {
      gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (tab->priv->box_scrolled),
				      GTK_POLICY_NEVER, GTK_POLICY_NEVER);
      scroll_width  = term_width;
      scroll_height = term_height;

      terminal_scrolled_get_border_size (tab, &border_width, &border_height);
      scroll_width  += border_width;
      scroll_height += border_height;
    }
  gtk_widget_set_size_request (tab->priv->box_scrolled, scroll_width, scroll_height);

  glook_play_tab_geometry_hint (GLOOK_TAB (tab));

  /* update window size to fit new terminal size. */
  if (tab->priv->resize_idle_id > 0)
    g_source_remove (tab->priv->resize_idle_id);

  tab->priv->resize_idle_id =
   g_idle_add_full (G_PRIORITY_HIGH_IDLE, resize_window_idle,
		    g_object_ref (tab), g_object_unref);
}

/**
 * glook_play_tab_set_encoding_ui:
 */
static void
glook_play_tab_set_encoding_ui (GLookPlayTab *tab,
				const gchar  *encoding_ui)
{
  gchar *new_ui;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  new_ui = g_strdup (encoding_ui);
  g_free (tab->priv->encoding_ui);
  tab->priv->encoding_ui = new_ui;

  /* remerge encoding ui. */
  if (tab->priv->encoding_merge_id > 0)
    {
      GtkWidget *window = gtk_widget_get_toplevel (GTK_WIDGET (tab));

      glook_window_unmerge_ui (GLOOK_WINDOW (window), NULL,
			       tab->priv->encoding_merge_id);
      if (tab->priv->encoding_ui)
	{
	  tab->priv->encoding_merge_id =
	   glook_window_merge_ui (GLOOK_WINDOW (window),
				  tab->priv->encoding_actions, NULL,
				  tab->priv->encoding_ui, -1);
	  if (tab->priv->encoding_merge_id < 0)
	    g_critical ("Failed to merge look tab encoding ui.");
	}
    }
}

/**
 * glook_play_tab_get_encoding_ui:
 */
const gchar*
glook_play_tab_get_encoding_ui (GLookPlayTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), NULL);
  return tab->priv->encoding_ui;
}

/**
 * remove_encoding_action:
 */
static void
remove_encoding_action (GtkAction      *action,
			GtkActionGroup *group)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GTK_IS_ACTION_GROUP (group));

  gtk_action_group_remove_action (group, action);
}

/**
 * on_encoding_action_activated:
 */
static void
on_encoding_action_activated (GtkAction    *action,
			      GLookPlayTab *tab)
{
  const gchar *encoding;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  encoding = g_object_get_data (G_OBJECT (action), "encoding");
  vte_terminal_set_encoding (VTE_TERMINAL (tab->priv->terminal), encoding);
}

/**
 * glook_play_tab_build_encoding_ui:
 *
 * convert selected-encoding settings to a mergable ui string.
 */
gboolean
glook_play_tab_build_encoding_ui (GValue   *value,
				  GVariant *variant,
				  gpointer  user_data)
{
  GLookPlayTab *tab = user_data;
  GList *actions;
  GString *ui_string;
  GVariantIter iter;
  gchar *language, *encoding;

  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), FALSE);

  /* clear encoding action group first. */
  actions = gtk_action_group_list_actions (tab->priv->encoding_actions);
  if (actions)
    {
      g_list_foreach (actions, (GFunc) remove_encoding_action,
		      tab->priv->encoding_actions);
      g_list_free (actions);
    }
  /* build encoding ui string. */
  ui_string = g_string_new ("<menubar name='Menubar'>"
			    "<menu action='TabMenu'>"
			    "<placeholder name='TabMenuPlaceholder'>"
			    "<menu action='TerminalEncodingMenu'>"
			    "<placeholder name='EncodingMenuPlaceholder'>");

  /* create encoding action and insert into encoding action group. */
  g_variant_iter_init (&iter, variant);
  while (g_variant_iter_next (&iter, "{ss}", &language, &encoding))
    {
      GtkAction *action;
      gchar name[256], label[512];

      g_snprintf (name, sizeof (name), "%sAction", encoding);
      g_snprintf (label, sizeof (label), "%s (_%s)", language, encoding);

      action = gtk_action_new (name, label, NULL, NULL);

      /* add ui string. */
      g_string_append_printf (ui_string, "<menuitem action='%s'/>", name);

      /* associated encoding with action. */
      g_object_set_data_full (G_OBJECT (action), "encoding", encoding, g_free);
      g_free (language);

      /* when action activated, changed terminal encoding. */
      g_signal_connect (action, "activate",
			G_CALLBACK (on_encoding_action_activated), tab);

      gtk_action_group_add_action (tab->priv->encoding_actions, action);
      g_object_unref (action);
    }
  g_string_append (ui_string, "</placeholder></menu></placeholder></menu></menubar>");

  g_value_set_string (value, ui_string->str);
  g_string_free (ui_string, TRUE);

  return TRUE;
}

/**
 * on_play_tab_mapped:
 */
static void
on_play_tab_mapped (GLookPlayTab *tab,
		    gpointer      user_data)
{
  GSettings *settings;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  /* disconnect to prevent bind more than once. */
  g_signal_handlers_disconnect_by_func (tab, on_play_tab_mapped, NULL);

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_GENERAL);

  /* when selected-encoding settings changed, we need rebuild
   * the encoding ui. */
  g_settings_bind_with_mapping (settings, "selected-encodings",
				tab, "encoding-ui", G_SETTINGS_BIND_GET,
				glook_play_tab_build_encoding_ui,
				NULL, tab, NULL);
  set_terminal_size_request (tab, 24, 80);
}

/**
 * glook_play_tab_fill_file_info:
 */
static inline void
glook_play_tab_fill_file_info (GLookPlayTab *tab,
			       const gchar  *filename,
			       GtkWidget    *dialog)
{
  GtkBuilder *builder;
  TRecordHead *head;
  GObject *object;
  time_t tim;
  struct tm *tmp;
  gchar buff[128], *p;
  gint n;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));
  g_return_if_fail (GTK_IS_DIALOG (dialog));

  builder = g_object_get_data (G_OBJECT (dialog), "builder");
  head = t_decoder_get_head (tab->priv->decoder);

  object = gtk_builder_get_object (builder, "Filename");
  p = g_path_get_basename (filename);
  gtk_label_set_text (GTK_LABEL (object), p);
  g_free (p);

  object = gtk_builder_get_object (builder, "Creator");
  gtk_label_set_text (GTK_LABEL (object), head->creator);

  tim = (time_t) head->starting;
  tmp = localtime (&tim);
  strftime (buff, sizeof (buff), "%x %X", tmp);
  object = gtk_builder_get_object (builder, "Starting");
  gtk_label_set_text (GTK_LABEL (object), buff);

  tim = (time_t) head->duration;
  tmp = gmtime (&tim);
  n = g_snprintf (buff, sizeof (buff), "%d ", tmp->tm_yday);
  strftime (buff + n, sizeof (buff) - n, "%X", tmp);
  object = gtk_builder_get_object (builder, "Duration");
  gtk_label_set_text (GTK_LABEL (object), buff);

  object = gtk_builder_get_object (builder, "Version");
  g_snprintf (buff, sizeof (buff), "%d", head->version);
  gtk_label_set_text (GTK_LABEL (object), buff);

  object = gtk_builder_get_object (builder, "Charset");
  gtk_label_set_text (GTK_LABEL (object), head->charset);

  object = gtk_builder_get_object (builder, "Compress");
  g_snprintf (buff, sizeof (buff), "%d", head->clevel);
  gtk_label_set_text (GTK_LABEL (object), buff);

  object = gtk_builder_get_object (builder, "Terminal");
  g_snprintf (buff, sizeof (buff), "%s  %d x %d",
	      head->term, head->ncols, head->nrows);
  gtk_label_set_text (GTK_LABEL (object), buff);

  object = gtk_builder_get_object (builder, "Encrypt");
  gtk_label_set_text (GTK_LABEL (object),
		      head->encrypted ? _("Yes") : _("No"));
}

/**
 * on_terminal_fileinfo_activated:
 */
static void
on_terminal_fileinfo_activated (GtkAction    *action,
				GLookPlayTab *tab)
{
  static GObject *dialog;
  const gchar *filename;
  GtkWidget *window;
  GtkBuilder *builder;
  GError *error = NULL;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  filename = t_decoder_get_filename (tab->priv->decoder);
  if (!filename)
    return;

  /* make sure has only one dialog exists. */
  if (dialog)
    {
      glook_play_tab_fill_file_info (tab, filename,
				     GTK_WIDGET (dialog));
      gtk_window_present (GTK_WINDOW (dialog));
      return;
    }
  builder = gtk_builder_new ();

  gtk_builder_add_from_file (builder, FILE_INFO_GLADE_FILE, &error);
  if (error)
    {
      g_critical ("%s", error->message);
      g_error_free (error);
      g_object_unref (builder);
      return;
    }
  dialog = gtk_builder_get_object (builder, "Dialog");
  g_object_set_data_full (dialog, "builder", builder,
			  g_object_unref);

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  gtk_window_set_transient_for (GTK_WINDOW (dialog),
				GTK_WINDOW (window));

  glook_play_tab_fill_file_info (tab, filename,
				 GTK_WIDGET (dialog));

  g_signal_connect (dialog, "response",
		    G_CALLBACK (gtk_widget_destroy), NULL);
  g_signal_connect (dialog, "destroy",
		    G_CALLBACK (gtk_widget_destroyed), &dialog);

  gtk_widget_show_all (GTK_WIDGET (dialog));
}

/**
 * on_terminal_playback_activated:
 */
static void
on_terminal_playback_activated (GtkAction    *action,
				GLookPlayTab *tab)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  if (!glook_play_tab_is_active (GLOOK_TAB (tab)))
    return;

  if (tab->priv->paused)	/* will play */
    {
      gtk_action_set_stock_id (action, GTK_STOCK_MEDIA_PAUSE);

      glook_tab_set_stock (GLOOK_TAB (tab), GTK_STOCK_MEDIA_PLAY);
      gtk_label_set_text (GTK_LABEL (tab->priv->status_state), _("Playing"));

      tab->priv->paused = FALSE;

      if (tab->priv->play_idle == 0)
	{
	  tab->priv->play_idle =
	   g_idle_add_full (G_PRIORITY_HIGH_IDLE,
			    glook_play_tab_play_idle, tab,
			    glook_play_tab_play_idle_destroy);
	}
    }
  else	/* will paused. */
    {
      gtk_action_set_stock_id (action, GTK_STOCK_MEDIA_PLAY);
      glook_tab_set_stock (GLOOK_TAB (tab), GTK_STOCK_MEDIA_PAUSE);
      gtk_label_set_text (GTK_LABEL (tab->priv->status_state), _("Paused"));

      if (tab->priv->play_idle > 0)
	g_source_remove (tab->priv->play_idle);

      if (tab->priv->play_timeout > 0)
	g_source_remove (tab->priv->play_timeout);

      tab->priv->paused = TRUE;
    }
}

/**
 * on_terminal_forward_activated:
 */
static void
on_terminal_forward_activated (GtkAction    *action,
			       GLookPlayTab *tab)
{
  gint value;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  if (!glook_play_tab_is_active (GLOOK_TAB (tab)))
    return;

  value = gtk_adjustment_get_value (tab->priv->elapsed_adjust);
  value += gtk_adjustment_get_page_increment (tab->priv->elapsed_adjust);
  value = MIN (value, gtk_adjustment_get_upper (tab->priv->elapsed_adjust));
  gtk_adjustment_set_value (tab->priv->elapsed_adjust, value);
}

/**
 * on_terminal_backward_activated:
 */
static void
on_terminal_backward_activated (GtkAction    *action,
				GLookPlayTab *tab)
{
  gint value;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  if (!glook_play_tab_is_active (GLOOK_TAB (tab)))
    return;

  value = gtk_adjustment_get_value (tab->priv->elapsed_adjust);
  value -= gtk_adjustment_get_page_increment (tab->priv->elapsed_adjust);
  gtk_adjustment_set_value (tab->priv->elapsed_adjust, MAX (value, 0));
}

/**
 * on_terminal_stop_activated:
 */
static void
on_terminal_stop_activated (GtkAction    *action,
			    GLookPlayTab *tab)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  glook_play_tab_stop (GLOOK_TAB (tab));
}

/**
 * unref_popup_info_idle:
 */
static gboolean
unref_popup_info_idle (gpointer data)
{
  GLookPlayTab *tab = data;

  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), FALSE);

  g_object_set_data (G_OBJECT (tab), "popup-info", NULL);
  return FALSE;
}

/**
 * popup_menu_deactivated:
 */
static void
popup_menu_deactivated (GtkWidget    *menu,
			GLookPlayTab *tab)
{
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  /* unref the popup info from a idle since we still need it
   * from the action callbacks which will run before idle. */
  g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
		   unref_popup_info_idle,
		   g_object_ref (tab), g_object_unref);

  /* disconnect the handler. */
  g_signal_handlers_disconnect_by_func (menu, popup_menu_deactivated, tab);
}

/**
 * popup_menu_clipboard_targets_received:
 */
static void
popup_menu_clipboard_targets_received (GtkClipboard *clipboard,
				       GdkAtom      *targets,
				       gint          ntargets,
				       gpointer      data)
{
  GLookPlayTab *tab = GLOOK_PLAY_TAB (data);
  GLookTerminalPopupInfo *info;
  GtkAction *open_link_action;
  GtkAction *copy_link_action;
  GtkAction *mailto_action;
  GtkAction *callto_action;
  GtkAction *action;
  GtkWidget *menu;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  info = g_object_get_data (G_OBJECT (tab), "popup-info");

  /* change link actions visible. */
  open_link_action = gtk_action_group_get_action (tab->priv->ui_actions,
						  "TerminalOpenLink");
  copy_link_action = gtk_action_group_get_action (tab->priv->ui_actions,
						  "TerminalCopyLink");
  mailto_action = gtk_action_group_get_action (tab->priv->ui_actions,
					       "TerminalMailto");
  callto_action = gtk_action_group_get_action (tab->priv->ui_actions,
					       "TerminalCallto");
  gtk_action_set_visible (open_link_action, FALSE);
  gtk_action_set_visible (copy_link_action, FALSE);
  gtk_action_set_visible (mailto_action, FALSE);
  gtk_action_set_visible (callto_action, FALSE);

  if (info->address)
    {
      switch (info->uri_type)
	{
	case TERMINAL_URI_TYPE_URI:
	case TERMINAL_URI_TYPE_WEB:
	  gtk_action_set_visible (open_link_action, TRUE);
	  gtk_action_set_visible (copy_link_action, TRUE);
	  break;
	case TERMINAL_URI_TYPE_MAIL:
	  gtk_action_set_visible (mailto_action, TRUE);
	  gtk_action_set_visible (copy_link_action, TRUE);
	  break;
	case TERMINAL_URI_TYPE_CALL:
	  gtk_action_set_visible (callto_action, TRUE);
	  gtk_action_set_visible (copy_link_action, TRUE);
	  break;
	default:
	  break;
	}
    }
  /* change copy/paste action sensitive. */
  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalCopy");
  if (vte_terminal_get_has_selection (VTE_TERMINAL (tab->priv->terminal)))
    gtk_action_set_sensitive (action, TRUE);
  else
    gtk_action_set_sensitive (action, FALSE);

  menu = gtk_ui_manager_get_widget (tab->priv->popup_menu_ui,
				    "/TerminalPopupMenu");
  g_signal_connect (menu, "deactivate",
		    G_CALLBACK (popup_menu_deactivated), tab);

  /* select the first item when popup activate by accelerator. */
  if (info->button == 0)
    gtk_menu_shell_select_first (GTK_MENU_SHELL (menu), FALSE);

  gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 
		  info->button, info->timestamp);
}

/**
 * on_terminal_show_popup_menu:
 */
static void
on_terminal_show_popup_menu (GLookTerminal          *terminal,
			     GLookTerminalPopupInfo *info,
			     GLookPlayTab           *tab)
{
  GtkClipboard *clipboard;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  g_object_set_data_full (G_OBJECT (tab), "popup-info",
			  glook_terminal_popup_info_ref (info),
			  (GDestroyNotify) glook_terminal_popup_info_unref);

  clipboard = gtk_widget_get_clipboard (GTK_WIDGET (terminal),
					GDK_SELECTION_CLIPBOARD);
  gtk_clipboard_request_targets (clipboard,
				 popup_menu_clipboard_targets_received, tab);
}

/**
 * on_terminal_size_changed:
 */
static void
on_terminal_size_changed (GLookTerminal *terminal,
			  gint           rows,
			  gint           cols,
			  GLookPlayTab  *tab)
{
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  set_terminal_size_request (tab, rows, cols);
}

/**
 * on_terminal_font_desc_notify:
 */
static void
on_terminal_font_desc_notify (GLookTerminal *terminal,
			      GParamSpec    *spec,
			      GLookPlayTab  *tab)
{
  glong rows, cols;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  glook_terminal_get_size (terminal, &cols, &rows);
  set_terminal_size_request (tab, rows, cols);
}

/**
 * on_password_entry_changed:
 */
static void
on_password_entry_changed (GtkEntry  *entry,
			   GtkWidget *dialog)
{
  gint n;

  g_return_if_fail (GTK_IS_ENTRY (entry));

  n = gtk_entry_get_text_length (entry);
  gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog),
				     GTK_RESPONSE_ACCEPT, (n > 0));
  g_object_set (entry, "secondary-icon-name", NULL, NULL);
}

/**
 * glook_play_tab_set_passwd:
 *
 * get password from user by dialog.
 */
static gboolean
glook_play_tab_set_passwd (TDecoder     *decoder,
			   gpointer      data)
{
  GLookPlayTab *tab = GLOOK_PLAY_TAB (data);
  GtkWidget *window;
  GtkWidget *dialog;
  GtkWidget *vbox;
  GtkWidget *label;
  GtkWidget *passwd_entry;
  gboolean ret = FALSE;

  g_return_val_if_fail (T_IS_DECODER (decoder), FALSE);
  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), FALSE);

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  dialog = gtk_dialog_new_with_buttons (_("Deocde password"),
					GTK_WINDOW (window),
					GTK_DIALOG_MODAL |
					GTK_DIALOG_NO_SEPARATOR |
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
					NULL);
  g_object_set (dialog, "border-width", 8, "default-width", 300,
		"default-height", 150, NULL);
  gtk_dialog_set_default_response (GTK_DIALOG (dialog),
				   GTK_RESPONSE_ACCEPT);
  gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog),
				     GTK_RESPONSE_ACCEPT, FALSE);

  vbox = gtk_dialog_get_content_area (GTK_DIALOG (dialog));

  label = gtk_label_new (_("Please give decode password:"));
  g_object_set (label, "xalign", 0.0, NULL);
  gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, TRUE, 0);

  passwd_entry = gtk_entry_new ();
  g_object_set (passwd_entry, "visibility", FALSE,
		"activates-default", TRUE,
		"primary-icon-name", GTK_STOCK_DIALOG_AUTHENTICATION,
		"primary-icon-activatable", FALSE, NULL);
  g_signal_connect (passwd_entry, "changed",
		    G_CALLBACK (on_password_entry_changed), dialog);
  gtk_box_pack_start (GTK_BOX (vbox), passwd_entry, FALSE, TRUE, 6);

  gtk_widget_show_all (dialog);

  /* loop until user give corrent password or cancel the dialog. */
  while (1)
    {
      const gchar *text;

      if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_ACCEPT)
	break;

      text = gtk_entry_get_text (GTK_ENTRY (passwd_entry));
      if (t_decoder_set_passwd (decoder, text, -1))
	{
	  ret = TRUE;
	  break;
	}
      /* set password entry and try again. */
      g_object_set (passwd_entry, "text", "",
		    "secondary-icon-name", GTK_STOCK_DIALOG_WARNING,
		    "secondary-icon-tooltip-text", _("Password incorrent"),
		    "secondary-icon-activatable", FALSE, NULL);
      gtk_widget_grab_focus (passwd_entry);
    }
  gtk_widget_destroy (dialog);

  return ret;
}

/**
 * glook_play_add_recent_item:
 *
 * add open file to recent manager.
 */
static void
glook_play_add_recent_item (const gchar *path)
{
  GtkRecentManager *manager;
  GtkRecentData data;
  gchar *groups[] = { "Ttyutils", NULL };
  gchar *uri;

  data.display_name = g_path_get_basename (path);
  data.description = NULL;
  data.mime_type = "application/x-ttyrcd";
  data.app_name = (gchar*) g_get_application_name ();
  data.app_exec = "ttyglook %u";
  data.groups = groups;
  data.is_private = FALSE;

  manager = gtk_recent_manager_get_default ();
  uri = g_strdup_printf ("file://%s", path);
  gtk_recent_manager_add_full (manager, uri, &data);
  g_free (uri);
  g_free (data.display_name);
}

/**
 * glook_play_get_path_from_uri:
 */
static gchar*
glook_play_get_path_from_uri (GLookUri *uri)
{
  GFile *file = NULL;
  const gchar *p;
  gchar *path;

  /* get playback filename from uri. */
  p = glook_uri_lookup_attr (uri, "uri");
  if (p)
    {
      file = g_file_new_for_uri (p);
      goto next;
    }
  p = glook_uri_lookup_attr (uri, "path");
  if (p)
    file = g_file_new_for_path (p);

next:
  if (!file)
    {
      g_warning (_("No file to playback"));
      return NULL;
    }
  /* can play local file only. */
  path = g_file_get_path (file);
  g_object_unref (file);

  if (!path)
    {
      g_warning (_("No file to playback"));
      return NULL;
    }
  return path;
}

/**
 * glook_play_tab_play_idle:
 */
static gboolean
glook_play_tab_play_idle (gpointer user_data)
{
  GLookPlayTab *tab = user_data;
  TRecordIndex *indexes;
  guint32 istart, iend, index;
  guint interval;
  GtkSpinButton *spin;
  gint value;

  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), FALSE);

  indexes = t_decoder_get_indexes (tab->priv->decoder,
				   &istart, &iend);
  g_assert (indexes);

  index = tab->priv->icurr - istart;
  interval = indexes[index].interval * 1000;

  /* don't wait for too long. */
  spin = GTK_SPIN_BUTTON (tab->priv->interval_spin);
  value = gtk_spin_button_get_value_as_int (spin);
  if (value > 0)
    interval = MIN (interval, value);
  else if (value < 0)
    interval = MAX (interval, ABS (value));

  if (tab->priv->play_timeout == 0)
    {
      tab->priv->play_timeout =
       g_timeout_add_full (G_PRIORITY_DEFAULT, interval,
			   glook_play_tab_play_timeout, tab,
			   glook_play_tab_play_timeout_destroy);
    }
  return FALSE;
}

/**
 * glook_play_tab_play_idle_destroy:
 */
static void
glook_play_tab_play_idle_destroy (gpointer user_data)
{
  GLookPlayTab *tab = user_data;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  tab->priv->play_idle = 0;
}

/**
 * update_status_elapsed:
 */
static void
update_status_elapsed (GLookPlayTab *tab)
{
  time_t elapsed;
  struct tm *tmp;
  gchar buff[64];
  guint32 nindexes;
  GError *error = NULL;

  nindexes = t_decoder_get_nindexes (tab->priv->decoder);
  if (tab->priv->icurr >= nindexes)
    return;

  elapsed = t_decoder_get_index_elapsed (tab->priv->decoder,
					 tab->priv->icurr,
					 &error);
  if (error)
    {
      g_warning ("%s", error->message);
      g_error_free (error);
      return;
    }
  tmp = gmtime (&elapsed);
  strftime (buff, sizeof (buff), "%X", tmp);

  gtk_label_set_text (GTK_LABEL (tab->priv->status_elapsed), buff);
}

/**
 * glook_play_tab_play_timeout:
 */
static gboolean
glook_play_tab_play_timeout (gpointer user_data)
{
  GLookPlayTab *tab = user_data;
  const gchar *data;
  guint16 size;
  guint32 nindexes;
  GError *error = NULL;

  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), FALSE);

  nindexes = t_decoder_get_nindexes (tab->priv->decoder);

  /* end of file. */
  if (tab->priv->icurr >= nindexes)
    {
      glook_play_tab_stop (GLOOK_TAB (tab));

      glook_terminal_show_play_over (GLOOK_TERMINAL (tab->priv->terminal),
				     g_timer_elapsed (tab->priv->timer, NULL));
      return FALSE;
    }
  data = t_decoder_get_index_data (tab->priv->decoder,
				   tab->priv->icurr,
				   &size, &error);
  if (error)
    {
      g_warning ("%s", error->message);
      g_error_free (error);
      return -1;
    }
  vte_terminal_feed (VTE_TERMINAL (tab->priv->terminal), data, size);

  tab->priv->icurr++;

  gtk_adjustment_set_value (tab->priv->elapsed_adjust,
			    (gdouble) tab->priv->icurr);
  update_status_elapsed (tab);

  if (!tab->priv->paused)
    {
      if (tab->priv->play_idle == 0)
	{
	  tab->priv->play_idle =
	   g_idle_add_full (G_PRIORITY_HIGH_IDLE,
			    glook_play_tab_play_idle, tab,
			    glook_play_tab_play_idle_destroy);
	}
    }
  return FALSE;
}

/**
 * glook_play_tab_play_timeout_destroy:
 */
static void
glook_play_tab_play_timeout_destroy (gpointer user_data)
{
  GLookPlayTab *tab = user_data;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  tab->priv->play_timeout = 0;
}

/**
 * glook_play_tab_status_timer_timeout:
 */
static gboolean
glook_play_tab_status_timer_timeout (gpointer user_data)
{
  GLookPlayTab *tab = user_data;
  time_t elapsed;
  struct tm *tmp;
  gchar buff[64];

  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), FALSE);

  elapsed = g_timer_elapsed (tab->priv->timer, NULL);
  tmp = gmtime (&elapsed);
  strftime (buff, sizeof (buff), "%X", tmp);

  gtk_label_set_text (GTK_LABEL (tab->priv->status_timer), buff);
  return TRUE;
}

/**
 * glook_play_tab_start_play:
 */
static void
glook_play_tab_start_play (GLookPlayTab *tab,
			   const gchar  *path)
{
  gchar *basename;
  GtkAction *action;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  tab->priv->paused = FALSE;

  /* update status timer. */
  g_timer_start (tab->priv->timer);
  tab->priv->status_timer_timeout =
   g_timeout_add_seconds (1, glook_play_tab_status_timer_timeout, tab);

  /* update tab title and icon. */
  basename = g_path_get_basename (path);
  g_object_set (tab, "stock", GTK_STOCK_MEDIA_PLAY,
		"label", basename, NULL);
  g_free (basename);

  /* change playback action image. */
  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalPlayback");
  gtk_action_set_stock_id (action, GTK_STOCK_MEDIA_PAUSE);

  gtk_label_set_text (GTK_LABEL (tab->priv->status_state), _("Playing"));

  /* playback. */
  tab->priv->play_idle =
   g_idle_add_full (G_PRIORITY_HIGH_IDLE,
		    glook_play_tab_play_idle, tab,
		    glook_play_tab_play_idle_destroy);
}

/**
 * apply_terminal_setting_idle:
 */
static gboolean
apply_terminal_setting_idle (gpointer data)
{
  GLookPlayTab *tab = GLOOK_PLAY_TAB (data);
  TRecordHead *head;

  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), FALSE);

  head = t_decoder_get_head (tab->priv->decoder);

#if 0
  /* we cannot modify emulation unless install other
   * termcaps file in vte search path or /etc/termcap
   * file is present(this is obsolescence by terminfo).
   */
  vte_terminal_set_emulation (terminal), info->term);
#endif
  vte_terminal_set_encoding (VTE_TERMINAL (tab->priv->terminal),
			     head->charset);
  set_terminal_size_request (tab, head->nrows, head->ncols);

  return FALSE;
}

/**
 * update_control_actions_sensitive:
 */
static void
update_control_actions_sensitive (GLookPlayTab *tab,
				  gboolean      sensitive)
{
  GtkAction *action;

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalPlayback");
  gtk_action_set_sensitive (action, sensitive);
  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalForward");
  gtk_action_set_sensitive (action, sensitive);
  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalBackward");
  gtk_action_set_sensitive (action, sensitive);
  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalStop");
  gtk_action_set_sensitive (action, sensitive);
  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalFileInfo");
  gtk_action_set_sensitive (action, sensitive);
}

/**
 * glook_play_tab_open:
 */
static void
glook_play_tab_open (GLookTab *tab)
{
  GLookPlayTab *play_tab = GLOOK_PLAY_TAB (tab);
  GLookUri *uri;
  gchar *path;
  guint32 nindexes;
  GError *error = NULL;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  uri = glook_tab_get_uri (tab);

  /* get playback filename. */
  if (!(path = glook_play_get_path_from_uri (uri)))
    return;

  /* stop current playback. */
  glook_play_tab_stop (tab);

  /* try to open the file. */
  t_decoder_open (play_tab->priv->decoder, path,
		  glook_play_tab_set_passwd,
		  g_object_ref (play_tab),
		  g_object_unref, &error);
  if (error)
    {
      g_free (path);

      /* remove uri from recent manager. */
      if (error->code != G_IO_ERROR_PERMISSION_DENIED)
	{
	  path = (gchar*) glook_uri_lookup_attr (uri, "uri");
	  if (path)
	    {
	      GtkRecentManager *manager;
	      manager = gtk_recent_manager_get_default ();
	      gtk_recent_manager_remove_item (manager, path, NULL);
	    }
	}
      g_warning ("%s", error->message);
      g_error_free (error);
      return;
    }
  /* add the path to recent manager. */
  glook_play_add_recent_item (path);

  /* update elapsed adjustment. */
  nindexes = t_decoder_get_nindexes (play_tab->priv->decoder);
  gtk_adjustment_configure (play_tab->priv->elapsed_adjust, 0, 0,
			    nindexes, 1, MAX (1, nindexes / 50), 0);

  /* apply terminal settings. */
  g_idle_add_full (G_PRIORITY_DEFAULT_IDLE, apply_terminal_setting_idle,
		   g_object_ref (tab), g_object_unref);

  /* update widget sensitive state. */
  gtk_widget_set_sensitive (play_tab->priv->control_box, TRUE);
  update_control_actions_sensitive (play_tab, TRUE); 

  /* start playback. */
  glook_play_tab_start_play (play_tab, path);
  g_free (path);

  /* update tab state. */
  g_signal_emit_by_name (tab, "finished");
}

/**
 * glook_play_tab_stop:
 */
static void
glook_play_tab_stop (GLookTab *tab)
{
  GLookPlayTab *play_tab = GLOOK_PLAY_TAB (tab);
  GtkAction *action;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  play_tab->priv->icurr = 0;

  if (play_tab->priv->status_timer_timeout > 0)
    {
      g_source_remove (play_tab->priv->status_timer_timeout);
      play_tab->priv->status_timer_timeout = 0;
    }
  g_timer_stop (play_tab->priv->timer);

  t_decoder_close (play_tab->priv->decoder);

  vte_terminal_reset (VTE_TERMINAL (play_tab->priv->terminal), TRUE, TRUE);

  g_object_set (tab, "stock", GTK_STOCK_MEDIA_STOP, NULL);

  action = gtk_action_group_get_action (play_tab->priv->ui_actions,
					"TerminalPlayback");
  gtk_action_set_stock_id (action, GTK_STOCK_MEDIA_PLAY);
  gtk_label_set_text (GTK_LABEL (play_tab->priv->status_state), _("Stop"));

  gtk_adjustment_configure (play_tab->priv->elapsed_adjust,
			    0, 0, 0, 0, 0, 0);
  gtk_widget_set_sensitive (play_tab->priv->control_box, FALSE);

  update_control_actions_sensitive (play_tab, FALSE); 
}

/**
 * glook_play_tab_is_active:
 */
static gboolean
glook_play_tab_is_active (GLookTab *tab)
{
  GLookPlayTab *play_tab = GLOOK_PLAY_TAB (tab);

  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), FALSE);

  if (t_decoder_get_filename (play_tab->priv->decoder))
    return TRUE;

  return FALSE;
}

/**
 * glook_play_tab_reload:
 */
static void
glook_play_tab_reload (GLookTab *tab)
{
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  if (glook_play_tab_is_active (tab))
    return;

  glook_play_tab_open (tab);
}

/**
 * glook_play_tab_zoom:
 *
 * override parent method.
 */
static gboolean
glook_play_tab_zoom (GLookTab *tab,
		     gint      factor)
{
  GLookPlayTab *play_tab = GLOOK_PLAY_TAB (tab);
  GLookTerminal *terminal;
  glong rows, cols;

  g_return_val_if_fail (GLOOK_IS_PLAY_TAB (tab), FALSE);
  g_return_val_if_fail (factor > 0, FALSE);

  terminal = GLOOK_TERMINAL (play_tab->priv->terminal);

  /* change terminal font size. */
  glook_terminal_zoom (terminal, factor);

  /* changed terminal size request. */
  glook_terminal_get_size (terminal, &cols, &rows);
  set_terminal_size_request (play_tab, rows, cols);

  return TRUE;
}

/**
 * glook_play_tab_merge_ui:
 */
static void
glook_play_tab_merge_ui (GLookTab  *tab,
			 GtkWidget *window)
{
  GLookPlayTab *play_tab = GLOOK_PLAY_TAB (tab);
  GtkWidget *statusbar;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));
  g_return_if_fail (GTK_IS_WINDOW (window));

  /* merge tab menu. */
  play_tab->priv->ui_merge_id =
   glook_window_merge_ui (GLOOK_WINDOW (window),
			  play_tab->priv->ui_actions,
			  play_tab->priv->ui_accel_group,
			  _tabmenu_ui_string, -1);
  if (play_tab->priv->ui_merge_id < 0)
    g_critical ("Failed to merge play tab ui.");

  /* merge encoding menu. */
  if (play_tab->priv->encoding_ui)
    {
      play_tab->priv->encoding_merge_id =
       glook_window_merge_ui (GLOOK_WINDOW (window),
			      play_tab->priv->encoding_actions, NULL,
			      play_tab->priv->encoding_ui, -1);
      if (play_tab->priv->encoding_merge_id < 0)
	g_critical ("Failed to merge look tab encoding ui.");
    }
  /* merge status box. */
  if (!gtk_widget_get_parent (play_tab->priv->status_box))
    {
      statusbar = glook_window_get_statusbar (GLOOK_WINDOW (window));
      gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR (statusbar), FALSE);
      gtk_box_pack_end (GTK_BOX (statusbar),
			play_tab->priv->status_box, TRUE, TRUE, 0);
    }
}

/**
 * glook_play_tab_unmerge_ui:
 */
static void
glook_play_tab_unmerge_ui (GLookTab  *tab,
			   GtkWidget *window)
{
  GLookPlayTab *play_tab = GLOOK_PLAY_TAB (tab);
  GtkWidget *statusbar;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));
  g_return_if_fail (GTK_IS_WINDOW (window));

  if (play_tab->priv->encoding_merge_id > 0)
    {
      glook_window_unmerge_ui (GLOOK_WINDOW (window), NULL,
			       play_tab->priv->encoding_merge_id);
      play_tab->priv->encoding_merge_id = 0;
    }
  if (play_tab->priv->ui_merge_id > 0)
    {
      glook_window_unmerge_ui (GLOOK_WINDOW (window),
			       play_tab->priv->ui_accel_group,
			       play_tab->priv->ui_merge_id);
      play_tab->priv->ui_merge_id = 0;
    }
  /* remove status box. */
  if (gtk_widget_get_parent (play_tab->priv->status_box))
    {
      statusbar = glook_window_get_statusbar (GLOOK_WINDOW (window));
      gtk_container_remove (GTK_CONTAINER (statusbar),
			    play_tab->priv->status_box);
      gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR (statusbar), TRUE);
    }
}

/**
 * glook_play_tab_geometry_hint:
 */
static void
glook_play_tab_geometry_hint (GLookTab *tab)
{
  GLookPlayTab *play_tab = GLOOK_PLAY_TAB (tab);
  VteTerminal *terminal;
  GtkWidget *window;
  GdkGeometry geometry;
  guint width, height;
  glong cols, rows;
  GtkBorder *inner_border;
  gint xpad = 0, ypad = 0;

  g_return_if_fail (GLOOK_IS_TAB (tab));

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  if (!gtk_widget_get_realized (window))
    return;

  terminal = VTE_TERMINAL (play_tab->priv->terminal);

  gtk_widget_style_get (GTK_WIDGET (terminal),
			"inner-border", &inner_border, NULL);
  if (inner_border)
    {
      xpad = inner_border->left + inner_border->right;
      ypad = inner_border->top + inner_border->bottom;
      gtk_border_free (inner_border);
    }
  width  = vte_terminal_get_char_width (terminal);
  height = vte_terminal_get_char_height (terminal);
  glook_terminal_get_size (GLOOK_TERMINAL (terminal), &cols, &rows);

  geometry.min_width   = xpad + width * cols;
  geometry.min_height  = ypad + height * rows;

  gtk_window_set_geometry_hints (GTK_WINDOW (window), GTK_WIDGET (terminal),
				 &geometry,
				 GDK_HINT_MIN_SIZE);
}

/**
 * glook_play_tab_forward_idle:
 */
static gboolean
glook_play_tab_forward_idle (gpointer user_data)
{
  GLookPlayTab *tab = user_data;
  VteTerminal *terminal;
  guint32 value, nindexes;
  const gchar *data;
  guint16 size;
  GError *error = NULL;

  nindexes = t_decoder_get_nindexes (tab->priv->decoder);
  if (tab->priv->icurr >= nindexes)
    return FALSE;

  data = t_decoder_get_index_data (tab->priv->decoder,
				   tab->priv->icurr,
				   &size, &error);
  if (error)
    {
      g_warning ("%s", error->message);
      g_error_free (error);
      return FALSE;
    }
  terminal = VTE_TERMINAL (tab->priv->terminal);
  vte_terminal_feed (terminal, data, size);

  tab->priv->icurr++;

  value = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (tab),
					       "forward-value"));
  return (tab->priv->icurr < value);
}

/**
 * glook_play_tab_forward_idle_destroy:
 */
static void
glook_play_tab_forward_idle_destroy (gpointer user_data)
{
  GLookPlayTab *tab = user_data;

  g_object_set_data (G_OBJECT (tab), "forward-value",
		     GUINT_TO_POINTER (0));
  g_object_set_data (G_OBJECT (tab), "forward-idle-id",
		     GUINT_TO_POINTER (0));

  update_status_elapsed (tab);

  if (!tab->priv->paused)
    {
      if (tab->priv->play_idle == 0)
	{
	  tab->priv->play_idle =
	   g_idle_add_full (G_PRIORITY_HIGH_IDLE,
			    glook_play_tab_play_idle, tab,
			    glook_play_tab_play_idle_destroy);
	}
    }
}

/**
 * glook_play_tab_forward:
 */
static void
glook_play_tab_forward (GLookPlayTab *tab,
			guint32       value)
{
  guint id;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  id = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (tab),
					    "forward-idle-id"));
  if (id > 0)
    return;

  g_object_set_data (G_OBJECT (tab), "forward-value",
		     GUINT_TO_POINTER (value));
  id = g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
			glook_play_tab_forward_idle, tab,
			glook_play_tab_forward_idle_destroy);
  g_object_set_data (G_OBJECT (tab), "forward-idle-id",
		     GUINT_TO_POINTER (id));
}

/**
 * glook_play_tab_backward:
 */
static void
glook_play_tab_backward (GLookPlayTab *tab,
			 guint32       value)
{
  const gchar *sequence = "\033[H\033[2J";
  TRecordIndex *indexes;
  guint32 i, j, istart, iend, index;
  gssize expect;
  glong cols, rows;
  GError *error = NULL;

  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  /* clear the screen. */
  vte_terminal_feed (VTE_TERMINAL (tab->priv->terminal), sequence, -1);

  glook_terminal_get_size (GLOOK_TERMINAL (tab->priv->terminal),
			   &cols, &rows);
  expect = cols * rows * 2;

  /* try to refresh screen with previous data. */
  for (i = 0, j = value; i < 300 && j > 0; i++, j--)
    {
      if (!t_decoder_has_index (tab->priv->decoder, j))
	{
	  t_decoder_seek (tab->priv->decoder, j, &error);
	  if (error)
	    {
	      g_warning ("%s", error->message);
	      g_error_free (error);
	      return;
	    }
	}
      indexes = t_decoder_get_indexes (tab->priv->decoder,
				       &istart, &iend);
      index = j - istart;
      expect -= indexes[index].size;
      if (expect <= 0)
	break;
    }
  tab->priv->icurr = j;

  glook_play_tab_forward (tab, value);
}

/**
 * on_elapsed_adjust_value_changed:
 */
static void
on_elapsed_adjust_value_changed (GtkAdjustment *adjust,
				 GLookPlayTab  *tab)
{
  gint value;

  g_return_if_fail (GTK_IS_ADJUSTMENT (adjust));
  g_return_if_fail (GLOOK_IS_PLAY_TAB (tab));

  value = gtk_adjustment_get_value (adjust);
  if (value == tab->priv->icurr)
    return;

  if (tab->priv->play_idle > 0)
    g_source_remove (tab->priv->play_idle);

  if (tab->priv->play_timeout > 0)
    g_source_remove (tab->priv->play_timeout);

  /* forward. */
  if (value > tab->priv->icurr)
    glook_play_tab_forward (tab, value);
  else	/* backward. */
    glook_play_tab_backward (tab, value);
}

/**
 * glook_play_tab_new:
 */
GtkWidget*
glook_play_tab_new (void)
{
  return g_object_new (GLOOK_TYPE_PLAY_TAB, NULL);
}
