/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-misc.h"
#include "glook-profile-dialog.h"

#ifdef DEBUG
# define PROFILE_DIALOG_GLADE_FILE	SRCDIR G_DIR_SEPARATOR_S "glade" G_DIR_SEPARATOR_S "terminal-profile-dialog.xml"
#else
# define PROFILE_DIALOG_GLADE_FILE	GLADEDIR G_DIR_SEPARATOR_S "terminal-profile-dialog.xml"
#endif

enum
{
  COLOR_SCHEME_BLACK_ON_YELLOW,
  COLOR_SCHEME_BLACK_ON_WHILTE,
  COLOR_SCHEME_GRAY_ON_BLACK,
  COLOR_SCHEME_GREEN_ON_BLACK,
  COLOR_SCHEME_WHITE_ON_BLACK,
  COLOR_SCHEME_CUSTOMIZE,
};

enum
{
  PALETTE_SCHEME_LINUX,
  PALETTE_SCHEME_TANGO,
  PALETTE_SCHEME_XTERM,
  PALETTE_SCHEME_RXVT,
  PALETTE_SCHEME_CUSTOMIZE,
};

enum
{
  COLOR_COL_TITLE,
  COLOR_COL_NUMBER,
  COLOR_COL_FORE,
  COLOR_COL_BACK,
  COLOR_COL_CURSOR,
  COLOR_COL_HIGHLIGHT,
  N_COLOR_COLS
};

enum
{
  PROP_0,
  PROP_BG_EFFECT,
  PROP_BG_IMAGE,
};

struct _GLookProfileDialogPriv
{
  GtkBuilder  *builder;
  gchar       *profile;
  GSettings   *settings;
  gchar       *image;
};
#define _PROFILE_DIALOG_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_PROFILE_DIALOG, GLookProfileDialogPriv))

G_DEFINE_TYPE (GLookProfileDialog, glook_profile_dialog, GTK_TYPE_DIALOG);

/* One slot in the ring buffer, plus the array which holds the data for
 * the line, plus about 80 vte_charcell structures. */
#define BYTES_PER_LINE (sizeof(gpointer) + sizeof(GArray) + (80 * (sizeof(gunichar) + 4)))

/* count of built-in palette schemes, should be equal from
 * COLOR_SCHEME_CURSTOMIZE above. */
#define N_PALETTE_SCHEMES	4

/* built-in color palettes.
 * take from gnome-terminal sources
 */
static const GdkColor _palette_schemes[N_PALETTE_SCHEMES][16] =
{
  /* Linux palette */
  {
    { 0, 0x0000, 0x0000, 0x0000 },
    { 0, 0xaaaa, 0x0000, 0x0000 },
    { 0, 0x0000, 0xaaaa, 0x0000 },
    { 0, 0xaaaa, 0x5555, 0x0000 },
    { 0, 0x0000, 0x0000, 0xaaaa },
    { 0, 0xaaaa, 0x0000, 0xaaaa },
    { 0, 0x0000, 0xaaaa, 0xaaaa },
    { 0, 0xaaaa, 0xaaaa, 0xaaaa },
    { 0, 0x5555, 0x5555, 0x5555 },
    { 0, 0xffff, 0x5555, 0x5555 },
    { 0, 0x5555, 0xffff, 0x5555 },
    { 0, 0xffff, 0xffff, 0x5555 },
    { 0, 0x5555, 0x5555, 0xffff },
    { 0, 0xffff, 0x5555, 0xffff },
    { 0, 0x5555, 0xffff, 0xffff },
    { 0, 0xffff, 0xffff, 0xffff }
  },

  /* Tango palette */
  {
    { 0, 0x2e2e, 0x3434, 0x3636 },
    { 0, 0xcccc, 0x0000, 0x0000 },
    { 0, 0x4e4e, 0x9a9a, 0x0606 },
    { 0, 0xc4c4, 0xa0a0, 0x0000 },
    { 0, 0x3434, 0x6565, 0xa4a4 },
    { 0, 0x7575, 0x5050, 0x7b7b },
    { 0, 0x0606, 0x9820, 0x9a9a },
    { 0, 0xd3d3, 0xd7d7, 0xcfcf },
    { 0, 0x5555, 0x5757, 0x5353 },
    { 0, 0xefef, 0x2929, 0x2929 },
    { 0, 0x8a8a, 0xe2e2, 0x3434 },
    { 0, 0xfcfc, 0xe9e9, 0x4f4f },
    { 0, 0x7272, 0x9f9f, 0xcfcf },
    { 0, 0xadad, 0x7f7f, 0xa8a8 },
    { 0, 0x3434, 0xe2e2, 0xe2e2 },
    { 0, 0xeeee, 0xeeee, 0xecec }
  },

  /* XTerm palette */
  {
    { 0, 0x0000, 0x0000, 0x0000 },
    { 0, 0xcdcb, 0x0000, 0x0000 },
    { 0, 0x0000, 0xcdcb, 0x0000 },
    { 0, 0xcdcb, 0xcdcb, 0x0000 },
    { 0, 0x1e1a, 0x908f, 0xffff },
    { 0, 0xcdcb, 0x0000, 0xcdcb },
    { 0, 0x0000, 0xcdcb, 0xcdcb },
    { 0, 0xe5e2, 0xe5e2, 0xe5e2 },
    { 0, 0x4ccc, 0x4ccc, 0x4ccc },
    { 0, 0xffff, 0x0000, 0x0000 },
    { 0, 0x0000, 0xffff, 0x0000 },
    { 0, 0xffff, 0xffff, 0x0000 },
    { 0, 0x4645, 0x8281, 0xb4ae },
    { 0, 0xffff, 0x0000, 0xffff },
    { 0, 0x0000, 0xffff, 0xffff },
    { 0, 0xffff, 0xffff, 0xffff }
  },

  /* RXVT palette */
  {
    { 0, 0x0000, 0x0000, 0x0000 },
    { 0, 0xcdcd, 0x0000, 0x0000 },
    { 0, 0x0000, 0xcdcd, 0x0000 },
    { 0, 0xcdcd, 0xcdcd, 0x0000 },
    { 0, 0x0000, 0x0000, 0xcdcd },
    { 0, 0xcdcd, 0x0000, 0xcdcd },
    { 0, 0x0000, 0xcdcd, 0xcdcd },
    { 0, 0xfafa, 0xebeb, 0xd7d7 },
    { 0, 0x4040, 0x4040, 0x4040 },
    { 0, 0xffff, 0x0000, 0x0000 },
    { 0, 0x0000, 0xffff, 0x0000 },
    { 0, 0xffff, 0xffff, 0x0000 },
    { 0, 0x0000, 0x0000, 0xffff },
    { 0, 0xffff, 0x0000, 0xffff },
    { 0, 0x0000, 0xffff, 0xffff },
    { 0, 0xffff, 0xffff, 0xffff }
  }
};

static void		glook_profile_dialog_response		(GtkDialog          *dialog,
								 gint                resp_id);
static void		glook_profile_dialog_set_bg_effect	(GLookProfileDialog *dialog,
								 guint               effect);
static guint		glook_profile_dialog_get_bg_effect	(GLookProfileDialog *dialog);
static void		glook_profile_dialog_set_bg_image	(GLookProfileDialog *dialog,
								 const gchar        *image);
static const gchar*	glook_profile_dialog_get_bg_image	(GLookProfileDialog *dialog);

/**
 * glook_profile_dialog_init:
 */
static void
glook_profile_dialog_init (GLookProfileDialog *dialog)
{
  GtkWidget *content_area;
  GObject *toplevel;
  GObject *chooser;
  GtkFileFilter *filter;
  GError *error = NULL;

  dialog->priv = _PROFILE_DIALOG_GET_PRIVATE (dialog);

  dialog->priv->profile = NULL;
  dialog->priv->settings = NULL;
  dialog->priv->image = NULL;

  gtk_dialog_add_buttons (GTK_DIALOG (dialog),
			  GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
			  GTK_STOCK_HELP, GTK_RESPONSE_HELP, NULL);
  gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_CLOSE);

  if (!g_file_test (PROFILE_DIALOG_GLADE_FILE, G_FILE_TEST_IS_REGULAR))
    {
      g_critical (_("File `%s' not found."), PROFILE_DIALOG_GLADE_FILE);
      return;
    }
  /* load ui from file. */
  dialog->priv->builder = gtk_builder_new ();
  gtk_builder_add_from_file (dialog->priv->builder,
			     PROFILE_DIALOG_GLADE_FILE, &error);
  if (error)
    {
      g_critical ("Failed to load `%s': %s",
		  PROFILE_DIALOG_GLADE_FILE, error->message);
      g_error_free (error);
      return;
    }
  /* add filter to image file chooser button. we cannot do it in the ui file. */
  chooser = gtk_builder_get_object (dialog->priv->builder,
				    "BackgroundImageButton");
  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("Image"));
  gtk_file_filter_add_pixbuf_formats (filter);
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (chooser), filter);
  gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (chooser), filter);

  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("All Files"));
  gtk_file_filter_add_pattern (filter, "*");
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (chooser), filter);

  /* connect widgets signals. */
  gtk_builder_connect_signals (dialog->priv->builder, dialog);

  /* packing toplevel widget to contents area. */
  content_area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
  toplevel = gtk_builder_get_object (dialog->priv->builder, "ToplevelBox");
  gtk_container_add (GTK_CONTAINER (content_area), GTK_WIDGET (toplevel));
}

/**
 * glook_profile_dialog_dispose:
 */
static void
glook_profile_dialog_dispose (GObject *object)
{
  GLookProfileDialog *dialog = GLOOK_PROFILE_DIALOG (object);

  if (dialog->priv->builder)
    {
      g_object_unref (dialog->priv->builder);
      dialog->priv->builder = NULL;
    }
  if (dialog->priv->settings)
    {
      g_object_unref (dialog->priv->settings);
      dialog->priv->settings = NULL;
    }
  G_OBJECT_CLASS (glook_profile_dialog_parent_class)->dispose (object);
}

/**
 * glook_profile_dialog_finalize:
 */
static void
glook_profile_dialog_finalize (GObject *object)
{
  GLookProfileDialog *dialog = GLOOK_PROFILE_DIALOG (object);

  if (dialog->priv->profile)
    {
      g_free (dialog->priv->profile);
      dialog->priv->profile = NULL;
    }
  if (dialog->priv->image)
    {
      g_free (dialog->priv->image);
      dialog->priv->image = NULL;
    }
  G_OBJECT_CLASS (glook_profile_dialog_parent_class)->finalize (object);
}

/**
 * glook_profile_dialog_get_property:
 */
static void
glook_profile_dialog_get_property (GObject    *object,
				   guint       property_id,
				   GValue     *value,
				   GParamSpec *pspec)
{
  GLookProfileDialog *dialog = GLOOK_PROFILE_DIALOG (object);

  switch (property_id)
    {
    case PROP_BG_EFFECT:
      g_value_set_int (value, glook_profile_dialog_get_bg_effect (dialog));
      break;
    case PROP_BG_IMAGE:
      g_value_set_string (value, glook_profile_dialog_get_bg_image (dialog));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_profile_dialog_set_property:
 */
static void
glook_profile_dialog_set_property (GObject      *object,
				   guint         property_id,
				   const GValue *value,
				   GParamSpec   *pspec)
{
  GLookProfileDialog *dialog = GLOOK_PROFILE_DIALOG (object);

  switch (property_id)
    {
    case PROP_BG_EFFECT:
      glook_profile_dialog_set_bg_effect (dialog, g_value_get_int (value));
      break;
    case PROP_BG_IMAGE:
      glook_profile_dialog_set_bg_image (dialog, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_profile_dialog_class_init:
 */
static void
glook_profile_dialog_class_init (GLookProfileDialogClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GtkDialogClass *dialog_class = GTK_DIALOG_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookProfileDialogPriv));

  gobject_class->get_property = glook_profile_dialog_get_property;
  gobject_class->set_property = glook_profile_dialog_set_property;
  gobject_class->dispose  = glook_profile_dialog_dispose;
  gobject_class->finalize = glook_profile_dialog_finalize;

  dialog_class->response = glook_profile_dialog_response;

  g_object_class_install_property (gobject_class, PROP_BG_EFFECT,
				   g_param_spec_int ("bg-effect", "bg-effect", "Background effect",
						     0, 2, 0, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
						     G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_BG_IMAGE,
				   g_param_spec_string ("bg-image", "bg-image", "Background image file", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
}

/**
 * glook_profile_dialog_response:
 */
static void
glook_profile_dialog_response (GtkDialog *dialog,
			       gint       resp_id)
{
  GtkDialogClass *dialog_class;

  g_return_if_fail (GTK_IS_DIALOG (dialog));

  if (resp_id == GTK_RESPONSE_HELP)
    {
      glook_open_url (GTK_WIDGET (dialog), WEBSITE);
      return;
    }
  dialog_class = GTK_DIALOG_CLASS (glook_profile_dialog_parent_class);
  if (dialog_class->response)
    dialog_class->response (dialog, resp_id);
}

/**
 * glook_profile_dialog_set_bg_effect:
 */
static void
glook_profile_dialog_set_bg_effect (GLookProfileDialog *dialog,
				    guint               effect)
{
  const gchar *name;
  GObject *radio;

  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  switch (effect)
    {
    case BACKGROUND_EFFECT_NONE:
      name = "BackgroundNoneRadio";
      break;
    case BACKGROUND_EFFECT_IMAGE:
      name = "BackgroundImageRadio";
      break;
    case BACKGROUND_EFFECT_TRANSPARENT:
      name = "BackgroundTransparentRadio";
      break;
    default:
      g_warning ("background effect %d out of range.", effect);
      return;
    }
  radio = gtk_builder_get_object (dialog->priv->builder, name);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio), TRUE);
  g_object_notify (G_OBJECT (dialog), "bg-effect");
}

/**
 * glook_profile_dialog_get_bg_effect:
 */
static guint
glook_profile_dialog_get_bg_effect (GLookProfileDialog *dialog)
{
  GObject *radio;

  g_return_val_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog), 0);

  radio = gtk_builder_get_object (dialog->priv->builder,
				  "BackgroundNoneRadio");
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radio)))
    return 0;

  radio = gtk_builder_get_object (dialog->priv->builder,
				  "BackgroundImageRadio");
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radio)))
    return 1;

  return 2;
}

/**
 * glook_profile_dialog_set_bg_image:
 */
static void
glook_profile_dialog_set_bg_image (GLookProfileDialog *dialog,
				   const gchar        *image)
{
  GObject *chooser;

  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  chooser = gtk_builder_get_object (dialog->priv->builder,
				    "BackgroundImageButton");
  if (image && image[0])
    gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (chooser), image);
  else
    gtk_file_chooser_unselect_all (GTK_FILE_CHOOSER (chooser));

  g_object_notify (G_OBJECT (dialog), "bg-image");
}

/**
 * glook_profile_dialog_get_bg_image:
 */
static const gchar*
glook_profile_dialog_get_bg_image (GLookProfileDialog *dialog)
{
  GObject *chooser;

  g_return_val_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog), NULL);

  chooser = gtk_builder_get_object (dialog->priv->builder,
				    "BackgroundImageButton");
  g_free (dialog->priv->image);
  dialog->priv->image =
   gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (chooser));

  return dialog->priv->image ? dialog->priv->image : "";
}

/**
 * glook_profile_dialog_new:
 */
GtkWidget*
glook_profile_dialog_new (void)
{
  return g_object_new (GLOOK_TYPE_PROFILE_DIALOG,
		       "title", _("Edit terminal profile"),
		       "has-separator", FALSE,
		       "border-width", 4, NULL);
}

typedef struct _ObjectSettingsBind	ObjectSettingsBind;
struct _ObjectSettingsBind
{
  const gchar       *settings_key;
  const gchar       *object_name;
  const gchar       *object_prop;
  GSettingsBindFlags flags;
};
static ObjectSettingsBind _object_settings_binds[] =
{
  /* Note the order is important since we want some widget's sensitive state get automatic update. */
    { "name",			"ProfileNameEntry",		"text",		G_SETTINGS_BIND_DEFAULT },
    { "description",		"ProfileDescriptionEntry",	"text",		G_SETTINGS_BIND_DEFAULT },
    { "audible-bell",		"AudibleBellCheck",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "visible-bell",		"VisibleBellCheck",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "allow-bold",		"BoldTextCheck",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "cursor-shape",		"CursorShapeCombo",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "cursor-blink",		"CursorBlinkCombo",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "word-chars",		"WordCharsEntry",		"text",		G_SETTINGS_BIND_DEFAULT },
    { "font",			"FontButton",			"font-name",	G_SETTINGS_BIND_DEFAULT },
    { "use-system-font",	"UseSystemFontCheck",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "scrollback",		"ScrollLinesSpin",		"value",	G_SETTINGS_BIND_DEFAULT },
    { "show-scrollbar",		"ShowScrollbarCheck",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "scroll-on-output",	"OutputScrollCheck",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "scroll-on-keystroke",	"KeystrokeScrollCheck",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "backspace-binding",	"BackspaceCombo",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "delete-binding",		"DeleteCombo",			"active",	G_SETTINGS_BIND_DEFAULT },
    { "use-system-color",	"UseSystemColorCheck",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "color-scheme",		"ColorSchemeCombo",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "palette-scheme",		"PaletteSchemeCombo",		"active",	G_SETTINGS_BIND_DEFAULT },
    { "background-effect",	NULL,				"bg-effect",	G_SETTINGS_BIND_DEFAULT },
    { "background-image",	NULL,				"bg-image",	G_SETTINGS_BIND_DEFAULT },
    { "background-scroll",	"BackgroundScrollCheck",	"active",	G_SETTINGS_BIND_DEFAULT },
    { "background-saturation",	"BackgroundSaturationAdjust",	"value",	G_SETTINGS_BIND_DEFAULT },
};

/**
 * gdk_color_from_string_mapping:
 */
static gboolean
gdk_color_from_string_mapping (GValue   *value,
			       GVariant *variant,
			       gpointer  user_data)
{
  gchar *color_name;
  GdkColor color;

  g_variant_get (variant, "s", &color_name);
  if (!gdk_color_parse (color_name, &color))
    return FALSE;

  g_value_set_boxed (value, &color);
  return TRUE;
}

/**
 * gdk_color_to_string_mapping:
 */
static GVariant*
gdk_color_to_string_mapping (const GValue       *value,
			     const GVariantType *expect,
			     gpointer            user_data)
{
  GVariant *result;
  GdkColor *color;
  gchar *color_name;

  color = g_value_get_boxed (value);
  color_name = gdk_color_to_string (color);
  if (!color_name)
    return NULL;

  result = g_variant_new_string (color_name);
  g_free (color_name);

  return result;
}

/**
 * glook_profile_dialog_bind_settings:
 */
static void
glook_profile_dialog_bind_settings (GLookProfileDialog *dialog,
				    GSettings          *settings)
{
  gint i;
  GObject *object;

  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  /* bind palette color buttons settings. */
  for (i = 1; i <= 16; i++)
    {
      gchar button[32], key[32];

      g_snprintf (button, sizeof (button), "PaletteButton%d", i);
      g_snprintf (key, sizeof (key), "palette%d", i);

      object = gtk_builder_get_object (dialog->priv->builder, button);
      g_settings_bind_with_mapping (settings, key, object, "color",
				    G_SETTINGS_BIND_DEFAULT,
				    gdk_color_from_string_mapping,
				    gdk_color_to_string_mapping, NULL, NULL);
    }
  for (i = 0; i < G_N_ELEMENTS (_object_settings_binds); i++)
    {
      if (_object_settings_binds[i].object_name)
	{
	  object = gtk_builder_get_object (dialog->priv->builder,
					   _object_settings_binds[i].object_name);
	}
      else
	{
	  object = G_OBJECT (dialog);
	}
      g_assert (G_IS_OBJECT (object));

      g_settings_bind (settings, _object_settings_binds[i].settings_key,
		       object, _object_settings_binds[i].object_prop,
		       _object_settings_binds[i].flags);
    }
  object = gtk_builder_get_object (dialog->priv->builder, "ForegroundButton");
  g_settings_bind_with_mapping (settings, "foreground-color", object, "color",
				G_SETTINGS_BIND_DEFAULT,
				gdk_color_from_string_mapping,
				gdk_color_to_string_mapping, NULL, NULL);
  object = gtk_builder_get_object (dialog->priv->builder, "BackgroundButton");
  g_settings_bind_with_mapping (settings, "background-color", object, "color",
				G_SETTINGS_BIND_DEFAULT,
				gdk_color_from_string_mapping,
				gdk_color_to_string_mapping, NULL, NULL);
}

/**
 * glook_profile_dialog_unbind_settings:
 */
static void
glook_profile_dialog_unbind_settings (GLookProfileDialog *dialog)
{
  gint i;
  GObject *object;

  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  /* unbind palette color buttons settings. */
  for (i = 1; i <= 16; i++)
    {
      gchar button[32];

      g_snprintf (button, sizeof (button), "PaletteButton%d", i);

      object = gtk_builder_get_object (dialog->priv->builder, button);
      g_settings_unbind (object, "color");
    }
  /* unbind all widget properties from settings. */
  for (i = 0; i < G_N_ELEMENTS (_object_settings_binds); i++)
    {
      object = gtk_builder_get_object (dialog->priv->builder,
				       _object_settings_binds[i].object_name);
      g_assert (G_IS_OBJECT (object));
      g_settings_unbind (object, _object_settings_binds[i].object_prop);
    }
  object = gtk_builder_get_object (dialog->priv->builder, "ForegroundButton");
  g_settings_unbind (object, "color");
  object = gtk_builder_get_object (dialog->priv->builder, "BackgroundButton");
  g_settings_unbind (object, "color");
}

/**
 * glook_profile_dialog_set_profile:
 */
void
glook_profile_dialog_set_profile (GLookProfileDialog *dialog,
				  const gchar        *profile,
				  GSettings          *settings)
{
  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));
  g_return_if_fail (profile);
  g_return_if_fail (settings);

  g_free (dialog->priv->profile);
  dialog->priv->profile = g_strdup (profile);

  if (dialog->priv->settings)
    {
      glook_profile_dialog_unbind_settings (dialog);
      g_object_unref (dialog->priv->settings);
    }
  dialog->priv->settings = g_object_ref (settings);
  glook_profile_dialog_bind_settings (dialog, settings);
}

/**
 * glook_profile_dialog_get_profile:
 */
const gchar*
glook_profile_dialog_get_profile (GLookProfileDialog *dialog)
{
  g_return_val_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog), NULL);
  return dialog->priv->profile;
}

/**
 * on_use_system_font_toggled:
 * FIXME: how to get system monospaced font from GSettings.
 * we just know how to use GConf to lookup value of key
 * '/desktop/gnome/interface/monospace_font_name'
 * in order to get system monospaced font. but that require link
 * the GConf library into application.
 */
void
on_use_system_font_toggled (GtkToggleButton    *button,
			    GLookProfileDialog *dialog)
{
  GObject *fontbutton;

  g_return_if_fail (GTK_IS_TOGGLE_BUTTON (button));
  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  fontbutton = gtk_builder_get_object (dialog->priv->builder,
				       "FontButton");
  if (gtk_toggle_button_get_active (button))
    {
      gtk_widget_set_sensitive (GTK_WIDGET (fontbutton), FALSE);
    }
  else
    {
      gtk_widget_set_sensitive (GTK_WIDGET (fontbutton), TRUE);
    }
}

/**
 * on_use_system_color_toggled:
 */
void
on_use_system_color_toggled (GtkToggleButton    *button,
			     GLookProfileDialog *dialog)
{
  GObject *table;
  gboolean active;

  g_return_if_fail (GTK_IS_TOGGLE_BUTTON (button));
  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  table = gtk_builder_get_object (dialog->priv->builder, "ColorSchemeTable");

  active = gtk_toggle_button_get_active (button);
  gtk_widget_set_sensitive (GTK_WIDGET (table), !active);
}

/**
 * on_color_scheme_combo_changed:
 */
void
on_color_scheme_combo_changed (GtkComboBox        *combo,
			       GLookProfileDialog *dialog)
{
  GtkTreeModel *model;
  GtkTreeIter iter;

  g_return_if_fail (GTK_IS_COMBO_BOX (combo));
  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  if (gtk_combo_box_get_active_iter (combo, &iter))
    {
      gchar *fore, *back;
      GdkColor fore_color, back_color;
      GObject *button;

      model = gtk_combo_box_get_model (combo);
      gtk_tree_model_get (model, &iter, COLOR_COL_FORE, &fore,
			  COLOR_COL_BACK, &back, -1);
      gdk_color_parse (fore, &fore_color);
      gdk_color_parse (back, &back_color);
      g_free (fore);
      g_free (back);

      button = gtk_builder_get_object (dialog->priv->builder,
				       "ForegroundButton");
      gtk_color_button_set_color (GTK_COLOR_BUTTON (button), &fore_color);

      button = gtk_builder_get_object (dialog->priv->builder,
				       "BackgroundButton");
      gtk_color_button_set_color (GTK_COLOR_BUTTON (button), &back_color);

    }
}

static void
glook_profile_dialog_set_customize_color (GLookProfileDialog *dialog,
					  GtkColorButton     *button,
					  gint                column)
{
  GObject *combo;
  GtkTreeModel *model;
  GtkTreeIter iter;
  gchar path[8];
  GdkColor color;
  gchar *color_string;

  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));
  g_return_if_fail (GTK_IS_COLOR_BUTTON (button));

  combo = gtk_builder_get_object (dialog->priv->builder, "ColorSchemeCombo");

  /* modify model foreground color. */
  model = gtk_combo_box_get_model (GTK_COMBO_BOX (combo));
  g_snprintf (path, sizeof (path), "%d", COLOR_SCHEME_CUSTOMIZE);
  if (!gtk_tree_model_get_iter_from_string (model, &iter, path))
    {
      g_critical ("cannot get iter for path `%s'", path);
      return;
    }
  gtk_color_button_get_color (button, &color);
  color_string = gdk_color_to_string (&color);

  gtk_list_store_set (GTK_LIST_STORE (model), &iter, column, color_string, -1);
  g_free (color_string);

  /* block the handler when change button color. */
  g_signal_handlers_block_by_func (combo, on_color_scheme_combo_changed, dialog);
  gtk_combo_box_set_active (GTK_COMBO_BOX (combo), COLOR_SCHEME_CUSTOMIZE);
  g_signal_handlers_unblock_by_func (combo, on_color_scheme_combo_changed, dialog);
}

/**
 * on_foreground_button_color_set:
 */
void
on_foreground_button_color_set (GtkColorButton     *button,
				GLookProfileDialog *dialog)
{
  g_return_if_fail (GTK_IS_COLOR_BUTTON (button));
  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  glook_profile_dialog_set_customize_color (dialog, button, COLOR_COL_FORE);
}

/**
 * on_background_button_color_set:
 */
void
on_background_button_color_set (GtkColorButton     *button,
				GLookProfileDialog *dialog)
{
  g_return_if_fail (GTK_IS_COLOR_BUTTON (button));
  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  glook_profile_dialog_set_customize_color (dialog, button, COLOR_COL_BACK);
}

/**
 * on_palette_scheme_combo_changed:
 */
void
on_palette_scheme_combo_changed (GtkComboBox        *combo,
				 GLookProfileDialog *dialog)
{
  gint active, i;
  GObject *button;

  g_return_if_fail (GTK_IS_COMBO_BOX (combo));
  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  active = gtk_combo_box_get_active (combo);

  if (active < N_PALETTE_SCHEMES)
    {
      for (i = 0; i < 16; i++)
	{
	  gchar name[32];
	  GdkColor color;

	  g_snprintf (name, sizeof (name), "PaletteButton%d", i + 1);
	  color = _palette_schemes[active][i];

	  button = gtk_builder_get_object (dialog->priv->builder, name);
	  gtk_color_button_set_color (GTK_COLOR_BUTTON (button), &color);
	}
    }
}

/**
 * on_palette_button_color_set:
 */
void
on_palette_button_color_set (GtkColorButton     *button,
			     GLookProfileDialog *dialog)
{
  GObject *combo;

  g_return_if_fail (GTK_IS_COLOR_BUTTON (button));
  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  combo = gtk_builder_get_object (dialog->priv->builder, "PaletteSchemeCombo");
  gtk_combo_box_set_active (GTK_COMBO_BOX (combo), PALETTE_SCHEME_CUSTOMIZE);
}

/**
 * on_compat_reset_button_clicked:
 */
void
on_compat_reset_button_clicked (GtkWidget          *button,
				GLookProfileDialog *dialog)
{
  GObject *combo;

  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  combo = gtk_builder_get_object (dialog->priv->builder, "BackspaceCombo");
  gtk_combo_box_set_active (GTK_COMBO_BOX (combo), 0);

  combo = gtk_builder_get_object (dialog->priv->builder, "DeleteCombo");
  gtk_combo_box_set_active (GTK_COMBO_BOX (combo), 0);
}

/**
 * on_scroll_lines_spin_value_changed:
 */
void
on_scroll_lines_spin_value_changed (GtkSpinButton      *button,
				    GLookProfileDialog *dialog)
{
  gdouble value, kb;
  GObject *kbspin;

  g_return_if_fail (GTK_IS_SPIN_BUTTON (button));
  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  value = gtk_spin_button_get_value (button);
  kb = value * BYTES_PER_LINE / 1024;

  kbspin = gtk_builder_get_object (dialog->priv->builder,
				   "ScrollKBytesSpin");
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (kbspin), kb);
}

/**
 * on_scroll_kbytes_spin_value_changed:
 */
void
on_scroll_kbytes_spin_value_changed (GtkSpinButton      *button,
				     GLookProfileDialog *dialog)
{
  gdouble value, lines;
  GObject *spin;

  g_return_if_fail (GTK_IS_SPIN_BUTTON (button));
  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  value = gtk_spin_button_get_value (button);
  lines = value * 1024 / BYTES_PER_LINE;

  spin = gtk_builder_get_object (dialog->priv->builder,
				 "ScrollLinesSpin");
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (spin), lines);
}

/**
 * on_background_effect_radio_toggled:
 */
void
on_background_effect_radio_toggled (GtkToggleButton    *button,
				    GLookProfileDialog *dialog)
{
  GObject *radio;
  GObject *image_align, *trans_align;
  gint effect;

  g_return_if_fail (GTK_IS_TOGGLE_BUTTON (button));
  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  if (!gtk_toggle_button_get_active (button))
    return;

  image_align = gtk_builder_get_object (dialog->priv->builder,
					"BackgroundImageAlignment");
  trans_align = gtk_builder_get_object (dialog->priv->builder,
					"BackgroundTransparentAlignment");

  radio = gtk_builder_get_object (dialog->priv->builder,
				  "BackgroundNoneRadio");
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radio)))
    {
      gtk_widget_set_sensitive (GTK_WIDGET (image_align), FALSE);
      gtk_widget_set_sensitive (GTK_WIDGET (trans_align), FALSE);
      effect = BACKGROUND_EFFECT_NONE;
      goto end;
    }
  radio = gtk_builder_get_object (dialog->priv->builder,
				  "BackgroundImageRadio");
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radio)))
    {
      gtk_widget_set_sensitive (GTK_WIDGET (image_align), TRUE);
      gtk_widget_set_sensitive (GTK_WIDGET (trans_align), TRUE);
      effect = BACKGROUND_EFFECT_IMAGE;
      goto end;
    }
  gtk_widget_set_sensitive (GTK_WIDGET (image_align), FALSE);
  gtk_widget_set_sensitive (GTK_WIDGET (trans_align), TRUE);
  effect = BACKGROUND_EFFECT_TRANSPARENT;

end:
  /* we must block the handler when set bg effect property. */
  g_signal_handlers_block_by_func (radio, on_background_effect_radio_toggled, dialog);
  glook_profile_dialog_set_bg_effect (dialog, effect);
  g_signal_handlers_unblock_by_func (radio, on_background_effect_radio_toggled, dialog);
}

/**
 * on_background_image_file_set:
 */
void
on_background_image_dialog_responsed (GtkFileChooser     *chooser,
				      gint                respid,
				      GLookProfileDialog *dialog)
{
  gchar *filename;

  g_return_if_fail (GTK_IS_FILE_CHOOSER (chooser));
  g_return_if_fail (GLOOK_IS_PROFILE_DIALOG (dialog));

  filename = gtk_file_chooser_get_filename (chooser);

  g_signal_handlers_block_by_func (chooser, on_background_image_dialog_responsed, dialog);
  glook_profile_dialog_set_bg_image (dialog, filename);
  g_free (filename);
  g_signal_handlers_unblock_by_func (chooser, on_background_image_dialog_responsed, dialog);
}
