/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-app.h"
#include "glook-statusbar.h"

struct _GLookStatusbarPriv
{
  GtkWidget *label;
#if 0
  GtkWidget *accel_image;
  GtkWidget *accel_viewport;
#endif
};
#define _STATUSBAR_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_STATUSBAR, GLookStatusbarPriv))

G_DEFINE_TYPE (GLookStatusbar, glook_statusbar, GTK_TYPE_STATUSBAR);

static void	glook_statusbar_text_pushed	(GtkStatusbar   *statusbar,
						 guint           context_id,
						 const gchar    *text);
static void	glook_statusbar_text_popped	(GtkStatusbar   *statusbar,
						 guint           context_id,
						 const gchar    *text);
/**
 * accelerator_sensitive_bind_mapping:
 */
#if 0
static gboolean
accelerator_sensitive_bind_mapping (GValue   *value,
				    GVariant *variant,
				    gpointer  user_data)
{
  GLookStatusbar *statusbar = user_data;
  gboolean enabled;

  g_return_val_if_fail (GLOOK_IS_STATUSBAR (statusbar), FALSE);

  enabled = g_variant_get_boolean (variant);
  if (enabled)
    {
      gtk_widget_set_tooltip_text (statusbar->priv->accel_image,
				   _("Accelerator is disabled, double click "
				     "here to turn on the accelerator"));
      g_value_set_boolean (value, FALSE);
    }
  else
    {
      gtk_widget_set_tooltip_text (statusbar->priv->accel_image,
				   _("Accelerator is enabled, double click "
				     "here to turn off the accelerator"));
      g_value_set_boolean (value, TRUE);
    }
  return TRUE;
}
#endif

#if 0
/**
 * on_accel_viewport_button_pressed:
 */
static gboolean
on_accel_viewport_button_pressed (GtkWidget      *viewport,
				  GdkEventButton *event,
				  GLookStatusbar *statusbar)
{
  GSettings *settings;

  if (event->type == GDK_2BUTTON_PRESS && event->button == 1)
    {
      settings = glook_app_get_settings (glook_app_get (),
					 SETTINGS_ACCELERATOR);
      if (g_settings_get_boolean (settings, "disable-accelerators"))
	g_settings_set_boolean (settings, "disable-accelerators", FALSE);
      else
	g_settings_set_boolean (settings, "disable-accelerators", TRUE);

      return TRUE;
    }
  return FALSE;
}
#endif

/**
 * glook_statusbar_init:
 */
static void
glook_statusbar_init (GLookStatusbar *statusbar)
{
  GList *childs;
  GtkWidget *area;
#if 0
  GtkWidget *alignment;
  GSettings *settings;
#endif

  gtk_box_set_spacing (GTK_BOX (statusbar), 4);

  statusbar->priv = _STATUSBAR_GET_PRIVATE (statusbar);

  /* save the statusbar label widget. */
  area = gtk_statusbar_get_message_area (GTK_STATUSBAR (statusbar));
  childs =  gtk_container_get_children (GTK_CONTAINER (area));
  statusbar->priv->label = childs->data;
  g_list_free (childs);

#if 0
  /* packing accelerator status widget. */
  statusbar->priv->accel_viewport = gtk_viewport_new (NULL, NULL);
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (statusbar->priv->accel_viewport),
				GTK_SHADOW_NONE);
  g_signal_connect (statusbar->priv->accel_viewport, "button-press-event",
		    G_CALLBACK (on_accel_viewport_button_pressed), statusbar);

  alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 4, 4, 4, 4);
  gtk_container_add (GTK_CONTAINER (statusbar->priv->accel_viewport), alignment);
  statusbar->priv->accel_image = gtk_image_new_from_icon_name ("ttyutils-accelerator",
							       GTK_ICON_SIZE_MENU);
  gtk_container_add (GTK_CONTAINER (alignment), statusbar->priv->accel_image);

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_ACCELERATOR);
  g_settings_bind_with_mapping (settings, "disable-accelerators",
				statusbar->priv->accel_image, "sensitive",
				G_SETTINGS_BIND_GET,
				accelerator_sensitive_bind_mapping,
				NULL, statusbar, NULL);

  gtk_widget_show_all (statusbar->priv->accel_viewport);
  gtk_box_pack_end (GTK_BOX (statusbar), statusbar->priv->accel_viewport,
		    FALSE, FALSE, 0);
#endif
}

/**
 * glook_statusbar_dispose:
 */
static void
glook_statusbar_dispose (GObject *object)
{
  //GLookStatusbar *statusbar = GLOOK_STATUSBAR (object);

  G_OBJECT_CLASS (glook_statusbar_parent_class)->dispose (object);
}

/**
 * glook_statusbar_finalize:
 */
static void
glook_statusbar_finalize (GObject *object)
{
  //GLookStatusbar *look_tip = GLOOK_STATUSBAR (object);

  G_OBJECT_CLASS (glook_statusbar_parent_class)->finalize (object);
}

/**
 * glook_statusbar_class_init:
 */
static void
glook_statusbar_class_init (GLookStatusbarClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GtkStatusbarClass *statusbar_class = GTK_STATUSBAR_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookStatusbarPriv));

  gobject_class->dispose  = glook_statusbar_dispose;
  gobject_class->finalize = glook_statusbar_finalize;

  statusbar_class->text_pushed = glook_statusbar_text_pushed;
  statusbar_class->text_popped = glook_statusbar_text_popped;
}

/**
 * glook_statusbar_text_pushed:
 */
static void
glook_statusbar_text_pushed (GtkStatusbar *widget,
			     guint         context_id,
			     const gchar  *text)
{
  GLookStatusbar *statusbar = GLOOK_STATUSBAR (widget);
  GtkStatusbarClass *statusbar_class;

  g_return_if_fail (GLOOK_IS_STATUSBAR (statusbar));

  /* pop the message first. */
  statusbar_class = GTK_STATUSBAR_CLASS (glook_statusbar_parent_class);
  if (statusbar_class->text_pushed)
    statusbar_class->text_pushed (widget, context_id, text);

  /* modify label property. */
  gtk_label_set_use_markup (GTK_LABEL (statusbar->priv->label), TRUE);
}

/**
 * glook_statusbar_text_popped:
 */
static void
glook_statusbar_text_popped (GtkStatusbar *widget,
			     guint         context_id,
			     const gchar  *text)
{
  GLookStatusbar *statusbar = GLOOK_STATUSBAR (widget);
  GtkStatusbarClass *statusbar_class;

  g_return_if_fail (GLOOK_IS_STATUSBAR (statusbar));

  /* pop the message first. */
  statusbar_class = GTK_STATUSBAR_CLASS (glook_statusbar_parent_class);
  if (statusbar_class->text_popped)
    statusbar_class->text_popped (widget, context_id, text);

  /* modify label property. */
  gtk_label_set_use_markup (GTK_LABEL (statusbar->priv->label), TRUE);
}

/**
 * glook_statusbar_new:
 */
GtkWidget*
glook_statusbar_new (void)
{
  return g_object_new (GLOOK_TYPE_STATUSBAR,
		       "has-resize-grip", TRUE, NULL);
}

/**
 * glook_statusbar_push:
 */
guint
glook_statusbar_push (GLookStatusbar *statusbar,
		      guint           context_id,
		      GtkMessageType  type,
		      const gchar    *format,
		      ...)
{
  va_list args;
  gchar buff[512], *markup;
  guint ret;

  g_return_val_if_fail (GLOOK_IS_STATUSBAR (statusbar), 0);

  va_start (args, format);
  g_vsnprintf (buff, sizeof (buff), format, args);
  va_end (args);

  switch (type)
    {
    case GTK_MESSAGE_WARNING:
      markup = g_markup_printf_escaped ("<span foreground='blue'>%s</span>", buff);
      ret = gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id, markup);
      g_free (markup);
      break;
    case GTK_MESSAGE_ERROR:
      markup = g_markup_printf_escaped ("<span foreground='red'>%s</span>", buff);
      ret = gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id, markup);
      g_free (markup);
      break;
    default:
      markup = g_markup_escape_text (buff, -1);
      ret = gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id, markup);
      g_free (markup);
      break;
    }
  return ret;
}
