/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-app.h"
#include "glook-window.h"
#include "glook-ui.h"
#include "glook-notebook.h"
#include "glook-tab.h"
#include "glook-blank-tab.h"

/* -1 means autosize. */
const gint	_zoom_factors[] =
{
  50, 60, 70, 80, 90, 100, 110, 120, 130, 140, 150, 200, 300,
};
const gint	n_zoom_factors = G_N_ELEMENTS (_zoom_factors);
#define _ZOOM_100_INDEX		5

enum
{
  PROP_0,
  PROP_LABEL,
  PROP_IMAGE,
  PROP_STOCK,
};

G_DEFINE_ABSTRACT_TYPE (GLookTab, glook_tab, GTK_TYPE_VBOX);

static gboolean		glook_tab_scroll_event		(GtkWidget      *widget,
							 GdkEventScroll *event);
static GtkWidget*	glook_tab_create_tab_head	(GLookTab       *tab);

static void		on_tab_widget_style_set		(GtkWidget      *hbox,
							 GtkStyle       *previous_style,
							 GtkWidget      *button);
static void		on_tab_close_button_clicked	(GtkWidget      *button,
							 GLookTab       *tab);
static gboolean		on_tab_head_button_pressed	(GtkWidget      *widget,
							 GdkEventButton *event,
							 GLookTab       *tab);

/**
 * glook_tab_get_zoom_100_index:
 */
gint
glook_tab_get_zoom_100_index (void)
{
  return _ZOOM_100_INDEX;
}

/* dummy functions to prevent core dump or provides default
 * act when subclass has not overwirte these method. */
static gboolean
glook_tab_is_active_real (GLookTab *tab)
{
  return FALSE;
}

/**
 * glook_tab_init:
 */
static void
glook_tab_init (GLookTab *tab)
{
  tab->uri = NULL;
  tab->menu = NULL;
  tab->progress = FALSE;
  tab->zoom_index = _ZOOM_100_INDEX;
  tab->tab_head = glook_tab_create_tab_head (tab);
}

/**
 * glook_tab_dispose:
 */
static void
glook_tab_dispose (GObject *object)
{
  GLookTab *tab = GLOOK_TAB (object);

  if (tab->menu)
    {
      gtk_widget_destroy (tab->menu);
      tab->menu = NULL;
    }
  G_OBJECT_CLASS (glook_tab_parent_class)->dispose (object);
}

/**
 * glook_tab_finalize:
 */
static void
glook_tab_finalize (GObject *object)
{
  GLookTab *tab = GLOOK_TAB (object);

  if (tab->uri)
    {
      glook_uri_free (tab->uri);
      tab->uri = NULL;
    }
  if (tab->tab_head)
    {
      gtk_widget_destroy (tab->tab_head);
      tab->tab_head = NULL;
    }
  G_OBJECT_CLASS (glook_tab_parent_class)->finalize (object);
}

/**
 * glook_tab_get_property:
 */
static void
glook_tab_get_property (GObject    *object,
			guint       property_id,
			GValue     *value,
			GParamSpec *pspec)
{
  GLookTab *tab = GLOOK_TAB (object);

  switch (property_id)
    {
    case PROP_LABEL:
      g_value_set_string (value, glook_tab_get_label (tab));
      break;
    case PROP_IMAGE:
      g_value_set_string (value, glook_tab_get_image (tab));
      break;
    case PROP_STOCK:
      g_value_set_string (value, glook_tab_get_stock (tab));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_tab_set_property:
 */
static void
glook_tab_set_property (GObject      *object,
			guint         property_id,
			const GValue *value,
			GParamSpec   *pspec)
{
  GLookTab *tab = GLOOK_TAB (object);

  switch (property_id)
    {
    case PROP_LABEL:
      glook_tab_set_label (tab, g_value_get_string (value));
      break;
    case PROP_IMAGE:
      glook_tab_set_image (tab, g_value_get_string (value));
      break;
    case PROP_STOCK:
      glook_tab_set_stock (tab, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_tab_progress:
 */
static void
glook_tab_progress (GLookTab *tab)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));

  glook_tab_set_image_visible (tab, FALSE);
  glook_tab_set_spinner_visible (tab, TRUE);
  glook_tab_set_progress (tab, TRUE);
}

/**
 * glook_tab_finished:
 */
static void
glook_tab_finished (GLookTab *tab)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));

  glook_tab_set_spinner_visible (tab, FALSE);
  glook_tab_set_image_visible (tab, TRUE);
  glook_tab_set_progress (tab, FALSE);
}

/**
 * glook_tab_class_init:
 */
static void
glook_tab_class_init (GLookTabClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  gobject_class->get_property = glook_tab_get_property;
  gobject_class->set_property = glook_tab_set_property;
  gobject_class->dispose  = glook_tab_dispose;
  gobject_class->finalize = glook_tab_finalize;

  widget_class->scroll_event = glook_tab_scroll_event;

  klass->progress      = glook_tab_progress;
  klass->finished      = glook_tab_finished;
  klass->merge_ui      = NULL;
  klass->unmerge_ui    = NULL;
  klass->open          = NULL;
  klass->stop          = NULL;
  klass->is_active     = glook_tab_is_active_real;
  klass->zoom          = NULL;
  klass->reload        = NULL;
  klass->find          = NULL;
  klass->geometry_hint = NULL;

  g_object_class_install_property (gobject_class, PROP_LABEL,
				   g_param_spec_string ("label", "Label", "Tab title", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_IMAGE,
				   g_param_spec_string ("image", "Image", "Tab image", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_STOCK,
				   g_param_spec_string ("stock", "Stock", "Tab image stock", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));

  g_signal_new ("progress", GLOOK_TYPE_TAB, G_SIGNAL_RUN_FIRST,
		G_STRUCT_OFFSET (GLookTabClass, progress),
		NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0);
  g_signal_new ("finished", GLOOK_TYPE_TAB, G_SIGNAL_RUN_FIRST,
		G_STRUCT_OFFSET (GLookTabClass, finished),
		NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0);

  /* make the close button as small as possible */
  gtk_rc_parse_string ("style \"glook-tab-close-button-style\"\n"
		       "{\n"
		       "GtkWidget::focus-padding = 0\n"
		       "GtkWidget::focus-line-width = 0\n"
		       "xthickness = 0\n"
		       "ythickness = 0\n"
		       "}\n"
		       "widget \"*.glook-tab-close-button\" style \"glook-tab-close-button-style\"");
}

/**
 * glook_tab_scroll_event:
 */
static gboolean
glook_tab_scroll_event (GtkWidget      *widget,
			GdkEventScroll *event)
{
  GtkWidget *window;
  GtkWidgetClass *widget_class;

  g_return_val_if_fail (GLOOK_IS_TAB (widget), FALSE);

  window = gtk_widget_get_toplevel (widget);

  /* Ctrl + wheel mice zoom in/out the tab. */
  if (GLOOK_IS_WINDOW (window) && event->state & GDK_CONTROL_MASK)
    {
      GtkUIManager *uiman;
      GtkAction *action;

      uiman = glook_window_get_uiman (GLOOK_WINDOW (window));

      if (event->direction == GDK_SCROLL_DOWN ||
	  event->direction == GDK_SCROLL_RIGHT)
	action = glook_ui_get_window_action (GLOOK_UI (uiman), "ZoomOut");
      else
	action = glook_ui_get_window_action (GLOOK_UI (uiman), "ZoomIn");

      gtk_action_activate (action);
      return TRUE;
    }
  widget_class = GTK_WIDGET_CLASS (glook_tab_parent_class);
  if (widget_class->scroll_event)
    return widget_class->scroll_event (widget, event);

  return FALSE;
}

/**
 * glook_tab_create_tab_head:
 */
static GtkWidget*
glook_tab_create_tab_head (GLookTab *tab)
{
  GtkWidget *head, *hbox, *label, *button, *spinner, *image;

  head = gtk_event_box_new ();
  gtk_widget_show (head);
  gtk_widget_set_can_focus (head, TRUE);
  gtk_event_box_set_visible_window (GTK_EVENT_BOX (head), FALSE);
  gtk_event_box_set_above_child (GTK_EVENT_BOX (head), FALSE);

  /* popup context menu */
  g_signal_connect (head, "button-press-event", 
		    G_CALLBACK (on_tab_head_button_pressed), tab);

  hbox = gtk_hbox_new (FALSE, 4);
  gtk_widget_show (hbox);
  gtk_container_add (GTK_CONTAINER (head), hbox);

  /* spinner */
  spinner = gtk_spinner_new ();
  gtk_box_pack_start (GTK_BOX (hbox), spinner, FALSE, FALSE, 0);

  /* image */
  image = gtk_image_new_from_stock (GTK_STOCK_MISSING_IMAGE,
				    GTK_ICON_SIZE_MENU);
  gtk_widget_show (image);
  gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);

  /* label widget */
  label = gtk_label_new (_("Untitle"));
  gtk_widget_show (label);
  g_object_set (label, "xalign", 0.0, "xpad", 2,
		"ellipsize", PANGO_ELLIPSIZE_END,
		"use-markup", TRUE, NULL);
  gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);

  /* close button */
  button = gtk_button_new ();
  gtk_widget_show (button);
  gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
  gtk_button_set_image (GTK_BUTTON (button),
			gtk_image_new_from_stock (GTK_STOCK_CLOSE,
						  GTK_ICON_SIZE_MENU));
  gtk_widget_set_name (button, "glook-tab-close-button");
  gtk_widget_set_tooltip_text (button, _("Close tab"));
  gtk_box_pack_end (GTK_BOX (hbox), button, FALSE, FALSE, 0);

  g_signal_connect (button, "clicked",
		    G_CALLBACK (on_tab_close_button_clicked), tab);
  g_signal_connect (head, "style-set",
                    G_CALLBACK (on_tab_widget_style_set), button);

  /* save the label widget as object data */
  g_object_set_data (G_OBJECT (head), "spinner", spinner);
  g_object_set_data (G_OBJECT (head), "image", image);
  g_object_set_data (G_OBJECT (head), "label-widget", label);
  g_object_set_data (G_OBJECT (head), "close-button", button);
  g_object_set_data (G_OBJECT (head), "tab", tab);

  return head;
}

/**
 * on_tab_widget_style_set:
 */
static void
on_tab_widget_style_set (GtkWidget *hbox,
			 GtkStyle  *previous_style,
			 GtkWidget *button)
{
  gint h, w;

  gtk_icon_size_lookup_for_settings (gtk_widget_get_settings (button),
				     GTK_ICON_SIZE_MENU, &w, &h);
  gtk_widget_set_size_request (button, w + 2, h + 2);
}

/**
 * glook_tab_close:
 */
void
glook_tab_close (GLookTab *tab)
{
  GtkWidget *notebook;

  g_return_if_fail (GLOOK_IS_TAB (tab));

  notebook = gtk_widget_get_parent (GTK_WIDGET (tab));

  if (GLOOK_IS_NOTEBOOK (notebook))
    glook_notebook_remove (GLOOK_NOTEBOOK (notebook), GTK_WIDGET (tab));
  else
    gtk_widget_destroy (GTK_WIDGET (tab));
}

/**
 * on_tab_close_button_clicked:
 */
static void
on_tab_close_button_clicked (GtkWidget *button,
			     GLookTab  *tab)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));
  glook_tab_close (tab);
}

/**
 * on_detach_tab_item_activated:
 *
 * create a new window and move tab to that window.
 */
static void
on_detach_tab_item_activated (GtkWidget *item,
			      GLookTab  *tab)
{
  GtkWidget *window;
  GtkWidget *notebook;
  gint x, y, n;

  g_return_if_fail (GLOOK_IS_TAB (tab));

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));

  gtk_window_get_position (GTK_WINDOW (window), &x, &y);
  notebook = glook_window_get_notebook (GLOOK_WINDOW (window));

  /* remove tab from old window. */
  g_object_ref (tab);
  n = gtk_notebook_page_num (GTK_NOTEBOOK (notebook), GTK_WIDGET (tab));
  gtk_notebook_remove_page (GTK_NOTEBOOK (notebook), n);

  /* create a new window to append tab. */
  window = glook_app_create_window (glook_app_get (), FALSE);
  notebook = glook_window_get_notebook (GLOOK_WINDOW (window));
  glook_notebook_append (GLOOK_NOTEBOOK (notebook), GTK_WIDGET (tab));
  g_object_unref (tab);

  gtk_window_move (GTK_WINDOW (window), x + 10, y + 10);
  gtk_widget_show (window);
}

/**
 * on_close_tab_item_activated:
 */
static void
on_close_tab_item_activated (GtkWidget *item,
			     GLookTab  *tab)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));
  glook_tab_close (tab);
}

/**
 * on_close_other_tabs_item_activated:
 */
static void
on_close_other_tabs_item_activated (GtkWidget *item,
				    GLookTab  *tab)
{
  GtkWidget *notebook;

  g_return_if_fail (GLOOK_IS_TAB (tab));

  /* assume already packing into notebook. */
  notebook = gtk_widget_get_parent (GTK_WIDGET (tab));
  if (!GLOOK_IS_NOTEBOOK (notebook))
    return;

  glook_notebook_remove_others (GLOOK_NOTEBOOK (notebook),
				GTK_WIDGET (tab));
}

/**
 * on_reload_item_activated:
 */
static void
on_reload_item_activated (GtkWidget *item,
			  GLookTab  *tab)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));

  /* call subclass method. */
  glook_tab_reload (tab);
}

/**
 * on_new_tab_item_activated:
 */
static void
on_new_tab_item_activated (GtkWidget *item,
			   GLookTab  *tab)
{
  GtkWidget *newtab;
  GtkWidget *notebook;
  gint n;

  g_return_if_fail (GLOOK_IS_TAB (tab));

  notebook = gtk_widget_get_parent (GTK_WIDGET (tab));
  if (!GLOOK_IS_NOTEBOOK (notebook))
    return;

  /* create a new blank tab and insert it just after
   * current active tab. */
  newtab = glook_blank_tab_new ();
  glook_blank_tab_set_uri (GLOOK_BLANK_TAB (newtab));
  gtk_widget_show (newtab);

  n = glook_notebook_get_nth (GLOOK_NOTEBOOK (notebook),
			      GTK_WIDGET (tab));
  glook_notebook_insert (GLOOK_NOTEBOOK (notebook), newtab, n + 1);
  glook_notebook_set_current_n (GLOOK_NOTEBOOK (notebook), n + 1);
}

/**
 * glook_tab_create_menu:
 */
static void
glook_tab_create_menu (GLookTab *tab)
{
  GtkWidget *item;

  if (!tab->menu)
    {
      tab->menu = gtk_menu_new ();

      /* new tab item */
      item = gtk_image_menu_item_new_from_stock (GTK_STOCK_NEW, NULL);
      g_object_set (item, "label", _("New _tab"), NULL);
      g_signal_connect (item, "activate",
			G_CALLBACK (on_new_tab_item_activated), tab);
      gtk_menu_shell_append (GTK_MENU_SHELL (tab->menu), item);

      /* reload item */
      item = gtk_image_menu_item_new_from_stock (GTK_STOCK_REFRESH, NULL);
      g_object_set (item, "label", _("_Reload"), NULL);
      g_signal_connect (item, "activate",
			G_CALLBACK (on_reload_item_activated), tab);
      g_object_set_data (G_OBJECT (tab->menu), "reload-item", item);
      gtk_menu_shell_append (GTK_MENU_SHELL (tab->menu), item);

      item = gtk_separator_menu_item_new ();
      gtk_menu_shell_append (GTK_MENU_SHELL (tab->menu), item);

      /* detach item */
      item = gtk_menu_item_new_with_mnemonic (_("_Detach tab"));
      g_signal_connect (item, "activate",
			G_CALLBACK (on_detach_tab_item_activated), tab);
      gtk_menu_shell_append (GTK_MENU_SHELL (tab->menu), item);

      /* close tab item */
      item = gtk_image_menu_item_new_with_mnemonic (_("_Close tab"));
      gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item),
				     gtk_image_new_from_stock (GTK_STOCK_CLOSE,
							       GTK_ICON_SIZE_MENU));
      g_signal_connect (item, "activate",
			G_CALLBACK (on_close_tab_item_activated), tab);
      gtk_menu_shell_append (GTK_MENU_SHELL (tab->menu), item);

      /* close other tabs item */
      item = gtk_menu_item_new_with_mnemonic (_("Close _other tabs"));
      g_signal_connect (item, "activate",
			G_CALLBACK (on_close_other_tabs_item_activated), tab);
      gtk_menu_shell_append (GTK_MENU_SHELL (tab->menu), item);

      gtk_menu_attach_to_widget (GTK_MENU (tab->menu),
				 GTK_WIDGET (tab), NULL);
    }
  /* not all subclass support reload. */
  item = g_object_get_data (G_OBJECT (tab->menu), "reload-item");
  gtk_widget_set_sensitive (item, GLOOK_TAB_REFRESH_SUPPORT (tab));

  gtk_widget_show_all (tab->menu);
}

/**
 * on_tab_head_button_pressed:
 * 
 * Returns: TRUE to stop other handlers from being invoked 
 * for the event. FALSE to propagate the event further. 
 */
static gboolean
on_tab_head_button_pressed (GtkWidget      *widget,
			    GdkEventButton *event,
			    GLookTab       *tab)
{
  g_return_val_if_fail (GLOOK_IS_TAB (tab), FALSE);

  /* ignore double-clicks and triple-clicks */
  if (event->button == 3 && event->type == GDK_BUTTON_PRESS)
    {
      glook_tab_create_menu (tab);

      gtk_menu_popup (GTK_MENU (tab->menu),
		      NULL, NULL, NULL, NULL, 
		      event ? event->button : 0,
		      event ? event->time :
		      gtk_get_current_event_time ());
      return TRUE;
    }
  return FALSE;
}

/**
 * glook_tab_get_tab_head:
 */
GtkWidget*
glook_tab_get_tab_head (GLookTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_TAB (tab), NULL);
  return tab->tab_head;
}

/**
 * glook_tab_get_label:
 */
const gchar*
glook_tab_get_label (GLookTab *tab)
{
  GtkWidget *label;

  g_return_val_if_fail (GLOOK_IS_TAB (tab), NULL);

  label = g_object_get_data (G_OBJECT (tab->tab_head),
			     "label-widget");
  return gtk_label_get_label (GTK_LABEL (label));
}

/**
 * glook_tab_get_label_text:
 */
const gchar*
glook_tab_get_label_text (GLookTab *tab)
{
  GtkWidget *label;

  g_return_val_if_fail (GLOOK_IS_TAB (tab), NULL);

  label = g_object_get_data (G_OBJECT (tab->tab_head),
			     "label-widget");
  return gtk_label_get_text (GTK_LABEL (label));
}

/**
 * glook_tab_set_label:
 *
 * the @string maybe plain text or markup language.
 */
void
glook_tab_set_label (GLookTab    *tab,
		     const gchar *string)
{
  GtkWidget *label;

  g_return_if_fail (GLOOK_IS_TAB (tab));
  g_return_if_fail (string);

  label = g_object_get_data (G_OBJECT (tab->tab_head), "label-widget");
  gtk_label_set_label (GTK_LABEL (label), string);

  g_object_notify (G_OBJECT (tab), "label");
}

/**
 * glook_tab_get_image:
 */
const gchar*
glook_tab_get_image (GLookTab *tab)
{
  GtkWidget *image;
  const gchar *name;

  g_return_val_if_fail (GLOOK_IS_TAB (tab), NULL);

  image = g_object_get_data (G_OBJECT (tab->tab_head), "image");
  gtk_image_get_icon_name (GTK_IMAGE (image), &name, NULL);

  return name;
}

/**
 * glook_tab_set_image:
 */
void
glook_tab_set_image (GLookTab    *tab,
		     const gchar *icon_name)
{
  GtkWidget *image;

  g_return_if_fail (GLOOK_IS_TAB (tab));
  g_return_if_fail (icon_name);

  image = g_object_get_data (G_OBJECT (tab->tab_head), "image");
  gtk_image_set_from_icon_name (GTK_IMAGE (image), icon_name,
				GTK_ICON_SIZE_MENU);
  g_object_notify (G_OBJECT (tab), "image");
}

/**
 * glook_tab_get_stock:
 */
const gchar*
glook_tab_get_stock (GLookTab *tab)
{
  GtkWidget *image;
  const gchar *stock;

  g_return_val_if_fail (GLOOK_IS_TAB (tab), NULL);

  image = g_object_get_data (G_OBJECT (tab->tab_head), "image");
  gtk_image_get_stock (GTK_IMAGE (image), (gchar**) &stock, NULL);

  return stock;
}

/**
 * glook_tab_set_stock:
 */
void
glook_tab_set_stock (GLookTab    *tab,
		     const gchar *stock)
{
  GtkWidget *image;

  g_return_if_fail (GLOOK_IS_TAB (tab));
  g_return_if_fail (stock);

  image = g_object_get_data (G_OBJECT (tab->tab_head), "image");
  gtk_image_set_from_stock (GTK_IMAGE (image), stock,
			    GTK_ICON_SIZE_MENU);
  g_object_notify (G_OBJECT (tab), "stock");
}

/**
 * glook_tab_set_image_visible:
 */
void
glook_tab_set_image_visible (GLookTab    *tab,
			     gboolean     visible)
{
  GtkWidget *image;

  g_return_if_fail (GLOOK_IS_TAB (tab));

  image = g_object_get_data (G_OBJECT (tab->tab_head), "image");
  gtk_widget_set_visible (image, visible);
}

/**
 * glook_tab_set_spinner_visible:
 */
void
glook_tab_set_spinner_visible (GLookTab *tab,
			       gboolean  visible)
{
  GtkWidget *spinner;

  g_return_if_fail (GLOOK_IS_TAB (tab));

  spinner = g_object_get_data (G_OBJECT (tab->tab_head), "spinner");
  if (visible)
    {
      gtk_spinner_start (GTK_SPINNER (spinner));
      gtk_widget_set_visible (spinner, TRUE);
    }
  else
    {
      gtk_spinner_stop (GTK_SPINNER (spinner));
      gtk_widget_set_visible (spinner, FALSE);
    }
}

/**
 * glook_tab_get_uri:
 */
GLookUri*
glook_tab_get_uri (GLookTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_TAB (tab), NULL);
  return tab->uri;
}

/**
 * glook_tab_set_uri:
 *
 * this function will transfer the uri's owner to tab.
 */
void
glook_tab_set_uri (GLookTab *tab,
		   GLookUri *uri)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));
  g_return_if_fail (uri);

  if (tab->uri)
    glook_uri_free (tab->uri);

  tab->uri = glook_uri_ref (uri);
}

/**
 * glook_tab_get_progress:
 */
gboolean
glook_tab_get_progress (GLookTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_TAB (tab), FALSE);
  return tab->progress;
}

/**
 * glook_tab_set_progress:
 */
void
glook_tab_set_progress (GLookTab *tab,
			gboolean  progress)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));
  tab->progress = progress;
}

/**
 * glook_tab_get_zoom_index:
 */
gint
glook_tab_get_zoom_index (GLookTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_TAB (tab), FALSE);
  return tab->zoom_index;
}

/**
 * glook_tab_get_zoom_factor:
 */
gint
glook_tab_get_zoom_factor (GLookTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_TAB (tab), FALSE);
  return _zoom_factors[tab->zoom_index];
}

/**
 * glook_tab_merge_ui:
 *
 * child class should override this method in order to
 * merge child special ui elements into window.
 */
void
glook_tab_merge_ui (GLookTab  *tab,
		    GtkWidget *window)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));

  if (GLOOK_TAB_GET_CLASS (tab)->merge_ui)
    GLOOK_TAB_GET_CLASS (tab)->merge_ui (tab, window);
}

/**
 * glook_tab_unmerge_ui:
 */
void
glook_tab_unmerge_ui (GLookTab  *tab,
		      GtkWidget *window)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));

  if (GLOOK_TAB_GET_CLASS (tab)->unmerge_ui)
    GLOOK_TAB_GET_CLASS (tab)->unmerge_ui (tab, window);
}

/**
 * glook_tab_open:
 */
void
glook_tab_open (GLookTab *tab)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));

  if (GLOOK_TAB_GET_CLASS (tab)->open)
    GLOOK_TAB_GET_CLASS (tab)->open (tab);
}

/**
 * glook_tab_stop:
 */
void
glook_tab_stop (GLookTab *tab)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));

  if (GLOOK_TAB_GET_CLASS (tab)->stop)
    GLOOK_TAB_GET_CLASS (tab)->stop (tab);
}

/**
 * glook_tab_is_active:
 */
gboolean
glook_tab_is_active (GLookTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_TAB (tab), FALSE);

  if (GLOOK_TAB_GET_CLASS (tab)->is_active)
    return GLOOK_TAB_GET_CLASS (tab)->is_active (tab);

  return FALSE;
}

/**
 * glook_tab_zoom:
 */
gboolean
glook_tab_zoom (GLookTab *tab,
		gint      index)
{
  g_return_val_if_fail (GLOOK_IS_TAB (tab), FALSE);

  if (index >= 0 && index < n_zoom_factors)
    {
      gint factor;

      factor = _zoom_factors[index];

      if (GLOOK_TAB_GET_CLASS (tab)->zoom)
	{
	  if (GLOOK_TAB_GET_CLASS (tab)->zoom (tab, factor))
	    {
	      tab->zoom_index = index;
	      return TRUE;
	    }
	}
    }
  return FALSE;
}

/**
 * glook_tab_reload:
 */
void
glook_tab_reload (GLookTab *tab)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));

  if (GLOOK_TAB_GET_CLASS (tab)->reload)
    GLOOK_TAB_GET_CLASS (tab)->reload (tab);
}

/**
 * glook_tab_find:
 */
void
glook_tab_find (GLookTab *tab)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));

  if (GLOOK_TAB_GET_CLASS (tab)->find)
    GLOOK_TAB_GET_CLASS (tab)->find (tab);
}

/**
 * glook_tab_geometry_hint:
 */
void
glook_tab_geometry_hint (GLookTab *tab)
{
  g_return_if_fail (GLOOK_IS_TAB (tab));

  if (GLOOK_TAB_GET_CLASS (tab)->geometry_hint)
    GLOOK_TAB_GET_CLASS (tab)->geometry_hint (tab);
}
