/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-app.h"
#include "glook-misc.h"
#include "glook-terminal.h"
#include "glook-local-tab.h"
#include "glook-local-preference.h"
#include "glook-window.h"
#include "glook-profile-dialog.h"

static void	on_terminal_open_link_activated		(GtkAction       *action,
							 GLookTerminal   *terminal);
static void	on_terminal_copy_link_activated		(GtkAction       *action,
							 GLookTerminal   *terminal);
static void	on_terminal_copy_activated		(GtkAction       *action,
							 GLookTerminal   *terminal);
static void	on_terminal_select_all_activated	(GtkAction       *action,
							 GLookTerminal   *terminal);
static void	on_reset_terminal_activated		(GtkAction       *action,
							 GLookTerminal   *terminal);
static void	on_reset_clear_terminal_activated	(GtkAction       *action,
							 GLookTerminal   *terminal);
static void	on_save_terminal_screen_activated	(GtkAction       *action,
							 GLookTerminal   *terminal);
static void	on_terminal_encoding_current_activated	(GtkAction       *action,
							 GLookTerminal   *terminal);
static void	on_terminal_encoding_setting_activated	(GtkAction       *action,
							 GLookTerminal   *terminal);
static void	on_terminal_edit_profile_activated	(GtkAction       *action,
							 GLookTerminal   *terminal);
static void	on_terminal_manage_profiles_activated	(GtkAction       *action,
							 GLookTerminal   *terminal);
static void	on_terminal_profile_action_changed	(GtkRadioAction  *action,
							 GtkRadioAction  *current,
							 GLookTerminal   *terminal);

GtkActionEntry _terminal_actions[] =
{
    { "TerminalEncodingMenu", NULL, N_("_Terminal encoding"), },
    { "TerminalProfileMenu", NULL, N_("Chan_ge terminal profile"), },

    { "TerminalOpenLink", NULL, N_("_Open link"), NULL,
      N_("Open link address"),
      G_CALLBACK (on_terminal_open_link_activated),
    },
    { "TerminalMailto", NULL, N_("Send e_mail"), NULL,
      N_("Send email to this address"),
      G_CALLBACK (on_terminal_open_link_activated),
    },
    { "TerminalCallto", NULL, N_("C_all"), NULL,
      N_("Call this address"),
      G_CALLBACK (on_terminal_open_link_activated),
    },
    { "TerminalCopyLink", NULL, N_("Copy _link address"), NULL,
      N_("Open link address"),
      G_CALLBACK (on_terminal_copy_link_activated),
    },
    { "TerminalCopy", GTK_STOCK_COPY, NULL, NULL,
      N_("Copy selected text to clipboard"),
      G_CALLBACK (on_terminal_copy_activated),
    },
    { "TerminalSelectAll", GTK_STOCK_SELECT_ALL, NULL, NULL,
      N_("selects all text within the terminal"),
      G_CALLBACK (on_terminal_select_all_activated),
    },
    { "TerminalReset", NULL, N_("_Reset terminal"), NULL,
      N_("Resets as much of the terminal's internal state"),
      G_CALLBACK (on_reset_terminal_activated),
    },
    { "TerminalResetClear", NULL, N_("Reset and _clear terminal"), NULL,
      N_("Resets terminal's internal state and clear screen"),
      G_CALLBACK (on_reset_clear_terminal_activated),
    },
    { "TerminalSaveScreen", NULL, N_("Sa_ve terminal screen"), NULL,
      N_("Resets terminal's internal state and clear screen"),
      G_CALLBACK (on_save_terminal_screen_activated),
    },
    { "TerminalEncodingCurrent", NULL, N_("Current _Locale"), NULL,
      N_("Set terminal encoding to current locale"),
      G_CALLBACK (on_terminal_encoding_current_activated),
    },
    { "TerminalEncodingSetting", NULL, N_("_Add or Remove"), NULL,
      N_("Add or remove selected encodings"),
      G_CALLBACK (on_terminal_encoding_setting_activated),
    },
    { "TerminalEditProfile", NULL, N_("Edit terminal pro_file"), NULL,
      N_("Edit current terminal profile"),
      G_CALLBACK (on_terminal_edit_profile_activated),
    },
    { "TerminalManageProfile", NULL, N_("_Manage profiles"), NULL,
      N_("Manage terminal profiles"),
      G_CALLBACK (on_terminal_manage_profiles_activated),
    },
};
static GtkRadioActionEntry _terminal_radio_actions[] =
{
    { "TerminalProfile1", NULL, "profile1", NULL,
      N_("Change terminal profile"), 1,
    },
    { "TerminalProfile2", NULL, "profile2", NULL,
      N_("Change terminal profile"), 2,
    },
    { "TerminalProfile3", NULL, "profile3", NULL,
      N_("Change terminal profile"), 3,
    },
    { "TerminalProfile4", NULL, "profile4", NULL,
      N_("Change terminal profile"), 4,
    },
    { "TerminalProfile5", NULL, "profile5", NULL,
      N_("Change terminal profile"), 5,
    },
};

/**
 * map_boolean_to_scrollbar_policy:
 */
static gboolean
map_boolean_to_scrollbar_policy (GValue   *result,
				 GVariant *variant,
				 gpointer  data)
{
  GtkPolicyType policy;

  if (g_variant_get_boolean (variant))
    policy = GTK_POLICY_ALWAYS;
  else
    policy = GTK_POLICY_NEVER;

  g_value_set_enum (result, policy);
  return TRUE;
}

/**
 * get_active_profile_action:
 */
static gint
get_active_profile_action (const gchar *profile)
{
  if (g_strcmp0 (profile, "profile1") == 0)
    return 1;
  if (g_strcmp0 (profile, "profile2") == 0)
    return 2;
  if (g_strcmp0 (profile, "profile3") == 0)
    return 3;
  if (g_strcmp0 (profile, "profile4") == 0)
    return 4;
  if (g_strcmp0 (profile, "profile5") == 0)
    return 5;

  return -1;
}

/**
 * bind_profile_action_settings:
 */
static void
bind_profile_action_settings (GtkActionGroup *group)
{
  GSettings *settings;
  GtkAction *action;

  g_return_if_fail (GTK_IS_ACTION_GROUP (group));

#if 0
  settings = glook_app_get_settings (glook_app_get (), SETTINGS_ACCELERATOR);

  action = gtk_action_group_get_action (group, "TerminalDisableAccelerator");
  g_settings_bind (settings, "disable-accelerators", action, "active",
		   G_SETTINGS_BIND_DEFAULT);
  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalDisableMnemonic");
  g_settings_bind (settings, "disable-mnemonics", action, "active",
		   G_SETTINGS_BIND_DEFAULT);
#endif

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_PROFILE1);
  action = gtk_action_group_get_action (group, "TerminalProfile1");
  g_settings_bind (settings, "name", action, "label", G_SETTINGS_BIND_GET);

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_PROFILE2);
  action = gtk_action_group_get_action (group, "TerminalProfile2");
  g_settings_bind (settings, "name", action, "label", G_SETTINGS_BIND_GET);

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_PROFILE3);
  action = gtk_action_group_get_action (group, "TerminalProfile3");
  g_settings_bind (settings, "name", action, "label", G_SETTINGS_BIND_GET);

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_PROFILE4);
  action = gtk_action_group_get_action (group, "TerminalProfile4");
  g_settings_bind (settings, "name", action, "label", G_SETTINGS_BIND_GET);

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_PROFILE5);
  action = gtk_action_group_get_action (group, "TerminalProfile5");
  g_settings_bind (settings, "name", action, "label", G_SETTINGS_BIND_GET);
}

/**
 * glook_terminal_add_actions:
 */
void
glook_terminal_add_actions (GtkActionGroup *group,
			    GtkWidget      *terminal,
			    GtkWidget      *scrolled,
			    const gchar    *profile)
{
  GtkAction *action;
  gchar label[256];
  const gchar *charset;
  GSettings *settings;

  g_return_if_fail (GTK_IS_ACTION_GROUP (group));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (GTK_IS_SCROLLED_WINDOW (scrolled));

  gtk_action_group_add_actions (group, _terminal_actions,
				G_N_ELEMENTS (_terminal_actions), terminal);

  gtk_action_group_add_radio_actions (group, _terminal_radio_actions,
				      G_N_ELEMENTS (_terminal_radio_actions),
				      get_active_profile_action (profile),
				      G_CALLBACK (on_terminal_profile_action_changed),
				      terminal);
  g_get_charset (&charset);
  g_snprintf (label, sizeof (label), _("Current _Locale (%s)"), charset);

  action = gtk_action_group_get_action (group, "TerminalEncodingCurrent");
  gtk_action_set_label (action, label);

  /* when settings changed, we have to update scrolled window policy. */
  settings = glook_app_get_profile_settings (glook_app_get (), profile);
  g_settings_bind_with_mapping (settings, "show-scrollbar",
				scrolled, "vscrollbar-policy",
				G_SETTINGS_BIND_GET,
				map_boolean_to_scrollbar_policy,
				NULL, NULL, NULL);

  bind_profile_action_settings (group);
}

/**
 * on_terminal_open_link_activated:
 */
static void
on_terminal_open_link_activated (GtkAction     *action,
				 GLookTerminal *terminal)
{
  GLookTerminalPopupInfo *info;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  //info = g_object_get_data (G_OBJECT (terminal), "popup-info");
  info = glook_terminal_get_popup_info (terminal);
  g_assert (info);

  if (info->address)
    {
      GtkWidget *window;

      window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));
      glook_open_url (window, info->address);
    }
}

/**
 * on_terminal_copy_link_activated:
 */
static void
on_terminal_copy_link_activated (GtkAction    *action,
				 GLookTerminal *terminal)
{
  GLookTerminalPopupInfo *info;
  GtkClipboard *clipboard;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  clipboard = gtk_widget_get_clipboard (GTK_WIDGET (terminal),
					GDK_SELECTION_CLIPBOARD);
  if (clipboard)
    {
      //info = g_object_get_data (G_OBJECT (terminal), "popup-info");
      info = glook_terminal_get_popup_info (terminal);
      g_assert (info);

      gtk_clipboard_set_text (clipboard, info->address, -1);
    }
}

/**
 * on_terminal_copy_activated:
 */
static void
on_terminal_copy_activated (GtkAction     *action,
			    GLookTerminal *terminal)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  vte_terminal_copy_clipboard (VTE_TERMINAL (terminal));
}

/**
 * on_terminal_select_all_activated:
 */
static void
on_terminal_select_all_activated (GtkAction     *action,
				  GLookTerminal *terminal)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  vte_terminal_select_all (VTE_TERMINAL (terminal));
}

/**
 * on_reset_terminal_activated:
 */
static void
on_reset_terminal_activated (GtkAction     *action,
			     GLookTerminal *terminal)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  vte_terminal_reset (VTE_TERMINAL (terminal), TRUE, FALSE);
}

/**
 * on_reset_clear_terminal_activated:
 */
static void
on_reset_clear_terminal_activated (GtkAction     *action,
				   GLookTerminal *terminal)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  vte_terminal_reset (VTE_TERMINAL (terminal), TRUE, TRUE);
}

/**
 * on_save_terminal_screen_activated:
 *
 * called when save terminal screen action activated.
 */
static void
on_save_terminal_screen_activated (GtkAction     *action,
				   GLookTerminal *terminal)
{
  GtkWidget *window;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));
  glook_terminal_save_screen_dialog (terminal, window);
}

/**
 * on_terminal_encoding_current_activated:
 */
static void
on_terminal_encoding_current_activated (GtkAction     *action,
					GLookTerminal *terminal)
{
  const gchar *encoding;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  g_get_charset (&encoding);
  vte_terminal_set_encoding (VTE_TERMINAL (terminal), encoding);
}

/**
 * on_terminal_encoding_setting_activated:
 */
static void
on_terminal_encoding_setting_activated (GtkAction     *action,
					GLookTerminal *terminal)
{
  GtkWidget *window;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));
  if (GTK_IS_WINDOW (window))
    {
      GtkWidget *tab;
      const gchar *uri = "local:path=preference.xml";

      tab = glook_window_open_uri_string (GLOOK_WINDOW (window), uri);

      /* ensure the tab is actually opened. */
      while (gtk_events_pending ())
	gtk_main_iteration ();

      glook_local_preference_select_encoding_page (GLOOK_LOCAL_TAB (tab));
    }
}

/**
 * on_profile_dialog_responsed:
 */
static void
on_profile_dialog_responsed (GtkWidget     *dialog,
			     gint           resp_id,
			     GLookTerminal *terminal)
{
  g_return_if_fail (GTK_IS_DIALOG (dialog));

  if (resp_id != GTK_RESPONSE_HELP)
    {
      gtk_widget_destroy (dialog);
    }
}

/**
 * on_terminal_edit_profile_activated:
 */
static void
on_terminal_edit_profile_activated (GtkAction     *action,
				    GLookTerminal *terminal)
{
  GtkWidget *dialog;
  GtkWidget *window;
  const gchar *profile;
  GSettings *settings;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  dialog = glook_profile_dialog_new ();

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));
  gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
  gtk_window_set_destroy_with_parent (GTK_WINDOW (dialog), TRUE);

  profile = glook_terminal_get_profile (terminal);
  settings = glook_app_get_profile_settings (glook_app_get (), profile);
  glook_profile_dialog_set_profile (GLOOK_PROFILE_DIALOG (dialog),
				    profile, settings);
  g_signal_connect (dialog, "response",
		    G_CALLBACK (on_profile_dialog_responsed), terminal);
  gtk_widget_show (dialog);
}

/**
 * on_terminal_manage_profiles_activated:
 */
static void
on_terminal_manage_profiles_activated (GtkAction     *action,
				       GLookTerminal *terminal)
{
  GtkWidget *window;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));
  if (GLOOK_IS_WINDOW (window))
    {
      GtkWidget *tab;
      const gchar *uri = "local:path=preference.xml";

      tab = glook_window_open_uri_string (GLOOK_WINDOW (window), uri);

      /* ensure the tab is actually opened. */
      while (gtk_events_pending ())
	gtk_main_iteration ();

      glook_local_preference_select_terminal_page (GLOOK_LOCAL_TAB (tab));
    }
}

/**
 * on_terminal_profile_action_changed:
 */
static void
on_terminal_profile_action_changed (GtkRadioAction *action,
				    GtkRadioAction *current,
				    GLookTerminal  *terminal)
{
  GSettings *settings;
  gchar profile[32];
  GtkWidget *scrolled;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GTK_IS_ACTION (current));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  scrolled = gtk_widget_get_parent (GTK_WIDGET (terminal));

  g_snprintf (profile, sizeof (profile), "profile%d",
	      gtk_radio_action_get_current_value (current));

  if (GTK_IS_SCROLLED_WINDOW (scrolled))
    {
      /* unbind first. */
      const gchar *prev_profile = glook_terminal_get_profile (terminal);
      settings = glook_app_get_profile_settings (glook_app_get (), prev_profile);
      g_settings_unbind (scrolled, "vscrollbar-policy");

      /* when settings changed, we have to update scrolled window policy. */
      settings = glook_app_get_profile_settings (glook_app_get (), profile);
      g_settings_bind_with_mapping (settings, "show-scrollbar",
				    scrolled, "vscrollbar-policy",
				    G_SETTINGS_BIND_GET,
				    map_boolean_to_scrollbar_policy,
				    NULL, NULL, NULL);
    }
  /* change terminal profile. */
  glook_terminal_set_profile (terminal, profile);
}

