/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-window.h"
#include "glook-tab.h"
#include "glook-ui.h"
#include "glook-actions.h"
#include "glook-addrbar.h"
#include "glook-toolbar.h"

struct _GLookToolbarPriv
{
  GtkWidget      *addrbar;
  GtkWidget      *zoom_in;
  GtkWidget      *zoom_out;
  GtkWidget      *zoom_combo;
};
#define _TOOLBAR_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_TOOLBAR, GLookToolbarPriv))

G_DEFINE_TYPE (GLookToolbar, glook_toolbar, GTK_TYPE_TOOLBAR);

/* defined in glook-tab.c */
extern const gint	_zoom_factors[];
extern const gint	n_zoom_factors;

/**
 * glook_toolbar_init:
 */
static void
glook_toolbar_init (GLookToolbar *toolbar)
{
  GtkToolItem *item;

  toolbar->priv = _TOOLBAR_GET_PRIVATE (toolbar);

  toolbar->priv->addrbar = glook_addrbar_new ();

  gtk_widget_show (toolbar->priv->addrbar);

  /* create a toolitem to hold the address bar */
  item = gtk_tool_item_new ();
  gtk_tool_item_set_expand (item, TRUE);
  gtk_tool_item_set_visible_vertical (item, FALSE);
  gtk_widget_show (GTK_WIDGET (item));

  /* insert the toolitem(address bar) to toolbar */
  gtk_container_add (GTK_CONTAINER (item), toolbar->priv->addrbar);
  gtk_widget_show (GTK_WIDGET (item));
  gtk_toolbar_insert (GTK_TOOLBAR (toolbar), item, 0);
}

/**
 * glook_toolbar_dispose:
 */
static void
glook_toolbar_dispose (GObject *object)
{
#if 0
  GLookToolbar *toolbar = GLOOK_TOOLBAR (object);

#endif
  G_OBJECT_CLASS (glook_toolbar_parent_class)->dispose (object);
}

/**
 * glook_toolbar_finalize:
 */
static void
glook_toolbar_finalize (GObject *object)
{
#if 0
  GLookToolbar *toolbar = GLOOK_TOOLBAR (object);

#endif
  G_OBJECT_CLASS (glook_toolbar_parent_class)->finalize (object);
}

/**
 * glook_toolbar_zoom_real:
 */
static void
glook_toolbar_zoom_real (GLookToolbar *toolbar,
			 gint          factor)
{
}

/**
 * glook_toolbar_class_init:
 */
static void
glook_toolbar_class_init (GLookToolbarClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookToolbarPriv));

  gobject_class->dispose  = glook_toolbar_dispose;
  gobject_class->finalize = glook_toolbar_finalize;

  klass->zoom = glook_toolbar_zoom_real;

  g_signal_new ("zoom", GLOOK_TYPE_TOOLBAR, G_SIGNAL_RUN_LAST,
		G_STRUCT_OFFSET (GLookToolbarClass, zoom),
		NULL, NULL,
		g_cclosure_marshal_VOID__INT,
		G_TYPE_NONE, 1, G_TYPE_INT);
}

/**
 * zoom_combo_row_separator_func:
 */
static gboolean
zoom_combo_row_separator_func (GtkTreeModel *model,
			       GtkTreeIter  *iter,
			       gpointer      data)
{
  gboolean separator;

  gtk_tree_model_get (model, iter, 2, &separator, -1);
  return separator;
}

/**
 * glook_toolbar_create_zoom_combo:
 */
static GtkWidget*
glook_toolbar_create_zoom_combo (GLookToolbar *toolbar)
{
  GtkWidget *combo;
  GtkListStore *store;
  GtkCellRenderer *renderer;
  gint i;

  /* create the list store with 3 columns:
   * 1. zoom factor;
   * 2. title for display;
   * 3. is a separator;
   */
  store = gtk_list_store_new (3, G_TYPE_INT, G_TYPE_STRING,
			      G_TYPE_BOOLEAN);
  for (i = 0; i < n_zoom_factors; i++)
    {
      gchar title[64];

      /* 0: separator */
      if (_zoom_factors[i] == 0)
	{
	  gtk_list_store_insert_with_values (store, NULL, -1, 0, 0,
					     1, "", 2, TRUE, -1);
	  continue;
	}
      g_snprintf (title, sizeof (title), "%%%d", _zoom_factors[i]);

      gtk_list_store_insert_with_values (store, NULL, -1,
					 0, _zoom_factors[i],
					 1, title, 2, FALSE, -1);
    }
  /* create the combo box */
  combo = gtk_combo_box_new_with_model (GTK_TREE_MODEL (store));
  g_object_unref (store);

  /* set row separator function. */
  gtk_combo_box_set_row_separator_func (GTK_COMBO_BOX (combo),
					zoom_combo_row_separator_func,
					toolbar, NULL);
  /* don't grab the focus when it is clicked with mouse */
  gtk_combo_box_set_focus_on_click (GTK_COMBO_BOX (combo), FALSE);

  /* trace this combo. */
  toolbar->priv->zoom_combo = combo;

  /* append renderer column */
  renderer = gtk_cell_renderer_text_new ();
  gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (combo), renderer, TRUE);
  gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (combo), renderer,
				  "text", 1, NULL);
  /* set default value. */
  gtk_combo_box_set_active (GTK_COMBO_BOX (combo),
			    glook_tab_get_zoom_100_index ());
  return combo;
}

/**
 * on_zoom_combo_changed:
 */
static void
on_zoom_combo_changed (GtkComboBox  *combo,
		       GLookToolbar *toolbar)
{
  gint active;

  g_return_if_fail (GTK_IS_COMBO_BOX (combo));
  g_return_if_fail (GLOOK_IS_TOOLBAR (toolbar));

  active = gtk_combo_box_get_active (combo);

  /* chanage zoom in/out sensitive */
  gtk_widget_set_sensitive (toolbar->priv->zoom_in, TRUE);
  gtk_widget_set_sensitive (toolbar->priv->zoom_out, TRUE);

  /* cannot zoom out more. */
  if (active == 0)
    gtk_widget_set_sensitive (toolbar->priv->zoom_out, FALSE);

  /* cannot zoom in more. */
  if (active == n_zoom_factors - 1)
    gtk_widget_set_sensitive (toolbar->priv->zoom_in, FALSE);

  g_signal_emit_by_name (toolbar, "zoom", active);
}

/**
 * on_play_file_button_clicked:
 */
static void
on_play_file_button_clicked (GtkButton *button,
			     GtkWidget *window)
{
  glook_open_record_file (GLOOK_WINDOW (window));
}

/**
 * create_play_recent_button:
 */
static GtkWidget*
create_play_recent_button (GtkWidget *window)
{
  GtkRecentManager *manager;
  GtkRecentFilter *filter;
  GtkWidget *menu;
  GtkToolItem *toolitem;

  manager = gtk_recent_manager_get_default ();
  menu = gtk_recent_chooser_menu_new_for_manager (manager);
  g_object_set (menu, "limit", 20, "show-numbers", TRUE, NULL);

  g_signal_connect (menu, "item-activated",
		    G_CALLBACK (on_app_play_recent_activated), window);

  filter = gtk_recent_filter_new ();
  gtk_recent_filter_set_name (filter, _("Terminal session record file"));
  gtk_recent_filter_add_mime_type (filter, "application/x-ttyrcd");
  gtk_recent_chooser_set_filter (GTK_RECENT_CHOOSER (menu), filter);

  toolitem = gtk_menu_tool_button_new_from_stock (GTK_STOCK_MEDIA_PLAY);
  gtk_menu_tool_button_set_menu (GTK_MENU_TOOL_BUTTON (toolitem), menu);
  gtk_tool_item_set_tooltip_text (toolitem, _("Playback selected file"));
  gtk_menu_tool_button_set_arrow_tooltip_text (GTK_MENU_TOOL_BUTTON (toolitem),
					       _("Playback recently used file"));
  g_signal_connect (toolitem, "clicked",
		    G_CALLBACK (on_play_file_button_clicked), window);
  gtk_widget_show (GTK_WIDGET (toolitem));
  return GTK_WIDGET (toolitem);
}

/**
 * glook_toolbar_new:
 */
GtkWidget*
glook_toolbar_new (GtkWidget *window)
{
  GtkWidget *toolbar;
  GtkUIManager *uiman;
  GtkAction *action;
  GtkWidget *item;
  GtkToolItem *toolitem;
  GtkWidget *combo;

  toolbar = g_object_new (GLOOK_TYPE_TOOLBAR, NULL);

  uiman = glook_window_get_uiman (GLOOK_WINDOW (window));
  g_assert (uiman);

  /* insert the refresh button */
  action = glook_ui_get_window_action (GLOOK_UI (uiman), "Reload");
  item = gtk_action_create_tool_item (action);
  gtk_toolbar_insert (GTK_TOOLBAR (toolbar), GTK_TOOL_ITEM (item), 0);

  toolitem = gtk_separator_tool_item_new ();
  gtk_widget_show (GTK_WIDGET (toolitem));
  gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolitem, 0);

  /* insert the terminal button */
#ifdef G_OS_UNIX
  action = glook_ui_get_window_action (GLOOK_UI (uiman), "Terminal");
  item = gtk_action_create_tool_item (action);
  gtk_toolbar_insert (GTK_TOOLBAR (toolbar), GTK_TOOL_ITEM (item), 0);
#endif

  /* insert the agent button */
  action = glook_ui_get_window_action (GLOOK_UI (uiman), "Look");
  item = gtk_action_create_tool_item (action);
  gtk_toolbar_insert (GTK_TOOLBAR (toolbar), GTK_TOOL_ITEM (item), 0);

  /* insert the play recent file button. */
  item = create_play_recent_button (window);
  gtk_toolbar_insert (GTK_TOOLBAR (toolbar), GTK_TOOL_ITEM (item), 0);

  /* append the zoom in button */
  action = glook_ui_get_window_action (GLOOK_UI (uiman), "ZoomIn");
  item = gtk_action_create_tool_item (action);
  GLOOK_TOOLBAR (toolbar)->priv->zoom_in = item;
  gtk_toolbar_insert (GTK_TOOLBAR (toolbar), GTK_TOOL_ITEM (item), -1);

  /* append the zoom combo box */
  combo = glook_toolbar_create_zoom_combo (GLOOK_TOOLBAR (toolbar));
  g_signal_connect (combo, "changed",
		    G_CALLBACK (on_zoom_combo_changed), toolbar);
  gtk_widget_show (combo);

  item = GTK_WIDGET (gtk_tool_item_new ());
  gtk_widget_show (item);
  gtk_container_add (GTK_CONTAINER (item), combo);
  gtk_toolbar_insert (GTK_TOOLBAR (toolbar), GTK_TOOL_ITEM (item), -1);

  /* append the zoom out button */
  action = glook_ui_get_window_action (GLOOK_UI (uiman), "ZoomOut");
  item = gtk_action_create_tool_item (action);
  GLOOK_TOOLBAR (toolbar)->priv->zoom_out = item;
  gtk_toolbar_insert (GTK_TOOLBAR (toolbar), GTK_TOOL_ITEM (item), -1);

  return toolbar;
}

/**
 * glook_toolbar_get_addrbar:
 */
GtkWidget*
glook_toolbar_get_addrbar (GLookToolbar *toolbar)
{
  g_return_val_if_fail (GLOOK_IS_TOOLBAR (toolbar), NULL);
  return toolbar->priv->addrbar;
}

/**
 * glook_toolbar_get_zoom_combo:
 */
GtkWidget*
glook_toolbar_get_zoom_combo (GLookToolbar *toolbar)
{
  g_return_val_if_fail (GLOOK_IS_TOOLBAR (toolbar), NULL);
  return toolbar->priv->zoom_combo;
}

/**
 * glook_toolbar_zoom_in:
 *
 * zoom in. and emit zoom signal.
 */
void
glook_toolbar_zoom_in (GLookToolbar *toolbar)
{
  GtkComboBox *combo;
  GtkTreeModel *model;
  GtkTreeIter iter;
  gboolean separator;

  g_return_if_fail (GLOOK_IS_TOOLBAR (toolbar));

  combo = GTK_COMBO_BOX (toolbar->priv->zoom_combo);
  model = gtk_combo_box_get_model (combo);

  /* get current active iter. */
  if (!gtk_combo_box_get_active_iter (combo, &iter))
    return;

  /* get next item. */
  if (!gtk_tree_model_iter_next (model, &iter))
    return;

  /* check whether it is a separator. */
  gtk_tree_model_get (model, &iter, 2, &separator, -1);
  if (!separator)
    {
      gtk_combo_box_set_active_iter (combo, &iter);
      return;
    }
  /* continue to get next item. */
  if (!gtk_tree_model_iter_next (model, &iter))
    {
      g_debug ("next iter invalid");
      return;
    }
  gtk_combo_box_set_active_iter (combo, &iter);
}

/**
 * glook_toolbar_zoom_out:
 *
 * zoom out. and emit zoom signal.
 */
void
glook_toolbar_zoom_out (GLookToolbar *toolbar)
{
  GtkComboBox *combo;
  GtkTreeModel *model;
  GtkTreePath *path;
  GtkTreeIter iter;
  gboolean separator;

  g_return_if_fail (GLOOK_IS_TOOLBAR (toolbar));

  combo = GTK_COMBO_BOX (toolbar->priv->zoom_combo);
  model = gtk_combo_box_get_model (combo);

  /* get current active iter. */
  if (!gtk_combo_box_get_active_iter (combo, &iter))
    return;

  /* get previous item. */
  path = gtk_tree_model_get_path (model, &iter);
  if (!gtk_tree_path_prev (path))
    return;

  /* convert back to iter from path. */
  gtk_tree_model_get_iter (model, &iter, path);
  gtk_tree_path_free (path);

  /* check whether it is a separator. */
  gtk_tree_model_get (model, &iter, 2, &separator, -1);
  if (!separator)
    {
      gtk_combo_box_set_active_iter (combo, &iter);
      return;
    }
  /* continue to get previous item. */
  path = gtk_tree_model_get_path (model, &iter);
  if (!gtk_tree_path_prev (path))
    return;

  /* convert back to iter from path. */
  gtk_tree_model_get_iter (model, &iter, path);
  gtk_tree_path_free (path);

  gtk_combo_box_set_active_iter (combo, &iter);
}

/**
 * glook_toolbar_zoom_100:
 *
 * set current zoom index to 100%, and emit a zoom signal.
 */
void
glook_toolbar_zoom_100 (GLookToolbar *toolbar)
{
  GtkComboBox *combo;
  GtkTreeModel *model;
  GtkTreeIter iter;
  gboolean valid;

  g_return_if_fail (GLOOK_IS_TOOLBAR (toolbar));

  combo = GTK_COMBO_BOX (toolbar->priv->zoom_combo);
  model = gtk_combo_box_get_model (combo);

  valid = gtk_tree_model_get_iter_first (model, &iter);
  while (valid)
    {
      gint factor;

      gtk_tree_model_get (model, &iter, 0, &factor, -1);
      if (factor == 100)
	{
	  gtk_combo_box_set_active_iter (combo, &iter);
	  break;
	}
      valid = gtk_tree_model_iter_next (model, &iter);
    }
}
