/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-accel-map.h"
#include "glook-window.h"
#include "glook-actions.h"
#include "glook-ui.h"

typedef struct _GLookActionAccel	GLookActionAccel;

struct _GLookUiPriv
{
  GtkActionGroup *window_actions;
  GHashTable     *merge_uis;
};
#define _UI_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_UI, GLookUiPriv))

struct _GLookActionAccel
{
  const gchar  *action_name;
  const gchar  *accel_path;
  gboolean      visible;
};

G_DEFINE_TYPE (GLookUi, glook_ui, GTK_TYPE_UI_MANAGER);

static void	on_ui_connect_proxy	(GtkUIManager *ui,
					 GtkAction    *action,
					 GtkWidget    *widget,
					 gpointer      user_data);
static void	on_ui_disconnect_proxy	(GtkUIManager *ui,
					 GtkAction    *action,
					 GtkWidget    *widget,
					 gpointer      user_data);

static GtkActionEntry	_menu_actions[] =
{
    { "AppMenu",	NULL, N_("_Application")	},
    { "TabMenu",	NULL, N_("_Tab")		},
    { "HelpMenu",	NULL, N_("_Help")		},
    { "ZoomMenu",	NULL, N_("_Zoom")		},
    { "PlayRecentMenu",	NULL, N_("Play recent _file")	},
    { "ManualMenu",	NULL, N_("User _Manual")	},
};

static GtkActionEntry	_app_actions[] =
{
    { "NewTab", NULL, N_("New _tab"), NULL, N_("Create new tab"),
      G_CALLBACK (on_app_new_tab_activated)
    },
    { "NewWindow", NULL, N_("New _window"), NULL, N_("Create new window"),
      G_CALLBACK (on_app_new_window_activated)
    },
    { "Look", "application-x-ttylook",
      N_("W_atch terminal session"), NULL, N_("Watch terminal session"),
      G_CALLBACK (on_app_open_agent_tab_activated),
    },
    { "PlayFile", GTK_STOCK_MEDIA_PLAY,
      N_("Pla_y file"), NULL, N_("Playback record file"),
      G_CALLBACK (on_app_play_file_activated),
    },
#ifdef G_OS_UNIX
    { "Terminal", "ttyutils-terminal",
      N_("Ter_minal"), NULL, N_("Open terminal"),
      G_CALLBACK (on_app_open_terminal_activated),
    },
#endif
    { "Bookmarks", NULL, N_("Manage _bookmarks"), NULL, N_("Manage bookmarks"),
      G_CALLBACK (on_app_bookmarks_activated)
    },
    { "Preferences", GTK_STOCK_PREFERENCES, NULL, NULL,
      N_("Configure application"),
      G_CALLBACK (on_app_preference_activated),
    },
    { "Log", NULL, N_("_Debug logs"), NULL,
      N_("View the debug console for log messages"),
      G_CALLBACK (on_app_log_activated)
    },
    { "CloseWindow", GTK_STOCK_CLOSE, N_("_Close window"), NULL,
      N_("Close this window"),
      G_CALLBACK (on_app_close_window_activated)
    },
    { "Quit", GTK_STOCK_QUIT, NULL, NULL, N_("Quit the application"),
      G_CALLBACK (on_app_quit_activated)
    },
};

static GtkActionEntry	_tab_actions[] =
{
    { "Reload", GTK_STOCK_REFRESH, N_("_Reload"), NULL, N_("Reload current tab"),
      G_CALLBACK (on_tab_reload_activated),
    },
    { "Copy", GTK_STOCK_COPY, NULL, NULL, N_("Copy selected data to clipboard"),
      G_CALLBACK (on_tab_copy_activated),
    },
    { "Paste", GTK_STOCK_PASTE, NULL, NULL, N_("Paste data from clipboard"),
      G_CALLBACK (on_tab_paste_activated),
    },
    { "SelectAll", GTK_STOCK_SELECT_ALL, NULL, NULL, N_("Select all data"),
      G_CALLBACK (on_tab_select_all_activated),
    },
    { "CloseTab", GTK_STOCK_CLOSE, N_("_Close tab"), NULL, N_("Close current tab"),
      G_CALLBACK (on_tab_close_tab_activated),
    },
    { "ZoomIn", GTK_STOCK_ZOOM_IN, NULL, NULL, N_("Enlarge font size"),
      G_CALLBACK (on_tab_zoom_in_activated),
    },
    { "ZoomOut", GTK_STOCK_ZOOM_OUT, NULL, NULL, N_("Reduce font size"),
      G_CALLBACK (on_tab_zoom_out_activated),
    },
    { "Zoom100", GTK_STOCK_ZOOM_100, NULL, NULL, N_("Normal font size"),
      G_CALLBACK (on_tab_zoom_100_activated),
    },
};
static GtkActionEntry	_help_actions[] =
{
    { "GetStart", GTK_STOCK_HELP, N_("Get _start"), NULL, N_("Getting started"),
      G_CALLBACK (on_help_getstart_activated)
    },
    { "Resources", GTK_STOCK_HOME, N_("_Online resources"), NULL,
      N_("View ttyutils online resources"),
      G_CALLBACK (on_help_online_activated)
    },
    { "ttyrcdMan", NULL, "tty_rcd", NULL,
      N_("View ttyrcd(1) manual page"),
      G_CALLBACK (on_help_ttyrcd_activated)
    },
    { "ttyagentMan", NULL, "tty_agent", NULL,
      N_("View ttyagent(1) manual page"),
      G_CALLBACK (on_help_ttyagent_activated)
    },
    { "ttylookMan", NULL, "tty_look", NULL,
      N_("View ttylook(1) manual page"),
      G_CALLBACK (on_help_ttylook_activated)
    },
    { "ttyplayMan", NULL, "tty_play", NULL,
      N_("View ttyplay(1) manual page"),
      G_CALLBACK (on_help_ttyplay_activated)
    },
    { "About", GTK_STOCK_ABOUT, N_("_About Ttyutils GLook"), NULL,
      N_("View application information"),
      G_CALLBACK (on_help_about_activated)
    },
};
static GtkActionEntry	_invisible_actions[] =
{
    { "AddressBar", NULL, N_("_Address bar"), NULL, N_("Activate address bar"),
      G_CALLBACK (on_app_address_bar_activated),
    },
    { "Tab1", NULL, N_("Tab _1"), NULL, N_("Switch to tab 1"),
      G_CALLBACK (on_tab_switch_activated),
    },
    { "Tab2", NULL, N_("Tab _2"), NULL, N_("Switch to tab 2"),
      G_CALLBACK (on_tab_switch_activated),
    },
    { "Tab3", NULL, N_("Tab _3"), NULL, N_("Switch to tab 3"),
      G_CALLBACK (on_tab_switch_activated),
    },
    { "Tab4", NULL, N_("Tab _4"), NULL, N_("Switch to tab 4"),
      G_CALLBACK (on_tab_switch_activated),
    },
    { "Tab5", NULL, N_("Tab _5"), NULL, N_("Switch to tab 5"),
      G_CALLBACK (on_tab_switch_activated),
    },
    { "Tab6", NULL, N_("Tab _6"), NULL, N_("Switch to tab 6"),
      G_CALLBACK (on_tab_switch_activated),
    },
    { "Tab7", NULL, N_("Tab _7"), NULL, N_("Switch to tab 7"),
      G_CALLBACK (on_tab_switch_activated),
    },
    { "Tab8", NULL, N_("Tab _8"), NULL, N_("Switch to tab 8"),
      G_CALLBACK (on_tab_switch_activated),
    },
    { "Tab9", NULL, N_("Tab _9"), NULL, N_("Switch to last tab"),
      G_CALLBACK (on_tab_switch_activated),
    },
    { "LeftTab", NULL, N_("_Left tab"), NULL, N_("Switch to left tab"),
      G_CALLBACK (on_tab_switch_left_activated),
    },
    { "RightTab", NULL, N_("_Right tab"), NULL, N_("Switch to right tab"),
      G_CALLBACK (on_tab_switch_right_activated),
    },
};
static GtkToggleActionEntry	_app_toggle_actions[] =
{
    { "ShowControl", NULL, N_("S_how control"), NULL,
      N_("Show or hide menubar, toolbar, statusbar, and notebook tab"),
      G_CALLBACK (on_app_show_control_toggled), TRUE,
    },
};

static const gchar	*_window_ui_string = 
"<ui>"
"  <menubar name='Menubar'>"
"    <menu action='AppMenu'>"
"      <menuitem action='AddressBar'/>"	/* invisible */
"      <menuitem action='NewTab'/>"
"      <menuitem action='NewWindow'/>"
"      <separator/>"
"      <menuitem action='Look'/>"
"      <separator/>"
"      <menuitem action='PlayFile'/>"
"      <menuitem action='PlayRecentMenu'/>"
#ifdef G_OS_UNIX
"      <separator/>"
"      <menuitem action='Terminal'/>"
#endif
"      <separator/>"
"      <menuitem action='ShowControl'/>"
"      <menuitem action='Bookmarks'/>"
"      <menuitem action='Preferences'/>"
"      <separator/>"
"      <menuitem action='Log'/>"
"      <separator/>"
"      <menuitem action='CloseWindow'/>"
"      <menuitem action='Quit'/>"
"    </menu>"
"    <menu action='TabMenu'>"
"      <menuitem action='Copy'/>"
"      <menuitem action='Paste'/>"
"      <menuitem action='SelectAll'/>"
"      <separator/>"
"      <menu action='ZoomMenu'>"
"        <menuitem action='ZoomIn'/>"
"        <menuitem action='ZoomOut'/>"
"        <menuitem action='Zoom100'/>"
"      </menu>"
"      <menuitem action='Reload'/>"
"      <separator/>"
"      <placeholder name='TabMenuPlaceholder'/>"
"      <separator/>"
"      <menuitem action='Tab1'/>"	/* invisible */
"      <menuitem action='Tab2'/>"	/* invisible */
"      <menuitem action='Tab3'/>"	/* invisible */
"      <menuitem action='Tab4'/>"	/* invisible */
"      <menuitem action='Tab5'/>"	/* invisible */
"      <menuitem action='Tab6'/>"	/* invisible */
"      <menuitem action='Tab7'/>"	/* invisible */
"      <menuitem action='Tab8'/>"	/* invisible */
"      <menuitem action='Tab9'/>"	/* invisible */
"      <menuitem action='LeftTab'/>"	/* invisible */
"      <menuitem action='RightTab'/>"	/* invisible */
"      <menuitem action='CloseTab'/>"
"    </menu>"
"    <placeholder name='MenuPlaceholder'/>"
"    <menu action='HelpMenu'>"
"      <menuitem action='GetStart'/>"
"      <menu action='ManualMenu'>"
"        <menuitem action='ttyrcdMan'/>"
"        <menuitem action='ttyagentMan'/>"
"        <menuitem action='ttylookMan'/>"
"        <menuitem action='ttyplayMan'/>"
"      </menu>"
"      <menuitem action='Resources'/>"
"      <separator/>"
"      <menuitem action='About'/>"
"    </menu>"
"  </menubar>"
"</ui>";

static GLookActionAccel		_action_accels[] =
{
    { "AddressBar",	_GLOOK_ACCEL_PATH_ADDRESS_BAR,	FALSE, },
    { "NewTab",		_GLOOK_ACCEL_PATH_NEW_TAB,	TRUE, },
    { "NewWindow",	_GLOOK_ACCEL_PATH_NEW_WINDOW,	TRUE, },
    { "PlayFile",	_GLOOK_ACCEL_PATH_PLAY_FILE,	TRUE, },
    { "ShowControl",	_GLOOK_ACCEL_PATH_SHOW_CONTROL,	TRUE, },
    { "CloseWindow",	_GLOOK_ACCEL_PATH_CLOSE_WINDOW,	TRUE, },
    { "Quit",		_GLOOK_ACCEL_PATH_QUIT,		TRUE, },
    { "Copy",		_GLOOK_ACCEL_PATH_COPY,		TRUE, },
    { "Paste",		_GLOOK_ACCEL_PATH_PASTE,	TRUE, },
    { "SelectAll",	_GLOOK_ACCEL_PATH_SELECT_ALL,	TRUE, },
    { "Reload",		_GLOOK_ACCEL_PATH_RELOAD,	TRUE, },
    { "ZoomIn",		_GLOOK_ACCEL_PATH_ZOOM_IN,	TRUE, },
    { "ZoomOut",	_GLOOK_ACCEL_PATH_ZOOM_OUT,	TRUE, },
    { "Zoom100",	_GLOOK_ACCEL_PATH_ZOOM_100,	TRUE, },
    { "Tab1",		_GLOOK_ACCEL_PATH_TAB_1,	FALSE, },
    { "Tab2",		_GLOOK_ACCEL_PATH_TAB_2,	FALSE, },
    { "Tab3",		_GLOOK_ACCEL_PATH_TAB_3,	FALSE, },
    { "Tab4",		_GLOOK_ACCEL_PATH_TAB_4,	FALSE, },
    { "Tab5",		_GLOOK_ACCEL_PATH_TAB_5,	FALSE, },
    { "Tab6",		_GLOOK_ACCEL_PATH_TAB_6,	FALSE, },
    { "Tab7",		_GLOOK_ACCEL_PATH_TAB_7,	FALSE, },
    { "Tab8",		_GLOOK_ACCEL_PATH_TAB_8,	FALSE, },
    { "Tab9",		_GLOOK_ACCEL_PATH_TAB_9,	FALSE, },
    { "LeftTab",	_GLOOK_ACCEL_PATH_LEFT_TAB,	FALSE, },
    { "RightTab",	_GLOOK_ACCEL_PATH_RIGHT_TAB,	FALSE, },
    { "CloseTab",	_GLOOK_ACCEL_PATH_CLOSE_TAB,	TRUE, },
    { "GetStart",	_GLOOK_ACCEL_PATH_GET_START,	TRUE, },
    { "Resources",	_GLOOK_ACCEL_PATH_RESOURCES,	TRUE, },
};

/**
 * glook_ui_init:
 */
static void
glook_ui_init (GLookUi *ui)
{
  ui->priv = _UI_GET_PRIVATE (ui);

  ui->priv->window_actions = gtk_action_group_new ("WindowActionGroup");
  gtk_action_group_set_translation_domain (ui->priv->window_actions,
					   GETTEXT_PACKAGE);

  ui->priv->merge_uis = g_hash_table_new (g_direct_hash, g_direct_equal);

  g_signal_connect (ui, "connect-proxy",
		    G_CALLBACK (on_ui_connect_proxy), NULL);
  g_signal_connect (ui, "disconnect-proxy",
		    G_CALLBACK (on_ui_disconnect_proxy), NULL);
}

/**
 * glook_ui_dispose:
 */
static void
glook_ui_dispose (GObject *object)
{
  GLookUi *ui = GLOOK_UI (object);

  if (ui->priv->window_actions)
    {
      g_object_unref (ui->priv->window_actions);
      ui->priv->window_actions = NULL;
    }
  G_OBJECT_CLASS (glook_ui_parent_class)->dispose (object);
}

/**
 * glook_ui_finalize:
 */
static void
glook_ui_finalize (GObject *object)
{
  GLookUi *ui = GLOOK_UI (object);

  if (ui->priv->merge_uis)
    {
      g_hash_table_destroy (ui->priv->merge_uis);
      ui->priv->merge_uis = NULL;
    }
  G_OBJECT_CLASS (glook_ui_parent_class)->finalize (object);
}

/**
 * glook_ui_class_init:
 */
static void
glook_ui_class_init (GLookUiClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookUiPriv));

  gobject_class->dispose  = glook_ui_dispose;
  gobject_class->finalize = glook_ui_finalize;
}

/**
 * on_ui_connect_proxy:
 */
static void
on_ui_connect_proxy (GtkUIManager *ui,
		     GtkAction    *action,
		     GtkWidget    *widget,
		     gpointer      user_data)
{
  g_return_if_fail (GTK_IS_UI_MANAGER (ui));
  g_return_if_fail (GTK_IS_ACTION (action));

  if (GTK_IS_MENU_ITEM (widget))
    {
      const gchar *tooltip;

      tooltip = gtk_action_get_tooltip (action);
      if (tooltip)
	gtk_widget_set_tooltip_text (widget, tooltip);
    }
}

/**
 * on_ui_disconnect_proxy:
 */
static void
on_ui_disconnect_proxy (GtkUIManager *ui,
			GtkAction    *action,
			GtkWidget    *widget,
			gpointer      user_data)
{
  g_return_if_fail (GTK_IS_UI_MANAGER (ui));
  g_return_if_fail (GTK_IS_ACTION (action));
}

/**
 * glook_ui_set_actions_accel_path:
 */
static void
glook_ui_set_actions_accel_path (GLookUi *ui)
{
  GtkAction *action;
  gint i;

  g_return_if_fail (GLOOK_IS_UI (ui));

  for (i = 0; i < G_N_ELEMENTS (_action_accels); i++)
    {
      action = gtk_action_group_get_action (ui->priv->window_actions,
					    _action_accels[i].action_name);
      gtk_action_set_visible (action, _action_accels[i].visible);
      gtk_action_set_accel_path (action, _action_accels[i].accel_path);
    }
}

/**
 * setup_play_recent_file_menu:
 */
static void
setup_play_recent_file_menu (GLookUi   *ui,
			     GtkWidget *window)
{
  GtkRecentManager *manager;
  GtkRecentFilter *filter;
  GtkWidget *menu;
  GtkWidget *item;

  g_return_if_fail (GLOOK_IS_UI (ui));

  manager = gtk_recent_manager_get_default ();
  menu = gtk_recent_chooser_menu_new_for_manager (manager);
  g_object_set (menu, "limit", 20, "show-numbers", TRUE, NULL);

  g_signal_connect (menu, "item-activated",
		    G_CALLBACK (on_app_play_recent_activated), window);

  filter = gtk_recent_filter_new ();
  gtk_recent_filter_set_name (filter, _("Terminal session record file"));
  gtk_recent_filter_add_mime_type (filter, "application/x-ttyrcd");
  gtk_recent_chooser_set_filter (GTK_RECENT_CHOOSER (menu), filter);

  item = gtk_ui_manager_get_widget (GTK_UI_MANAGER (ui),
				    "/Menubar/AppMenu/PlayRecentMenu");
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), menu);
}

/**
 * glook_ui_new:
 */
GLookUi*
glook_ui_new (GtkWidget *window)
{
  GLookUi *ui;
  GError *error = NULL;

  g_return_val_if_fail (GTK_IS_WINDOW (window), NULL);

  ui = g_object_new (GLOOK_TYPE_UI, NULL);

  /* add all built-ins actions */
  gtk_action_group_add_actions (ui->priv->window_actions, _menu_actions,
				G_N_ELEMENTS (_menu_actions), window);
  gtk_action_group_add_actions (ui->priv->window_actions, _app_actions,
				G_N_ELEMENTS (_app_actions), window);
  gtk_action_group_add_toggle_actions (ui->priv->window_actions, _app_toggle_actions,
				       G_N_ELEMENTS (_app_toggle_actions), window);
  gtk_action_group_add_actions (ui->priv->window_actions, _tab_actions,
				G_N_ELEMENTS (_tab_actions), window);
  gtk_action_group_add_actions (ui->priv->window_actions, _help_actions,
				G_N_ELEMENTS (_help_actions), window);
  gtk_action_group_add_actions (ui->priv->window_actions, _invisible_actions,
				G_N_ELEMENTS (_invisible_actions), window);

  /* insert windows action group to ui mananger */
  gtk_ui_manager_insert_action_group (GTK_UI_MANAGER (ui),
				      ui->priv->window_actions, 0);
  g_object_ref (ui->priv->window_actions);

  /* create widgets from string */
  if (!gtk_ui_manager_add_ui_from_string (GTK_UI_MANAGER (ui),
					  _window_ui_string, -1, &error))
    {
      g_error ("Unable to load ui: %s", error->message);
      return NULL;
    }
  /* change actions accel path. */
  glook_ui_set_actions_accel_path (ui);

  /* set play recent file submenu. */
  setup_play_recent_file_menu (ui, window);

  return ui;
}

/**
 * glook_ui_get_menubar:
 */
GtkWidget*
glook_ui_get_menubar (GLookUi *ui)
{
  g_return_val_if_fail (GLOOK_IS_UI (ui), NULL);

  return gtk_ui_manager_get_widget (GTK_UI_MANAGER (ui), "/Menubar");
}

/**
 * glook_ui_get_action:
 */
GtkAction*
glook_ui_get_window_action (GLookUi     *ui,
			    const gchar *name)
{
  g_return_val_if_fail (GLOOK_IS_UI (ui), NULL);
  g_return_val_if_fail (name, NULL);

  return gtk_action_group_get_action (ui->priv->window_actions, name);
}

/**
 * glook_ui_merge:
 */
gint
glook_ui_merge (GLookUi        *ui,
		GtkActionGroup *action_group,
		const gchar    *string,
		gint            length)
{
  guint merge_id;
  GError *error = NULL;

  g_return_val_if_fail (GLOOK_IS_UI (ui), -1);
  g_return_val_if_fail (GTK_IS_ACTION_GROUP (action_group), -1);
  g_return_val_if_fail (string, -1);

  /* insert action group to ui doesn't increase the refcount
   * of action group, but when later remove from ui will
   * decrease the refcount of action group.
   * to keep external still has a reference. we need increase
   * the refcount after insert. */
  gtk_ui_manager_insert_action_group (GTK_UI_MANAGER (ui),
				      g_object_ref (action_group), 0);
  /* parse the ui string and merge it. */
  merge_id = gtk_ui_manager_add_ui_from_string (GTK_UI_MANAGER (ui),
						string, length, &error);
  if (error)
    {
      g_critical (_("Unable to load ui: %s"), error->message);
      g_error_free (error);
      return -1;
    }
  /* insert the merge_id to hash table. */
  g_hash_table_insert (ui->priv->merge_uis,
		       GUINT_TO_POINTER (merge_id), action_group);
  return merge_id;
}

/**
 * glook_ui_unmerge:
 */
void
glook_ui_unmerge (GLookUi *ui,
		  guint    merge_id)
{
  GtkActionGroup *action_group;

  g_return_if_fail (GLOOK_IS_UI (ui));

  action_group = g_hash_table_lookup (ui->priv->merge_uis,
				      GUINT_TO_POINTER (merge_id));
  if (!action_group)
    {
      g_warning ("ID %d has not merge to ui.", merge_id);
      return;
    }
  /* remove the merge action group. it will decrease the
   * refcount of action group. */
  gtk_ui_manager_remove_action_group (GTK_UI_MANAGER (ui), action_group);

  /* unmerge the ui */
  gtk_ui_manager_remove_ui (GTK_UI_MANAGER (ui), merge_id);

  /* remove it from hash table. */
  g_hash_table_remove (ui->priv->merge_uis, GUINT_TO_POINTER (merge_id));
}

/**
 * glook_get_merge_id:
 *
 * get the merge id by associated action group.
 */
guint
glook_get_merge_id (GLookUi        *ui,
		    GtkActionGroup *action_group)
{
  guint merge_id = -1;
  GList *keys, *node;

  g_return_val_if_fail (GLOOK_IS_UI (ui), -1);
  g_return_val_if_fail (GTK_IS_ACTION_GROUP (action_group), -1);

  /* we get all keys first, then use these keys to lookup
   * the hash table, if found a value match @action_group,
   * return this key. */
  keys = g_hash_table_get_keys (ui->priv->merge_uis);
  if (!keys)
    return -1;

  for (node = keys; node; node = node->next)
    {
      GtkActionGroup *value;

      value = g_hash_table_lookup (ui->priv->merge_uis, node->data);
      if (value == action_group)
	{
	  merge_id = GPOINTER_TO_UINT (node->data);
	  break;
	}
    }
  g_list_free (keys);

  return merge_id;
}
