-- Copyright (c) 2010 Lodes Network Inc.
-- Written by: xiaohu <xiaohu417@gmail.com>
--
require "_ttyutils"

------------------------------------------------------
-- enter interactive, like a stand alone interpreter
------------------------------------------------------
function shell (hisfile)
	local line, chunk, errmsg, errcode

	print ("type `os.exit()' or press ^D to interrupt.")

	-- read from history file
	--
	hisfile = hisfile or os.getenv ("HOME") .. "/" .. ".ttyutils.history";
	history.read (hisfile)

	while true do
		line = readline.readline (string.format ("%s %s> ", _APPNAME, _APPVER))
		if not line then print ("\n"); break end

		line = history.expand (line)
		if not line then print ("\n"); break end

		chunk, errmsg = loadstring (line)
		if not chunk then
			print (errmsg)
		else
			history.add (line)
			status, msg = pcall (chunk)
			if not status then print (msg) end
		end
	end

	-- write back to history file for next read
	--
	history.write (hisfile)
end

------------------------------------------------------
-- alignemnt print
------------------------------------------------------
local function align_print(name, value)
	local name = tostring (name)
	if (string.len (name) < 8) then
		print (name .. "\t\t\t- " .. tostring (value))
	elseif (string.len (name) < 16) then
		print (name .. "\t\t- " .. tostring (value))
	else
		print (name .. "\t- " .. tostring (value))
	end
end

------------------------------------------------------
-- list package contents
------------------------------------------------------
function ls (package)
	local package = package or _G

	if type (package) ~= "table" then 
		print (tostring (package))
	else
		local count = 0
		local t = {}
		-- sort it
		for n in pairs (package) do table.insert (t, n) end
		table.sort (t)
		for _, n in ipairs (t) do
			align_print (n, package[n])
			count = count + 1
		end
		print (string.format ("Total: %d", count))
	end
end

------------------------------------------------------
-- find name in packages
------------------------------------------------------
local function grep_package(pattern, name, package)
	for n, v in pairs (package) do
		if string.find (tostring (n), pattern) then
			align_print (name .. "." .. n, v)
		end

		if type (v) == "table" then 
			-- we must ignore some package to prevent
			-- infinite loop
			if n ~= "package" and n ~= "_G" then
				grep_package (pattern, n, v)
			end
		end
	end
end

------------------------------------------------------
-- find all names match @pattern
------------------------------------------------------
function grep (pattern)
	if not pattern then
		error ("Usage: grep (pattern)")
	end
	grep_package (pattern, "_G", _G)
end

------------------------------------------------------
-- clear screen
------------------------------------------------------
function clear ()
	os.execute ("clear")
end
