/* Copyright (c) 2010 Lodes Network Inc.
 * Written by: xiaohu <xiaohu417@gmail.com>
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

#include <glib.h>
#include <glib/gi18n-lib.h>

static const gchar *_domain = "package-name";

/** gettext_bindtextdomain:
 */
static gint
lua_gettext_bindtextdomain (lua_State *L)
{
#ifdef ENABLE_NLS
  const gchar *domain = luaL_optstring (L, 1, GETTEXT_PACKAGE);
  const gchar *localedir = luaL_optstring (L, 2, LOCALEDIR);

  /* save package domain */
  lua_pushlightuserdata (L, (gpointer) &_domain);
  lua_pushstring (L, domain);
  lua_settable (L, LUA_REGISTRYINDEX);

  bindtextdomain (domain, localedir);
#endif
  return 0;
}

/** gettext_translate:
 */
static gint
lua_gettext_translate (lua_State *L)
{
  const gchar *string = luaL_checkstring (L, 1);
#ifdef ENABLE_NLS
  const gchar *domain;

  lua_pushlightuserdata (L, (gpointer) &_domain);
  lua_gettable (L, LUA_REGISTRYINDEX);
  domain = lua_tostring (L, -1);

  lua_pushstring (L, dgettext (domain, string));
#else
  lua_pushstring (L, string);
#endif
  return 1;
}

luaL_Reg _lua_gettext[] =
{
    { "bindtextdomain",	lua_gettext_bindtextdomain	},
    { "translate",	lua_gettext_translate		},

    { NULL, NULL }
};
