/* Copyright (c) 2010 Lodes Network Inc.
 * Written by: xiaohu <xiaohu417@gmail.com>
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <malloc.h>

#ifdef HAVE_READLINE_READLINE_H
# include <readline/readline.h>
#endif
#ifdef HAVE_READLINE_HISTORY_H
# include <readline/history.h>
#endif

#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

#include <glib.h>

/**
 * history_read:
 */
gint
lua_history_read (lua_State *L)
{
#ifdef HAVE_READLINE_HISTORY_H
  using_history ();
  read_history (luaL_checkstring (L, 1));
#endif
  return 0;
}

/**
 * history_write:
 */
gint
lua_history_write (lua_State *L)
{
#ifdef HAVE_READLINE_HISTORY_H
  write_history (luaL_checkstring (L, 1));
#endif
  return 0;
}

/**
 * history_add:
 */
gint
lua_history_add (lua_State *L)
{
#ifdef HAVE_READLINE_READLINE_H
  add_history (luaL_checkstring (L, 1));
#endif
  return 0;
}

/**
 * history_expand:
 */
gint
lua_history_expand (lua_State *L)
{
  const gchar *line = luaL_checkstring (L, 1);
#ifdef HAVE_READLINE_HISTORY_H
  gchar *express = NULL;
  gint result = history_expand ((gchar*) line, &express);
  if (result == 2 || result < 0)
    {
      if (express)
	{
	  g_warning ("%s", express);
	  free (express);
	}
      lua_pushnil (L);
    }
  else
    {
      lua_pushstring (L, express);
      free (express);
    }
#else
  lua_pushstring (L, line);
#endif
  return 1;
}

/**
 * readline_readline:
 */
gint
lua_readline_readline (lua_State *L)
{
#ifdef HAVE_READLINE_READLINE_H
  gchar *line = readline (luaL_checkstring (L, 1));
  lua_pushstring (L, line);
  free (line);
#else
  gchar line[1024];
  if (fgets (line, sizeof (line), stdin))
    {
      line[strlen (line) - 1] = '\0';
      lua_pushstring (L, line);
    }
#endif
  return 1;
}

luaL_Reg _lua_history[] =
{
    { "read",		lua_history_read	},
    { "write",		lua_history_write	},
    { "add",		lua_history_add	},
    { "expand",		lua_history_expand	},
    { NULL, NULL }
};

luaL_Reg _lua_readline[] =
{
    { "readline",	lua_readline_readline	},
    { NULL, NULL }
};
