/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <pwd.h>
#include <errno.h>

#include <glib/gi18n.h>

#include <ttyutils/ttyutils.h>

#include "record-agent.h"

static const gchar	*_introspection_xml =
"<node>"
"  <interface name='net.ttyutils.Agent.RecorderManager'>"
"    <method name='Login'>"
"      <arg name='time'     type='t' direction='in'/>"
"      <arg name='hostname' type='s' direction='in'/>"
"      <arg name='username' type='s' direction='in'/>"
"      <arg name='pid'      type='u' direction='in'/>"
"      <arg name='ttyname'  type='s' direction='in'/>"
"      <arg name='ptsname'  type='s' direction='in'/>"
"      <arg name='command'  type='s' direction='in'/>"
"      <arg name='lookport' type='q' direction='in'/>"
"    </method>"
"    <method name='Logout'>"
"      <arg name='hostname' type='s' direction='in'/>"
"      <arg name='pid'      type='u' direction='in'/>"
"      <arg name='ttyname'  type='s' direction='in'/>"
"      <arg name='ptsname'  type='s' direction='in'/>"
"    </method>"
#if 0
"    <method name='GetList'>"
"      <arg name='record-servers'  type='a(tssusssq)' direction='out'/>"
"    </method>"
"    <signal name='Added'>"
"      <arg name='time'     type='t' direction='out'/>"
"      <arg name='hostname' type='s' direction='out'/>"
"      <arg name='username' type='s' direction='out'/>"
"      <arg name='pid'      type='u' direction='out'/>"
"      <arg name='ttyname'  type='s' direction='out'/>"
"      <arg name='ptsname'  type='s' direction='out'/>"
"      <arg name='command'  type='s' direction='out'/>"
"      <arg name='lookport' type='q' direction='out'/>"
"    </signal>"
"    <signal name='Removed'>"
"      <arg name='hostname' type='s' direction='out'/>"
"      <arg name='pid'      type='u' direction='out'/>"
"      <arg name='ttyname'  type='s' direction='out'/>"
"      <arg name='ptsname'  type='s' direction='out'/>"
"    </signal>"
"    <property name='Count' type='u' access='read'/>"
#endif
"  </interface>"
"</node>";

enum
{
  PORP_0,
  PROP_LOGIN,
};

typedef struct _RecordServer	RecordServer;

struct _RecordServer
{
  guint64    time;
  gchar     *hostname;
  gchar     *username;
  guint32    pid;
  gchar     *ttyname;
  gchar     *ptsname;
  gchar     *command;
  guint16    lookport;
};

struct _RecordAgentPrivate
{
  RecordServer *server;
  gboolean        login;
};
#define RECORD_AGENT_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), RECORD_TYPE_AGENT, RecordAgentPrivate))

G_DEFINE_TYPE (RecordAgent, record_agent, G_TYPE_DBUS_PROXY);

extern	time_t	record_get_uptime	(void);


/**
 * record_agent_free_server:
 */
static void
record_agent_free_server (RecordServer *server)
{
  if (server)
    {
      g_free (server->hostname);
      g_free (server->username);
      g_free (server->ttyname);
      g_free (server->ptsname);
      g_free (server->command);

      g_slice_free (RecordServer, server);
    }
}

/**
 * record_agent_get_interface_info:
 */
static GDBusInterfaceInfo*
record_agent_get_interface_info (const gchar *interface_name)
{
  static GDBusNodeInfo *node = NULL;
  static gsize inited = 0;
  gint i;

  g_return_val_if_fail (interface_name, NULL);

  /* we need load node info from xml only once. */
  if (g_once_init_enter (&inited))
    {
      node = g_dbus_node_info_new_for_xml (_introspection_xml, NULL);
      g_once_init_leave (&inited, 1);
    }
  /* get interface info by interface name. */
  for (i = 0; node->interfaces[i]; i++)
    {
      if (g_strcmp0 (node->interfaces[i]->name, interface_name) == 0)
	return node->interfaces[i];
    } 
  return NULL;
}

/**
 * record_agent_init:
 */
void
record_agent_init (RecordAgent *agent)
{
  agent->priv = RECORD_AGENT_PRIVATE (agent);

  agent->priv->server = NULL;
  agent->priv->login = FALSE;
}

/**
 * record_agent_finalize:
 */
static void
record_agent_finalize (GObject *object)
{
  RecordAgent *agent = RECORD_AGENT (object);

  if (agent->priv->server)
    {
      record_agent_free_server (agent->priv->server);
      agent->priv->server = NULL;
    }
  G_OBJECT_CLASS (record_agent_parent_class)->finalize (object);
}

/**
 * record_agent_get_property:
 */
static void
record_agent_get_property (GObject    *object,
			   guint       property_id,
			   GValue     *value,
			   GParamSpec *pspec)
{
  RecordAgent *agent = RECORD_AGENT (object);

  switch (property_id)
    {
    case PROP_LOGIN:
      g_value_set_boolean (value, record_agent_get_login (agent));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * record_agent_set_property:
 */
static void
record_agent_set_property (GObject      *object,
			   guint         property_id,
			   const GValue *value,
			   GParamSpec   *pspec)
{
  //RecordAgent *agent = RECORD_AGENT (object);

  switch (property_id)
    {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * record_agent_g_signal:
 */
static void
record_agent_g_signal (GDBusProxy   *proxy,
		       const gchar  *sender_name,
		       const gchar  *signal_name,
		       GVariant     *parameters)
{
  RecordAgent *agent = RECORD_AGENT (proxy);

  g_return_if_fail (RECORD_IS_AGENT (agent));

  _debug ("receive signal %s from %s", signal_name, sender_name);

#if 0
  if (g_strcmp0 (signal_name, "Changed") == 0)
    g_signal_emit (user, signals[CHANGED_SIGNAL], 0);
#endif
}

/**
 * record_agent_g_properties_changed:
 */
static void
record_agent_g_properties_changed (GDBusProxy         *proxy,
				   GVariant           *changed_properties,
				   const gchar* const *invalidated_properties)
{
  RecordAgent *agent = RECORD_AGENT (proxy);

  g_return_if_fail (RECORD_IS_AGENT (agent));

#ifdef DEBUG
  if (changed_properties != NULL)
    {
      GVariantIter *iter;
      const gchar *key;

      g_variant_get (changed_properties, "a{sv}", &iter);

      while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
	{
	  _debug ("agent proerty `%s' changed", key);
	}
      g_variant_iter_free (iter);
    }
#endif
}

/**
 * record_agent_class_init:
 */
static void
record_agent_class_init (RecordAgentClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GDBusProxyClass *proxy_class = G_DBUS_PROXY_CLASS (klass);

  g_type_class_add_private (klass, sizeof (RecordAgentPrivate));

  gobject_class->set_property = record_agent_set_property;
  gobject_class->get_property = record_agent_get_property;
  gobject_class->finalize = record_agent_finalize;

  proxy_class->g_signal = record_agent_g_signal;
  proxy_class->g_properties_changed  = record_agent_g_properties_changed;

  g_object_class_install_property (gobject_class, PROP_LOGIN,
				   g_param_spec_boolean ("login", "Login", "Whether or not login", FALSE,
							 G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							 G_PARAM_STATIC_BLURB | G_PARAM_READABLE));
}

/**
 * record_agent_new:
 */
void
record_agent_new (GDBusConnection     *connection,
		  const gchar         *bus_name,
		  const gchar         *object_path,
		  const gchar         *interface_name,
		  GCancellable        *cancellable,
		  gint                 timeout,
		  GAsyncReadyCallback  callback,
		  gpointer             user_data)
{
  GDBusInterfaceInfo *interface_info;
  GDBusProxyFlags flags;

  g_return_if_fail (G_IS_DBUS_CONNECTION (connection));

  interface_info = record_agent_get_interface_info (interface_name);
  g_assert (interface_info);

  /* we don't care proxy properties and signal. */
#ifdef DEBUG
  flags = G_DBUS_PROXY_FLAGS_NONE;
#else
  flags = G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES;
  flags |= G_DBUS_PROXY_FLAGS_DO_NOT_CONNECT_SIGNALS;
#endif

  g_async_initable_new_async (RECORD_TYPE_AGENT,
			      G_PRIORITY_DEFAULT,
			      cancellable, callback, user_data,
			      "g-name", bus_name,
			      "g-object-path", object_path,
			      "g-interface-name", interface_name,
			      "g-interface-info", interface_info,
			      "g-flags", flags,
			      "g-connection", connection,
			      "g-default-timeout", timeout, NULL);
}

/**
 * record_agent_new_finish:
 */
RecordAgent*
record_agent_new_finish (GAsyncResult  *res,
			 GError       **error)
{
  GObject *agent;
  GObject *source_object;

  source_object = g_async_result_get_source_object (res);
  g_assert (source_object != NULL);

  agent = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object),
				       res, error);
  g_object_unref (source_object);

  return agent ? RECORD_AGENT (agent) : NULL;
}

/**
 * record_agent_login:
 */
gboolean
record_agent_login (RecordAgent         *agent,
		    const gchar         *hostname,
		    const gchar         *ptsname,
		    const gchar         *command,
		    guint16              lookport,
		    GCancellable        *cancellable,
		    GAsyncReadyCallback  callback,
		    gpointer             user_data)
{
  RecordServer *server;
  struct passwd *pswd;
  GVariant *params;

  g_return_val_if_fail (RECORD_IS_AGENT (agent), FALSE);

  pswd = getpwuid (getuid ());
  if (!pswd)
    {
      g_warning ("getpwuid() error: %s", g_strerror (errno));
      return FALSE;
    }
  server = g_slice_new0 (RecordServer);

  server->time     = (guint64) record_get_uptime ();
  server->hostname = g_strdup (hostname);
  server->username = g_strdup (pswd->pw_name);
  server->pid      = (guint32) getpid ();
  server->ttyname  = g_strdup (ttyname (STDOUT_FILENO));
  server->ptsname  = g_strdup (ptsname);
  server->command  = g_strdup (command);
  server->lookport = lookport;

  params = g_variant_new ("(tssusssq)", server->time, server->hostname,
			  server->username, server->pid, server->ttyname,
			  server->ptsname, server->command, server->lookport);

  g_dbus_proxy_call (G_DBUS_PROXY (agent), "Login",
		     params, G_DBUS_CALL_FLAGS_NONE,
		     -1, cancellable, callback, user_data);

  if (agent->priv->server)
    record_agent_free_server (agent->priv->server);

  agent->priv->server = server;
  return TRUE;
}

/**
 * record_agent_login_finish:
 */
gboolean
record_agent_login_finish (RecordAgent   *agent,
			   GAsyncResult  *res,
			   GError       **error)
{
  GVariant *result;

  g_return_val_if_fail (RECORD_IS_AGENT (agent), FALSE);

  result = g_dbus_proxy_call_finish (G_DBUS_PROXY (agent), res, error);
  if (result)
    {
      g_variant_unref (result);
      agent->priv->login = TRUE;
      return TRUE;
    }
  return FALSE;
}

/**
 * record_agent_logout:
 */
void
record_agent_logout (RecordAgent         *agent,
		     GCancellable        *cancellable,
		     GAsyncReadyCallback  callback,
		     gpointer             user_data)
{
  GVariant *params;

  g_return_if_fail (RECORD_IS_AGENT (agent));

  if (!agent->priv->login || !agent->priv->server)
    return;

  params = g_variant_new ("(suss)", agent->priv->server->hostname,
			  agent->priv->server->pid,
			  agent->priv->server->ttyname,
			  agent->priv->server->ptsname);

  g_dbus_proxy_call (G_DBUS_PROXY (agent), "Logout",
		     params, G_DBUS_CALL_FLAGS_NONE,
		     -1, cancellable, callback, user_data);
}

/**
 * record_agent_logout_finish:
 */
void
record_agent_logout_finish (RecordAgent   *agent,
			    GAsyncResult  *res,
			    GError       **error)
{
  GVariant *result;

  g_return_if_fail (RECORD_IS_AGENT (agent));

  result = g_dbus_proxy_call_finish (G_DBUS_PROXY (agent), res, error);
  if (result)
    {
      g_variant_unref (result);
      agent->priv->login = FALSE;
    }
}

/**
 * record_agent_get_login:
 */
gboolean
record_agent_get_login (RecordAgent *agent)
{
  g_return_val_if_fail (RECORD_IS_AGENT (agent), FALSE);
  return agent->priv->login;
}
