/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <unistd.h>

#include <glib/gi18n.h>
#include <gio/gio.h>

#include <ttyutils/ttyutils.h>

#include "record-lookd.h"

enum
{
  PROP_0,
  PROP_PORT,
  PROP_MAXCONN,
  PROP_WRITABLE
};

typedef struct _LookConnection	LookConnection;

struct _LookConnection
{
  RecordLookd        *lookd;
  GSocketConnection  *connection;
  GCancellable       *cancellable;
  gchar               read_buff[512];
  TRing              *write_ring;
  gchar               write_buff[512];
};

struct _RecordLookdPriv
{
  guint16         port;
  guint           max_conn;
  gboolean        writable;
  TLookdInfo      info;
  GSocketService *service;
  GHashTable     *connections;
};
#define _LOOKD_GET_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), RECORD_TYPE_LOOKD, RecordLookdPriv))

G_DEFINE_TYPE (RecordLookd, record_lookd, G_TYPE_OBJECT);

/**
 * record_look_connection_free:
 */
static void
record_look_connection_free (gpointer data)
{
  LookConnection *look_connection = data;

  if (look_connection)
    {
      if (look_connection->connection)
	g_object_unref (look_connection->connection);

      if (look_connection->cancellable)
	g_object_unref (look_connection->cancellable);

      if (look_connection->write_ring)
	t_ring_free (look_connection->write_ring);

      g_free (look_connection);
    }
}

/**
 * record_lookd_init:
 */
static void
record_lookd_init (RecordLookd *lookd)
{
  const gchar *charset;

  lookd->priv = _LOOKD_GET_PRIVATE (lookd);

  lookd->priv->port = 0;
  lookd->priv->max_conn = 3;
  lookd->priv->writable = FALSE;
  lookd->priv->connections =
   g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL,
			  record_look_connection_free);
  lookd->priv->service = g_socket_service_new ();
  g_socket_listener_set_backlog (G_SOCKET_LISTENER (lookd->priv->service), 100);
  g_socket_service_stop (lookd->priv->service);

  /* get terminal configure information. */
  g_get_charset (&charset);
  g_strlcpy (lookd->priv->info.charset, charset,
	     sizeof (lookd->priv->info.charset));

  g_strlcpy (lookd->priv->info.term,
	     g_getenv ("TERM") ? g_getenv ("TERM") : "dumb",
	     sizeof (lookd->priv->info.term));
}

/**
 * record_lookd_dispose:
 */
static void
record_lookd_dispose (GObject *object)
{
  RecordLookd *lookd = RECORD_LOOKD (object);

  if (lookd->priv->service)
    {
      g_object_unref (lookd->priv->service);
      lookd->priv->service = NULL;
    }
  if (lookd->priv->connections)
    {
      g_hash_table_destroy (lookd->priv->connections);
      lookd->priv->connections = NULL;
    }
  G_OBJECT_CLASS (record_lookd_parent_class)->dispose (object);
}

/**
 * record_lookd_finalize:
 */
static void
record_lookd_finalize (GObject *object)
{
  //RecordLookd *lookd = RECORD_LOOKD (object);

  G_OBJECT_CLASS (record_lookd_parent_class)->finalize (object);
}

/**
 * record_lookd_get_property:
 */
static void
record_lookd_get_property (GObject    *object,
			   guint       property_id,
			   GValue     *value,
			   GParamSpec *pspec)
{
  RecordLookd *lookd = RECORD_LOOKD (object);

  switch (property_id)
    {
    case PROP_PORT:
      g_value_set_uint (value, record_lookd_get_port (lookd));
      break;
    case PROP_MAXCONN:
      g_value_set_uint (value, record_lookd_get_maxconn (lookd));
      break;
    case PROP_WRITABLE:
      g_value_set_boolean (value, record_lookd_get_writable (lookd));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * record_lookd_set_property:
 */
static void
record_lookd_set_property (GObject      *object,
			   guint         property_id,
			   const GValue *value,
			   GParamSpec   *pspec)
{
  RecordLookd *lookd = RECORD_LOOKD (object);

  switch (property_id)
    {
    case PROP_PORT:
      record_lookd_set_port (lookd, g_value_get_uint (value));
      break;
    case PROP_MAXCONN:
      record_lookd_set_maxconn (lookd, g_value_get_uint (value));
      break;
    case PROP_WRITABLE:
      record_lookd_set_writable (lookd, g_value_get_boolean (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * record_lookd_class_init:
 */
static void
record_lookd_class_init (RecordLookdClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (RecordLookdPriv));

  gobject_class->get_property = record_lookd_get_property;
  gobject_class->set_property = record_lookd_set_property;
  gobject_class->dispose      = record_lookd_dispose;
  gobject_class->finalize     = record_lookd_finalize;

  g_object_class_install_property (gobject_class, PROP_PORT,
				   g_param_spec_uint ("port", "Port", "Listen port",
						      0, G_MAXUINT16, 0,
						      G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
						      G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_MAXCONN,
				   g_param_spec_uint ("max-conn", "Maxconn", "maximum connection count",
						      0, G_MAXUINT16, 3,
						      G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
						      G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_WRITABLE,
				   g_param_spec_boolean ("writable", "Writable", "Writable", FALSE,
							 G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							 G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_signal_new ("received",
		G_OBJECT_CLASS_TYPE (klass),
		G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		t_marshal_VOID__STRING_UINT,
		G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_INT);
}

/**
 * record_lookd_new:
 */
RecordLookd*
record_lookd_new (void)
{
  return g_object_new (RECORD_TYPE_LOOKD, NULL);
}

/**
 * record_lookd_get_port:
 */
guint16
record_lookd_get_port (RecordLookd *lookd)
{
  g_return_val_if_fail (RECORD_IS_LOOKD (lookd), 0);
  return lookd->priv->port;
}

/**
 * record_lookd_set_port:
 */
void
record_lookd_set_port (RecordLookd *lookd,
		       guint16      port)
{
  g_return_if_fail (RECORD_IS_LOOKD (lookd));

  lookd->priv->port = port;
  g_object_notify (G_OBJECT (lookd), "port");
}

/**
 * record_lookd_get_maxconn:
 */
guint
record_lookd_get_maxconn (RecordLookd *lookd)
{
  g_return_val_if_fail (RECORD_IS_LOOKD (lookd), 0);
  return lookd->priv->max_conn;
}

/**
 * record_lookd_set_maxconn:
 */
void
record_lookd_set_maxconn (RecordLookd *lookd,
			  guint        max_conn)
{
  g_return_if_fail (RECORD_IS_LOOKD (lookd));

  lookd->priv->max_conn = max_conn;
  g_object_notify (G_OBJECT (lookd), "max-conn");
}

/**
 * record_lookd_get_writable:
 */
gboolean
record_lookd_get_writable (RecordLookd *lookd)
{
  g_return_val_if_fail (RECORD_IS_LOOKD (lookd), 0);
  return lookd->priv->writable;
}

/**
 * record_lookd_set_writable:
 */
void
record_lookd_set_writable (RecordLookd *lookd,
			   gboolean     writable)
{
  g_return_if_fail (RECORD_IS_LOOKD (lookd));

  lookd->priv->writable = writable;
  g_object_notify (G_OBJECT (lookd), "writable");
}

/**
 * record_lookd_read_ready:
 *
 * callback for read data from client
 */
static void
record_lookd_read_ready (GObject      *object,
			 GAsyncResult *res,
			 gpointer      user_data)
{
  LookConnection *look_connection = user_data;
  GInputStream *istream = G_INPUT_STREAM (object);
  gssize size;
  GError *error = NULL;

  g_return_if_fail (G_IS_INPUT_STREAM (istream));
  g_return_if_fail (look_connection);
  g_return_if_fail (RECORD_IS_LOOKD (look_connection->lookd));

  size = g_input_stream_read_finish (istream, res, &error);
  if (error)
    {
      g_warning (_("Read data from look client error: %s"),
		 error->message);
      g_error_free (error);

      /* remove(disconnect) this client connection */
      g_hash_table_remove (look_connection->lookd->priv->connections,
			   look_connection->connection);
      return;
    }
  _debug ("read %d bytes from client.", size);

  /* end of file(disconnected) */
  if (size == 0)
    {
      _debug ("Connection bas been closed by client.");

      g_hash_table_remove (look_connection->lookd->priv->connections,
			   look_connection->connection);
      return;
    }
  g_signal_emit_by_name (look_connection->lookd, "received",
			 look_connection->read_buff, size, 0);
  /* continue read */
  g_input_stream_read_async (istream,
			     look_connection->read_buff,
			     sizeof (look_connection->read_buff),
			     G_PRIORITY_DEFAULT,
			     look_connection->cancellable,
			     record_lookd_read_ready,
			     look_connection);
}

/**
 * record_lookd_written_lookd_info_ready:
 */
static void
record_lookd_written_lookd_info_ready (GObject      *object,
				       GAsyncResult *res,
				       gpointer      data)
{
  GOutputStream *ostream = G_OUTPUT_STREAM (object);
  GInputStream *istream;
  LookConnection *look_connection = data;
  gsize nwritten;
  GError *error = NULL;

  g_return_if_fail (G_IS_OUTPUT_STREAM (ostream));
  g_return_if_fail (look_connection);

  nwritten = g_output_stream_write_finish (ostream, res, &error);
  if (error)
    {
      g_warning (_("Send data to look client error: %s"),
		 error->message);
      g_error_free (error);

      /* remove(disconnect) this client connection */
      g_hash_table_remove (look_connection->lookd->priv->connections,
			   look_connection->connection);
      return;
    }
  istream = g_io_stream_get_input_stream (G_IO_STREAM (look_connection->connection));

  /* read data from client async */
  g_input_stream_read_async (istream,
			     look_connection->read_buff,
			     sizeof (look_connection->read_buff),
			     G_PRIORITY_DEFAULT,
			     look_connection->cancellable,
			     record_lookd_read_ready,
			     look_connection);
}

/**
 * record_lookd_connected:
 *
 * accept connecting from client
 */
static void
record_lookd_connected (GSocketService    *service,
			GSocketConnection *connection,
			GObject           *source_object,
			RecordLookd       *lookd)
{
  GOutputStream *ostream;
  GSocket *socket;
  LookConnection *look_connection;

  g_return_if_fail (G_IS_SOCKET_SERVICE (service));
  g_return_if_fail (G_IS_SOCKET_CONNECTION (connection));
  g_return_if_fail (RECORD_IS_LOOKD (lookd));

  /* check whether maximum connection counts is reached */
  if (lookd->priv->max_conn > 0 &&
      g_hash_table_size (lookd->priv->connections) >= lookd->priv->max_conn)
    {
      g_message (_("Maximum connection count %d reached, disconnecting ..."),
		 lookd->priv->max_conn);
      g_object_unref (connection);
      return;
    }
  /* tell system attempt to verify that the remote socket endpoint is
   * still present if a sufficiently long period of time passes with no data
   * being exchanged. */
  socket = g_socket_connection_get_socket (connection);
  g_socket_set_keepalive (socket, TRUE);

  /* save this new connection */
  look_connection = g_new0 (LookConnection, 1);
  look_connection->lookd = lookd;
  /* the new connection is owned by GIO */
  look_connection->connection = g_object_ref (connection);
  look_connection->cancellable = g_cancellable_new ();
  look_connection->write_ring = t_ring_new ();

  /* append to connection list */
  g_hash_table_insert (lookd->priv->connections,
		       look_connection->connection, look_connection);

  _debug ("connection count: %d",
	  g_hash_table_size (lookd->priv->connections));

  ostream = g_io_stream_get_output_stream (G_IO_STREAM (connection));

  t_termio_get_winsize (STDOUT_FILENO,
			&lookd->priv->info.termrows,
			&lookd->priv->info.termcols);
  lookd->priv->info.termrows = g_htons (lookd->priv->info.termrows);
  lookd->priv->info.termcols = g_htons (lookd->priv->info.termcols);

  /* send terminal configure info to client. */
  g_output_stream_write_async (ostream, &lookd->priv->info,
			       sizeof (TLookdInfo),
			       G_PRIORITY_DEFAULT,
			       look_connection->cancellable,
			       record_lookd_written_lookd_info_ready,
			       look_connection);
}

/**
 * record_lookd_start:
 *
 * start accept client connect.
 */
gboolean
record_lookd_start (RecordLookd *lookd,
		    const gchar *ptsname)
{
  static guint incoming_id = 0;
  guint16 port;
  GError *error = NULL;

  g_return_val_if_fail (RECORD_IS_LOOKD (lookd), FALSE);
  g_return_val_if_fail (ptsname, FALSE);

  if (g_socket_service_is_active (lookd->priv->service))
    {
      _critical ("look service already start.");
      return TRUE;
    }
  /* when restart, we just need allow new connect can be accepted */
  if (incoming_id != 0)
    {
      g_socket_service_start (lookd->priv->service);
      return TRUE;
    }
  port = lookd->priv->port;

  while (1)
    {
      /* try to select a random port if port is 0 */
      if (port == 0)
	port = g_random_int_range (1024, G_MAXUINT16);

      g_socket_listener_add_inet_port (G_SOCKET_LISTENER (lookd->priv->service),
				       port, NULL, &error);
      if (error)
	{
	  /* port already been used, and is just random selected.
	   * try another */
	  if (error->code == G_IO_ERROR_ADDRESS_IN_USE &&
	      lookd->priv->port == 0)
	    {
	      g_error_free (error);
	      error = NULL;
	      port = 0;
	      continue;
	    }
	  g_warning (_("Failed to start look service: %s"), error->message);
	  g_error_free (error);
	  return FALSE;
	}
      break;
    }
  /* save just alloced random port */
  if (lookd->priv->port == 0)
    lookd->priv->port = port;

  /* save ptsname. */
  g_strlcpy (lookd->priv->info.ptsname, ptsname,
	     sizeof (lookd->priv->info.ptsname));

  /* accept client connect */
  incoming_id = g_signal_connect (lookd->priv->service, "incoming",
				  G_CALLBACK (record_lookd_connected), lookd);
  g_socket_service_start (lookd->priv->service);

  return TRUE;
}

/**
 * record_lookd_connection_cancel:
 */
static void
record_lookd_connection_cancel (gpointer key,
				gpointer value,
				gpointer user_data)
{
  LookConnection *look_connection = value;

  g_return_if_fail (look_connection);

  if (G_IS_CANCELLABLE (look_connection->cancellable))
    g_cancellable_cancel (look_connection->cancellable);
}

/**
 * record_lookd_stop:
 *
 * if @close_connections is TRUE, all connected client will
 * be disconnect.
 */
void
record_lookd_stop (RecordLookd *lookd,
		   gboolean     close_connections)
{
  g_return_if_fail (RECORD_IS_LOOKD (lookd));

  /* disallow new connecting */
  if (g_socket_service_is_active (lookd->priv->service))
    g_socket_service_stop (lookd->priv->service);

  if (close_connections)
    {
      /* cancel all connection op, this will cause read error,
       * then close the connection */
      g_hash_table_foreach (lookd->priv->connections,
			    record_lookd_connection_cancel, lookd);
    }
}

/**
 * record_lookd_written_ready:
 */
static void
record_lookd_written_ready (GObject      *object,
			    GAsyncResult *res,
			    gpointer      data)
{
  GOutputStream *ostream = G_OUTPUT_STREAM (object);
  LookConnection *look_connection = data;
  GError *error = NULL;
  gsize nwritten;

  g_return_if_fail (G_IS_OUTPUT_STREAM (ostream));
  g_return_if_fail (look_connection);

  nwritten = g_output_stream_write_finish (ostream, res, &error);
  if (error)
    {
      g_warning (_("Send data to look client error: %s"),
		 error->message);
      g_error_free (error);
    }
  /* continue send cached data */
  if (t_ring_get_size (look_connection->write_ring) > 0)
    {
      gint nread = t_ring_read (look_connection->write_ring,
				look_connection->write_buff,
				sizeof (look_connection->write_buff));

      g_output_stream_write_async (ostream,
				   look_connection->write_buff, nread,
				   G_PRIORITY_DEFAULT,
				   look_connection->cancellable,
				   record_lookd_written_ready,
				   look_connection);
    }
}

/**
 * record_lookd_connection_write:
 */
static void
record_lookd_connection_write (gpointer key,
			       gpointer value,
			       gpointer user_data)
{
  RecordLookd *lookd = RECORD_LOOKD (user_data);
  GSocketConnection *connection = G_SOCKET_CONNECTION (key);
  LookConnection *look_connection = value;
  GOutputStream *ostream;
  const gchar *data;
  gsize size;

  g_return_if_fail (RECORD_IS_LOOKD (lookd));
  g_return_if_fail (G_IS_SOCKET_CONNECTION (connection));
  g_return_if_fail (look_connection);

  data = g_object_get_data (G_OBJECT (lookd), "write-data");
  size = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (lookd), "write-size"));

  /* save to write buffer */
  if (!t_ring_can_write (look_connection->write_ring, size))
    {
      g_warning ("Cannot write %d bytes to look client.", size);
      return;
    }
  t_ring_write (look_connection->write_ring, data, size);

  ostream = g_io_stream_get_output_stream (G_IO_STREAM (connection));

  /* can send data only if connection if free */
  if (!g_output_stream_has_pending (ostream))
    {
      gint nread = t_ring_read (look_connection->write_ring,
				look_connection->write_buff,
				sizeof (look_connection->write_buff));

      g_output_stream_write_async (ostream,
				   look_connection->write_buff, nread,
				   G_PRIORITY_DEFAULT,
				   look_connection->cancellable,
				   record_lookd_written_ready,
				   look_connection);
    }
}

/**
 * record_lookd_write:
 *
 * write data to all connected clients
 */
void
record_lookd_write (RecordLookd *lookd,
		    const gchar *data,
		    gssize       size)
{
  g_return_if_fail (RECORD_IS_LOOKD (lookd));
  g_return_if_fail (data);

  if (size < 0)
    size = strlen (data);

  g_object_set_data (G_OBJECT (lookd), "write-data", (gpointer) data);
  g_object_set_data (G_OBJECT (lookd), "write-size", GUINT_TO_POINTER (size));

  /* send data to all clients */
  g_hash_table_foreach (lookd->priv->connections,
			record_lookd_connection_write, lookd);

  g_object_set_data (G_OBJECT (lookd), "write-data", NULL);
  g_object_set_data (G_OBJECT (lookd), "write-size", NULL);
}

/**
 * record_lookd_resize:
 *
 * send terminal size to client.
 */
void
record_lookd_resize (RecordLookd *lookd,
		     guint16      rows,
		     guint16      cols)
{
  gchar seq[32];

  g_return_if_fail (RECORD_IS_LOOKD (lookd));

  g_snprintf (seq, sizeof (seq), "\033[8;%d;%dt", rows, cols);
  record_lookd_write (lookd, seq, -1);
}
