/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include <ttyutils/ttyutils.h>

#include "record-plugin.h"

enum
{
  PORP_0,
  NAME,
};

struct _RecordPluginPrivate
{
  gchar    *name;
};
#define _PLUGIN_GET_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), RECORD_TYPE_PLUGIN, RecordPluginPrivate))

G_DEFINE_TYPE (RecordPlugin, record_plugin, G_TYPE_OBJECT);

/**
 * record_plugin_init:
 */
void
record_plugin_init (RecordPlugin *plugin)
{
  plugin->priv = _PLUGIN_GET_PRIVATE (plugin);

  plugin->priv->name = NULL;
}

/**
 * record_plugin_finalize:
 */
void
record_plugin_finalize (GObject *object)
{
  RecordPlugin *plugin = RECORD_PLUGIN (object);

  plugin->priv->name = (g_free (plugin->priv->name), NULL);

  G_OBJECT_CLASS (record_plugin_parent_class)->finalize (object);
}

/**
 * record_plugin_get_property:
 */
static void
record_plugin_get_property (GObject    *object,
			    guint       property_id,
			    GValue     *value,
			    GParamSpec *pspec)
{
  RecordPlugin *plugin = RECORD_PLUGIN (object);

  switch (property_id)
    {
    case NAME:
      g_value_set_string (value, record_plugin_get_name (plugin));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * record_plugin_set_property:
 */
static void
record_plugin_set_property (GObject      *object,
			    guint         property_id,
			    const GValue *value,
			    GParamSpec   *pspec)
{
  RecordPlugin *plugin = RECORD_PLUGIN (object);

  switch (property_id)
    {
    case NAME:
      record_plugin_set_name (plugin, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * record_plugin_class_init:
 */
void
record_plugin_class_init (RecordPluginClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (RecordPluginPrivate));

  gobject_class->set_property = record_plugin_set_property;
  gobject_class->get_property = record_plugin_get_property;
  gobject_class->finalize = record_plugin_finalize;

  g_object_class_install_property (gobject_class, NAME,
				   g_param_spec_string ("name", "name", _("Plugin name"), NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
}

/**
 * record_plugin_new:
 */
RecordPlugin*
record_plugin_new (void)
{
  return g_object_new (RECORD_TYPE_PLUGIN, NULL);
}

/**
 * record_plugin_get_name:
 */
const gchar*
record_plugin_get_name (RecordPlugin *plugin)
{
  g_return_val_if_fail (RECORD_IS_PLUGIN (plugin), NULL);
  return plugin->priv->name;
}

/**
 * record_plugin_set_name:
 */
void
record_plugin_set_name (RecordPlugin *plugin,
			  const gchar    *name)
{
  gchar *newname = NULL;

  g_return_if_fail (RECORD_IS_PLUGIN (plugin));

  if (name)
    newname = g_strdup (name);

  g_free (plugin->priv->name);
  plugin->priv->name = newname;

  g_object_notify (G_OBJECT (plugin), "name");
}
