/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifndef __RECORD_PROCESS_H__
#define __RECORD_PROCESS_H__

#include <glib-object.h>

#ifdef HAVE_TERMIOS_H
# include <termios.h>
#endif

G_BEGIN_DECLS

#define RECORD_TYPE_PROCESS		(record_process_get_type ())
#define RECORD_PROCESS(obj)		(G_TYPE_CHECK_INSTANCE_CAST ((obj), RECORD_TYPE_PROCESS, RecordProcess))
#define RECORD_PROCESS_CLASS(klass)	(G_TYPE_CHECK_CLASS_CAST ((klass), RECORD_TYPE_PROCESS, RecordProcessClass))
#define RECORD_IS_PROCESS(obj)		(G_TYPE_CHECK_INSTANCE_TYPE ((obj), RECORD_TYPE_PROCESS))
#define RECORD_IS_PROCESS_CLASS(klass)	(G_TYPE_CHECK_CLASS_TYPE ((klass), RECORD_TYPE_PROCESS))
#define RECORD_PROCESS_GET_CLASS(obj)	(G_TYPE_INSTANCE_GET_CLASS ((obj), RECORD_TYPE_PROCESS, RecordProcessClass))

typedef struct _RecordProcess		RecordProcess;
typedef struct _RecordProcessClass	RecordProcessClass;
typedef struct _RecordProcessPriv	RecordProcessPriv;
typedef struct _RecordProcessEnv	RecordProcessEnv;

struct _RecordProcessEnv
{
  gchar    *name;
  gchar    *value;
  gboolean  overwrite;
};

struct _RecordProcess
{
  GObject parent;
  RecordProcessPriv *priv;
};

struct _RecordProcessClass
{
  GObjectClass parent_class;

  void	(*hangup)	(RecordProcess    *process);
  void	(*child_exit)	(RecordProcess    *process,
			 GPid              pid,
			 gint              status);
};

GType		record_process_get_type		(void);

RecordProcess*	record_process_new		(void);

const gchar*	record_process_get_ptsname	(RecordProcess   *process);
GPid		record_process_get_child_pid	(RecordProcess   *process);

const gchar*	record_process_get_workdir	(RecordProcess   *process);
void		record_process_set_workdir	(RecordProcess   *process,
						 const gchar     *workdir);
const gchar**	record_process_get_argv		(RecordProcess   *process);
void		record_process_set_argv		(RecordProcess   *process,
						 const gchar    **argv);

gboolean	record_process_env_set		(RecordProcess   *process,
						 const gchar     *name,
						 const gchar     *value,
						 gboolean         overwrite);
void		record_process_env_unset	(RecordProcess   *process,
						 const gchar     *name);
const GList*	record_process_env_list		(RecordProcess   *process);

void		record_process_set_login_log	(RecordProcess    *process,
						 gboolean          login_log);
gboolean	record_process_get_login_log	(RecordProcess    *process);

gboolean	record_process_start		(RecordProcess    *process,
						 struct termios   *term_attr);
void		record_process_stop		(RecordProcess    *process);

gboolean	record_process_write		(RecordProcess   *process,
						 const gchar     *buff,
						 gssize           size);

gboolean	record_process_set_size		(RecordProcess   *process,
						 guint16          lines,
						 guint16          columns);
gboolean	record_process_get_size		(RecordProcess   *process,
						 guint16         *lines,
						 guint16         *columns);
void		record_process_set_utf8		(RecordProcess   *process,
						 gboolean         utf8);

G_END_DECLS

#endif /* __RECORD_PROCESS_H__ */
