#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "ttyutils/ttyutils.h"

static void
on_console_readable (TConsole    *console,
		     const gchar *data,
		     guint        size,
		     GMainLoop   *loop)
{
  g_assert (T_IS_CONSOLE (console));
  g_assert (data);
  g_assert (size > 0);
  g_assert (loop);

  g_print ("Incoming length: %d\r\n", size);
  g_main_loop_quit (loop);
}

static void
on_console_hangup (TConsole    *console,
		   GMainLoop   *loop)
{
  t_console_write (console, "Hangup\r\n", -1);
  g_main_loop_quit (loop);
}

/**
 * test_console_quick:
 */
static void
test_console_quick (void)
{
  TConsole *console = t_console_new ();
  GMainLoop *loop;

  loop = g_main_loop_new (NULL, FALSE);

  g_signal_connect (console, "readable",
		    G_CALLBACK (on_console_readable), loop);
  g_signal_connect (console, "hangup",
		    G_CALLBACK (on_console_hangup), loop);

  t_console_write (console, "Please press any key to continue ...\r\n", -1);

  t_console_start (console);

  g_main_loop_run (loop);
  g_main_loop_unref (loop);

  g_object_unref (console);
}

/**
 * test_console_perf:
 */
static void
test_console_perf (void)
{
  TConsole *console = t_console_new ();
  gchar buff[2048];
  gint i;

  memset (buff, '8', sizeof (buff));

  /* write 2M data */
  for (i = 0; i < 2000; i++)
    t_console_write (console, buff, sizeof (buff));

  g_object_unref (console);
}

static void
test_console_slow (void)
{
}

/**
 * test_console:
 */
void
test_console (void)
{
  struct termios ttmode;

  if (!isatty (STDIN_FILENO))
    {
      g_test_message ("STDIN is not a tty device, skip console test.");
      return;
    }
  tcgetattr (STDIN_FILENO, &ttmode);

  t_termio_makeraw (STDIN_FILENO);

  if (g_test_quick ())
    test_console_quick ();

  if (g_test_perf ())
    test_console_perf ();

  if (g_test_slow ())
    test_console_slow ();

  if (g_test_thorough ())
    {
      test_console_quick ();
      test_console_perf ();
      test_console_slow ();
    }
  tcsetattr (STDIN_FILENO, TCSANOW, &ttmode);
}
