/* Copyright (C) 2010 xiaohu <xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <locale.h>
#include <ctype.h>
#include <errno.h>
#ifdef HAVE_CURSES_H
# include <curses.h>
#endif
#include <zip.h>
#include <gcrypt.h>

#include <glib/gi18n.h>
#include <glib/gstdio.h>
#include <gtk/gtk.h>

#include <ttyutils/ttyutils.h>

#ifdef DEBUG
# define ASSAYER_GLADE_FILE	SRCDIR G_DIR_SEPARATOR_S "assayer.xml"
#else
# define ASSAYER_GLADE_FILE	GLADEDIR G_DIR_SEPARATOR_S "assayer.xml"
#endif

/* column of archive file list stroe.
 * must keep sync with .glade file defined. */
enum
{
  FILE_LIST_COL_NAME,
  FILE_LIST_COL_SIZE,
  FILE_LIST_COL_COMP_SIZE,
  FILE_LIST_COL_COMMENT,
  FILE_LIST_COL_COMP_METHOD,
  FILE_LIST_COL_MTIME,
  FILE_LIST_NCOLUMNS,
};

enum
{
  INDEX_COL_NO,
  INDEX_COL_INTERVAL,
  INDEX_COL_ELAPSED,
  INDEX_COL_SIZE,
  INDEX_NCOLUMNS
};

extern void		assay_gtk_static_init		(void);

static gboolean		assay_open_file			(GtkWidget        *window,
							 GFile            *file);
static void		assay_close_file		(GtkWidget        *window);

/* defined in assay-convert.c. */
extern void		assay_convert_assistant_run	(GtkWidget        *window);

static void		on_index_tree_selection_changed	(GtkTreeSelection *selection,
							 GtkWidget        *window);

static gboolean		 _opt_version = FALSE;

static GOptionEntry	option_entries[] =
{
    { "version", 'v', 0, G_OPTION_ARG_NONE, &_opt_version,
      N_("Output version information"), NULL },

    { NULL, }
};

/**
 * assay_show_message:
 *
 * show message to info bar.
 */
static void
assay_show_message (GtkWidget      *window,
		    GtkMessageType  type,
		    const gchar    *message)
{
  GtkBuilder *builder;
  GObject *infobar;
  GObject *infolabel;

  builder = g_object_get_data (G_OBJECT (window), "builder");

  infolabel = gtk_builder_get_object (builder, "InfoLabel");
  gtk_label_set_label (GTK_LABEL (infolabel), message);

  infobar = gtk_builder_get_object (builder, "Infobar");
  g_object_set (infobar, "no-show-all", FALSE, "message-type", type, NULL);
  gtk_widget_show_all (GTK_WIDGET (infobar));
}

/**
 * assay_hide_message:
 *
 * hide info bar.
 */
static void
assay_hide_message (GtkWidget *window)
{
  GtkBuilder *builder;
  GObject *infobar;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");

  infobar = gtk_builder_get_object (builder, "Infobar");
  gtk_widget_hide (GTK_WIDGET (infobar));
}

/**
 * on_infobar_hide_button_clicked:
 */
void
on_infobar_hide_button_clicked (GtkWidget *button,
				GtkWidget *window)
{
  g_return_if_fail (GTK_IS_WINDOW (window));
  assay_hide_message (window);
}

/**
 * assay_set_widgets_sensitive:
 */
static void
assay_set_widgets_sensitive (GtkWidget *window,
			     gboolean   sensitive)
{
  GtkBuilder *builder;
  GObject *notebook;
  GObject *action;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");

  notebook = gtk_builder_get_object (builder, "Notebook");
  gtk_widget_set_sensitive (GTK_WIDGET (notebook), sensitive);

  action = gtk_builder_get_object (builder, "CloseAction");
  gtk_action_set_sensitive (GTK_ACTION (action), sensitive);
  action = gtk_builder_get_object (builder, "PlayAction");
  gtk_action_set_sensitive (GTK_ACTION (action), sensitive);
}


/**
 * on_recent_item_activated:
 *
 * open recect file.
 */
void
on_recent_item_activated (GtkRecentChooser *chooser,
			  GtkWidget        *window)
{
  gchar *uri;
  GFile *file;

  g_return_if_fail (GTK_IS_WINDOW (window));

  uri = gtk_recent_chooser_get_current_uri (chooser);
  file = g_file_new_for_uri (uri);

  if (!assay_open_file (window, file))
    {
      /* remove the recent item if failed to open it. */
      GtkRecentManager *manager = gtk_recent_manager_get_default ();
      gtk_recent_manager_remove_item (manager, uri, NULL);
    }
  g_free (uri);
  g_object_unref (file);
}

/**
 * on_open_button_clicked:
 *
 * open file by dialog.
 */
void
on_open_button_clicked (GtkWidget *button,
			GtkWidget *window)
{
  GtkBuilder *builder;
  GObject *action;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");

  action = gtk_builder_get_object (builder, "OpenAction");
  gtk_action_activate (GTK_ACTION (action));
}

/**
 * assay_create_recent_button:
 *
 * prepend a recent option menu button into toolbar.
 */
static void
assay_create_recent_button (GObject *window,
			    GObject *toolbar)
{
  GtkRecentManager *manager;
  GtkWidget *menu;
  GtkRecentFilter *filter;
  GtkToolItem *item;

  manager = gtk_recent_manager_get_default ();
  menu = gtk_recent_chooser_menu_new_for_manager (manager);
  g_object_set (menu, "limit", 20, "show-numbers", TRUE, NULL);

  g_signal_connect (menu, "item-activated",
		    G_CALLBACK (on_recent_item_activated), window);

  filter = gtk_recent_filter_new ();
  gtk_recent_filter_set_name (filter, _("Terminal session record file"));
  gtk_recent_filter_add_mime_type (GTK_RECENT_FILTER (filter),
				   "application/x-ttyrcd");
  gtk_recent_chooser_set_filter (GTK_RECENT_CHOOSER (menu), filter);

  item = gtk_menu_tool_button_new_from_stock (GTK_STOCK_OPEN);
  gtk_menu_tool_button_set_menu (GTK_MENU_TOOL_BUTTON (item), menu);
  gtk_tool_item_set_tooltip_text (item, _("Open a file"));
  gtk_menu_tool_button_set_arrow_tooltip_text (GTK_MENU_TOOL_BUTTON (item),
					       _("Open a recently used file"));
  g_signal_connect (item, "clicked",
		    G_CALLBACK (on_open_button_clicked), window);

  gtk_toolbar_insert (GTK_TOOLBAR (toolbar), item, 0);
}

/**
 * main:
 */
gint
main (gint argc, gchar **argv)
{
  GOptionContext *option;
  GtkBuilder *builder;
  TDecoder *decoder;
  GObject *window;
  GObject *toolbar;
  GObject *treeview;
  GtkTreeSelection *selection;
  GObject *text_buffer;
  GError *error = NULL;

  g_set_application_name ("Ttyutils Assayer");

  setlocale (LC_ALL, "");
  bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);

  gtk_init (&argc, &argv);

#ifdef ENABLE_STATIC
  assay_gtk_static_init ();
#endif

  /* initialize libgcrypt subsystems. */
#ifdef DEBUG
  gcry_control (GCRYCTL_SET_VERBOSITY, 1);
  gcry_control (GCRYCTL_SET_DEBUG_FLAGS, 0);
  gcry_control (GCRYCTL_SET_DEBUG_FLAGS, 1);
#else
  gcry_control (GCRYCTL_SET_VERBOSITY, 0);
  gcry_control (GCRYCTL_CLEAR_DEBUG_FLAGS, 0);
  gcry_control (GCRYCTL_CLEAR_DEBUG_FLAGS, 1);
#endif

  gcry_set_log_handler (t_gcrypt_log_handler, NULL);
  if (!gcry_check_version (GCRYPT_VERSION))
    {
      g_warning (_("libgcrypt version %s or above is required."),
		 GCRYPT_VERSION);
      return EXIT_FAILURE;
    }
  gcry_control (GCRYCTL_DISABLE_SECMEM, 0);
#ifdef DEBUG
  gcry_control (GCRYCTL_PRINT_CONFIG, NULL);
#endif
  gcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);

  option = g_option_context_new (_("Record file"));
  g_option_context_set_summary (option, _("Terminal session record assayer"));
  g_option_context_set_description (option, WEBSITE);
  g_option_context_add_main_entries (option, option_entries, GETTEXT_PACKAGE);
  g_option_context_parse (option, &argc, &argv, &error);
  if (error)
    {
      g_warning (_("Parse command line error: %s"), error->message);
      g_error_free (error);
      g_option_context_free (option);
      return EXIT_FAILURE;
    }
  g_option_context_free (option);

  if (_opt_version)
    {
      t_print_version ();
      return EXIT_SUCCESS;
    }
  /* load ui from file. */
  if (!g_file_test (ASSAYER_GLADE_FILE, G_FILE_TEST_EXISTS))
    {
      g_critical ("`%s' not found.", ASSAYER_GLADE_FILE);
      return 1;
    }
  builder = gtk_builder_new ();
  gtk_builder_add_from_file (builder, ASSAYER_GLADE_FILE, &error);
  if (error)
    {
      g_critical ("%s", error->message);
      g_error_free (error);
      return 1;
    }
  decoder = t_decoder_new ();

  window = gtk_builder_get_object (builder, "Window");
  toolbar = gtk_builder_get_object (builder, "Toolbar");

  assay_create_recent_button (window, toolbar);

  /* create text tag to format data. */
  text_buffer = gtk_builder_get_object (builder, "DataTextbuffer");
  gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (text_buffer), "blue",
			      "foreground", "blue", "scale", 0.9, NULL);
  gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (text_buffer), "font",
			      "font", "Courier 10 Pitch 10.5", NULL);

  g_object_set_data_full (window, "builder", builder, g_object_unref);
  g_object_set_data_full (window, "decoder", decoder, g_object_unref);

  gtk_builder_connect_signals (builder, window);

  /* when index tree selection changed, we need update the data text
   * view buffer. */
  treeview = gtk_builder_get_object (builder, "IndexTreeview");
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
  g_signal_connect (selection, "changed",
		    G_CALLBACK (on_index_tree_selection_changed), window);

  assay_set_widgets_sensitive (GTK_WIDGET (window), FALSE);

  gtk_widget_show_all (GTK_WIDGET (window));

  /* if has remains arguments, treat it as file to open. */
  if (argc > 1 && argv[1])
    {
      GFile *file = g_file_new_for_path (argv[1]);
      assay_open_file (GTK_WIDGET (window), file);
      g_object_unref (file);
    }
  gtk_main ();
  return 0;
}

/**
 * assay_show_summary_info:
 */
static void
assay_show_summary_info (GtkWidget   *window,
			 GFile       *file,
			 TRecordHead *head)
{
  GtkBuilder *builder;
  GObject *entry;
  GFileInfo *info;
  gchar buff[256];
  gchar *pathname, *filename, *dirname;
  guint64 total_size, file_size, indexes_size;
  gdouble compressibility;
  TDecoder *decoder;
  struct zip *archive;
  const gchar *comment;
  gint length, i;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");
  decoder = g_object_get_data (G_OBJECT (window), "decoder");

  pathname = g_file_get_path (file);
  filename = g_path_get_basename (pathname);
  dirname = g_path_get_dirname (pathname);

  entry = gtk_builder_get_object (builder, "FilenameEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), filename);
  entry = gtk_builder_get_object (builder, "PathnameEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), dirname);

  g_free (pathname);
  g_free (filename);
  g_free (dirname);

  info = g_file_query_info (file, G_FILE_ATTRIBUTE_STANDARD_SIZE,
			    G_FILE_QUERY_INFO_NONE, NULL, NULL);
  file_size = g_file_info_get_size (info);

  indexes_size = 0;

  archive = t_decoder_get_archive (decoder);
  for (i = 0; i < head->nfiles; i++)
    {
      struct zip_stat sb;

      g_snprintf (buff, sizeof (buff), "indexes/%d", i);
      if (zip_stat (archive, buff, 0, &sb) != 0)
	{
	  assay_show_message (window, GTK_MESSAGE_WARNING,
			      zip_strerror (archive));
	  continue;
	}
      indexes_size += sb.size;
    }
  total_size = sizeof (TRecordHead);
  total_size += indexes_size;
  total_size += head->ndata;
  compressibility = (gdouble) file_size / (gdouble) total_size;

  g_snprintf (buff, sizeof (buff), "%.2f%% (%'lld vs %'lld[%'d + %'lld + %'lld])",
	      compressibility * 100,
	      file_size, total_size, sizeof (TRecordHead),
	      indexes_size, head->ndata);
  entry = gtk_builder_get_object (builder, "CompressibilityEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), buff);
  gtk_entry_set_progress_fraction (GTK_ENTRY (entry),
				   MIN (1.0, compressibility));
  g_object_unref (info);

  entry = gtk_builder_get_object (builder, "CommentEntry");
  comment = t_decoder_get_comment (decoder, &length);
  g_strlcpy (buff, comment, MIN (length + 1, sizeof (buff)));
  gtk_entry_set_text (GTK_ENTRY (entry), buff);
}

/**
 * assay_show_head_info:
 */
static void
assay_show_head_info (GtkWidget   *window,
		      TRecordHead *head)
{
  GtkBuilder *builder;
  GObject *entry;
  gchar buff[128];
  time_t tim;
  struct tm *tmp;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");

  g_snprintf (buff, sizeof (buff), "%d", head->version);
  entry = gtk_builder_get_object (builder, "HeadVersionEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), buff);

  entry = gtk_builder_get_object (builder, "HeadCharsetEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), head->charset);

  entry = gtk_builder_get_object (builder, "HeadTerminalEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), head->term);
  g_snprintf (buff, sizeof (buff), "%d x %d", head->ncols, head->nrows);
  entry = gtk_builder_get_object (builder, "HeadTermsizeEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), buff);

  g_snprintf (buff, sizeof (buff), "%d", head->clevel);
  entry = gtk_builder_get_object (builder, "HeadCLevelEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), buff);

  tim = (time_t) head->starting;
  tmp = localtime (&tim);
  strftime (buff, sizeof (buff), "%x %X", tmp);
  entry = gtk_builder_get_object (builder, "HeadStartingEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), buff);

  tim = (time_t) head->duration;
  tmp = gmtime (&tim);
  strftime (buff, sizeof (buff), "%X", tmp);
  entry = gtk_builder_get_object (builder, "HeadDurationEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), buff);

  g_snprintf (buff, sizeof (buff),
	      "%02x %02x %02x %02x %02x %02x %02x %02x "
	      "%02x %02x %02x %02x %02x %02x %02x %02x "
	      "%02x %02x %02x %02x %02x %02x %02x %02x",
	      head->passwd[0],  head->passwd[1],
	      head->passwd[2],  head->passwd[3],
	      head->passwd[4],  head->passwd[5],
	      head->passwd[6],  head->passwd[7],
	      head->passwd[8],  head->passwd[9],
	      head->passwd[10], head->passwd[11],
	      head->passwd[12], head->passwd[13],
	      head->passwd[14], head->passwd[15],
	      head->passwd[16], head->passwd[17],
	      head->passwd[18], head->passwd[19],
	      head->passwd[20], head->passwd[21],
	      head->passwd[22], head->passwd[23]);
  entry = gtk_builder_get_object (builder, "HeadPasswordEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), buff);

  g_snprintf (buff, sizeof (buff), "%d", head->nindexes);
  entry = gtk_builder_get_object (builder, "HeadNIndexesEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), buff);

  g_snprintf (buff, sizeof (buff), "%d (1 + %d * 2)",
	      1 + head->nfiles * 2, head->nfiles);
  entry = gtk_builder_get_object (builder, "HeadNFilesEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), buff);

  entry = gtk_builder_get_object (builder, "HeadCreatorEntry");
  gtk_entry_set_text (GTK_ENTRY (entry), head->creator);
}

/**
 * assay_str_compress_method:
 */
static const gchar*
assay_str_compress_method (gint method)
{
  switch (method)
    {
    case -1:
      return "better of deflate or store";
    case 0:
      return "stored(uncompressed)";
    case 1:
      return "shrunk";
    case 2:
      return "reduced with factor 1";
    case 3:
      return "reduced with factor 2";
    case 4:
      return "reduced with factor 3";
    case 5:
      return "reduced with factor 4";
    case 6:
      return "imploded";
    case 8:
      return "deflated";
    case 9:
      return "deflated64";
    case 10:
      return "PKWARE imploding";
    case 12:
      return "BZIP2";
    default:
      return "unknown";
    }
}

/**
 * assay_show_file_list:
 */
static void
assay_show_file_list (GtkWidget   *window,
		      const gchar *path)
{
  GtkBuilder *builder;
  TDecoder *decoder;
  GObject *treeview;
  GObject *model;
  struct zip *archive;
  gint n, i;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");
  decoder = g_object_get_data (G_OBJECT (window), "decoder");

  archive = t_decoder_get_archive (decoder);

  model = gtk_builder_get_object (builder, "FileListModel");
  gtk_list_store_clear (GTK_LIST_STORE (model));

  n = zip_get_num_files (archive);
  for (i = 0; i < n; i++)
    {
      struct zip_stat sb;
      struct tm *tmp;
      gchar comp_size[64], method[64];
      gchar strtime[256], comment[256] = { '\0', };
      const gchar *p;
      gint length;

      zip_stat_index (archive, i, 0, &sb);

      p = zip_get_file_comment (archive, i, &length, 0);
      if (p && length > 0)
	strncpy (comment, p, MIN (length, sizeof (comment)));

      tmp = localtime (&sb.mtime);
      strftime (strtime, sizeof (strtime), "%x %X", tmp);

      g_snprintf (comp_size, sizeof (comp_size), "%lu (%.2f)",
		  sb.comp_size, (gdouble) sb.comp_size / (gdouble) sb.size);
      g_snprintf (method, sizeof (method), "%d - %s", sb.comp_method,
		  assay_str_compress_method (sb.comp_method));

      gtk_list_store_insert_with_values (GTK_LIST_STORE (model), NULL, -1,
					 FILE_LIST_COL_NAME, sb.name,
					 FILE_LIST_COL_SIZE, sb.size,
					 FILE_LIST_COL_COMP_SIZE, comp_size,
					 FILE_LIST_COL_COMMENT, comment,
					 FILE_LIST_COL_COMP_METHOD, method,
					 FILE_LIST_COL_MTIME, strtime, -1);
    }
  treeview = gtk_builder_get_object (builder, "FileListTreeview");
  gtk_tree_view_columns_autosize (GTK_TREE_VIEW (treeview));
}

/**
 * assay_read_indexes_into_treeview:
 */
static void
assay_read_indexes_into_treeview (GtkWidget *window)
{
  GtkBuilder *builder;
  TDecoder *decoder;
  GObject *model;
  GObject *treeview;
  GObject *scale;
  GObject *adjust;
  TRecordIndex *indexes;
  guint32 istart, iend, i;
  gdouble elapsed = 0.0;
  gchar *markup;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");
  decoder = g_object_get_data (G_OBJECT (window), "decoder");

  indexes = t_decoder_get_indexes (decoder, &istart, &iend);

  /* index scale lower, upper and current value. */
  adjust = gtk_builder_get_object (builder, "IndexScaleAdjustment");
  gtk_adjustment_configure (GTK_ADJUSTMENT (adjust), 0.0, (gdouble) istart,
			    (gdouble) (iend - 1), 1.0, 1.0, 0.0);
  /* index scale mark. */
  scale = gtk_builder_get_object (builder, "IndexScale");
  gtk_scale_clear_marks (GTK_SCALE (scale));
  markup = g_markup_printf_escaped ("<small>%d</small>", istart);
  gtk_scale_add_mark (GTK_SCALE (scale), (gdouble) istart,
		      GTK_POS_BOTTOM, markup);
  g_free (markup);
  markup = g_markup_printf_escaped ("<small>%d</small>", iend - 1);
  gtk_scale_add_mark (GTK_SCALE (scale), (gdouble) (iend - 1),
		      GTK_POS_BOTTOM, markup);
  g_free (markup);

  model = gtk_builder_get_object (builder, "IndexTreeModel");
  gtk_list_store_clear (GTK_LIST_STORE (model));

  for (i = 0; istart < iend; i++, istart++)
    {
      GError *error = NULL;

      //elapsed += indexes[i].interval;
      elapsed = t_decoder_get_index_elapsed (decoder, istart, &error);
      if (error)
	{
	  g_warning ("%s", error->message);
	  g_error_free (error);
	}
      gtk_list_store_insert_with_values (GTK_LIST_STORE (model), NULL, -1,
					 INDEX_COL_NO, istart,
					 INDEX_COL_INTERVAL, indexes[i].interval,
					 INDEX_COL_ELAPSED, elapsed,
					 INDEX_COL_SIZE, indexes[i].size,
					 -1);
    }
  treeview = gtk_builder_get_object (builder, "IndexTreeview");
  gtk_tree_view_columns_autosize (GTK_TREE_VIEW (treeview));
}

/**
 * assay_show_contents:
 */
static void
assay_show_contents (GtkWidget *window)
{
  //  GtkBuilder *builder;

  g_return_if_fail (GTK_IS_WINDOW (window));

  //builder = g_object_get_data (G_OBJECT (window), "builder");

  assay_read_indexes_into_treeview (window);
}

/**
 * assay_open_passwd_func:
 *
 * callback for set decode password.
 */
static gboolean
assay_open_passwd_func (TDecoder *decoder,
			gpointer  user_data)
{
  GtkWidget *window = user_data;
  GtkBuilder *builder;
  GObject *dialog;
  GObject *entry;

  g_return_val_if_fail (T_IS_DECODER (decoder), FALSE);
  g_return_val_if_fail (GTK_IS_WINDOW (window), FALSE);

  builder = g_object_get_data (G_OBJECT (window), "builder");
  dialog = gtk_builder_get_object (builder, "PasswordDialog");
  entry = gtk_builder_get_object (builder, "PasswordDialogEntry");

  gtk_dialog_set_default_response (GTK_DIALOG (dialog),
				   GTK_RESPONSE_OK);
  gtk_widget_show_all (GTK_WIDGET (dialog));

  while (1)
    {
      const gchar *passwd;

      gtk_entry_set_text (GTK_ENTRY (entry), "");
      gtk_widget_grab_focus (GTK_WIDGET (entry));

      /* give up. just read head and indexes info(without encrypted data). */
      if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_OK)
	break;

      passwd = gtk_entry_get_text (GTK_ENTRY (entry));
      if (t_decoder_set_passwd (decoder, passwd, -1))
	break;
    }
  gtk_widget_hide (GTK_WIDGET (dialog));
  return TRUE;
}

/**
 * assay_add_recent_item:
 */
static void
assay_add_recent_item (GtkWidget   *window,
		       const gchar *path)
{
  GtkRecentData data;
  gchar *groups[] = { "Ttyutils", NULL };
  gchar *uri;

  data.display_name = g_path_get_basename (path);
  data.description = NULL;
  data.mime_type = "application/x-ttyrcd";
  data.app_name = (gchar*) g_get_application_name ();
  data.app_exec = "ttyglook %u";
  data.groups = groups;
  data.is_private = FALSE;

  uri = g_strdup_printf ("file://%s", path);
  gtk_recent_manager_add_full (gtk_recent_manager_get_default (), uri, &data);

  g_free (uri);
  g_free (data.display_name);
}

/**
 * assay_open_file:
 */
static gboolean
assay_open_file (GtkWidget *window,
		 GFile     *file)
{
  TDecoder *decoder;
  TRecordHead *head;
  gchar *path, title[256], *base;
  GError *error = NULL;

  g_return_val_if_fail (GTK_IS_WINDOW (window), FALSE);
  g_return_val_if_fail (G_IS_FILE (file), FALSE);

  /* close file first. */
  assay_close_file (window);

  path = g_file_get_path (file);
  g_return_val_if_fail (path, FALSE);

  decoder = g_object_get_data (G_OBJECT (window), "decoder");

  t_decoder_open (decoder, path, assay_open_passwd_func,
		  g_object_ref (window), g_object_unref, &error);
  if (error)
    {
      assay_show_message (window, GTK_MESSAGE_WARNING, error->message);
      g_error_free (error);
      g_free (path);
      return FALSE;
    }
  assay_set_widgets_sensitive (window, TRUE);

  head = t_decoder_get_head (decoder);
  assay_show_summary_info (window, file, head);
  assay_show_head_info (window, head);
  assay_show_file_list (window, path);
  assay_show_contents (window);

  /* change window title. */
  base = g_path_get_basename (path);
  g_snprintf (title, sizeof (title), "%s - Assayer", base);
  gtk_window_set_title (GTK_WINDOW (window), title);
  g_free (base);

  assay_add_recent_item (window, path);
  g_free (path);

  return TRUE;
}

/**
 * on_open_action_activated:
 */
void
on_open_action_activated (GtkAction *action,
			  GtkWidget *window)
{
  static gchar *last_dir = NULL;
  GtkWidget *dialog;
  GtkFileFilter *filter;
  gchar *path = NULL;

  g_return_if_fail (GTK_IS_WINDOW (window));

  dialog = gtk_file_chooser_dialog_new (_("Open record file"),
					GTK_WINDOW (window),
					GTK_FILE_CHOOSER_ACTION_OPEN,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
					NULL);
  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("Terminal session record file"));
  gtk_file_filter_add_mime_type (filter, "application/x-ttyrcd");
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (dialog), filter);
  gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (dialog), filter);

  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("All files"));
  gtk_file_filter_add_pattern (filter, "*");
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (dialog), filter);

  if (last_dir)
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), last_dir);

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      path = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    }
  gtk_widget_destroy (dialog);

  if (path)
    {
      GFile *file = g_file_new_for_path (path);
      if (assay_open_file (window, file))
	{
	  /* save last access directory. */
	  last_dir = g_path_get_dirname (path);
	}
      g_object_unref (file);
      g_free (path);
    }
}

/**
 * assay_close_file:
 */
void
assay_close_file (GtkWidget *window)
{
  GtkBuilder *builder;
  TDecoder *decoder;
  GSList *objects, *node;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");
  decoder = g_object_get_data (G_OBJECT (window), "decoder");

  t_decoder_close (decoder);

  assay_set_widgets_sensitive (window, FALSE);
  assay_hide_message (window);

  objects = gtk_builder_get_objects (builder);
  for (node = objects; node; node = node->next)
    {
      if (GTK_IS_ENTRY (node->data))
	{
	  gtk_entry_set_text (GTK_ENTRY (node->data), "");
	  gtk_entry_set_progress_fraction (GTK_ENTRY (node->data), 0.0);
	}
      else if (GTK_IS_LIST_STORE (node->data))
	{
	  gtk_list_store_clear (GTK_LIST_STORE (node->data));
	}
      else if (GTK_IS_ADJUSTMENT (node->data))
	{
	  gtk_adjustment_configure (GTK_ADJUSTMENT (node->data),
				    0, 0, 0, 0, 0 ,0);
	}
      else if (GTK_IS_SCALE (node->data))
	{
	  gtk_scale_clear_marks (GTK_SCALE (node->data));
	}
      else if (GTK_IS_TEXT_BUFFER (node->data))
	{
	  gtk_text_buffer_set_text (GTK_TEXT_BUFFER (node->data), "", -1);
	}
    }
  g_slist_free (objects);
}

/**
 * on_close_action_activated:
 */
void
on_close_action_activated (GtkAction *action,
			   GtkWidget *window)
{
  g_return_if_fail (GTK_IS_WINDOW (window));

  assay_close_file (window);
}

/**
 * on_play_action_activated:
 */
void
on_play_action_activated (GtkAction *action,
			  GtkWidget *window)
{
  TDecoder *decoder;
  static gchar command[512];
  gchar *progname;
  GError *error = NULL;

  g_return_if_fail (GTK_IS_WINDOW (window));

  decoder = g_object_get_data (G_OBJECT (window), "decoder");
  if (!decoder)
    return;

  progname = g_find_program_in_path ("ttyplay");
  if (!progname)
    {
      assay_show_message (window, GTK_MESSAGE_ERROR,
			  "ttyplay not found.");
      return;
    }
  g_snprintf (command, sizeof (command),
	      "%s --newwin --interval=1000 "
	      "--logfile=/tmp/ttyplay.log %s",
	      progname, t_decoder_get_filename (decoder));
  g_free (progname);

  g_spawn_command_line_async (command, &error);
  if (error)
    {
      assay_show_message (window, GTK_MESSAGE_ERROR,
			  error->message);
      g_error_free (error);
    }
}

/**
 * on_about_action_activated:
 */
void
on_about_action_activated (GtkAction *action,
			   GtkWidget *window)
{
  const gchar *copyright = _("Copyright (C) 2010 wuxiaohu, "
			     "all rights reserved.");
  const gchar *authors[] = { "xiaohu<xiaohu417@gmail.com>", NULL };

  g_return_if_fail (GTK_IS_WINDOW (window));

  gtk_show_about_dialog (GTK_WINDOW (window),
			 "logo-icon-name", GTK_STOCK_MEDIA_RECORD,
			 "version", VERSION,
			 "comments", _("Deagnostic and repair tool for Ttyutils record file"),
			 "copyright", copyright,
			 "website", WEBSITE,
			 "authors", authors,
			 "license", "This software is released under BSD license, "
			 "a copied of the COPYING file is includes in the release.",
			 "wrap-license", TRUE,
			 NULL);
}


/**
 * on_index_tree_selection_changed:
 */
static void
on_index_tree_selection_changed (GtkTreeSelection *selection,
				 GtkWidget        *window)
{
  GtkBuilder *builder;
  TDecoder *decoder;
  GtkTreeModel *model;
  GtkTreeIter iter;
  guint n, i;
  GObject *adjust;
  GObject *text_buffer;
  GObject *action;
  const gchar *data;
  guint16 size;
  GError *error = NULL;

  g_return_if_fail (GTK_IS_TREE_SELECTION (selection));
  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");
  decoder = g_object_get_data (G_OBJECT (window), "decoder");

  if (!t_decoder_get_archive (decoder))
    return;

  if (!gtk_tree_selection_get_selected (selection, &model, &iter))
    return;

  gtk_tree_model_get (GTK_TREE_MODEL (model), &iter, 0, &n, -1);
  if (!t_decoder_has_index (decoder, n))
    return;

  text_buffer = gtk_builder_get_object (builder, "DataTextbuffer");
  gtk_text_buffer_set_text (GTK_TEXT_BUFFER (text_buffer), "", -1);

  data = t_decoder_get_index_data (decoder, n, &size, &error);
  if (error)
    {
      if (error->code == G_IO_ERROR_PERMISSION_DENIED)
	gtk_text_buffer_set_text (GTK_TEXT_BUFFER (text_buffer),
				  _("Entryped"), -1);
      else
	assay_show_message (window, GTK_MESSAGE_ERROR, error->message);

      g_error_free (error);

      action = gtk_builder_get_object (builder, "DisplayInTermAction");
      gtk_action_set_sensitive (GTK_ACTION (action), FALSE);
      action = gtk_builder_get_object (builder, "DisplayInTextEditorAction");
      gtk_action_set_sensitive (GTK_ACTION (action), FALSE);
      action = gtk_builder_get_object (builder, "DisplayInHexEditorAction");
      gtk_action_set_sensitive (GTK_ACTION (action), FALSE);
      return;
    }
  action = gtk_builder_get_object (builder, "DisplayInTermAction");
  gtk_action_set_sensitive (GTK_ACTION (action), TRUE);
  action = gtk_builder_get_object (builder, "DisplayInTextEditorAction");
  gtk_action_set_sensitive (GTK_ACTION (action), TRUE);
  action = gtk_builder_get_object (builder, "DisplayInHexEditorAction");
  gtk_action_set_sensitive (GTK_ACTION (action), TRUE);

  for (i = 0; i < size; i++)
    {
      GtkTextIter iter;
      gchar buff[8];

      gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (text_buffer), &iter);

      if (isprint (data[i]))
	{
	  gtk_text_buffer_insert_with_tags_by_name (GTK_TEXT_BUFFER (text_buffer),
						    &iter, data + i, 1, "font", NULL);
	}
      else
	{
	  snprintf (buff, sizeof (buff), "\\%03o", data[i]);
	  gtk_text_buffer_insert_with_tags_by_name (GTK_TEXT_BUFFER (text_buffer),
						    &iter, buff, -1, "blue", "font", NULL);
	}
    }
  /* update index scale. */
  adjust = gtk_builder_get_object (builder, "IndexScaleAdjustment");
  if (n != (guint32) gtk_adjustment_get_value (GTK_ADJUSTMENT (adjust)))
    gtk_adjustment_set_value (GTK_ADJUSTMENT (adjust), (gdouble) n);
}

/**
 * on_index_scale_value_changed:
 */
void
on_index_scale_value_changed (GtkAdjustment *adjust,
			      GtkWidget     *window)
{
  GtkBuilder *builder;
  guint32 value, start;
  gchar path[32];
  GtkTreePath *treepath;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");

  value = (guint32) gtk_adjustment_get_value (adjust);
  start = (guint32) gtk_adjustment_get_lower (adjust);
  g_snprintf (path, sizeof (path), "%d", value - start);

  treepath = gtk_tree_path_new_from_string (path);
  if (treepath)
    {
      GObject *treeview;
      GtkTreeSelection *selection;

      treeview = gtk_builder_get_object (builder, "IndexTreeview");
      selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));

      if (!gtk_tree_selection_path_is_selected (selection, treepath))
	{
	  gtk_tree_selection_select_path (selection, treepath);
	  gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (treeview), treepath,
					NULL, TRUE, 0.5, 0.0);
	}
      gtk_tree_path_free (treepath);
    }
}

/**
 * on_index_first_button_clicked:
 */
void
on_index_first_button_clicked (GtkWidget *button,
			       GtkWidget *window)
{
  GtkBuilder *builder;
  GObject *adjust;
  guint32 value;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");

  adjust = gtk_builder_get_object (builder, "IndexScaleAdjustment");
  value = gtk_adjustment_get_lower (GTK_ADJUSTMENT (adjust));
  gtk_adjustment_set_value (GTK_ADJUSTMENT (adjust), (gdouble) value);
}

/**
 * on_index_last_button_clicked:
 */
void
on_index_last_button_clicked (GtkWidget *button,
			      GtkWidget *window)
{
  GtkBuilder *builder;
  GObject *adjust;
  guint32 value;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");

  adjust = gtk_builder_get_object (builder, "IndexScaleAdjustment");
  value = gtk_adjustment_get_upper (GTK_ADJUSTMENT (adjust));
  gtk_adjustment_set_value (GTK_ADJUSTMENT (adjust), (gdouble) value);
}

/**
 * on_index_top_button_clicked:
 */
void
on_index_top_button_clicked (GtkWidget *button,
			     GtkWidget *window)
{
  GtkBuilder *builder;
  TDecoder *decoder;
  GObject *adjust;
  GObject *widget;
  guint32 value;
  GError *error = NULL;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");
  decoder = g_object_get_data (G_OBJECT (window), "decoder");

  if (!t_decoder_get_archive (decoder))
    return;

  adjust = gtk_builder_get_object (builder, "IndexScaleAdjustment");
  value = gtk_adjustment_get_lower (GTK_ADJUSTMENT (adjust));

  widget = gtk_builder_get_object (builder, "IndexBottomButton");
  gtk_widget_set_sensitive (GTK_WIDGET (widget), TRUE);

  /* nothing to do. */
  if (value == 0)
    {
      gtk_widget_set_sensitive (button, FALSE);
      return;
    }
  /* tell decoder to read next index file. */
  t_decoder_seek (decoder, value - 1, &error);
  if (error)
    {
      assay_show_message (window, GTK_MESSAGE_ERROR, error->message);
      g_error_free (error);
      return;
    }
  /* reload indexes into treeview */
  assay_read_indexes_into_treeview (window);

  gtk_adjustment_set_value (GTK_ADJUSTMENT (adjust), (gdouble) (value - 1));
}

/**
 * on_index_bottom_button_clicked:
 */
void
on_index_bottom_button_clicked (GtkWidget *button,
				GtkWidget *window)
{
  GtkBuilder *builder;
  TDecoder *decoder;
  GObject *adjust;
  GObject *widget;
  guint32 value;
  GError *error = NULL;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");
  decoder = g_object_get_data (G_OBJECT (window), "decoder");

  if (!t_decoder_get_archive (decoder))
    return;

  adjust = gtk_builder_get_object (builder, "IndexScaleAdjustment");
  value = gtk_adjustment_get_upper (GTK_ADJUSTMENT (adjust));

  widget = gtk_builder_get_object (builder, "IndexTopButton");
  gtk_widget_set_sensitive (GTK_WIDGET (widget), TRUE);

  /* nothing to do. */
  if (value == t_decoder_get_nindexes (decoder) - 1)
    {
      gtk_widget_set_sensitive (button, FALSE);
      return;
    }
  /* tell decoder to read next index file. */
  t_decoder_seek (decoder, value + 1, &error);
  if (error)
    {
      assay_show_message (window, GTK_MESSAGE_ERROR, error->message);
      g_error_free (error);
      return;
    }
  /* reload indexes into treeview */
  assay_read_indexes_into_treeview (window);

  gtk_adjustment_set_value (GTK_ADJUSTMENT (adjust), (gdouble) (value + 1));
}

/**
 * on_index_treeview_button_press:
 */
gboolean
on_index_treeview_button_press (GtkTreeView    *treeview,
				GdkEventButton *event,
				GtkWidget      *window)
{
  GtkBuilder *builder;
  GtkTreeSelection *selection;
  GtkTreePath *path;
  GObject *menu;

  g_return_val_if_fail (GTK_IS_TREE_VIEW (treeview), FALSE);
  g_return_val_if_fail (GTK_IS_WINDOW (window), FALSE);

  if (event->button != 3 || event->type != GDK_BUTTON_PRESS)
    return FALSE;

  builder = g_object_get_data (G_OBJECT (window), "builder");

  selection = gtk_tree_view_get_selection (treeview);

  /* get tree path for row which behind mouse point. */
  if (gtk_tree_view_get_path_at_pos (treeview, (gint) event->x, (gint) event->y,
				     &path, NULL, NULL, NULL))
    {
      if (!gtk_tree_selection_path_is_selected (selection, path))
	{
	  gtk_tree_selection_unselect_all (selection);
	  gtk_tree_selection_select_path (selection, path);
	}
      gtk_tree_path_free (path);

      menu = gtk_builder_get_object (builder, "IndexPopupMenu");
      gtk_widget_show_all (GTK_WIDGET (menu));
      gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 
		      event->button, event->time);
      return TRUE;
    }
  return FALSE;
}


typedef const gchar*	(*FillCommandLineCB)	(GtkWidget   *window,
						 const gchar *filename);

/**
 * unlink_tmpfile_timeout:
 */
static gboolean
unlink_tmpfile_timeout (gpointer data)
{
  g_unlink ((gchar*) data);
  return FALSE;
}

/**
 * assay_display_selected_data:
 */
static void
assay_display_selected_data (GtkWidget        *window,
			     FillCommandLineCB cb)
{
  GtkBuilder *builder;
  TDecoder *decoder;
  GObject *treeview;
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;
  guint n;
  const gchar *data, *command;
  guint16 size;
  gchar tmpname[128];
  gint fd;
  GdkScreen *screen;
  GError *error = NULL;

  g_return_if_fail (GTK_IS_WINDOW (window));

  builder = g_object_get_data (G_OBJECT (window), "builder");
  decoder = g_object_get_data (G_OBJECT (window), "decoder");

  treeview = gtk_builder_get_object (builder, "IndexTreeview");
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
  if (!gtk_tree_selection_get_selected (selection, &model, &iter))
    return;

  /* get the selected row. */
  gtk_tree_model_get (GTK_TREE_MODEL (model), &iter, 0, &n, -1);
  if (!t_decoder_has_index (decoder, n))
    return;

  /* get the selected data. */
  data = t_decoder_get_index_data (decoder, n, &size, &error);
  if (error)
    {
      if (error->code != G_IO_ERROR_PERMISSION_DENIED)
	assay_show_message (window, GTK_MESSAGE_ERROR, error->message);

      g_error_free (error);
      return;
    }
  /* create temporary file. */
  g_snprintf (tmpname, sizeof (tmpname), "%s/ttyassay.XXXXXX", g_get_tmp_dir ());
  fd = g_mkstemp (tmpname);
  if (fd < 0)
    {
      g_warning ("Unable to create temporary file.");
      return;
    }
  close (fd);

  /* save data into temporary file. */
  g_file_set_contents (tmpname, data, size, &error);
  if (error)
    {
      assay_show_message (window, GTK_MESSAGE_ERROR, error->message);
      g_error_free (error);
      return;
    }
  /* get command line. */
  command = cb (window, tmpname);
  if (!command)
    {
      g_unlink (tmpname);
      return;
    }
  /* startup the program. */
  screen = gtk_window_get_screen (GTK_WINDOW (window));
  gdk_spawn_command_line_on_screen (screen, command, &error);
  if (error)
    {
      assay_show_message (window, GTK_MESSAGE_ERROR, error->message);
      g_error_free (error);
      g_unlink (tmpname);
      return;
    }
  /* if application quit before timeout, the termporary file
   * will not be removed. */
  g_timeout_add_seconds_full (G_PRIORITY_DEFAULT, 3,
			      unlink_tmpfile_timeout,
			      g_strdup (tmpname), g_free);
}

/**
 * fill_terminal_command_cb:
 */
static const gchar*
fill_terminal_command_cb (GtkWidget   *window,
			  const gchar *filename)
{
  static gchar command[512];
  gchar *progname;

  /* find terminal program. */
  progname = g_find_program_in_path ("gnome-terminal");
  if (progname)
    {
      TDecoder *decoder;
      TRecordHead *head;

      decoder = g_object_get_data (G_OBJECT (window), "decoder");
      head = t_decoder_get_head (decoder);

      g_snprintf (command, sizeof (command),
		  "%s --geometry=%dx%d+0+0 -e \"cat %s -\"",
		  progname, head->ncols, head->nrows, filename);
      g_free (progname);
      return command;
    }
  progname = g_find_program_in_path ("xterm");
    {
      g_snprintf (command, sizeof (command),
		  "%s -e \"cat %s -\"", progname, filename);
      g_free (progname);
      return command;
    }
  assay_show_message (window, GTK_MESSAGE_ERROR,
		      _("Terminal application not found."));
  return NULL;
}

/**
 * on_display_in_terminal_activated:
 */
void
on_display_in_terminal_activated (GtkAction *action,
				  GtkWidget *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GTK_IS_WINDOW (window));

  assay_display_selected_data (window, fill_terminal_command_cb);
}

/**
 * fill_text_editor_command_cb:
 */
static const gchar*
fill_text_editor_command_cb (GtkWidget   *window,
			     const gchar *filename)
{
  static gchar command[512];
  gchar *progname;

  /* find text editor program. */
  progname = g_find_program_in_path ("gvim");
  if (progname)
    {
      g_snprintf (command, sizeof (command), "%s %s", progname, filename);
      g_free (progname);
      return command;
    }
  progname = g_find_program_in_path ("emacs");
    {
      g_snprintf (command, sizeof (command), "%s %s", progname, filename);
      g_free (progname);
      return command;
    }
  assay_show_message (window, GTK_MESSAGE_ERROR,
		      _("Text editor application not found."));
  return NULL;
}

/**
 * on_display_in_text_editor_activated:
 */
void
on_display_in_text_editor_activated (GtkAction *action,
				     GtkWidget *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GTK_IS_WINDOW (window));

  assay_display_selected_data (window, fill_text_editor_command_cb);
}

/**
 * fill_hex_editor_command_cb:
 */
static const gchar*
fill_hex_editor_command_cb (GtkWidget   *window,
			    const gchar *filename)
{
  static gchar command[512];
  gchar *progname;

  /* find hex editor program. */
  progname = g_find_program_in_path ("ghex2");
  if (progname)
    {
      g_snprintf (command, sizeof (command), "%s %s", progname, filename);
      g_free (progname);
      return command;
    }
  assay_show_message (window, GTK_MESSAGE_ERROR,
		      _("Hex editor application not found."));
  return NULL;
}

/**
 * on_display_in_hex_editor_activated:
 */
void
on_display_in_hex_editor_activated (GtkAction *action,
				    GtkWidget *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GTK_IS_WINDOW (window));

  assay_display_selected_data (window, fill_hex_editor_command_cb);
}

/**
 * on_convert_script_action_activated:
 *
 * convert script file.
 */
void
on_convert_script_action_activated (GtkAction *action,
				    GtkWidget *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GTK_IS_WINDOW (window));

  /* open convert assistant. */
  assay_convert_assistant_run (window);
}

/**
 * on_view_escape_action_activated:
 */
void
on_view_escape_action_activated (GtkAction *action,
				 GtkWidget *window)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GTK_IS_WINDOW (window));

  g_debug ("TODO");
  /* open convert assistant. */
  //assay_convert_assistant_run (window);
}
